# button.tcl --
#
#	Wrappers around several tk buttons and the tk label. Adds
#	some aliases for widget options and parses the label setting
#	for "&" to generate key bindings.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/button.tcl,v 1.13 1999/10/26 13:12:14 fjl Exp $

package provide twoButton 1.0
package require twoUtil

namespace eval ::two::button {
    namespace export two_*
    variable Button
    
    namespace import ::two::util::two_parseArgs
}

# ::two::button::two_button --
#
#	extends button; see AnyButton

proc ::two::button::two_button {args} {
    return [eval AnyButton button $args]
}

# ::two::button::two_label --
#
#	extends label; see AnyButton

proc ::two::button::two_label {args} {
    return [eval AnyButton label $args]
}

# ::two::button::two_checkbutton --
#
#	extends checkbutton; see AnyButton

proc ::two::button::two_checkbutton {args} {
    return [eval AnyButton checkbutton $args]
}

# ::two::button::AnyButton --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".b"); the original widget procedure
#	is renamed to "_two$w" (like "_two.b") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#
# Arguments:
#	cmd		What command should be used? (button, checkbutton, label)
#	w		Widget path
#	args		Widget arguments and options

proc ::two::button::AnyButton {cmd w args} {
    # several ways to specify button text

    set idx  -1
    set text ""
    two_parseArgs args text "-text"
    two_parseArgs args text "-label"
    two_parseArgs args text "-l"
    two_parseArgs args text "-t"
    if {[string length $text]} {
	set idx [string first "&" $text]
	if {$idx >= 0} {
	    set underline $idx
	    # ignore "&&"
	    while {[string match "&" [string index $text [expr {$idx + 1}]]]} {
		set base [expr {$idx + 2}]
		set idx  [string first "&" [string range $text $base end]]
		if {$idx < 0} {
		    break
		} else {
		    set underline [expr {$underline + $idx + 1}]
		    incr idx $base
		}
	    }
	}
	if {$idx >= 0} {
	    regsub -all -- {&([^&])} $text {\1} text
	    lappend args -text $text -underline $idx
	} else {
	    lappend args -text $text
	}
    }

    # Create button
    uplevel [list $cmd $w] $args

    # Create binding
    if {$idx >= 0} {
        set key "[string index $text $idx]"
	set tl [winfo toplevel $w]
	bind $tl <Alt-Key-[string tolower $key]> [list two_invoke $w]
	bind $tl <Alt-Key-[string toupper $key]> [list two_invoke $w]
    }

    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "

    # catch widget destruction in order to clean up Combobox
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # Return the correct result...
    return $w
}

# ::two::button::Widget --
#	replacement for widget command
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::button::Widget {w opt args} {
    variable Button

    # See if this option is for us...
    switch -exact -- $opt {
	cget {
	    switch -exact -- [lindex $args 0] {
	        -label -
		-l       { return [eval [list _two$w $opt] [lreplace $args 0 0 "-text"]] }
		default  { return [eval [list _two$w $opt] $args] }
	    }
	}
	configure {
	    set text ""
	    two_parseArgs args text "-text"
	    two_parseArgs args text "-label"
	    two_parseArgs args text "-l"
	    if {[string length $text]} {
        	set idx [string first "&" $text]
		if {$idx >= 0} {
		    # remove old binding
    		    if {[info exists Button($w,bind)]} {
		        bind [winfo toplevel $w] <$Button($w,bind)> {#nothing}
		    }
		    regsub "&" $text "" text
		    lappend args -text $text -underline $idx
		    # Create binding
		    set key "Alt-Key-[string tolower [string index $text $idx]]"
                    bind [winfo toplevel $w] <$key> [list two_invoke $w]
		    set Button($w,bind) $key
		} else {
		    lappend args -text $text
		}
	    }
	    return [eval [list _two$w $opt] $args]
	}
	default {
            return [eval [list _two$w $opt] $args]
	}
    }
    return
}

# ::two::button::_Destroy --
#
#	Call this procedure when the two_menu widget is destroyed.
#	Will destroy the original widget command.
#
# Arguments:
#	w		Path to Widget

proc ::two::button::_Destroy {w} {
    # remove the original widget command
    rename $w ""
    return
}

namespace import ::two::button::*
