# combobox.tcl --
#
#	Combobox widget. A combobox is like an editable option menu.
#	Consists of an entry, a button and a dropdown list.
#
# Copyright (c) 1998, 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/combobox.tcl,v 1.43 2000/01/13 14:51:18 fjl Exp $

package provide twoCombobox 1.0
package require twoUtil
package require twoAutosb

option add *TwoComboBox.borderWidth               2       widgetDefault
option add *TwoComboBox.drop.borderWidth          0
option add *TwoComboBox.drop.highlightThickness   1
option add *TwoComboBox.drop.highlightColor       black
option add *TwoComboBox.drop.relief               flat
option add *TwoComboBox.relief                    sunken  widgetDefault
option add *TwoComboBox.Button.highlightThickness 1
option add *TwoComboBox.Entry.borderWidth         0

# Workaround: Tk needs these procedures for focus traversion
# but won't load it form inside the namespace (bug in tcl8.0 up to p1)
catch {auto_load tk_focusNext}
catch {auto_load tk_focusPrev}

namespace eval ::two::combobox {
    namespace export two_combobox
    variable Combobox
    
    namespace import ::two::util::two_parseArgs
    
    # variable Combobox
    #
    # current -	path to currently opened combobox
    # buttonPress -	remember if buttonPress-1 has happend after opening drop
    # $w,command -	command that will be executed when opening the combobox
    # $w,state -	current state {normal|fixed|disabled}
    # $w,takefocus - should the combobox take part in keyboard focus traversal?
    # $w,style -	look of the combobox {windows|unix}
    
    array set Combobox {
        Drag	0
        buttonPress 0
        current	""
    }
}

image create bitmap comboBoxArrowWin -data {
\#define cbwin_width 9
\#define cbwin_height 10
static unsigned char cbwin_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x7c, 0x00, 0x38, 0x00,
  0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
}

image create bitmap comboBoxArrowX -data {
\#define cbarrow_width 9
\#define cbarrow_height 12
static unsigned char cbarrow_bits[] = {
  0x38, 0x00, 0x38, 0x00, 0x38, 0x00, 0x38, 0x00, 0xff, 0x01, 0xfe, 0x00,
  0x7c, 0x00, 0x38, 0x00, 0x10, 0x00, 0x00, 0x00, 0xff, 0x01, 0xff, 0x01,
  };
}

# ::two::combobox::two_combobox --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.cb"); the original widget procedure
#	is renamed to "_two$w" (like "_two.tl.cb") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#	combobox creates a toplevel widget "$w.drop" for the drop box.
#	Arguments are passed to entry widget.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::combobox::two_combobox {w args} {
    variable Combobox
    global tcl_platform

    # Initialise the globals
    array set Combobox "
	$w,command  {}
	$w,state    normal
	$w,style    $tcl_platform(platform)
	$w,takefocus 1
    "
    
    # This is a multi-style widget
    set ::two::SetStyleCmds($w) [list $w configure -style]

    # See if this option is for us...
    two_parseArgs args Combobox($w,command)   "-command"
    two_parseArgs args Combobox($w,state)     "-state"
    two_parseArgs args Combobox($w,style)     "-style"
    two_parseArgs args Combobox($w,takefocus) "-takefocus"

    # Container frame
    frame $w -class TwoComboBox
    set tl [winfo toplevel $w]
    uplevel [list entry $w.entry] $args
    if {![string match "unix" $Combobox($w,style)]} {
        button $w.button -image comboBoxArrowWin
    } else {
        button $w.button -image comboBoxArrowX
    }
    switch -exact -- $Combobox($w,state) {
	fixed {
            $w.entry configure  -state "disabled"
	    $w.button configure -takefocus $Combobox($w,takefocus)
	}
	disabled {
            $w.entry  configure -state "disabled"
            $w.button configure -state "disabled"
	}
	normal -
	default {
            $w.entry  configure -takefocus $Combobox($w,takefocus)
	    $w.button configure -takefocus 0
	}
    }
    pack $w.button -fill both -side right
    pack $w.entry -expand true -fill both -side left
    # Use our own bindings
    bindtags $w.entry  [list $w.entry  TwoComboBoxEntry  Entry $tl all]
    bindtags $w.button [list $w.button TwoComboBoxButton       $tl all]

    # Accompaining dropdown (toplevel widget)
    set d $w.drop
    toplevel $d -class TwoComboBox
    wm withdraw $d
    wm overrideredirect $d 1
    two_autosb $d.sx -command [list $d.lb xview] -orient horizontal
    two_autosb $d.sy -command [list $d.lb yview]
    listbox $d.lb -exportselection 0 -bd 0 \
        -xscrollcommand [list $d.sx set] \
        -yscrollcommand [list $d.sy set]
    frame $d.handle -height 4 -cursor bottom_side -bd 0

    grid $d.lb $d.sy -sticky nsew
    grid $d.sx       -sticky ew
    grid $d.handle   -sticky ew -columnspan 2
    grid rowconfigure    $d 0 -weight 1
    grid columnconfigure $d 0 -weight 1

    # Use our own bindings
    # Caveat: combobox bindings have to appear after normal widget bindings
    bindtags $d.lb [list $d.lb Listbox TwoComboBoxList TwoComboBox $tl all]
    # Handle for drop down list uses only its own bindings
    bindtags $d.handle [list $d.handle TwoComboBoxHandle $tl all]

    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Combobox
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # Return the correct result...
    return $w
}

# ::two::combobox::Widget --
#
#	replacement for widget command
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::combobox::Widget {w opt args} {
    variable Combobox

    # See if this option is for us...
    switch -exact -- $opt {
	configure {
	    two_parseArgs args Combobox($w,command)   "-command"
	    two_parseArgs args Combobox($w,state)     "-state"
	    two_parseArgs args Combobox($w,takefocus) "-takefocus"
	    switch -exact -- $Combobox($w,state) {
		fixed {
        	    $w.entry  configure -state "disabled" -takefocus 0
        	    $w.button configure -state "normal"   -takefocus $Combobox($w,takefocus)
		}
		disabled {
        	    $w.entry  configure -state "disabled" -takefocus 0
        	    $w.button configure -state "disabled" -takefocus 0
		}
		normal -
	        default {
        	    $w.entry  configure -state "normal" -takefocus $Combobox($w,takefocus)
        	    $w.button configure -state "normal" -takefocus 0
		}
	    }
	    if {[two_parseArgs args Combobox($w,style) "-style"]} {
		if {![string match "unix" $Combobox($w,style)]} {
		    $w.button configure -image comboBoxArrowWin
		} else {
		    $w.button configure -image comboBoxArrowX
		}
	    }

	    if {[llength $args]} {
		return [eval [list _two$w $opt] $args]
	    }
	}
	cget {
	    switch -exact -- $args {
	        -command { return $Combobox($w,command) }
	        -state   { return $Combobox($w,state)   }
	        -style   { return $Combobox($w,style)   }
		default  { return [eval [list _two$w $opt] $args] }
	    }
	}
	setFocus {
	    switch -exact -- $Combobox($w,state) {
		fixed {
		    focus $w.button
		}
		disabled {
		    focus $w
		}
		normal -
	        default {
        	    focus $w.entry
		}
	    }
	}
	get {
	    return [$w.entry get]
	}
	set {
    	    set state [$w.entry cget -state]
	    $w.entry configure -state normal
	    $w.entry delete 0 end
	    $w.entry insert 0 [lindex $args 0]
	    $w.entry configure -state $state
	}
	setlist {
	    $w.drop.lb delete 0 end
	    eval [list $w.drop.lb insert end] $args
	}
	getlist {
	    return [$w.drop.lb get 0 end]
	}
	add {
	    eval [list $w.drop.lb insert end] $args
	}
	open {
	    Open $w
	}
	entry {
	    return $w.entry
	}
	listbox {
	    return $w.drop.lb
	}
	default {
            return [eval [list _two$w $opt] $args]
	}
    }
    return
}

# ::two::combobox::_Destroy --
#
#	Call this procedure when the combobox widget is destroyed.
#	Will clean widget-specific Elements of Combobox.
#
# Arguments:
#	w		Path to Widget

proc ::two::combobox::_Destroy {w} {
    variable Combobox

    unset ::two::SetStyleCmds($w)

    # clear Combobox
    foreach key [array names Combobox "$w,*"] {
        unset Combobox($key)
    }
    # remove the original widget command
    rename $w ""
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoComboBox-class bindings
#-------------------------------------------------------------------------

# Entry widget in combobox

bind TwoComboBoxEntry <1> {
    if {[string match "fixed" $::two::combobox::Combobox([winfo parent %W],state)]} {
        ::two::combobox::Open [winfo parent %W]
    }
}
bind TwoComboBoxEntry <Next> {
    ::two::combobox::Open [winfo parent %W]
}

# drop down button

bind TwoComboBoxButton <1> {
    if {![string match "disabled" [%W cget -state]]} {
        %W configure -relief sunken
        ::two::combobox::Open [winfo parent %W]
        set ::two::combobox::Combobox(buttonPress) 0
    }
}
bind TwoComboBoxButton <space> {
    %W configure -relief sunken
    ::two::combobox::Open [winfo parent %W]
    set ::two::combobox::Combobox(buttonPress) 0
}
bind TwoComboBoxButton <Next> {
    %W configure -relief sunken
    ::two::combobox::Open [winfo parent %W]
    set ::two::combobox::Combobox(buttonPress) 0
}
bind TwoComboBoxButton <Enter> {
    if {![string match "disabled" [%W cget -state]]} {
        %W configure -state active
    }
}
bind TwoComboBoxButton <Leave> {
    if {![string match "disabled" [%W cget -state]]} {
        %W configure -state normal
    }
}

# Listbox widget in dropped down combobox

bind TwoComboBoxList <Return> {
    ::two::combobox::Close $::two::combobox::Combobox(current)
    break
}
bind TwoComboBoxList <Escape> {
    %W selection clear 0 end
    ::two::combobox::Close $::two::combobox::Combobox(current)
    break
}

# Handle for resizing drop-down list

bind TwoComboBoxHandle <Enter> {
    ::two::combobox::HandleMotion %W %x
}
bind TwoComboBoxHandle <Motion> {
    ::two::combobox::HandleMotion %W %x
}
bind TwoComboBoxHandle <1> {
    ::two::combobox::HandleB1 %W %x %y
}
bind TwoComboBoxHandle <ButtonRelease-1> {
    set ::two::combobox::Combobox(Drag) 0
}

bind TwoComboBoxHandle <B1-Motion> {
    ::two::combobox::HandleB1Motion %W %x %X %y %Y
}

# Bindings to the frames only trigger when the listbox is
# opened (dropped down)

bind TwoComboBox <ButtonPress> {
    set w [winfo containing [winfo pointerx .] [winfo pointery .]]
    if {[string match $w $::two::combobox::Combobox(current).button]} {
        set ::two::combobox::Combobox(buttonPress) 1
    }
}
bind TwoComboBox <ButtonRelease-1> {
    set w [winfo containing [winfo pointerx .] [winfo pointery .]]
    if {$::two::combobox::Combobox(buttonPress) ||
        ![string match $w $::two::combobox::Combobox(current).button]} {
        if {[winfo exists $::two::combobox::Combobox(current)]} {
            ::two::combobox::Close $::two::combobox::Combobox(current)
        }
    }
}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::combobox::HandleMotion --
#
#	handle mouse cursor motion events within handle frame. Set
#	the mouse cursor shape according to the relative position.
#
# Arguments:
#	h		Path to handle
#	x		relative X Position of mouse cursor within handle (%x)

proc ::two::combobox::HandleMotion {h x} {
    variable Combobox

    if {!$Combobox(Drag)} {
        if {$x > [winfo width $h] - 10} {
            $h configure -cursor bottom_right_corner
        } elseif {$x < 10} {
            $h configure -cursor bottom_left_corner
        } else {
            $h configure -cursor bottom_side
        }
    }
    return
}

# ::two::combobox::HandleB1 --
#
#	handle mosue button press within within handle frame.
#
# Arguments:
#	h		Path to handle
#	x		relative X Position of mouse cursor within handle (%x)
#	y		relative Y Position (%y)

proc ::two::combobox::HandleB1 {h x y} {
    variable Combobox

    set p [winfo toplevel $h]
    set ht [$p cget -highlightthickness]
    set Combobox(Drag) 1
    set Combobox(Cursor) [$h cget -cursor]
    set Combobox(dxl) [expr {$x + $ht}]
    set Combobox(dxr) [expr {[winfo width  $h] - $x + $ht}]
    set Combobox(x) [expr {[winfo rootx $p] + [winfo width  $h] + $x + 3 * $ht}]
    set Combobox(y) [expr {[winfo rooty $p] - [winfo height $h] + $y - $ht}]

    return
}

# ::two::combobox::HandleB1Motion --
#
#	handle interactive resizing of dropdown list.
#	This relies on the mouse cursor's shape set in the
#	HandleMotion procedure!
#
# Arguments:
#	h		Path to handle
#	relX		relative X Position of mouse cursor within border (%x)
#	absX		absolute X Positino (%X)
#	relY		relative Y Position (%y)
#	absY		absolute Y Position (%Y)

proc ::two::combobox::HandleB1Motion {h relX absX relY absY} {
    variable Combobox

    if {$Combobox(Drag)} {
        set p [winfo parent $h]

	if {[string match "*right*" $Combobox(Cursor)]} {
	    set width [expr {$relX + $Combobox(dxr)}]
	} else {
	    set width [winfo width $p]
	}
	if {[string match "*left*" $Combobox(Cursor)]} {
	    set x [expr {$absX - $Combobox(dxl)}]
	    set width [expr {$Combobox(x) - $absX}]
	} else {
	    set x [winfo rootx $p]
	}
        set y [winfo rooty $p]
        set height [expr {$absY - $Combobox(y)}]
        if {$width < 50} {
            set width 50
        }
        if {$height < 50} {
            set height 50
        }
        wm geometry $p ${width}x${height}+${x}+${y}
    }
    return
}

# ::two::combobox::Open --
#
#	Open the combobox, i.e. show the dropdown listbox.
#
# Arguments:
#	c		Pathname of combobox widget

proc ::two::combobox::Open {c} {
    variable Combobox
    
    if {[string match "disabled" $Combobox($c,state)]} {
        return
    }

    set Combobox(current) $c
    if {[string length $Combobox($c,command)]} {
        namespace eval :: $Combobox($c,command)
    }

    set x [winfo rootx $c]
    set y [expr {[winfo rooty $c] + [winfo height $c]}]
    set w [winfo width $c]
    set h 200

    wm geometry $c.drop "${w}x${h}+${x}+${y}"
    update idletasks
    wm deiconify $c.drop
    raise $c.drop
    focus $c.drop.lb
    grab set -global $c.drop
    return
}

# ::two::combobox::Close --
#
#	Close the combobox, i.e. withdraw the dropdown listbox and
#	insert the current selection into the entry field.
#
# Arguments:
#	c		Pathname of combobox widget

proc ::two::combobox::Close {c} {
    variable Combobox
    
    if {[string length [$c.drop.lb curselection]]} {
    	set state [$c.entry cget -state]
	$c.entry configure -state normal
        $c.entry delete 0 end
        $c.entry insert end [$c.drop.lb get active]
	$c.entry configure -state $state
    }
    grab release $c.drop
    wm withdraw  $c.drop
    focus [tk_focusNext $c]
    $c.button configure -relief raised
    return
}

namespace import ::two::combobox::*
