# menu.tcl --
#
#	Enhanced menu command: alias names for menu entries, handling
#	of "&" to set "underline", tooltips.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/menu.tcl,v 1.21 1999/10/26 13:12:25 fjl Exp $

package provide twoMenu 1.0
package require twoUtil
package require twoTooltip

namespace eval ::two::menu {
    namespace export two_menu
    variable Menu
    
    namespace import ::two::util::two_parseMultiArgs
    
    # variable Menu
    #
    # $w,alias -	index position of alias "alias"
}

# ::two::menu::two_menu --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.xm"); the original widget procedure
#	is renamed to "_two$w" (like "_two.tl.xm") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::menu::two_menu {w args} {
    # Create menu
    uplevel [list menu $w] $args

    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Combobox
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # Return the correct result...
    return $w
}

# ::two::menu::Widget --
#
#	replacement for widget command
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

variable ::two::menu::MenuSpec
array set ::two::menu::MenuSpec {
    -alias	alias
    -tip	tip
    -info	info
    -x		xaccel
    -xaccel	xaccel
    -text	label
    -l		label
    -label	label
}

proc ::two::menu::Widget {w opt args} {
    variable Menu
    # TCL Bug? Using variable would break upvar in called proc:-(
    upvar MenuSpec MenuSpec

    set alias  ""
    set tip    ""
    set info   ""
    set xaccel ""
    set label ""
    # See if this option is for us...
    switch -exact -- $opt {
	add -
	insert -
	entryconfigure {
	    two_parseMultiArgs MenuSpec args
	    # label was specified
	    if {[string length $label]} {
	        set idx [string first "&" $label]
		if {$idx >= 0} {
		    set underline $idx
		    # ignore "&&"
		    while {[string match "&" [string index $label [expr {$idx + 1}]]]} {
			set base [expr {$idx + 2}]
		        set idx  [string first "&" [string range $label $base end]]
			if {$idx < 0} {
			    break
			} else {
			    set underline [expr {$underline + $idx + 1}]
			    incr idx $base
			}
		    }
		}
		if {$idx >= 0} {
		    regsub -all -- {&([^&])} $label {\1} label
		    lappend args -label $label -underline $underline
		} else {
		    lappend args -label $label
		}
	    }
	    # calculate index for insert/entryconfigure
	    if {![string match "add" $opt]} {
	        set index [lindex $args 0]
		if {[string length $alias]} {
	            # new alias
	            set Menu($w,$alias) $index
		} else {
	            if [info exists Menu($w,$index)] {
                        # there already is an alias
			set index $Menu($w,$index)
			set args [concat $index [lrange $args 1 end]]
		    }
		}
	    }
	    # create entry
	    if {[llength $args]} {
		set retval [eval [list _two$w $opt] $args]
	    }
            
	    # calculate index for "add"
	    if {[string match "add" $opt]} {
	        set index [_two$w index end]
		if {[string length $alias]} {
	            set Menu($w,$alias) $index
		}
            }

	    if {[string length $xaccel] == 1} {
	        _two$w entryconfigure $index -accelerator "Ctrl+$xaccel"
                # eval postcommand to set proper state of entries
		# then invoke entry (invoke respects state!)
		# Can't use [winfo toplevel $w] as this would return
		# a false toplevel:-(
		bind . <Control-Key-$xaccel> "
		    eval \[[list $w] cget -postcommand\]
		    [list $w invoke $index]
		    break
		"
	    } elseif {[string match "F*" $xaccel]} {
	        _two$w entryconfigure $index -accelerator "$xaccel"
		bind . <$xaccel> "
		    eval \[[list $w] cget -postcommand\]
		    [list $w invoke $index]
		    break
		"
	    }
	    if {[string length $tip] || [string length $info]} {
	        two_tooltip $w -index $index -tip $tip -info $info
	    }
	    return $retval
	}
	entrycget {
	    switch -exact -- [lindex $args 1] {
	        -tip    { return -code error "Sorry, not yet implemented" }
	        -info   { return -code error "Sorry, not yet implemented" }
		-alias  { return -code error "Sorry, not yet implemented" }
	        -text   { return [eval [list _two$w $opt] [lreplace $args 1 1 "-label"]] }
		default { return [eval [list _two$w $opt] $args] }
	    }
	}
	activate -
	delete -
	index -
	invoke -
	postcascade -
	type -
	yposition {
	    set arglist ""
	    foreach a $args {
                # is this an alias?
		if {[info exists Menu($w,$a)]} {
		    lappend arglist $Menu($w,$a)
		} else {
		    lappend arglist $a
		}
	    }
	    return [uplevel [list _two$w $opt] $arglist]
	}
	default {
            return [uplevel [list _two$w $opt] $args]
	}
    }
    return
}

# ::two::menu::_Destroy --
#
#	Call this procedure when the two_menu widget is destroyed.
#	Will destroy the original widget command.
#
# Arguments:
#	w		Path to Widget

proc ::two::menu::_Destroy {w} {
    # remove the original widget command
    rename $w ""
    return
}

namespace import ::two::menu::*
