# pre.tcl --
#
#	Tcl-Only implementation of the two_pre mechanism. This provides
#	conditinal execution of portions of code, depending on settings
#	within the code and special command line arguments.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/pre.tcl,v 1.18 1999/12/14 13:47:14 fjl Exp $

package provide twoPre 1.0

namespace eval ::two::pre {
    namespace export two_*
    variable Pre
    variable PreFixed
    variable PreStack
    
    # Variables:
    #
    # Pre -		Hash array; keys are all defined macros; value is ignored.
    # PreFixed -	Override two_pre define/undef in files
    # PreStack -	List with "0" and "1": 1 = previous ifdef or ifndef
    #			was true. Implements a stack for proper nesting.
}

# ::two::pre::ParseOptions --
#
#	Parse command line.
#	This has to be done exactly once, so there is no need for a procedure.
#	It should happen transparently to gain compatibility with the
#	preprocessor implementation

proc ::two::pre::ParseOptions {} {
    variable Pre
    variable PreFixed
    global argc argv

    set newargv ""

    for {set i 0} {$i < $argc} {incr i} {
        set arg [lindex $argv $i]
        switch -exact -- $arg {
            -D -
            -DEFINE {
                if {$i == $argc - 1} {
                    return -code error "Usage: -DEFINE name"
                }
                incr i
                set arg [lindex $argv $i]
                set PreFixed($arg) 1
                set Pre($arg) 1
            }
            -U -
            -UNDEF {
                if {$i == $argc - 1} {
                    return -code error "Usage: -UNDEF name"
                }
                incr i
                set arg [lindex $argv $i]
                set PreFixed($arg) 0
                if {[info exists Pre($arg)]} {
                    unset Pre($arg)
                }
            }
            default {
                lappend newargv $arg
            }
        }
    }
    set argv $newargv
    set argc [llength $newargv]
}
::two::pre::ParseOptions

# ::two::pre::two_define --
#
#	define a macro. The name "macro" is only to make it look like
#	the C preprocessor; this is only a flag here. Note that -UNDEFINE on the
#	command line overrides this function.
#
# Arguments:
#	macro		The macro which should be define.

proc ::two::pre::two_define {macro} {
    variable Pre
    variable PreFixed

    if {![string length $macro]} {
	return -code error "wrong # args: should be \"two_define name\""
    }
    if {![info exists PreFixed($macro)]} {
        set Pre($macro) 1
    }
    return
}

# ::two::pre::two_undef --
#
#	undefine a macro. Note that -DEFINE on the command line overrides
#	this function.
#
# Arguments:
#	macro		The macro which should be define.

proc ::two::pre::two_undef {macro} {
    variable Pre
    variable PreFixed

    if {![string length $macro]} {
	return -code error "wrong # args: should be \"two_undef name\""
    }
    if {![info exists PreFixed($macro)]} {
        if {[info exists Pre($macro)]} {
            unset Pre($macro)
	}
    }
    return
}

# ::two::pre::two_ifdef --
#
#	Evaluate script only if macro has been defined.
#	If not, evaluate the "else" script.
#
# Arguments:
#	macro		Macro string which should have been defined before
#	script		Script to evaluate if macro was defined
#	args		1 or 3 elements: "<two_else script> two_endif"

proc ::two::pre::two_ifdef {macro script args} {
    variable Pre
    global errorCode errorInfo

    if {([llength $args] != 1 && [llength $args] != 3) ||
        ([llength $args] == 3 &&
	![string match "two_else"  [lindex $args 0]] &&
	![string match "two_endif" [lindex $args 2]])
    } {
	return -code error "wrong # args: should be \"two_ifdef name script ?two_else script? two_endif\""
    }
    if {![info exists Pre($macro)]} {
        # take the else script
        set script [lindex $args 1]
    }
    set retcode [catch {uplevel $script} retval]
    if {$retcode} {
        return -code $retcode -errorcode $errorCode -errorinfo $errorInfo $retval
    } else {
        return $retval
    }
}

# ::two::pre::two_ifndef --
#
#	Evaluate script only if macro has not been defined.
#	If it has been defined, evaluate the "else" script.
#
# Arguments:
#	macro		Macro string which should not have been defined before
#	script		Script to evaluate if macro was not defined
#	args		1 or 3 elements: "<two_else script> two_endif"

proc ::two::pre::two_ifndef {macro script args} {
    variable Pre
    global errorCode errorInfo

    if {([llength $args] != 1 && [llength $args] != 3) ||
        ([llength $args] == 3 &&
	![string match "two_else"  [lindex $args 0]] &&
	![string match "two_endif" [lindex $args 2]])
    } {
	return -code error "wrong # args: should be \"two_ifndef macro script ?two_else script? two_endif\""
    }
    if {[info exists Pre($macro)]} {
        # take the else script
        set script [lindex $args 1]
    }
    set retcode [catch {uplevel $script} retval]
    if {$retcode} {
        return -code $retcode -errorcode $errorCode -errorinfo $errorInfo $retval
    } else {
        return $retval
    }
}

namespace import ::two::pre::*
