# split.tcl --
#
#	Provides an "interactive geometry manager". 2 Slave widgets are
#	arranged in rows or columns, with the user being able to modify
#	the size of each such pane.
#
# Copyright (c) 1998, 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/split.tcl,v 1.5 1999/10/26 13:12:27 fjl Exp $

package provide twoSplit 1.0
package require twoUtil

option add *TwoSplitKnob.borderWidth         0
option add *TwoSplitKnob.relief              raised
option add *TwoSplitKnob.borderWidth         1

namespace eval ::two::split {
    namespace export two_split
    variable Split
    
    namespace import ::two::util::two_parseArgs
    
    # variable Split
    #
    # $m,style -	style (unix|windows)
    # $m,opaque -	opaque resizing on/off
    # $m,split -	current position of split bar
    # $m,anchor -	side (n/s/e/w)
    # $m,slaves -	list of paths to slaves
    # $m,handle -	path to handle
    # style -	default style
    # split -	default position of split bar
    # anchor -	default side
    # cursor -	cursor for non-dragable area on unix-style handle
    # drag -	currently dragging any split
    # mod -		the currently dragged split has been modified
    
    array set Split "
        style	$tcl_platform(platform)
        opaque	1
        anchor	w
        split	10
        cursor	left_ptr
        drag	0
        mod		0
    "
}

# ::two::split::two_split --
#
#	The split is implemented as a layout manager; it doesn't subclass
#	any existing widget. But this really is a very limited version;
#	it takes exactly 2 slaves:-)
#
# Arguments:
#	opt		Widget command...
#	args		optional args

proc ::two::split::two_split {opt args} {
    variable Split

    switch -glob -- $opt {
	.* {
	    set m [winfo parent $opt]

	    if {![info exists Split($m,slaves)]} {
                array set Split "
                    $m,anchor $Split(anchor)
                    $m,style  $Split(style)
                    $m,opaque $Split(opaque)
                    $m,split  $Split(split)
                    $m,slaves {}
                "
	    }
	    two_parseArgs args Split($m,anchor) "-anchor"
	    two_parseArgs args Split($m,style)  "-style"
	    two_parseArgs args Split($m,opaque) "-opaque"
	    two_parseArgs args Split($m,split)  "-split"
	    
	    if {![string match {[ewns]} $Split($m,anchor)]} {
	        return -code error "Wrong anchor \"$Split($m,anchor)\": Only n, s, e or w"
	    }

	    foreach slave [concat [list $opt] $args] {
	        # remove duplicates => change order!
                set idx [lsearch -exact $Split($m,slaves) $slave]
                if {$idx >= 0} {
                    set Split($m,slaves) [lreplace $Split($m,slaves) $idx $idx]
                } else {
                    bind $slave <Destroy> "+two_split forget $slave"
                }
                # append new slave
	        lappend Split($m,slaves) $slave
	    }

	    if {[llength $Split($m,slaves)] > 2} {
	        return -code error "too many children: only 2 allowed"
	    }

            # This is a multi-style widget
            set slave [lindex $Split($m,slaves) 0]
            set ::two::SetStyleCmds($slave) [list two_split configure $slave -style]
            set slave [lindex $Split($m,slaves) 1]
            if {[info exists ::two::SetStyleCmds($slave)]} {
                unset ::two::SetStyleCmds($slave)
            }

            # create split handle
            if {[llength $Split($m,slaves)] == 2} {
                AddHandle $m
            }

	    # re-set master
	    Config $m
	    
	    # add binding for config changes
	    if {[llength $Split($m,slaves)] == 2} {
                bind $m <Configure> [namespace code "CheckConfig %W %w %h"]
	    }
	}
	forget {
	    if {[llength $args] != 1} {
	        return -code error "wrong # args: should be \"two_split forget path \""
	    }

	    set m [winfo parent [lindex $args 0]]

	    if {![info exists Split($m,slaves)]} {
	        return
	    }

	    set slave [lindex $args 0]
            if {[info exists ::two::SetStyleCmds($slave)]} {
                unset ::two::SetStyleCmds($slave)
            }

            set idx [lsearch -exact $Split($m,slaves) $slave]
            if {$idx >= 0} {
                set Split($m,slaves) [lreplace $Split($m,slaves) $idx $idx]
                place forget $slave
            }
	    
	    if {[llength $Split($m,slaves)] < 2} {
	        catch {
                    set h $Split($m,handle)
                    destroy ${h}_knob
                    destroy $h
                }
	    }
	    if {[llength $Split($m,slaves)] == 0} {
                foreach key [array names Split "$m,*"] {
                    unset Split($key)
		}
            } else {
                Config $m
                set slave [lindex $Split($m,slaves) 0]
                set ::two::SetStyleCmds($slave) [list two_split configure $slave -style]
	    }
	}
	configure {
	    if {[llength $args] < 1} {
	        return -code error "wrong # args: should be \"two_split configure path ?options?\""
	    }
	    set w  [lindex $args 0]
	    set m  [winfo parent $w]
	    if {[llength $args] == 1} {
	        return [list \
		    [list -anchor $Split($m,anchor)] \
		    [list -style  $Split($m,style)]  \
		    [list -opaque $Split($m,opaque)] \
		    [list -split  $Split($m,split)]  \
		]
	    }
	    set args [lrange $args 1 end]

	    set oldAnchor $Split($m,anchor)
	    set oldStyle  $Split($m,style)
	    two_parseArgs args Split($m,anchor) "-anchor"
	    two_parseArgs args Split($m,style)  "-style"
	    two_parseArgs args Split($m,opaque) "-opaque"
	    two_parseArgs args Split($m,split)  "-split"

	    if {[llength $args]} {
	        return -code error "unknown option \"$args\""
	    }

	    if {([string compare $oldAnchor $Split($m,anchor)] != 0 ||
	         [string compare $oldStyle  $Split($m,style)]  != 0)
	        && [llength $Split($m,slaves)] == 2
	    } {
	        AddHandle $m
	    }

	    Config $m
            if {[llength $Split($m,slaves)] == 2} {
                PlaceHandle $m
            }
	}
	cget {
	    if {[llength $args] != 2} {
	        return -code error "wrong # args: should be \"two_split cget option\""
	    }
	    set w [lindex $args 0]
	    set m [winfo parent $w]
	    set opt [lindex $args 1]
	    switch -exact -- $opt {
	        -anchor  { return $Split($m,anchor) }
	        -style   { return $Split($m,style)  }
	        -opaque  { return $Split($m,opaque) }
		-split   { return $Split($m,split)  }
		default  { return -code error "unknown option \"$opt\"" }
	    }
	}
	slaves {
	    set m  [lindex $args 0]
	    return [list $Split($m,slave1) $Split($m,slave2)]
	}
	hide {
	    if {[llength $args] != 2} {
	        return -code error "wrong # args: should be \"two_split hide widget\""
	    }
	    error TODO
	}
	default {
            return -code error "bad option \"$opt\": must be cget, configure, forget, slaves or a widget path"
	}
    }
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoSplit-class bindings
#-------------------------------------------------------------------------

# Button 1: record state
bind TwoSplit <1> {
    ::two::split::SplitB1 %W %x %y
}

# B1-Release: drag finished
bind TwoSplit <B1-ButtonRelease> {
    ::two::split::SplitB1Release %W
}

# B1-Motion: resize slaves and/or drag handle
bind TwoSplit <B1-Motion> {
    ::two::split::SplitB1Motion %W %X %Y
}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::split::SplitB1 --
#
#	handle button press within split frame.
#
# Arguments:
#	h		Path splitter
#	x		relative X Position of mouse cursor within split (%x)
#	y		relative Y Position (%y)

proc ::two::split::SplitB1 {h x y} {
    variable Split

    set m [winfo parent $h]

    switch -exact -- $Split($m,anchor) {
        e - w {
            set Split(off) $x
            set Split(pos) [winfo rootx $m]
        }
        n - s {
            set Split(off) $y
            set Split(pos) [winfo rooty $m]
        }
    }
    set Split(drag)   1
    set Split(mod)    0
    set Split(m)      $m
    if {![string match "unix" $Split($m,style)]} {
	set Split(bg) [$Split($m,handle) cget -bg]
	$Split($m,handle) configure -bg black
    }
    return
}

# ::two::split::SplitB1Motion --
#
#	handle mouse motion while button pressed within split frame.
#
# Arguments:
#	h		Path split
#	absX		relative X Position of mouse cursor within split (%x)
#	absY		relative Y Position (%y)

proc ::two::split::SplitB1Motion {h absX absY} {
    variable Split

    if {!$Split(drag)} {
        return
    }

    set m $Split(m)
    set Split(mod) 1

    switch -exact -- $Split($m,anchor) {
        w {
            set max [winfo width $m]
            set splitpos [expr {$absX - $Split(pos) - $Split(off)} ]
        }
        e {
            set max [winfo width $m]
            set splitpos [expr {$max - ($absX - $Split(pos) - $Split(off)) - 4}]
        }
        n {
            set max [winfo height $m]
            set splitpos [expr {$absY - $Split(pos) - $Split(off)}]
        }
        s {
            set max [winfo height $m]
            set splitpos [expr {$max - ($absY - $Split(pos) - $Split(off)) - 4}]
        }
    }

    incr max -14
    if {$splitpos > $max} {
        set splitpos $max
    }
    if {$splitpos < 10} {
        set splitpos 10
    }
    set Split($m,split) $splitpos

    PlaceHandle $m
    if {$Split($m,opaque)} {
        PlaceSlaves $m
    }
    return
}

# ::two::split::SplitB1Release --
#
#	handle button release within split frame.
#
# Arguments:
#	h		Path split

proc ::two::split::SplitB1Release {h} {
    variable Split

    set m $Split(m)
    set Split(drag) 0

    if {![string match "unix" $Split($m,style)]} {
	$Split($m,handle) configure -bg $Split(bg)
    }
    if {!$Split(mod)} {
        return
    }
    if {!$Split($m,opaque)} {
        PlaceSlaves $m
    }
    return
}

# ::two::split::Config --
#
#	recalculate settings of panned frame
#
# Arguments:
#	m		Path to master widget

proc ::two::split::Config {m} {
    variable Split

    set anchor $Split($m,anchor)

    switch -exact -- [llength $Split($m,slaves)] {
        1 {
            place [lindex $Split($m,slaves) 0] \
                -x 0 -relx 0.0 -width  0 -relwidth  1.0 \
                -y 0 -rely 0.0 -height 0 -relheight 1.0 \
                -anchor nw
        }
        2 {
            set h $Split($m,handle)
            foreach {s1 s2} $Split($m,slaves) {break}
            # basiacally clear all options not set by PlaceSlaves
            # and set those which remain fixed
            switch -exact -- $anchor {
                w {
                    place $s1 \
                        -x 0 -relx 0.0 -width  0 -relwidth  0.0 \
                        -y 0 -rely 0.0 -height 0 -relheight 1.0 \
                        -anchor nw
                    place $s2 \
                        -x 0 -relx 0.0 -width  0 -relwidth  1.0 \
                        -y 0 -rely 0.0 -height 0 -relheight 1.0 \
                        -anchor nw
                }
                e {
                    place $s1 \
                        -x 0 -relx 1.0 -width  0 -relwidth  0.0 \
                        -y 0 -rely 0.0 -height 0 -relheight 1.0 \
                        -anchor ne
                    place $s2 \
                        -x 0 -relx 1.0 -width  0 -relwidth  1.0 \
                        -y 0 -rely 0.0 -height 0 -relheight 1.0 \
                        -anchor ne
                }
                n {
                    place $s1 \
                        -y 0 -rely 0.0 -height 0 -relheight 0.0 \
                        -x 0 -relx 0.0 -width  0 -relwidth  1.0 \
                        -anchor nw
                    place $s2 \
                        -y 0 -rely 0.0 -height 0 -relheight 1.0 \
                        -x 0 -relx 0.0 -width  0 -relwidth  1.0 \
                        -anchor nw
                }
                s {
                    place $s1 \
                        -y 0 -rely 1.0 -height 0 -relheight 0.0 \
                        -x 0 -relx 0.0 -width  0 -relwidth  1.0 \
                        -anchor sw
                    place $s2 \
                        -y 0 -rely 1.0 -height 0 -relheight 1.0 \
                        -x 0 -relx 0.0 -width  0 -relwidth 1.0 \
                        -anchor sw
                }
            }
            PlaceSlaves $m
        }
    }

    return
}

# ::two::split::CheckConfig --
#
#	Check if (after resizing the master) there is still enough
#	room for the split handle to be visible. If not, adopt
#	the split position.
#
# Arguments:
#	m		Path to master widget
#	w		Width of master widget
#	h		Height of master widget

proc ::two::split::CheckConfig {m w h} {
    variable Split

    if {![info exists Split($m,slaves)] || [llength $Split($m,slaves)] != 2} {
        bind $m <Config> ""
        return
    }

    switch $Split($m,anchor) {
        w - e {
            if {$Split($m,split) > $w - 14} {
                set Split($m,split) [expr {$w - 14 > 10 ? $w - 14 : 10}]
                PlaceHandle $m
                PlaceSlaves $m
            }
        }
        n - s {
            if {$Split($m,split) > $h - 14} {
                set Split($m,split) [expr {$h - 14 > 10 ? $h - 14 : 10}]
                PlaceHandle $m
                PlaceSlaves $m
            }
        }
    }
}

# ::two::split::AddHandle --
#
#	Create a new handle
#
# Arguments:
#	m		Path to master widget

proc ::two::split::AddHandle {m} {
    variable Split

    if {[string match "." $m]} {
        set h ".__TwoSplitH"
    } else {
        set h "$m.__TwoSplitH"
    }
    set Split($m,handle) $h
    
    set anchor $Split($m,anchor)
    
    catch {
        destroy ${h}_knob
        destroy $h
    }

    switch -exact -- $Split($m,anchor) {
        w - e {
            set cursor "sb_h_double_arrow"
        }
        n - s {
            set cursor "sb_v_double_arrow"
        }
    }

    if {[string match "unix" $Split($m,style)]} {
        frame ${h}      -class TwoSplitSep -bd 1 -relief sunken
        frame ${h}_knob -class TwoSplitKnob -cursor $cursor
        raise ${h}_knob
        bindtags ${h}_knob [list $h TwoSplit [winfo toplevel $m] all]
        switch -exact -- $Split($m,anchor) {
            w {
                place ${h}_knob -width 8 -height 8 -x -4 -y -12 -rely 1.0 -in $h -anchor nw
                place $h -width 2 -relheight 1.0 -anchor nw
            }
            e {
                place ${h}_knob -width 8 -height 8 -x 4 -y -12 -rely 1.0 -in $h -anchor ne
                place $h -width 2 -relheight 1.0 -relx 1.0 -anchor ne
            }
            n {
                place ${h}_knob -width 8 -height 8 -x -12 -relx 1.0 -y -4 -in $h -anchor nw
                place $h -height 2 -relwidth 1.0 -anchor nw
            }
            s {
                place ${h}_knob -width 8 -height 8 -x -12 -relx 1.0 -y 4 -in $h -anchor sw
                place $h -height 2 -relwidth 1.0 -rely 1.0 -anchor sw
            }
        }
    } else {
        frame $h -class TwoSplit -cursor $cursor
        switch -exact -- $Split($m,anchor) {
            w {
                place $h -width 4 -relheight 1.0 -anchor nw
            }
            e {
                place $h -width 4 -relheight 1.0 -relx 1.0 -anchor ne
            }
            n {
                place $h -height 4 -relwidth 1.0 -anchor nw
            }
            s {
                place $h -height 4 -relwidth 1.0 -rely 1.0 -anchor sw
            }
        }
    }
    PlaceHandle $m
}

# ::two::split::PlaceHandle --
#
#	Place the handle according to the current split location
#
# Arguments:
#	m		Path to master widget

proc ::two::split::PlaceHandle {m} {
    variable Split

    set h     $Split($m,handle)
    set delta $Split($m,split)

    if {[string match "unix" $Split($m,style)]} {
        incr delta
    }

    switch -exact -- $Split($m,anchor) {
        w {
            place configure $h -x $delta
        }
        e {
            place configure $h -x -$delta
        }
        n {
            place configure $h -y $delta
        }
        s {
            place configure $h -y -$delta
        }
    }
}

# ::two::split::PlaceSlaves --
#
#	Place the slaves according to the current split location
#
# Arguments:
#	m		Path to master widget


proc ::two::split::PlaceSlaves {m} {
    variable Split

    foreach {s1 s2} $Split($m,slaves) {break}

    set splitpos $Split($m,split)
    switch -exact -- $Split($m,anchor) {
        w {
            place configure $s1 -width $splitpos
            incr splitpos 4
            place configure $s2 -x $splitpos -width -$splitpos
        }
        e {
            place configure $s1 -width $splitpos
            incr splitpos 4
            place configure $s2 -x -$splitpos -width -$splitpos
        }
        n {
            place configure $s1 -height $splitpos
            incr splitpos 4
            place configure $s2 -y $splitpos -height -$splitpos
        }
        s {
            place configure $s1 -height $splitpos
            incr splitpos 4
            place configure $s2 -y -$splitpos -height -$splitpos
        }
    }
}

namespace import ::two::split::*
