# splitter.tcl --
#
#	Provides an "interactive geometry manager". Slave widgets are
#	arranged in rows or columns, with the user being able to modify
#	the size of each such pane.
#
# Copyright (c) 1998, 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/splitter.tcl,v 1.23 1999/10/26 13:12:28 fjl Exp $

package provide twoSplitter 1.0
package require twoUtil

#option add *Splitter.background                grey51
#option add *Splitter.background                black
option add *TwoSplitterHandle.borderWidth         0
option add *TwoSplitterHandle.relief              raised
option add *TwoSplitterHandle.borderWidth         1

namespace eval ::two::splitter {
    namespace export two_splitter
    variable Splitter
    
    namespace import ::two::util::two_parseArgs
    
    # variable Splitter
    #
    # $w,s,style -	style of slave (unix|windows)
    # $w,s,dyn -	dynamic resizing on/off
    # $w,s,h -	path to handle for this slave
    # $w,s,fract -	current fraction
    # $m,m,orient -	orientation (horizontal or vertical)
    # $m,m,slaves -	list of slaves
    # $m,m,len -	number of slaves
    # $h -		slave associated with this handle
    # style -	default style
    # width -	default width
    # orient -	default orientation
    # cursor -	cursor for non-dragable area on unix-style handle
    # drag -	currently dragging any splitter
    # mod -		the currently dragged splitter has been modified
    
    array set Splitter "
        style	$tcl_platform(platform)
        dynamic	1
        orient	vertical
        cursor	left_ptr
        drag	0
        mod		0
    "
}

# ::two::splitter::two_splitter --
#
#	The splitter is implemented as a layout manager; it doesn't subclass
#	any existing widget.
#
# Arguments:
#	opt		Widget command...
#	args		optional args

proc ::two::splitter::two_splitter {opt args} {
    variable Splitter

    switch -glob -- $opt {
	.* {
	    set orient  ""
	    set before  ""
	    set after   ""
	    set style   $Splitter(style)
	    set dynamic $Splitter(dynamic)
	    two_parseArgs args orient  "-orient"
	    two_parseArgs args style   "-style"
	    two_parseArgs args dynamic "-dynamic"
	    two_parseArgs args after   "-after"
	    two_parseArgs args before  "-before"
	    set m [winfo parent $opt]
	    set new [linsert $args 0 $opt]

	    # settings for master
	    if {[string length $orient]} {
	        set Splitter($m,m,orient) $orient
	    } elseif {![info exists Splitter($m,m,orient)]} {
	        set Splitter($m,m,orient) $Splitter(orient)
	    }

	    # New splitter?
	    if {![info exists Splitter($m,m,slaves)]} {
	        set Splitter($m,m,slaves) ""
	    } else {
	        # any of those widgets already managed by splitter?
		foreach w $new {
		    if {[info exists Splitter($w,s,fract)]} {
		        two_splitter forget $w
		    }
		}
	    }

	    # look for position where to insert new widgets
	    set idx [llength $Splitter($m,m,slaves)]
	    if {[string length $before]} {
	        if {![winfo exists $before]} {
		    return -code error "bad window path name \"$before\""
		}
	        set idx [lsearch -exact $Splitter($m,m,slaves) $before]
		if {$idx < 0} {
		    return -code error "window \"$before\" isn't managed by splitter"
		}
	    }
	    if {[string length $after]} {
	        if {![winfo exists $after]} {
		    return -code error "bad window path name \"$after\""
		}
	        set idx [lsearch -exact $Splitter($m,m,slaves) $after]
		if {$idx >= 0} {
		    incr idx
		} else {
		    return -code error "window \"$after\" isn't managed by splitter"
		}
	    }

            # are there already some slaves?
	    if {[llength $Splitter($m,m,slaves)]} {
	        set len [llength [linsert $args 0 $opt]]
	        set mul [expr {(double($Splitter($m,m,len)) / double($Splitter($m,m,len) + $len))}]
	        foreach w [lrange $Splitter($m,m,slaves) 0 [expr {$idx - 1}]] {
		    # reduce fraction
		    set Splitter($w,s,fract) [expr {$Splitter($w,s,fract) * $mul}]
		}
	        foreach w [lrange $Splitter($m,m,slaves) $idx [expr {$Splitter($m,m,len) - 2}]] {
		    # increase fraction
		    set Splitter($w,s,fract) [expr {1 - (1-$Splitter($w,s,fract))*$mul}]
		}
	    }
	    
            # process all widgets on command line
	    foreach w $new {
	        # set widget config
		set Splitter($w,s,style)  $style
		set Splitter($w,s,dyn)    $dynamic
		set Splitter($m,m,slaves) [linsert $Splitter($m,m,slaves) $idx $w]
		incr idx
		bind $w <Destroy> "+two_splitter forget $w"
		
                # This is a multi-style widget
                set ::two::SetStyleCmds($w) [list two_splitter configure $w -style]
	    }
	    # re-set master
	    Config $m
	}
	configure {
	    if {[llength $args] < 1} {
	        return -code error "wrong # args: should be \"two_splitter configure path ?options?\""
	    }
	    set w    [lindex $args 0]
	    set m    [winfo parent $w]
	    if {[llength $args] == 1} {
	        return [list \
		    [list -orient  $Splitter($m,m,orient)]  \
		    [list -style   $Splitter($w,s,style)]   \
		    [list -dynamic $Splitter($w,s,dyn)] \
		    [list -fract   $Splitter($w,s,fract)]   \
		]
	    }
	    set args [lrange $args 1 end]
	    set or   [two_parseArgs args Splitter($m,m,orient) "-orient"]
	    set st   [two_parseArgs args Splitter($w,s,style)  "-style"]
	    two_parseArgs args Splitter($w,s,dyn) "-dynamic"
	    two_parseArgs args Splitter($w,s,fract)   "-fract"
	    if {[llength $args]} {
	        return -code error "unknown option \"$args\""
	    }
	    if {$or} {
	        foreach slave [lrange $Splitter($m,m,slaves) 0 [expr $Splitter($m,m,len) - 2]] {
	            if {[info exists Splitter($slave,s,h)]} {
		        unset Splitter([lindex $Splitter($slave,s,h) 0])
	        	foreach h $Splitter($slave,s,h) {
			    destroy $h
			}
		    }
		}
	    } elseif {$st} {
	        if {[info exists Splitter($w,s,h)]} {
	            unset Splitter([lindex $Splitter($w,s,h) 0])
	            foreach h $Splitter($w,s,h) {
			destroy $h
		    }
		}
	    }
	    Config $m
	}
	cget {
	    if {[llength $args] != 2} {
	        return -code error "wrong # args: should be \"two_splitter cget option\""
	    }
	    set w   [lindex $args 0]
	    set m   [winfo  parent $w]
	    set opt [lindex $args 1]
	    switch -exact -- $opt {
	        -orient  { return $Splitter($m,m,orient)  }
	        -style   { return $Splitter($w,s,style)   }
	        -dynamic { return $Splitter($w,s,dyn) }
		-fract   { return $Splitter($w,s,fract)   }
		default  { return -code error "unknown option \"$opt\"" }
	    }
	}
	slaves {
	    set m    [lindex $args 0]
	    set args [lrange $args 1 end]
	    return $Splitter($m,m,slaves)
	}
	forget {
	    foreach w $args {
                unset ::two::SetStyleCmds($w)
	        place forget $w
	
		if {[regsub "two_splitter forget $w" [bind $w <Destroy>] "" newbinding]} {
		    bind $w <Destroy> $newbinding
		}
	        set m [winfo parent $w]
		# remove all handles (if any remain)
	        foreach slave [lrange $Splitter($m,m,slaves) 0 [expr $Splitter($m,m,len) - 2]] {
	            if {[info exists Splitter($slave,s,h)]} {
		        unset Splitter([lindex $Splitter($slave,s,h) 0])
			foreach h $Splitter($slave,s,h) {
		            destroy $h
			}
			unset Splitter($slave,s,h)
		    }
		}
		# remove slave from master's list of slaves
		set pos [lsearch -exact $Splitter($m,m,slaves) $w]
                set Splitter($m,m,slaves) [lreplace $Splitter($m,m,slaves) $pos $pos]
		# correct fractions of remaining slaves
	        set len [llength $Splitter($m,m,slaves)]
		if {$len == 1} {
		    set slave [lindex $Splitter($m,m,slaves) 0]
		    set Splitter($slave,s,fract) 1.0
		} elseif {$len>0 && $pos >= 0} {
		    set prev [lindex $Splitter($m,m,slaves) [expr $pos - 1]]
		    set mul [expr {(double($len) + 1) / double($len) }]
		    foreach slave [lrange $Splitter($m,m,slaves) 0 [expr {$pos - 1}]] {
			set Splitter($slave,s,fract) [expr {$Splitter($slave,s,fract) * $mul}]
		    }
		    foreach slave [lrange $Splitter($m,m,slaves) $pos [expr {$Splitter($m,m,len) - 1}]] {
			set Splitter($slave,s,fract) [expr {1 - ((1- $Splitter($slave,s,fract)) * $mul)}]
		    }
		}
		# remove state of slave
		foreach key [array names Splitter "$w,s,*"] {
        	    unset Splitter($key)
		}
		# last slave?
		if {$len == 0} {
		    foreach key [array names Splitter "$m,m,*"] {
        		unset Splitter($key)
		    }
		}
		# re-set master
		if {[info exists Splitter($m,m,slaves)]} {
		    Config $m
		}
	    }
	}
	default {
            return -code error "bad option \"$opt\": must be cget, configure, forget, slaves or a widget path"
	}
    }
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoSplitter-class bindings
#-------------------------------------------------------------------------

# Button 1: record state
bind TwoSplitter <1> {
    ::two::splitter::SplitterB1 %W %x %y
}

# B1-Release: drag finished
bind TwoSplitter <B1-ButtonRelease> {
    ::two::splitter::SplitterB1Release %W
}

# B1-Motion: resize slaves and/or drag handle
bind TwoSplitter <B1-Motion> {
    ::two::splitter::SplitterB1Motion %W %X %Y
}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::splitter::SplitterB1 --
#
#	handle button press within splitter frame.
#
# Arguments:
#	h		Path splitter
#	x		relative X Position of mouse cursor within splitter (%x)
#	y		relative Y Position (%y)

proc ::two::splitter::SplitterB1 {h x y} {
    variable Splitter

    set w $Splitter($h)
    set m [winfo parent $w]
    set Splitter(drag) 1
    set Splitter(mod)  0
    set Splitter(w)    $w
    set Splitter(m)    $m
    set Splitter(hor)  [string match "$Splitter($m,m,orient)*" "horizontal"]
    set Splitter(size) [expr {$Splitter(hor) ? [winfo width $m] : [winfo height $m]}]
    set Splitter(min)  [expr {int($Splitter($w,s,min) * $Splitter(size) + 10)}]
    set Splitter(max)  [expr {int($Splitter($w,s,max) * $Splitter(size) - 10)}]
    set Splitter(off)  [expr {$Splitter(hor) ? $x - 2 : $y - 2}]
    set Splitter(pos)  [expr {$Splitter(hor) ? [winfo rootx $m] : [winfo rooty $m]}]
    if {![string match "unix" $Splitter($w,s,style)]} {
	set Splitter(bg) [$h cget -bg]
	$h configure -bg black
    }
    return
}

# ::two::splitter::SplitterB1Motion --
#
#	handle mouse motion while button pressed within splitter frame.
#
# Arguments:
#	h		Path splitter
#	absX		relative X Position of mouse cursor within splitter (%x)
#	absY		relative Y Position (%y)

proc ::two::splitter::SplitterB1Motion {h absX absY} {
    variable Splitter

    if {!$Splitter(drag)} {
        return
    }

    set Splitter(mod) 1
    set w $Splitter(w)
    set m $Splitter(m)
    if {$Splitter(hor)} {
        set x [expr {$absX - $Splitter(pos) - $Splitter(off)}]
	if {$x < $Splitter(min)} {
	    set Splitter(x) $Splitter(min)
	} elseif {$x > $Splitter(max)} {
	    set Splitter(x) $Splitter(max)
	} else {
	    set Splitter(x) $x
	}
        if {!$Splitter($w,s,dyn)} {
	    if {[string match "unix" $Splitter($w,s,style)]} {
		place $h -x [expr {$Splitter(x) - 2}] -relx 0.0
        	place ${h}_sep -x [expr {$Splitter(x) + 1}] -relx 0.0
	    } else {
		place $h -x $Splitter(x) -relx 0.0
	    }
	}
    } else {
        set y [expr {$absY - $Splitter(pos) - $Splitter(off)}]
	if {$y < $Splitter(min)} {
	    set Splitter(y) $Splitter(min)
	} elseif {$y > $Splitter(max)} {
	    set Splitter(y) $Splitter(max)
	} else {
	    set Splitter(y) $y
	}
        if {!$Splitter($w,s,dyn)} {
	    if {[string match "unix" $Splitter($w,s,style)]} {
		place %W -y [expr {$Splitter(y) - 2}] -rely 0.0
        	place %W_sep -y [expr {$Splitter(y) + 1}] -rely 0.0
	    } else {
		place %W -y $Splitter(y) -rely 0.0
	    }
	}
    }
    if {$Splitter($w,s,dyn)} {
        set w $Splitter(w)
	if {$Splitter(hor)} {
	    set Splitter($w,s,fract) [expr {double($Splitter(x)) / double([winfo width $m])}]
	} else {
	    set Splitter($w,s,fract) [expr {double($Splitter(y)) / double([winfo height $m])}]
	}
        # TODO: only place the 2 neighbour slaves
	Config $m
    }
    return
}

# ::two::splitter::SplitterB1Release --
#
#	handle button release within splitter frame.
#
# Arguments:
#	h		Path splitter

proc ::two::splitter::SplitterB1Release {h} {
    variable Splitter

    set w $Splitter(w)
    set m $Splitter(m)
    set Splitter(drag) 0
    if {![string match "unix" $Splitter($w,s,style)]} {
	$h configure -bg $Splitter(bg)
    }
    if {!$Splitter(mod)} {
        return
    }
    if {$Splitter(hor)} {
        if {[string match "unix" $Splitter($w,s,style)] && !$Splitter($w,s,dyn)} {
	    incr Splitter(x) 2
	}
        set Splitter($w,s,fract) [expr {double($Splitter(x)) / double([winfo width $m])}]
    } else {
        if {[string match "unix" $Splitter($w,s,style)] && !$Splitter($w,s,dyn)} {
	    incr Splitter(y) 2
	}
        set Splitter($w,s,fract) [expr {double($Splitter(y)) / double([winfo height $m])}]
    }
    # TODO: only place the 2 neighbour slaves
    Config $m
    return
}

# ::two::splitter::Config --
#
#	recalculate settings of panned frame
#
# Arguments:
#	m		Path to master widget

proc ::two::splitter::Config {m} {
    variable Splitter

    set len [llength $Splitter($m,m,slaves)]
    set hor [string match "$Splitter($m,m,orient)*" "horizontal"]
    set idx 0
    set fract [expr 1.0 / $len]
    set Splitter($m,m,len) $len
    set oldfract 0.0
    foreach w [lrange $Splitter($m,m,slaves) 0 [expr $len - 1]] {
        # calculate required fraction...
        if {![info exists Splitter($w,s,fract)]} {
	    set Splitter($w,s,fract) $fract
	} else {
	    set fract $Splitter($w,s,fract)
	}
	if {$idx == $len - 1} {
	    set fract 1.0
	}
	# remember index
	set Splitter($w,s,idx) $idx
	# build/configure handle
	if {$idx != 0} {
	    if {[string match "." $m]} {
	        set h ".__h$idx"
	    } else {
	        set h "$m.__h$idx"
	    }
	    # splitter has wrong type?
	    if {[string match "unix" $Splitter($oldw,s,style)]} {
	        if {![winfo exists ${h}_sep] && [winfo exists $h]} {
	            destroy $h
		}
	    } else {
	        if {[winfo exists ${h}_sep]} {
	            destroy $h
		}
	    } 
	    if {![winfo exists $h]} {
		set cursor [expr {$hor ? "sb_h_double_arrow" : "sb_v_double_arrow"}]

                if {[string match "unix" $Splitter($oldw,s,style)]} {
	            frame $h -class TwoSplitterHandle -cursor $cursor
		    frame ${h}_sep -class TwoSplitterSep -bd 1 -relief sunken
		    raise $h
		    bindtags $h [list $h TwoSplitter [winfo toplevel $m] all]
		    set Splitter($oldw,s,h) [list $h ${h}_sep]
		} else {
	            frame $h -class TwoSplitter -cursor $cursor
		    set Splitter($oldw,s,h) [list $h]
		}
	    }
	    set Splitter($h) $oldw
	}
        # offset for pane sizes depending on style and position
	if {$idx == $len - 1} {
            # last pane
	    set off1 0
	    set off2 0
	    set off3 -1
	    set off4 -2
	} else {
	    set off1 -1 ;# first pane in unix style
	    set off2 -2 ;# first pane in windows style
	    set off3 -2 ;# subsequent pane in unix style
	    set off4 -4 ;# subsequent pane in windows style
	}
	# place children and splitter handles
	# o.k., this is lengthy code with a lot of cut-and-paste:-(
	if {$hor} {
	    if {$idx == 0} {
	        # the first pane
        	if {[string match "unix" $Splitter($w,s,style)]} {
	            place $w -rely 0.0 -relheight 1.0 -relx $oldfract -relwidth [expr {$fract - $oldfract}] -x 0 -width $off1
		} else {
	            place $w -rely 0.0 -relheight 1.0 -relx $oldfract -relwidth [expr {$fract - $oldfract}] -x 0 -width $off2
		}
	    } else {
	        # subsequent panes
        	if {[string match "unix" $Splitter($oldw,s,style)]} {
	            place $w -rely 0.0 -relheight 1.0 -relx $oldfract -relwidth [expr {$fract - $oldfract}] -x 1 -width $off3
		    place $h -y -12 -rely 1.0 -height 8 -relx $oldfract -x -4 -width 8
		    place ${h}_sep -relheight 1.0 -relx $oldfract -x -1 -width 2 -height 4
		} else {
	            place $w -rely 0.0 -relheight 1.0 -relx $oldfract -relwidth [expr {$fract - $oldfract}] -x 2 -width $off4
		    place $h -rely 0.0 -relheight 1.0 -relx $oldfract -x -2 -width 4
		}
	    }
	} else {
	    if {$idx == 0} {
	        # the first pane
        	if {[string match "unix" $Splitter($w,s,style)]} {
	            place $w -relx 0.0 -relwidth 1.0 -rely $oldfract -relheight [expr {$fract - $oldfract}] -y 0 -height $off1
		} else {
	            place $w -relx 0.0 -relwidth 1.0 -rely $oldfract -relheight [expr {$fract - $oldfract}] -y 0 -height $off2
		}
	    } else {
	        # subsequent panes
        	if {[string match "unix" $Splitter($oldw,s,style)]} {
	            place $w -relx 0.0 -relwidth 1.0 -rely $oldfract -relheight [expr {$fract - $oldfract}] -y 1 -height $off3
		    place $h -x -12 -relx 1.0 -width 8 -rely $oldfract -y -4 -height 8
		    place ${h}_sep -relwidth 1.0 -rely $oldfract -y -1 -height 2 -width 4
		} else {
	            place $w -relx 0.0 -relwidth 1.0 -rely $oldfract -relheight [expr {$fract - $oldfract}] -y 2 -height $off4
		    place $h -relx 0.0 -relwidth 1.0 -rely $oldfract -y -2 -height 4
		}
	    }
	}
	# set limits
	if {$len == 1} {
	    set Splitter($w,s,min) 0.0
	    set Splitter($w,s,max) 1.0
	} elseif {$idx == 0} {
            set Splitter($w,s,min) 0.0
	} elseif {$idx == $Splitter($m,m,len) - 1} {
	    set Splitter($w,s,min) $oldfract
            set Splitter($oldw,s,max) 1.0
	} else {
	    set Splitter($w,s,min) $oldfract
            set Splitter($oldw,s,max) $fract
	}
	# remember for next loop
	set oldw     $w
	set oldfract $fract
	
	# calculate fract for next pane (all subsequent panes are equally sized by default)
	if {$idx == $len - 1} {
	    set fract 1.0
	} else {
	    set fract [expr {$fract + (1 - $fract) / ($len - $idx - 1)}]
	}

	incr idx
    }
    return
}

namespace import ::two::splitter::*
