# text.tcl --
#
#	Modified text widget bindings. This widget behaves more like
#	windows text widgets. No emacs bindings are used to be able
#	to use those keys for the application's shortcuts. Has unlimited
#	undo and redo capability.
#
# Copyright (c) 1998, 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# parts of this code are based on the tk4.2 library file "text.tcl"
# SCCS: @(#) text.tcl 1.46 96/08/23 14:07:32
# Copyright (c) 1992-1994 The Regents of the University of California.
# Copyright (c) 1994-1996 Sun Microsystems, Inc.
#
# $Header: /home/ldv/cvs/two/lib/text.tcl,v 1.70 2000/01/03 16:01:27 fjl Exp $

package provide twoText 1.0
package require twoUtil
package require twoFramework

namespace eval ::two::text {
    namespace export two_text
    variable Tooltip
    
    # variable Text
    #
    # $w,cursor -	Original cursor while dragging with control-B2
    # $w,undoSp -	Pointer to the undo stack
    # $w,undo0Sp -	Pointer to the stack frame where the document is
    #			considered unmodified
    # $w,u_a<n> -	Undo command (d = delete, i = insert, r = replace)
    # $w,u_b<n> -	First undo argument (start)
    # $w,u_c<n> -	Second undo argument (end / text)
    # $w,r_?<n> -	Dito, but redo command; ? = a,b,c
    # $w,p_? -   	   Dito, but previous command and *no* count (no stack here);
    #			? = a,b,c
    # $w,redoSp -	Pointer to the redo stack
    # $w,mod -		True if editor has been modified
    # $w,nappend -	Don't append next key to undo stack (needed after save)
    # $w,ai -		If 1 then any Return will indent the new line just like
    #			the previous one
    # $w,istep -	Number of chars for indenting
    # $w,istring -	String consisting of $w,istep number of spaces
    # $w,xsel -		Export selection setting; we don't use tk's exportselection directly as
    #			we might have the same selection in several cloned texts!
    # $w,clones -	List of Paths to all cloned widgets, including the current widget
    # $w,prevPos -	Used when moving up or down lines via the keyboard.
    #			Keeps track of the previous insert position, so
    #			we can distinguish a series of ups and downs, all
    #			in a row, from a new up or down.
    # $w,prevCol -	Same as above, but only the column
    # word -		Characters allowed in a word (used by control-left/right)
    # afterId -		If non-null, it means that auto-scanning is underway
    #			and it gives the "after" id for the next auto-scan
    #			command to be executed.
    # char -		Character position on the line;  kept in order
    #			to allow moving up or down past short lines while
    #			still remembering the desired position.
    # mouseMoved -	Non-zero means the mouse has moved a significant
    #			amount since the button went down (so, for example,
    #			start dragging out a selection).
    # pressX - 		Horizontal position of last ButtonPress-1
    # selectMode -	The style of selection currently underway:
    #			char, word, or line.
    # x, y -		Last known mouse coordinates for scanning and auto-scanning.
    
    array set Text {
        word		{[_a-zA-Z0-9]}
        nonword		{[^_a-zA-Z0-9]}
        afterId		{}
        pressX		0
        char		0
        mouseMoved	0
        selectMode	char
    }
}

# Notice:
# -------
# In many places, the original widget proc "::two::text::_two$w" is
# used instead of the replacement "$w". Often, this is just for performance
# reasons (no need to go through another level of command parsing).
# In the two_text widget, there is another reason for doing this: the
# widget may have clones, and using the $w interface will usually change
# the settings in all of the clones. This is undesiable in many places
# in internal widget code and thus is worked around through directly
# acessing the original widget.


# ::two::text::two_text --
#
#	Create widget, parse arguments and install our own widget procedure.
#	To increase speed, "_two$w" is called directly (instead of "$w") whenever
#	it is known that only native options are used.
#
# Arguments:
#	w -		Widget path
#	args -		Widget arguments and options

two_widget ::two::text::two_text Text text {
    {-ai         -autoindent}
    {-istep      -indentstep}
    {-exportsel  -exportselection}
    {-xsel       -exportselection}
    {-autoindent
        autoIndent
	AutoIndent
	0
	{}
	{UpdateClones $w -autoindent}
    }
    {-overwrite
        overWrite
	OverWrite
	0
	{SetOverwrite $w; ShowOverwrite $w}
	{UpdateClones $w -overwrite}
    }
    {-indentstep
        indentStep
	IndentStep
	4
	{SetIndent $w}
	{UpdateClones $w -indentstep istring}
    }
    {-exportselection
        exportSelection
	ExportSelection
	1
	{}
	{UpdateClones $w -exportselection}
    }
    {-hardtab
        hardTab
	HardTab
	0
	{}
	{UpdateClones $w -hardtab}
    }
    {unknown Configure}
} {
    # indent step defaults to 4
    if {![info exists Text($w,istring)]} {
        set Text($w,istring) "    "
    }

    # Initialise the "globals"
    array set Text "
        $w,undoSp	0
	$w,undo0Sp	0
	$w,redoSp	0
	$w,p_a		{}
	$w,p_b		{}
	$w,p_c		{}
	$w,mod		0
	$w,nappend	0
	$w,clones	$w
        $w,prevPos	1.0
        $w,prevCol	0
    "

    # use our own bindings
    bindtags $w [list $w TwoText [winfo toplevel $w] all]
    
    # set the anchor mark
    $w mark set anchor insert

    # introduce the insert tag (distinct from the insert mark!)
    # used to highlight the character which is being overwritten
    # in overwrite mode
    $w tag configure insert -relief solid -borderwidth 1

    # Install our own selection handle
    proc _two$w,handle {offset length} "
        return \[[namespace code Selection] $w \$offset \$length\]
    "
    selection handle $w [namespace code _two$w,handle]
} {
    afterLoad {
	foreach clone $Text($w,clones) {
	    UndoReset $clone
	    RedoReset $clone
            set Text($clone,mod) 0
	    event generate $clone <<ModOff>>
	    event generate $clone <<SelOff>>
            event generate $clone <<UndoOff>>
	}
    }
    afterSave {
	foreach clone $Text($w,clones) {
	    set Text($clone,undo0Sp) $Text($clone,undoSp)
	    set Text($clone,nappend) 1
            set Text($clone,mod) 0
	    event generate $clone <<ModOff>>
	    event generate $clone <<SelOff>>
	}
    }
    clone {
	if {[llength $args] != 1} {
	    return -code error "wrong # args: should be \"$w clone newTextName\""
	}
	set clone [lindex $args 0]
	lappend Text($w,clones) $clone

	# create clone widget
	uplevel [list two_text $clone]

	# copy config
	foreach conf [_two$w configure] {
	    if {[llength $conf] == 5} {
		if {[string compare [lindex $conf 3] [lindex $conf 4]] != 0} {
		    if {![string match *scrollcommand [lindex $conf 0]]} {
		        _two$clone configure [lindex $conf 0] [lindex $conf 4]
		    }
		}
	    }
	}

	# copy state
	foreach key [array names Text "$w,*"] {
	    regsub "^$w,(.*)\$" $key "$clone,\\1" cloneKey
            set Text($cloneKey) $Text($key)
	}

	# copy contents
	_two$clone delete 1.0 end
	_two$clone insert 1.0 [_two$w get 1.0 end-1c]

	# copy marks
	eval [list _two$clone mark unset] [_two$clone mark names]
	foreach mark [_two$w mark names] {
	    _two$clone mark set     $mark [_two$w index        $mark]
	    _two$clone mark gravity $mark [_two$w mark gravity $mark]
	}

	# copy tags
	eval [list _two$clone tag delete] [_two$clone tag names]
	foreach tag [_two$w tag names] {
	    foreach cfg [_two$w tag configure $tag] {
	        # only copy the default value:
	        # $cfg consists of 2 or 5 elements
	        # 2 -> a synonym: argvName, dbName(synonym name)
	        # 5 -> argvName, dbName, dbClass, defValue, currentValue
	        if {[string length [lindex $cfg 4]]} {
		    _two$clone tag configure $tag [lindex $cfg 0] [lindex $cfg 4]
		}
	    }
	    if {[llength [_two$w tag ranges $tag]]} {
		eval [list _two$clone tag add $tag] [_two$w tag ranges $tag]
	    }
	}
	return $clone
    }
    delete {
	set oldPos [_two$w index insert]
	set oldSel [llength [_two$w tag ranges sel]]
	set oldSp  $Text($w,undoSp)
	set oldSp0 $Text($w,undo0Sp)
	set i1 [_two$w index [lindex $args 0]]
	if {[llength $args] < 2} {
	    set i2 $i1
	} else {
	    set i2 [_two$w index [lindex $args 1]]
	}
	foreach clone $Text($w,clones) {
	    set retval [Delete $clone $i1 $i2]
	    if {$oldSp == $oldSp0 && $Text($clone,undoSp) != $oldSp} {
        	set Text($clone,mod) 1
        	event generate $clone <<ModOn>>
	    }
	    if {!$oldSp && $Text($clone,undoSp)} {
        	event generate $clone <<UndoOn>>
	    }
	    if {$oldSel && ![llength [_two$clone tag ranges sel]]} {
                event generate $clone <<SelOff>>
	    }
	}
	if {![string match $oldPos [_two$w index insert]]} {
	    SendPos $w
	} elseif {$Text($w,-overwrite)} {
            ShowOverwrite $w
        }

	return $retval
    }
    goto {
	set oldPos [_two$w index insert]
	set retval [$w mark set insert [lindex $args 0]]
        if {[llength [_two$w tag ranges sel]]} {
            $w tag remove sel 1.0 end
	}
	$w see insert
        if {![string match $oldPos [_two$w index insert]]} {
            SendPos $w
        } elseif {$Text($w,-overwrite)} {
            ShowOverwrite $w
        }
	return $retval
    }
    indent {
	# get numerical representation of insert
	set pos [_two$w index insert]
	set oldPos $pos
	set oldSel [llength [_two$w tag ranges sel]]
	set oldSp  $Text($w,undoSp)
	set oldSp0 $Text($w,undo0Sp)
	foreach clone $Text($w,clones) {
	    Indent $clone $pos [lindex $args 0]
	    if {$oldSp == $oldSp0 && $Text($clone,undoSp) != $oldSp} {
        	set Text($clone,mod) 1
        	event generate $clone <<ModOn>>
	    }
	    if {!$oldSp && $Text($clone,undoSp)} {
        	event generate $clone <<UndoOn>>
	    }
	    if {$oldSel && ![llength [_two$clone tag ranges sel]]} {
                event generate $clone <<SelOff>>
	    }
	}
	if {![string match $oldPos [_two$w index insert]]} {
	    SendPos $w
	} elseif {$Text($w,-overwrite)} {
            ShowOverwrite $w
        }
    }
    insert {
	set oldPos [_two$w index insert]
	set oldSp  $Text($w,undoSp)
	set oldSp0 $Text($w,undo0Sp)
	# get numerical representation of position
	set i [_two$w index [lindex $args 0]]
	set s [lrange $args 1 end]
	if {![string length [lindex $s 0]] && [llength $s] == 1} {
            return
	}
	set i [_two$w index $i]
	foreach clone $Text($w,clones) {
	    set retval [eval [list Insert $clone $i] $s]
	    if {$oldSp == $oldSp0 && $Text($clone,undoSp) != $oldSp} {
        	set Text($clone,mod) 1
        	event generate $clone <<ModOn>>
	    }
	    if {!$oldSp && $Text($clone,undoSp)} {
        	event generate $clone <<UndoOn>>
	    }
	}
	if {![string match $oldPos [_two$w index insert]]} {
	    SendPos $w
	} elseif {$Text($w,-overwrite)} {
            ShowOverwrite $w
        }
	return $retval
    }
    insertReplace {
        # anything selected? if yes: trying to insert text *within* selection
	# will replace the whole selection
        if {[llength [_two$w tag ranges sel]] &&
	    [string length [lindex $args 1]] &&
    	    [_two$w compare [lindex $args 0] >= sel.first] &&
	    [_two$w compare [lindex $args 0] <= sel.last]
	} {
	    return [eval [list $w replace sel.first sel.last] [lrange $args 1 end]]
	}
	# nothing selected...
	if {$Text($w,-overwrite) && [$w compare insert != "insert lineend"]} {
	    # we are overwriting and NOT at the end of a line
	    # get number of characters that are to be inserted
	    set s ""
	    foreach {txt tag} [lrange $args 1 end] {
	        append s $txt
	    }
	    set len [string length [string trimright $s "\n"]]
	    return [eval [list $w replace insert "insert + $len c"] [lrange $args 1 end]]
	}
	# nothing special: just insert the text
	return [eval [list $w insert] $args]
    }
    mark {
	set oldPos [_two$w index insert]
	# only some of the mark subcommands apply to the other clones
	foreach clone $Text($w,clones) {
            if {[string compare $clone $w] != 0} {
		switch -exact -- [lindex $args 0] {
		    set {
		        if {![string match insert [lindex $args 1]]} {
                            _two$clone mark set [lindex $args 1] \
                                [_two$w index [lindex $args 2]]
			}
		    }
		    gravity -
		    unset {
		        eval [list _two$clone mark] $args
		    }
		}
	    }
	}
	set retval [eval [list _two$w mark] $args]
	if {[string match "set"    [lindex $args 0]] &&
	    [string match "insert" [lindex $args 1]]
	} {
	    if {![string match $oldPos [_two$w index insert]]} {
		SendPos $w
	    } elseif {$Text($w,-overwrite)} {
                ShowOverwrite $w
            }
	}
	return $retval
    }
    modState {
        return $Text($w,mod)
    }
    paste {
        # what text to paste?
        if {[llength $args] > 1} {
            # given on command line
            set txt [join [lrange $args 1 end] ""]
        } else {
            # none given, so take primary selection
            set txt [selection get -displayof $w]
        }
        # anything selected? if yes: trying to paste text *within* the selection
	# will replace the whole selection
        if {[llength [_two$w tag ranges sel]] &&
	    [string length [lindex $args 1]] &&
    	    [_two$w compare [lindex $args 0] >= sel.first] &&
	    [_two$w compare [lindex $args 0] <= sel.last]
	} {
	    return [$w replace sel.first sel.last $txt]
	}
	# nothing selected...
	return [$w insert [lindex $args 0] $txt]
    }
    redo {
        if {!$Text($w,redoSp)} {
            return 0
        }
	set oldPos [_two$w index insert]
	set oldSp  $Text($w,undoSp)
	set oldSel [llength [_two$w tag ranges sel]]
	foreach clone $Text($w,clones) {
	    Redo $clone
	    if {$Text($clone,undoSp) == $Text($clone,undo0Sp)} {
        	set Text($clone,mod) 0
		event generate $clone <<ModOff>>
	    } elseif {$Text($clone,undoSp) == $Text($clone,undo0Sp) + 1} {
        	set Text($clone,mod) 1
		event generate $clone <<ModOn>>
	    }
	    if {!$Text($clone,redoSp)} {
		event generate $clone <<RedoOff>>
	    }
	    if {!$oldSp} {
		event generate $clone <<UndoOn>>
	    }
	    if {$oldSel && ![llength [_two$clone tag ranges sel]]} {
                event generate $clone <<SelOff>>
	    }
	}
        # show cursor
        _two$w see insert

	if {![string match $oldPos [_two$w index insert]]} {
	    SendPos $w
	} elseif {$Text($w,-overwrite)} {
            ShowOverwrite $w
        }
    }
    redoState {
        if {$Text($w,redoSp) > 0} {
	    return 1
	} else {
	    return 0
	}
    }
    replace {
	set oldPos [_two$w index insert]
	set oldSel [llength [_two$w tag ranges sel]]
	set oldSp  $Text($w,undoSp)
	set oldSp0 $Text($w,undo0Sp)
	set i1 [_two$w index [lindex $args 0]]
	if {[llength $args] < 2} {
	    set i2 $i1
	} else {
	    set i2 [_two$w index [lindex $args 1]]
	}
	set s [lrange $args 2 end]
	foreach clone $Text($w,clones) {
	    set retval [eval [list Replace $clone $i1 $i2] $s]
	    if {$oldSp == $oldSp0 && $Text($clone,undoSp) != $oldSp} {
        	set Text($clone,mod) 1
        	event generate $clone <<ModOn>>
	    }
	    if {!$oldSp && $Text($clone,undoSp)} {
        	event generate $clone <<UndoOn>>
	    }
	    if {$oldSel && ![llength [_two$clone tag ranges sel]]} {
                event generate $clone <<SelOff>>
	    }
	}
	if {![string match $oldPos [_two$w index insert]]} {
	    SendPos $w
	} elseif {$Text($w,-overwrite)} {
            ShowOverwrite $w
        }
	return $retval
    }
    see {
        set pos [lindex $args 0]
        # make sure the line isn't the bottom line...
        _two$w see "$pos + 1 line"
        _two$w see $pos
    }
    selState {
	if {[llength [_two$w tag ranges sel]]} {
	    return 1
	} else {
	    return 0
	}
    }
    tag {
        set sel1 [llength [_two$w tag ranges sel]]
	foreach clone $Text($w,clones) {
            # only some of the tag subcommands apply to the other clones
            if {[string compare $clone $w] != 0} {
		switch -exact -- [lindex $args 0] {
		    add {
		        set newargs [lrange $args 0 1]
		        foreach idx [lrange $args 2 end] {
		            lappend newargs [_two$w index $idx]
		        }
                        eval [list _two$clone tag] $newargs
		    }
		    remove -
		    bind -
		    configure -
		    lower -
		    raise -
		    delete {
			eval [list _two$clone tag] $args
		    }
		}
	    }
	}
	# but all apply to the primary widget
	set retval [eval [list _two$w tag] $args]
        set sel2 [llength [_two$w tag ranges sel]]
	foreach clone $Text($w,clones) {
	    if {$sel1 && !$sel2} {
                event generate $clone <<SelOff>>
	    } elseif {!$sel1 && $sel2} {
                event generate $clone <<SelOn>>
	    }
	}
	if {$sel2 && $Text($w,-exportselection)} {
	    selection own $w
	}
	return $retval
    }
    undo {
	if {!$Text($w,undoSp)} {
	    return 0
	}
	set oldPos [_two$w index insert]
	set oldSel [llength [_two$w tag ranges sel]]
	set oldSp  $Text($w,redoSp)
	foreach clone $Text($w,clones) {
	    Undo $clone
	    if {$Text($clone,undoSp) == $Text($clone,undo0Sp)} {
        	set Text($clone,mod) 0
		event generate $clone <<ModOff>>
	    } elseif {$Text($clone,undoSp) == $Text($clone,undo0Sp) - 1} {
        	set Text($clone,mod) 1
		event generate $clone <<ModOn>>
	    }
	    if {!$Text($clone,undoSp)} {
		event generate $clone <<UndoOff>>
	    }
	    if {!$oldSp} {
		event generate $clone <<RedoOn>>
	    }
	    if {$oldSel && ![llength [_two$clone tag ranges sel]]} {
                event generate $clone <<SelOff>>
	    }
	}
        # show cursor
        _two$w see insert

	if {![string match $oldPos [_two$w index insert]]} {
	    SendPos $w
	} elseif {$Text($w,-overwrite)} {
            ShowOverwrite $w
        }
    }
    undoState {
	if {$Text($w,undoSp) > 0} {
	    return 1
	} else {
	    return 0
	}
    }
} {
    # remove reference from cloned widgets
    set pos [lsearch -exact $Text($w,clones) $w]
    set others [lreplace $Text($w,clones) $pos $pos]
    foreach other $others {
        set Text($other,clones) $others
    }
}

#-------------------------------------------------------------------------
# The code below creates the TwoText-class bindings
#-------------------------------------------------------------------------

bind TwoText <FocusIn> {
    ::two::text::SendPos %W
    ::two::text::SetOverwrite %W
}
bind TwoText <1> {
    set ::two::text::Text(selectMode)  char
    set ::two::text::Text(mouseMoved)  0
    set ::two::text::Text(pressX)      %x

    ::two::text::SetCursor %W [::two::text::ClosestGap %W %x %y]
    %W mark set anchor [::two::text::_two%W index insert]

    if {[string match "normal" [::two::text::_two%W cget -state]]} {
        focus %W
    }
    if {[llength [::two::text::_two%W tag ranges sel]]} {
        %W tag remove sel 1.0 end
    }
}
bind TwoText <B1-Motion> {
    set ::two::text::Text(x) %x
    set ::two::text::Text(y) %y
    ::two::text::SelectTo %W %x %y
}
bind TwoText <Double-1> {
    set ::two::text::Text(selectMode) word
    ::two::text::SelectTo %W %x %y
}
bind TwoText <Triple-1> {
    set ::two::text::Text(selectMode) line
    ::two::text::SelectTo %W %x %y
}
bind TwoText <Shift-1> {
    ::two::text::SelectTo %W %x %y
}
bind TwoText <Double-Shift-1> {
    ::two::text::SelectTo %W %x %y
}
bind TwoText <Triple-Shift-1> {
    ::two::text::SelectTo %W %x %y
}
bind TwoText <B1-Leave> {
    set ::two::text::Text(x) %x
    set ::two::text::Text(y) %y
    ::two::text::AutoScan %W
}
bind TwoText <B1-Enter> {
    ::two::text::CancelRepeat
}
bind TwoText <ButtonRelease-1> {
    ::two::text::CancelRepeat
}
bind TwoText <2> {
    if {!$tk_strictMotif} {
        ::two::text::_two%W scan mark %x %y
        set ::two::text::Text(x) %x
        set ::two::text::Text(y) %y
        set ::two::text::Text(mouseMoved) 0
    }
}
bind TwoText <B2-Motion> {
    if {!$tk_strictMotif} {
        if {(%x != $::two::text::Text(x)) || (%y != $::two::text::Text(y))} {
            set ::two::text::Text(mouseMoved) 1
        }
    }
}
bind TwoText <ButtonRelease-2> {
    if {[string match "normal" [::two::text::_two%W cget -state]]} {
        focus %W
    }
    if {!$::two::text::Text(mouseMoved) || $tk_strictMotif} {
        set ::two::text::Text(%W,nappend) 1
        ::two::text::_two%W mark set insert [::two::text::ClosestGap %W %x %y]
        %W mark set anchor [::two::text::_two%W index insert]
        ::two::text::SendPos %W
	if {[llength [::two::text::_two%W tag ranges sel]]} {
	    %W insert insert \
	        [::two::text::_two%W get sel.first sel.last]
	} else {
	    catch {%W insert insert \
	        [selection get -displayof %W]
            }
	}
    }
}
bind TwoText <Control-2> {
    if {!$tk_strictMotif} {
        set ::two::text::Text(%W,cursor) [%W cget -cursor]
        %W configure -cursor fleur
        ::two::text::_two%W scan mark %x %y
        set ::two::text::Text(x) %x
        set ::two::text::Text(y) %y
        set ::two::text::Text(mouseMoved) 0
    }
}
bind TwoText <Control-B2-Motion> {
    if {!$tk_strictMotif} {
        if {(%x != $::two::text::Text(x)) || (%y != $::two::text::Text(y))} {
            set ::two::text::Text(mouseMoved) 1
        }
        if {$::two::text::Text(mouseMoved)} {
            ::two::text::_two%W scan dragto %x %y
        }
    }
}
bind TwoText <Control-ButtonRelease-2> {
    if {!$tk_strictMotif} {
        %W configure -cursor $::two::text::Text(%W,cursor)
    }
}
bind TwoText <Left> {
    ::two::text::SetCursor %W insert-1c
}
bind TwoText <Right> {
    ::two::text::SetCursor %W insert+1c
}
bind TwoText <Up> {
    ::two::text::SetCursor %W [::two::text::UpDownLine %W -1]
}
bind TwoText <Down> {
    ::two::text::SetCursor %W [::two::text::UpDownLine %W 1]
}
bind TwoText <Shift-Left> {
    ::two::text::KeySelect %W [::two::text::_two%W index {insert - 1c}]
}
bind TwoText <Shift-Right> {
    ::two::text::KeySelect %W [::two::text::_two%W index {insert + 1c}]
}
bind TwoText <Shift-Up> {
    ::two::text::KeySelect %W [::two::text::UpDownLine %W -1]
}
bind TwoText <Shift-Down> {
    ::two::text::KeySelect %W [::two::text::UpDownLine %W 1]
}
bind TwoText <Control-Left> {
    ::two::text::SetCursor %W [::two::text::WordLeftStart %W insert]
}
bind TwoText <Control-Right> {
    ::two::text::SetCursor %W [::two::text::WordRightStart %W insert]
}
bind TwoText <Control-Up> {
    ::two::text::SetCursor %W [::two::text::PrevPara %W insert]
}
bind TwoText <Control-Down> {
    ::two::text::SetCursor %W [::two::text::NextPara %W insert]
}
bind TwoText <Shift-Control-Left> {
    if {
        ![llength [::two::text::_two%W tag ranges sel]] ||
        [::two::text::_two%W compare insert == sel.first]
    } {
        # no selection or cursor at start of selection, extending it to the left
        # -> expand sel to start of next word to the left
        ::two::text::KeySelect %W [::two::text::WordLeftStart %W insert]
    } elseif {[::two::text::_two%W compare insert == sel.last]} {
        # currently at end of selection; trying to shorten from the right
        # -> go to end of next word to the left
        set newpos [::two::text::WordLeftEnd %W insert]
        if {[::two::text::_two%W compare $newpos > sel.first]} {
            # o.k., not gone past the beginning
            ::two::text::KeySelect %W $newpos
        } else {
            # stop at start of selection
            ::two::text::KeySelect %W [::two::text::_two%W index sel.first]
        }
    } else {
        # sel present, but cursor is not at start or end of selection,
        # extending the selection to the left
        # -> go to start of current word and extend sel to start of next word to the left
        ::two::text::KeySelect %W [::two::text::WordLeftStart %W \
            [::two::text::WordStart %W insert]]
    }
}
bind TwoText <Shift-Control-Right> {
    if {
        ![llength [::two::text::_two%W tag ranges sel]] ||
        [::two::text::_two%W compare insert == sel.last]
    } {
        ::two::text::KeySelect %W [::two::text::WordRightEnd %W insert]
    } elseif {[::two::text::_two%W compare insert == sel.first]} {
        set newpos [::two::text::WordRightStart %W insert]
        if {[::two::text::_two%W compare $newpos < sel.last]} {
            ::two::text::KeySelect %W $newpos
        } else {
            ::two::text::KeySelect %W [::two::text::_two%W index sel.last]
        }
    } else {
        ::two::text::KeySelect %W [::two::text::WordRightEnd %W \
            [::two::text::WordEnd %W insert]]
    }
}
bind TwoText <Shift-Control-Up> {
    ::two::text::KeySelect %W [::two::text::PrevPara %W insert]
}
bind TwoText <Shift-Control-Down> {
    ::two::text::KeySelect %W [::two::text::NextPara %W insert]
}
bind TwoText <Prior> {
    ::two::text::SetCursor %W [::two::text::ScrollPages %W -1]
}
bind TwoText <Shift-Prior> {
    ::two::text::KeySelect %W [::two::text::ScrollPages %W -1]
}
bind TwoText <Next> {
    ::two::text::SetCursor %W [::two::text::ScrollPages %W 1]
}
bind TwoText <Shift-Next> {
    ::two::text::KeySelect %W [::two::text::ScrollPages %W 1]
}
bind TwoText <Control-Prior> {
    ::two::text::_two%W xview scroll -1 page
}
bind TwoText <Control-Next> {
    ::two::text::_two%W xview scroll 1 page
}

bind TwoText <Home> {
    ::two::text::SetCursor %W {insert linestart}
}
bind TwoText <Shift-Home> {
    ::two::text::KeySelect %W {insert linestart}
}
bind TwoText <End> {
    ::two::text::SetCursor %W {insert lineend}
}
bind TwoText <Shift-End> {
    ::two::text::KeySelect %W {insert lineend}
}
bind TwoText <Control-Home> {
    ::two::text::SetCursor %W 1.0
}
bind TwoText <Control-Shift-Home> {
    ::two::text::KeySelect %W 1.0
}
bind TwoText <Control-End> {
    ::two::text::SetCursor %W {end - 1 char}
}
bind TwoText <Control-Shift-End> {
    ::two::text::KeySelect %W {end - 1 char}
}
bind TwoText <Control-KeyPress-plus> {
    %W indent 1
}
bind TwoText <Control-KeyPress-greater> {
    %W indent 1
}
bind TwoText <Control-KeyPress-minus> {
    %W indent -1
}
bind TwoText <Control-KeyPress-less> {
    %W indent -1
}
bind TwoText <Tab> {
    if {[llength [::two::text::_two%W tag ranges sel]]} {
        %W indent 1
    } else {
	%W insertReplace insert \t
    }
    # don't change focus
    break
}
bind TwoText <Shift-Tab> {
    if {[llength [::two::text::_two%W tag ranges sel]]} {
        %W indent -1
    } else {
	%W insertReplace insert \t
    }
    # don't change focus
    break
}
bind TwoText <Control-Tab> {
    # Needed only to keep <Tab> binding from triggering;  doesn't
    # have to actually do anything.
}
bind TwoText <Shift-Control-Tab> {
    # Needed only to keep <Tab> binding from triggering;  doesn't
    # have to actually do anything.
}
bind TwoText <Return> {
    ::two::text::Newline %W
}
bind TwoText <Insert> {
    ::two::text::ToggleOverwrite %W
}
bind TwoText <Delete> {
    if {[llength [::two::text::_two%W tag ranges sel]] &&
        [::two::text::_two%W compare insert >= sel.first] &&
        [::two::text::_two%W compare insert <= sel.last]
    } {
	%W delete sel.first sel.last
    } elseif {[::two::text::_two%W compare insert != end-1c]} {
	%W delete insert insert+1c
    }
    ::two::text::_two%W see insert
}
bind TwoText <BackSpace> {
    if {[llength [::two::text::_two%W tag ranges sel]] &&
        [::two::text::_two%W compare insert >= sel.first] &&
        [::two::text::_two%W compare insert <= sel.last]
    } {
	%W delete sel.first sel.last
    } elseif {[::two::text::_two%W compare insert != 1.0]} {
	%W delete insert-1c insert
    }
    ::two::text::_two%W see insert
}
bind TwoText <Control-Delete> {
    if {[llength [::two::text::_two%W tag ranges sel]] &&
        [::two::text::_two%W compare insert >= sel.first] &&
        [::two::text::_two%W compare insert < sel.last]
    } {
	%W delete insert sel.last
    } elseif {[::two::text::_two%W compare insert != end-1c]} {
	%W delete insert "insert lineend"
    }
    ::two::text::_two%W see insert
}
bind TwoText <Control-BackSpace> {
    if {[llength [::two::text::_two%W tag ranges sel]] &&
        [::two::text::_two%W compare insert > sel.first] &&
        [::two::text::_two%W compare insert <= sel.last]
    } {
	%W delete sel.first insert
    } elseif {[::two::text::_two%W compare insert != 1.0]} {
	%W delete "insert linestart" insert
    }
    ::two::text::_two%W see insert
}
bind TwoText <KeyPress> {
    if {[string length %A]} {
	%W insertReplace insert %A
        if {[llength [::two::text::_two%W tag ranges sel]]} {
            %W tag remove sel 1.0 end
	}
        ::two::text::_two%W see insert
    }
}

# Ignore all Alt, Meta, and Control keypresses unless explicitly bound.
# Otherwise, if a widget binding for one of these is defined, the
# <KeyPress> class binding will also fire and insert the character,
# which is wrong.  Ditto for <Escape>.

bind TwoText <Alt-KeyPress>     { #nothing }
bind TwoText <Meta-KeyPress>    { #nothing }
bind TwoText <Control-KeyPress> { #nothing }
bind TwoText <Escape>           { #nothing }
bind TwoText <KP_Enter>         { #nothing }

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------
# Please note that calling the widget function (as returned by %W)
# will modify all cones as well and properly set undo/redo etc.,
# while using _two%W applies directly to the widget itself.
# This is useful for "mark set insert" only!

# ::two::text::SetIndent --
#
#	Set the string representation of the indentation step.
#	This is used when indenting a portion of text; it
#	consists of n whitespaces, where n ist the indent step.
#
# Arguments:
#	w		Path to widget

proc ::two::text::SetIndent {w} {
    variable Text

    set Text($w,istring) ""
    for {set x 0} {$x < $Text($w,-indentstep)} {incr x} {
	append Text($w,istring) " "
    }
}

# ::two::text::Newline --
#
#	Insert a newline character.
#
# Arguments:
#	w		Path to widget

proc ::two::text::Newline {w} {
    variable Text

    $w insertReplace insert \n

    if {$Text($w,-autoindent)} {
        set prevLine [_two$w index {insert - 1 lines linestart} ]
	set nrSpace 0
	_two$w search -count nrSpace -regexp "^( |\t)*" $prevLine insert
    	if {$nrSpace > 0} {
    	    $w insert insert [_two$w get $prevLine "$prevLine + $nrSpace chars"]
    	}
    }
    _two$w see insert
}


# ::two::text::UpdateClones --
#
#	Copy a specific setting to all clones
#
# Arguments:
#	w -		Path to widget
#	args -		Name of all options which should be updated all clones

proc ::two::text::UpdateClones {w args} {
    variable Text

    foreach opt $args {
	foreach clone $Text($w,clones) {
            set Text($clone,$opt) $Text($w,$opt)
	}
    }
}

# ::two::text::Configure --
#
#	Set configuration options on all clones. Only used for
#	default text-widget-Options. This is a bit hairy, as
#	the scrollbar settings should not be identical on
#	all clones!
#
# Arguments:
#	w -		Path to widget
#	args -		remaining options string....

proc ::two::text::Configure {w args} {
    variable Text

    # loop through all options, but break at "--"
    set delimit [lsearch -exact $args "--"]
    if {$delimit > 0} {
        incr delimit -1
        set argList [lrange $args 0 $delimit]
    } else {
        set argList $args
    }

    # remove scrollbar settings....
    set remaining ""
    foreach {arg value} $argList {
        if {[string match -xs* $arg] || [string match -ys* $arg]} {
	    set retval [_two$w configure $arg $value]
	} else {
	    lappend remaining $arg $value
	}
    }
    # the rest....
    if {$delimit > 0} {
        lappend remaining [lrange $args $delimit end]
    }

    if {[llength $remaining]} {
	if {[info exists Text($w,clones)]} {
            foreach clone $Text($w,clones) {
	        set retval [eval [list _two$clone configure] $remaining -exportselection 0]
	    }
	} else {
            set retval [eval [list _two$w configure] $remaining -exportselection 0]
	}
    }
    return $retval
}

# ::two::text::ClosestGap --
#
#	Given x and y coordinates, this procedure finds the closest boundary
#	between characters to the given coordinates and returns the index
#	of the character just after the boundary.
#
# Arguments:
#	w		The text window.
#	x		X-coordinate within the window.
#	y		Y-coordinate within the window.

proc ::two::text::ClosestGap {w x y} {
    set pos [_two$w index @$x,$y]
    set bbox [_two$w bbox $pos]
    if {![string compare $bbox ""]} {
	return $pos
    }
    if {($x - [lindex $bbox 0]) < ([lindex $bbox 2]/2)} {
	return $pos
    }
    return [_two$w index "$pos + 1 char"]
}

# ::two::text::SelectTo --
#
#	This procedure is invoked to extend the selection, typically when
#	dragging it with the mouse.  Depending on the selection mode (character,
#	word, line) it selects in different-sized units.  This procedure
#	ignores mouse motions initially until the mouse has moved from
#	one character to another or until there have been multiple clicks.
#
# Arguments:
#	w		The text window in which the button was pressed.
#	x		Mouse x position.
#	y 		Mouse y position.

proc ::two::text::SelectTo {w x y} {
    variable Text

    set cur [ClosestGap $w $x $y]

    if {![llength [_two$w tag ranges sel]]} {
        set anchor [_two$w index insert]
        $w mark set anchor $anchor
	if {[_two$w compare $cur < insert]} {
	    $w tag add sel $cur [_two$w index insert]
	} elseif {[_two$w compare $cur > insert]} {
	    $w tag add sel [_two$w index insert] $cur
	}
    } else {
        set anchor [_two$w index anchor]
    }

    if {[_two$w compare $cur != $anchor] || (abs($Text(pressX) - $x) >= 3)} {
	set Text(mouseMoved) 1
    }

    switch -exact -- $Text(selectMode) {
	char {
	    if {[_two$w compare $cur < anchor]} {
		set first $cur
		set last $anchor
	    } else {
		set first $anchor
		set last $cur
	    }
	}
	word {
	    if {[_two$w compare $cur < $anchor]} {
		set first [WordStart $w $cur]
		set last  [WordEnd   $w $anchor]
	    } else {
		set first [WordStart $w $anchor]
		set last  [WordEnd   $w $cur]
	    }
	}
	line {
	    if {[_two$w compare $cur < $anchor]} {
		set first [_two$w index "$cur linestart"]
		set last  [_two$w index "$anchor -1c lineend +1c"]
	    } else {
		set first [_two$w index "$anchor linestart"]
		set last  [_two$w index "$cur lineend +1c"]
	    }
	}
    }
    if {$Text(mouseMoved) || ![string match "char" $Text(selectMode)]} {
        if {[llength [_two$w tag ranges sel]]} {
            if {[_two$w compare $cur > $anchor]} {
                _two$w mark set insert $last
            } else {
                _two$w mark set insert $first
            }
            SendPos $w
        }
        if {[_two$w compare $first != $last]} {
            $w tag remove sel 1.0 $first
            $w tag add    sel $first $last
            $w tag remove sel $last end
	} else {
            $w tag remove sel 1.0 end
	}
    }
    return
}

# ::two::text::AutoScan --
#
#	This procedure is invoked when the mouse leaves a text window
#	with button 1 down.  It scrolls the window up, down, left, or right,
#	depending on where the mouse is (this information was saved in
#	Text(x) and Text(y)), and reschedules itself as an "after"
#	command so that the window continues to scroll until the mouse
#	moves back into the window or the mouse button is released.
#
# Arguments:
#	w		The text window.

proc ::two::text::AutoScan {w} {
    variable Text
    if {![winfo exists $w]} return
    if {$Text(y) >= [winfo height $w]} {
	_two$w yview scroll 2 units
    } elseif {$Text(y) < 0} {
	_two$w yview scroll -2 units
    } elseif {$Text(x) >= [winfo width $w]} {
	_two$w xview scroll 2 units
    } elseif {$Text(x) < 0} {
	_two$w xview scroll -2 units
    } else {
	return
    }
    SelectTo $w $Text(x) $Text(y)
    set Text(afterId) [after 50 [namespace code AutoScan] $w]
    return
}

# ::two::text::SetCursor --
#
#	Move the insertion cursor to a given position in a text.  Also
#	clears the selection, if there is one in the text, and makes sure
#	that the insertion cursor is visible.  Also, don't let the insertion
#	cursor appear on the dummy last line of the text.
#
# Arguments:
#	w		The text window.
#	pos		The desired new position for the cursor in the window.

proc ::two::text::SetCursor {w pos} {
    variable Text

    set Text(selectMode) char

    if {[_two$w compare $pos == end]} {
	set pos "end -1c"
    }
    _two$w mark set insert $pos
    _two$w see insert
    SendPos $w

    if {[llength [_two$w tag ranges sel]]} {
	$w tag remove sel 1.0 end
    }
    return
}

# ::two::text::KeySelect --
#
#	This procedure is invoked when stroking out selections using the
#	keyboard.  It moves the cursor to a new position, then extends
#	the selection to that position.
#
# Arguments:
#	w		The text window.
#	new		A new position for the insertion cursor (the cursor hasn't
#			actually been moved to this position yet).

proc ::two::text::KeySelect {w new} {
    variable Text

    if {![llength [_two$w tag ranges sel]]} {
	$w mark set anchor insert
    }
    # move anchor to same position as if selecting with the mouse
    # (those select modes only are started by double or triple clicking)
    switch -exact -- $Text(selectMode) {
        word {
            if {[_two$w compare $new < anchor]} {
                set anchor [WordEnd $w anchor]
            } else {
                set anchor [WordStart $w anchor]
            }
        }
        line {
            if {[_two$w compare $new < anchor]} {
                set anchor [_two$w index "anchor -1c lineend +1c"]
            } else {
                set anchor [_two$w index "anchor linestart"]
            }
        }
        default {
            set anchor anchor
        }
    }

    if {[_two$w compare $new < $anchor]} {
        set first $new
        set last  $anchor
    } else {
        set first $anchor
        set last  $new
    }
    $w tag remove sel 1.0    $first
    $w tag add    sel $first $last
    $w tag remove sel $last  end

    _two$w mark set insert $new
    _two$w see insert
    SendPos $w
    return
}

# ::two::text::Delete --
#
#	Delete a string from the text, ranging from start to end
#
# Arguments:
#	w		The text window in which to delete text
#	start		The first character to delete
#	end		The last character to delete

proc ::two::text::Delete {w start end} {
    variable Text

    # should we really delete the text?
    if {[string match "disabled" [_two$w cget -state]]} {
        return
    }

    # Workaround: selection in empty editor...aaarrgh :-(
    if {[string match "1.0" [_two$w index end-1c]]} {
        # delete selection
        _two$w delete 1.0 end
        return
    }

    # kill redo stack
    if {$Text($w,redoSp) != 0} {
        RedoReset $w
    }

    # Convert start and end into their numerical representation
    set start [_two$w index $start]
    if {[string match $start [_two$w index end]]} {
	set start [_two$w index "end - 1c"]
    }
    set end [_two$w index $end]
    if {[string match $end [_two$w index end]]} {
	set end [_two$w index "end - 1c"]
    }

    # if deleting single characters: only put single lines on undo stack
    if {[string match "\n" [_two$w get $start $end]]} {
        set Text($w,nappend) 1
    }

    # push reverse command on the undo stack; d=delete, i=insert
    if { [string match "i"  $Text($w,p_a)] &&
         [string match $end $Text($w,p_b)] &&
	 !$Text($w,nappend)
    } {
    	incr Text($w,undoSp) -1
	set Text($w,u_b$Text($w,undoSp)) $start
	set Text($w,u_c$Text($w,undoSp)) [_two$w get $start $end]$Text($w,u_c$Text($w,undoSp))
    	set Text($w,p_b) "$start"
    } else {
        set Text($w,nappend) 0
	set Text($w,u_a$Text($w,undoSp)) "i"
	set Text($w,u_b$Text($w,undoSp)) $start
	set Text($w,u_c$Text($w,undoSp)) [_two$w get $start $end]
    	set Text($w,p_a) "i"
    	set Text($w,p_b) $start
    }
    incr Text($w,undoSp)

    # Finally delete the text...
    _two$w delete $start $end
    return
}

# ::two::text::Insert --
#
#	Insert a string into a text at the point of the insertion cursor.
#	If there is a selection in the text, and it covers the point of the
#	insertion cursor, then delete it
#
# Arguments:
#	w		The text window in which to insert the string
#	pos		Numerical Position of where to insert the text
#	args		List of character - tag pairs. Usually just a single char

proc ::two::text::Insert {w pos args} {
    variable Text

    # should we really insert the string?
    if {[string match "disabled" [_two$w cget -state]]} {
        return
    }

    # kill redo stack
    if {$Text($w,redoSp)} {
        RedoReset $w
    }

    # Convert start into numerical representation, mark it,
    # insert the text and finally retrieve the new end.
    set start [_two$w index $pos]
    if {[string match $start [_two$w index end]]} {
	set start [_two$w index "end - 1c"]
    }
    _two$w mark set _twoPriv $start
    eval [list _two$w insert $start] $args

    set end [_two$w index _twoPriv]
    if {[string match $end [_two$w index end]]} {
	set end [_two$w index "end - 1c"]
    }
    _two$w mark unset _twoPriv

    # if inserting single characters: only put single lines on undo stack
    if {[lsearch -exact $args "\n"] > -1} {
        set Text($w,nappend) 1
    }

    # push reverse command on the undo stack; d=delete, i=insert, r=replace
    if {[string match "d"    $Text($w,p_a)] &&
        [string match $start $Text($w,p_c)] &&
	!$Text($w,nappend)
    } {
    	incr Text($w,undoSp) -1
    	set Text($w,u_b$Text($w,undoSp)) $Text($w,p_b)
    	set Text($w,u_c$Text($w,undoSp)) $end
    	set Text($w,p_c) "$end"
    } elseif {[string match "r" $Text($w,p_a)] &&
        [string match $start $Text($w,p_c)] &&
	!$Text($w,nappend)
    } {
    	incr Text($w,undoSp) -1
    	set Text($w,u_b$Text($w,undoSp)) [list $Text($w,p_b) $end]
    	set Text($w,p_c) $end
    } else {
        set Text($w,nappend) 0
	set Text($w,u_a$Text($w,undoSp)) "d"
	set Text($w,u_b$Text($w,undoSp)) $start
	set Text($w,u_c$Text($w,undoSp)) $end
    	set Text($w,p_a) "d"
    	set Text($w,p_b) $start
    	set Text($w,p_c) $end
    }
    incr Text($w,undoSp)
    return
}

# ::two::text::Replace --
#
#	Replace a portion of the text at the point of the insertion cursor.
#	If there is a selection in the text, and it covers the point of the
#	insertion cursor, then delete it
#
# Arguments:
#	w		The text window in which to insert the string
#	start		Index of first character to be replaced
#	end		Index of last character to be replaced
#	args		List of character - tag pairs. Usually just a single char

proc ::two::text::Replace {w start end args} {
    variable Text

    # should we really insert the string?
    if {[string match "disabled" [_two$w cget -state]]} {
        return
    }

    # kill redo stack
    if {$Text($w,redoSp)} {
        RedoReset $w
    }

    # Convert start into numerical representation, delete the text,
    # mark it the start position,
    # insert the text and finally retrieve the new end.
    set start [_two$w index $start]
    if {[string match $start [_two$w index end]]} {
	set start [_two$w index "end - 1c"]
    }
    set orig [_two$w get $start $end]
    _two$w delete $start $end
    _two$w mark set _twoPriv $start
    eval [list _two$w insert $start] $args
    set end [_two$w index _twoPriv]
    if {[string match $end [_two$w index end]]} {
	set end [_two$w index "end - 1c"]
    }
    _two$w mark unset _twoPriv

    # push reverse command on the undo stack; r=replace
    # Don't append anything here...
    set Text($w,nappend) 0
    set Text($w,u_a$Text($w,undoSp)) "r"
    set Text($w,u_b$Text($w,undoSp)) [list $start $end]
    set Text($w,u_c$Text($w,undoSp)) $orig
    set Text($w,p_a) "r"
    set Text($w,p_b) $start
    set Text($w,p_c) $end
    incr Text($w,undoSp)
    return
}

# ::two::text::UpDownLine --
#
#	Returns the index of the character one line above or below the
#	insertion cursor.  There are two or three tricky things here.  First,
#	we want to maintain the original column across repeated operations,
#	even though some lines that will get passed through don't have
#	enough characters to cover the original column.  Second, don't
#	try to scroll past the beginning or end of the text. And third,
#	if user selected to virtually expand hard-tabs, we have to calculate
#	the right column.
#
# Arguments:
#	w		The text window in which the cursor is to move.
#	n		The number of lines to move: -1 for up one line,
#			+1 for down one line.

proc ::two::text::UpDownLine {w n} {
    variable Text

    # Current Position
    set i [_two$w index insert]
    scan $i "%d.%d" line col

    # calculate new position
    set newLine [expr {$line + $n}]
    if {$Text($w,-hardtab)} {
        # We need to calculate the virtual position by expanding hard tabs
        # to 8 spaces (or less).
        
        # Cursor still where it was last time? -> use cached column
        if {[string match $i $Text($w,prevPos)]} {
            set col $Text($w,prevCol)
        } else {
            # get column with tabs expanded
            set col 0
            foreach char [split [_two$w get "insert linestart" insert] ""] {
                if {[string compare "\t" $char] == 0} {
                    incr col [expr {8 - ($col % 8)}]
                } else {
                    incr col
                }
            }
            # cache result
            set Text($w,prevCol) $col
        }

	# Now get real column within new line:
	set maxCol [expr {$col * 8}]
	set ncol 0 ;# new column number with tabs expanded
	set rcol 0 ;# real column number
	foreach char [split [_two$w get $newLine.0 $newLine.$maxCol] ""] {
	    if {[string compare "\t" $char] == 0} {
	        incr ncol [expr {8 - ($ncol % 8)}]
	    } else {
	        incr ncol
	    }
	    incr rcol
	    if {$ncol >= $col} break
	}
	set new [_two$w index [expr {$line + $n}].$rcol]
    } else {
        # Cursor still where it was last time? -> use cached column
        if {[string match $i $Text($w,prevPos)]} {
            set col $Text($w,prevCol)
        } else {
            set Text($w,prevCol) $col
        }
        # calculate new column
        set new [_two$w index $newLine.$col]
    }

    # did we go past the bounds? That case, we want to remember the
    # previous position (basically the row, the col ain't modified)
    if {[_two$w compare $new == end] || [_two$w compare $new == "insert linestart"]} {
	set new $i
    }
    set Text($w,prevPos) $new
    return $new
}

# ::two::text::PrevPara --
#
#	Returns the index of the beginning of the paragraph just before a given
#	position in the text (the beginning of a paragraph is the first non-blank
#	character after a blank line).
#
# Arguments:
#	w		The text window in which the cursor is to move.
#	pos		Position at which to start search.

proc ::two::text::PrevPara {w pos} {
    set pos [_two$w index "$pos linestart"]
    while {1} {
	if {([string match "\n" [_two$w get "$pos - 1 line"]] &&
	    ![string match "\n" [_two$w get $pos]]) ||
	    [string match "1.0" $pos]
        } {
	    if {[regexp -indices {^[ 	]+(.)} [_two$w get $pos "$pos lineend"] dummy index]} {
		set pos [_two$w index "$pos + [lindex $index 0] chars"]
	    }
	    if {[_two$w compare $pos != insert] || [string match "1.0" $pos]} {
		return $pos
	    }
	}
	set pos [_two$w index "$pos - 1 line"]
    }
    return
}

# ::two::text::NextPara --
#
#	Returns the index of the beginning of the paragraph just after a given
#	position in the text (the beginning of a paragraph is the first non-blank
#	character after a blank line).
#
# Arguments:
#	w		The text window in which the cursor is to move.
#	start		Position at which to start search.

proc ::two::text::NextPara {w start} {
    set pos [_two$w index "$start linestart + 1 line"]
    while {![string match "\n" [_two$w get $pos]]} {
	if {[_two$w compare $pos == end]} {
	    return [_two$w index "end - 1c"]
	}
	set pos [_two$w index "$pos + 1 line"]
    }
    while {[string match "\n" [_two$w get $pos]]} {
	set pos [_two$w index "$pos + 1 line"]
	if {[_two$w compare $pos == end]} {
	    return [_two$w index "end - 1c"]
	}
    }
    if {[regexp -indices {^[ 	]+(.)} [_two$w get $pos "$pos lineend"] dummy index]} {
	return [_two$w index "$pos + [lindex $index 0] chars"]
    }
    return $pos
}

# ::two::text::ScrollPages --
#
#	This is a utility procedure used in bindings for moving up and down
#	pages and possibly extending the selection along the way.  It scrolls
#	the view in the widget by the number of pages, and it returns the
#	index of the character that is at the same position in the new view
#	as the insertion cursor used to be in the old view.
#
# Arguments:
#	w		The text window in which the cursor is to move.
#	count		Number of pages forward to scroll;  may be negative
#			to scroll backwards.

proc ::two::text::ScrollPages {w count} {
    set bbox [_two$w bbox insert]
    set oldyview [_two$w yview]
    _two$w yview scroll $count pages
    if {![string length $bbox]} {
	return [_two$w index @[expr [winfo height $w]/2],0]
    }
    # did we really scroll?
    if {[string compare $oldyview [_two$w yview]]==0} {
        # no scrolling occured
        if {$count < 0} {
            return 1.0
        } else {
            return [_two$w index "end -1c"]
        }
    } else {
        return [_two$w index @[lindex $bbox 0],[lindex $bbox 1]]
    }
}

# ::two::text::CancelRepeat --
#
#	This procedure is invoked to cancel an auto-repeat action described
#	by Text(afterId).  It's used by several widgets to auto-scroll
#	the widget when the mouse is dragged out of the widget with a
#	button pressed.

proc ::two::text::CancelRepeat {} {
    variable Text
    after cancel $Text(afterId)
    set Text(afterId) {}
}

# ::two::text::Undo --
#
#	This procedure is used to undo the last text-modifying action
#
# Arguments:
#	w		Path to text widget

proc ::two::text::Undo {w} {
    variable Text

    # should we really undo changes?
    if {[string match "disabled" [_two$w cget -state]]} {
        return
    }

    if {!$Text($w,undoSp)} {
        return
    }

    incr Text($w,undoSp) -1
    set start $Text($w,u_b$Text($w,undoSp))

    # Do Undo and push reverse command on the redo stack
    switch $Text($w,u_a$Text($w,undoSp)) {
	i {
	    set text $Text($w,u_c$Text($w,undoSp))
	    _two$w mark set insert $start
	    _two$w insert $start $text
	    set end [_two$w index insert]

	    set Text($w,r_a$Text($w,redoSp)) "d"
	    set Text($w,r_b$Text($w,redoSp)) $start
	    set Text($w,r_c$Text($w,redoSp)) $end
	}
	d {
	    set end $Text($w,u_c$Text($w,undoSp))
	    set text [_two$w get $start $end]
	    _two$w mark set insert $start
	    _two$w delete $start $end

	    set Text($w,r_a$Text($w,redoSp)) "i"
	    set Text($w,r_b$Text($w,redoSp)) $start
	    set Text($w,r_c$Text($w,redoSp)) $text
	}
	r {
	    set start [lindex $Text($w,u_b$Text($w,undoSp)) 0]
	    set end   [lindex $Text($w,u_b$Text($w,undoSp)) 1]
	    set orig  [_two$w get $start $end]
	    _two$w delete $start $end
	    _two$w mark set insert $start
	    _two$w insert insert $Text($w,u_c$Text($w,undoSp))
	    set end [_two$w index insert]

	    set Text($w,r_a$Text($w,redoSp)) "r"
	    set Text($w,r_b$Text($w,redoSp)) [list $start $end]
	    set Text($w,r_c$Text($w,redoSp)) $orig
	}
    }

    # finish pop...
    unset Text($w,u_a$Text($w,undoSp))
    unset Text($w,u_b$Text($w,undoSp))
    unset Text($w,u_c$Text($w,undoSp))

    incr Text($w,redoSp)
    set Text($w,p_a) "u"
    return
}

# ::two::text::Redo --
#
#	This one is used to redo the last undo
#
# Arguments:
#	w		Path to text widget

proc ::two::text::Redo {w} {
    variable Text

    # should we really undo changes?
    if {[string match "disabled" [_two$w cget -state]]} {
        return
    }

    if {!$Text($w,redoSp)} {
        return
    }

    incr Text($w,redoSp) -1
    set start $Text($w,r_b$Text($w,redoSp))

    # Do Redo and push reverse command on the undo stack
    switch $Text($w,r_a$Text($w,redoSp)) {
	i {
	    set text $Text($w,r_c$Text($w,redoSp))
	    _two$w mark set insert $start
	    _two$w insert insert $text
	    set end [_two$w index insert]

	    set Text($w,u_a$Text($w,undoSp)) "d"
	    set Text($w,u_b$Text($w,undoSp)) $start
	    set Text($w,u_c$Text($w,undoSp)) $end
	}
	d {
	    set end $Text($w,r_c$Text($w,redoSp))
	    set text [_two$w get $start $end]
	    _two$w mark set insert $start
	    _two$w delete $start $end

	    set Text($w,u_a$Text($w,undoSp)) "i"
	    set Text($w,u_b$Text($w,undoSp)) $start
	    set Text($w,u_c$Text($w,undoSp)) $text
	}
	r {
	    set start [lindex $Text($w,r_b$Text($w,redoSp)) 0]
	    set end   [lindex $Text($w,r_b$Text($w,redoSp)) 1]
	    set orig  [_two$w get $start $end]
	    _two$w delete $start $end
	    _two$w mark set insert $start
	    _two$w insert insert $Text($w,r_c$Text($w,redoSp))
	    set end [_two$w index insert]

	    set Text($w,u_a$Text($w,undoSp)) "r"
	    set Text($w,u_b$Text($w,undoSp)) [list $start $end]
	    set Text($w,u_c$Text($w,undoSp)) $orig
	}
    }

    # finish pop...
    unset Text($w,r_a$Text($w,redoSp))
    unset Text($w,r_b$Text($w,redoSp))
    unset Text($w,r_c$Text($w,redoSp))

    incr Text($w,undoSp)
    set Text($w,p_a) "u"
    return
}

# ::two::text::UndoReset --
#
#	clears all the contents of the undo stack
#
# Arguments:
#	w		Path to text widget

proc ::two::text::UndoReset {w} {
    variable Text

    foreach entry [array names Text "$w,u_*"] {
        unset Text($entry)
    }
    set Text($w,undoSp) 0
    set Text($w,undo0Sp) 0
    set Text($w,nappend) 1
    return
}

# ::two::text::RedoReset --
#
#	clears all the contents of the redo stack
#
# Arguments:
#	w		Path to text widget

proc ::two::text::RedoReset {w} {
    variable Text

    foreach entry [array names Text "$w,r_*"] {
        unset Text($entry)
    }
    set Text($w,redoSp) 0
    event generate $w <<RedoOff>>
    return
}

# ::two::text::WordLeftStart --
#
#	Calculate position where to move the insertion cursor
#	if we want to go one word to the left. This is basically
#	the start of the word currently covering the insertion
#	cursor, or if there is none, the start of the next word
#	to the left.
#
# Arguments:
#	w		Path to text widget
#	idx		Current position
#
# Returns:
#	New position

proc ::two::text::WordLeftStart {w idx} {
    variable Text

    if {[_two$w compare $idx == "1.0"]} {
        return 1.0
    } elseif {
        [_two$w compare $idx == "$idx linestart"] ||
        [regexp $Text(word) [_two$w get $idx]]
    } {
        set idx "$idx -1c"
    }

    set cnt -1
    set newpos [_two$w search -count cnt -regexp -backwards -- \
        "(^|$Text(nonword))$Text(word)" $idx 1.0]

    if {$cnt == 1} {
        return $newpos
    } elseif {$cnt == 2} {
        return [_two$w index "$newpos +1c"]
    }
    return 1.0
}

# ::two::text::WordLeftEnd --
#
#	Calculate index of position immediatly after the end of the
#	next word to the left.
#
# Arguments:
#	w		Path to text widget
#	idx		Current position
#
# Returns:
#	New position

proc ::two::text::WordLeftEnd {w idx} {
    variable Text

    if {[_two$w compare $idx == "1.0"]} {
        return 1.0
    } elseif {
        [_two$w compare $idx == "$idx lineend"] ||
        [regexp $Text(word) [_two$w get "$idx -1c"]]
    } {
        set idx "$idx -1c"
    }

    set cnt -1
    set newpos [_two$w search -count cnt -regexp -backwards -- \
        "$Text(word)($Text(nonword)|\$)" $idx 1.0]

    if {$cnt > 0} {
        return [_two$w index "$newpos +1c"]
    }
    return 1.0
}

# ::two::text::WordRightStart --
#
#	Calculate position where to move the insertion cursor
#	if we want to go one word to the right. This is the start
#	of the next word to the right.
#
# Arguments:
#	w		Path to text widget
#	idx		Current position
#
# Returns:
#	New position

proc ::two::text::WordRightStart {w idx} {
    variable Text

    if {[_two$w compare $idx == "end -1c"]} {
        return [_two$w index "end -1c"]
    } elseif {[_two$w compare $idx == "$idx linestart"]} {
        set idx "$idx +1c"
    }

    set cnt -1
    set newpos [_two$w search -count cnt -regexp -- \
        "(^|$Text(nonword))$Text(word)" $idx "end -1c"]

    if {$cnt == 1} {
        return $newpos
    } elseif {$cnt == 2} {
        return [_two$w index "$newpos +1c"]
    }
    return [_two$w index "end -1c"]
}

# ::two::text::WordRightEnd --
#
#	Calculate index of position immediatly after the end of the
#	next word to the right.
#
# Arguments:
#	w		Path to text widget
#	idx		Current position
#
# Returns:
#	New position

proc ::two::text::WordRightEnd {w idx} {
    variable Text

    if {[_two$w compare $idx == "end -1c"]} {
        return [_two$w index "end -1c"]
    } elseif {[_two$w compare $idx == "$idx lineend"]} {
        set idx "$idx +1c"
    }

    set cnt -1
    set newpos [_two$w search -count cnt -regexp -- \
        "$Text(word)($Text(nonword)|\$)" $idx "end -1c"]

    if {$cnt > 0} {
        return [_two$w index "$newpos +1c"]
    }
    return [_two$w index "end -1c"]
}

# ::two::text::WordStart --
#
#	Calculate beginning of current word. This is used for selecting
#	whole words.
#
# Arguments:
#	w		Path to text widget
#	idx		Current position
#
# Returns:
#	New position

proc ::two::text::WordStart {w idx} {
    variable Text

    if {[_two$w compare $idx == 1.0]} {
        return 1.0
    } elseif {! [regexp $Text(word) [_two$w get "$idx -1c"]]} {
        return [_two$w index $idx]
    } elseif {[_two$w compare $idx == "$idx linestart"]} {
        return [_two$w index "$idx linestart"]
    }

    set cnt -1
    set newpos [_two$w search -count cnt -regexp -backwards -- \
        "(^|$Text(nonword))$Text(word)" $idx 1.0]

    if {$cnt == 1} {
        return $newpos
    } elseif {$cnt == 2} {
        return [_two$w index "$newpos +1c"]
    }
    return 1.0
}

# ::two::text::WordEnd --
#
#	Calculate end of current word. This is used for selecting
#	whole words.
#
# Arguments:
#	w		Path to text widget
#	idx		Current position
#
# Returns:
#	New position

proc ::two::text::WordEnd {w idx {next 0}} {
    variable Text

    if {[_two$w compare $idx == "end -1c"]} {
        return [_two$w index "end -1c"]
    } elseif {! [regexp $Text(word) [_two$w get $idx]]} {
        return [_two$w index $idx]
    } elseif {[_two$w compare $idx == "$idx lineend"]} {
        return [_two$w index "$idx lineend"]
    }

    set cnt -1
    set newpos [_two$w search -count cnt -regexp -- \
        "$Text(word)($Text(nonword)|\$)" "$idx -1c" "end -1c"]

    if {$cnt >= 0} {
        return [_two$w index "$newpos +1c"]
    }
    return [_two$w index "end -1c"]
}

# ::two::text::Indent --
#
#	Indent the current line (or lines if there is a multi-line selection)
#	or remove indentation, depending on $direction (value and sign).
#
# Arguments:
#	w		Path to text widget
#	pos		Index of position to indent
#	direction	Indent (>0) or remove indentation (<0)

proc ::two::text::Indent {w pos direction} {
    variable Text

    # Do we have a selection
    if {[llength [_two$w tag ranges sel]]} {
        # Store position of selection and cursor
        set first  [ _two$w index sel.first ]
        set last   [ _two$w index sel.last  ]
        set cursor [ _two$w index insert    ]
        # Retrieve line numbers
        scan [_two$w index "sel.first linestart"]-[_two$w index "sel.last-1c linestart"] "%d.0-%d.0" start end
        # Temporarily remove selection
        _two$w tag remove sel 1.0 end
        if { $direction > 0 } {
            # Indent all lines
            for {set line $start} {$line <= $end} {incr line} {
                TabToSpace $w $line.0
                Insert $w "$line.0" $Text($w,istring)
            }
        } else {
            # Remove a maximum of $count spaces from each line
            for {set line $start} {$line <= $end} {incr line} {
                TabToSpace $w $line.0
                set count 0
                _two$w search -count count -regexp \
                    "^ +" "$line.0" "$line.0 lineend"
                if {$count > $Text($w,-indentstep)} {
		    set count $Text($w,-indentstep)
		}
		if {$count > 0} {
                    Delete $w "$line.0" "$line.$count"
                }
            }
        }
	# restore selection and cursor position
	_two$w tag add sel $first $last
	_two$w mark set insert $cursor
    } else {
        TabToSpace $w $pos
        if {$direction > 0} {
            Insert $w "$pos linestart" $Text($w,istring)
        } else {
            set count 0
            _two$w search -count count -regexp \
                -- "^ +" "$pos linestart" "$pos lineend"
            if {$count > $Text($w,-indentstep)} {
	        set count $Text($w,-indentstep)
	    }
            if {$count > 0} {
                Delete $w "$pos linestart" "$pos linestart + $count chars"
            }
        }
    }
    return
}

# ::two::text::TabToSpace --
#
#	All tabs up to the first non-whitespace character are
#	replaced by spaces: used to make indentation increase/
#	decrease handle tabs in a proper way...
#
# Arguments:
#	w		Text Widget
#	pos		index if line which should be modified

proc ::two::text::TabToSpace {w pos} {
    # any tabs in here?
    set count 0
    _two$w search -count count -regexp \
        -- "^ *\t\[ \t\]*" "$pos linestart" "$pos lineend"
    if {$count > 0 } {
        # what string should we replace?
        set start [_two$w get "$pos linestart" "$pos linestart + $count c"]
        # loop through all tabs and replace them with the right
        # number of spaces, given a hardtab length of 8 chars.
        set tab 0
        while {[set first [string first \t $start]] >= 0} {
            incr tab
            set num   [expr {8 - ($first % 8)}]
            set rest  [string range $start [expr {$first + 1}] end]
            set start [string range $start 0 [expr {$first - 1}]]
            for {set idx 0} {$idx < $num} {incr idx} {
                append start " "
            }
            append start $rest
        }
        if {$tab > 0} {
            Replace $w "$pos linestart" "$pos linestart + $count c" $start
        }
    }
}


# ::two::text::ToggleOverwrite --
#
#	Toggle the overwriting state
#
# Arguments:
#	w		Widget

proc ::two::text::ToggleOverwrite {w} {
    variable Text

    if {$Text($w,-overwrite)} {
        set Text($w,-overwrite) 0
        event generate $w <<OverwriteOff>>
        _two$w tag remove insert 1.0 end
    } else {
        set Text($w,-overwrite) 1
        event generate $w <<OverwriteOn>>
        SendPos $w
    }
}

# ::two::text::SendPos --
#
#	Send a virtual event with x and y set to the current position of
#	the insertion cursor
#
# Arguments:
#	w		Widget

proc ::two::text::SendPos {w} {
    variable Text

    scan [_two$w index insert] "%d.%d" y x
    incr x
    event generate $w <<PosChanged>> -x $x -y $y
    if {$Text($w,-overwrite)} {
        ShowOverwrite $w
    }
    return
}

# ::two::text::ShowOverwrite
#
#	Show the overwrite cursor at the proper position
#
# Arguments:
#	w		Widget

proc ::two::text::ShowOverwrite {w} {
    variable Text

    _two$w tag remove insert 1.0 end

    if {$Text($w,-overwrite)} {
        # no overwrite cursor at end of line...
        if {[_two$w compare insert != "insert lineend"]} {
            # no overwrite cursor if within selection
            if {[llength [_two$w tag ranges sel]] == 0 ||
                [_two$w compare insert < sel.first] ||
                [_two$w compare insert > sel.last]
            } {
                _two$w tag add insert insert "insert +1c"
            }
        }
    }
    return
}

# ::two::text::SetOverwrite
#
#	Set the overwrite state according to current settings
#
# Arguments:
#	w		Widget

proc ::two::text::SetOverwrite {w} {
    variable Text

    if {$Text($w,-overwrite)} {
        event generate $w <<OverwriteOn>>
    } else {
        event generate $w <<OverwriteOff>>
    }
    return
}


# ::two::text::Selection --
#
#	Our own replacement selection handler. This is needed as we
#	simly set exportSelection to no in order to be able to have
#	a selection within the original widget and the clone
#	simultaneously. Without this workaround, we would only be
#	able to add the tag "sel" to one widget at a time.
#	CAVEAT: There seems to be no way to detect a "selection
#	lost" event - the text widget itself handles those events
#	with no interface at the TCL level (binding). There only
#	is the "selection own" command; thus we can only claim
#	the selection, but not remove the sel tag when we should.
#
# Arguments:
#	w		Widget

proc ::two::text::Selection {w offset length} {
    set sel ""
    foreach {start stop} [_two$w tag ranges sel] {
       append sel [_two$w get $start $stop]
    }
    return [string range $sel $offset [expr {$offset + $length}]]
}

namespace import ::two::text::*
