# tooltip.tcl --
#
#	Showing tooltips... Tooltips can be applied on buttons and
#	Menu entries. They can provide a short tip message wich appears
#	after the mouse resting for some time over the button (not
#	available on menu entries!). Additionally, the global variable
#	two_ttInfo is used to hold a more detailed description which
#	could be used in the status line of an application.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/tooltip.tcl,v 1.24 1999/10/26 13:12:30 fjl Exp $

package provide twoTooltip 1.0
package require twoUtil

option add *TwoToolTip.borderWidth       1
option add *TwoToolTip.background        black
option add *TwoToolTip.Label.foreground  black
option add *TwoToolTip.Label.background  lightYellow
option add *TwoToolTip.Label.wrapLength  250
option add *TwoToolTip.Label.padX        3
option add *TwoToolTip.Label.padY        3
option add *TwoToolTip.Label.justify     center

set two_ttInfo ""

namespace eval ::two::tooltip {
    namespace export two_tooltip
    variable Tooltip
    global two_ttInfo
    
    namespace import ::two::util::two_parseArgs
    
    # variable Tooltip
    #
    # enabled -	true if tooltips are enabled (only affects toplevel)
    # visible -	true if a tooltip is currently visible
    # offset -	y-distance from widget
    # delay -	delay between mouse enter event and showing of tooltip
    # duration -    how long is the tooltip visible
    # a_delay -	id returned by "after $Tooltip(delay)"
    # a_duration -	id returned by "after $Tooltip(duration)"
    # scr_w -	screen width (ensure that whole tooltip is on screen)
    # scr_h -	screen height (ensure that whole tooltip is on screen)
    # $w,tip -	tooltip for widget $w
    # $w,info -	info text for widget $w
    # $w,$id -	info text for index $id if $w is a menu
    
    array set Tooltip "
        enabled	1
        visible	0
        offset	4
        delay	500
        duration	5000
        scr_w	[winfo screenwidth .]
        scr_h	[winfo screenheight .]
        visible     0
        a_delay     0
        a_duration  0
    "
}

# ::two::tooltip::two_tooltip --
#
#	Add tooltip to widget w. Create toplevel "._two_tooltip" if it doesn't exist.
#	Enable/disable tooltip. Configure tooltips.
#
# Arguments:
#	w		Widget path
#	args		Tooltip arguments and options (see manpage)

proc ::two::tooltip::two_tooltip {w args} {
    variable Tooltip

    switch -exact -- $w {
	delay {
	    if {[llength $args] > 1} {
	    	return -code error "wrong # args: should be \"two_tooltip delay ?ms?\""
	    } elseif {[llength $args] == 1} {
		return [set Tooltip(delay) [lindex $args 0]]
	    } else {
		return $Tooltip(delay)
	    }
	}
	duration {
	    if {[llength $args] > 1} {
	    	return -code error "wrong # args: should be \"two_tooltip duration ?ms?\""
	    } elseif {[llength $args] == 1} {
		return [set Tooltip(duration) [lindex $args 0]]
	    } else {
		return $Tooltip(duration)
	    }
	}
	1 -
	on -
	enable {
	    set Tooltip(enabled) 1
	}
	0 -
	off -
	disable {
	    set Tooltip(enabled) 0
	    Hide 1
	}
	delete {
	    Hide
            foreach key [array names Tooltip "$w,*"] {
                unset Tooltip($key)
            }
	}
	add {
	    Create
	    eval Add $w [lrange $args 1 end]
	}
	default {
	    Create
	    eval Add $w $args
	}
    }
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoToolTipButton-class bindings
#-------------------------------------------------------------------------

bind TwoToolTipButton <Enter>      {::two::tooltip::Enter %W}
bind TwoToolTipButton <Leave>      {::two::tooltip::Hide 1}
bind TwoToolTipButton <Any-Key>    {::two::tooltip::Hide}
bind TwoToolTipButton <Any-Button> {::two::tooltip::Hide}

# And a binding to the tooltip itself:
bind TwoToolTip <Enter> {::two::tooltip::Hide 1}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::tooltip::Create --
#
#	Create the toplevel if it doesn't yet exist
#

proc ::two::tooltip::Create {} {
    if {![winfo exists ._two_tooltip]} {
        toplevel            ._two_tooltip -class TwoToolTip
        wm overrideredirect ._two_tooltip 1
        wm positionfrom     ._two_tooltip program
        wm withdraw         ._two_tooltip
        pack [label         ._two_tooltip.l -highlightthickness 0 -bd 0]
    }
}

# ::two::tooltip::Add --
#
#	Add a tooltip to a widget or menu. Uses bindtags TwoToolTipButton for
#	buttons, but needs to bind directly to menus.
#
# Arguments:
#	w		Widget path
#	args		?-index index? ?-tip m1? ?-info m2?

proc ::two::tooltip::Add {w args} {
    variable Tooltip

    set id  ""
    set tag ""
    set m1  ""
    set m2  ""
    set istext 0
    two_parseArgs args id  "-index"
    two_parseArgs args tag "-tag"
    two_parseArgs args m1  "-tip"
    two_parseArgs args m2  "-info"
    if {[llength $args]} {
	return -code error "bad argument \"$args\" for \"two_tooltip\": must be -index index, -tip text or -info text"
    }
    if {![string length $m2]} {
        set m2 $m1
    }
    if {[string length $id]} {
        # try to find out if this really is a menu
	if {[catch {$w cget -tearoff}]} {
	    return -code error "widget \"$w\" is no menu!"
	}
	# get numercial index
	set id [$w index $id]
	# set tip
	set Tooltip($w,$id)  $m2
    } elseif {[string length $tag]} {
        # try to find out if this really is a text or a canvas widget
	if {![catch {$w tag bind dummy}]} {
	    # Appears to be a text widget
	    set istext 1
	} elseif {[catch {$w bind dummy}]} {
	    return -code error "widget \"$w\" is no text or canvas widget!"
	}
	# set tip
	set Tooltip($w,tip,$tag)  $m1
	set Tooltip($w,info,$tag) $m2
    } else {
	set Tooltip($w,tip)  $m1
	set Tooltip($w,info) $m2
    }
    
    set Tooltip(visible) 0

    if {[string length $id]} {
        # Menu
	# We have to bind directly to the menu and use $w instead of %W:
	# The tk8.0 menu system uses clones with varying widget names!
	# Overwrite binding (don't use {+ code}) to avoid multiple binding
	bind $w <Motion> [namespace code "MenuSelect $w \[%W index @%x,%y\]"]
    } elseif {[string length $tag]} {
        if {$istext} {
            $w tag bind $tag <Enter>      [namespace code [list Enter %W $tag]]
            $w tag bind $tag <Leave>      [namespace code {Hide 1}]
            $w tag bind $tag <Any-Key>    [namespace code Hide]
            $w tag bind $tag <Any-Button> [namespace code Hide]
        } else {
            $w bind $tag <Enter>      [namespace code [list Enter %W $tag]]
            $w bind $tag <Leave>      [namespace code {Hide 1}]
            $w bind $tag <Any-Key>    [namespace code Hide]
            $w bind $tag <Any-Button> [namespace code Hide]
        }
    } else {
        # Normal button
	bindtags $w [concat TwoToolTipButton [bindtags $w]]
    }
    return
}

# ::two::tooltip::MenuSelect --
#
#	Binding for changing the selection within a menu
#
# Arguments:
#	w1		Widget path for original widget
#	idx		active index

proc ::two::tooltip::MenuSelect {w idx} {
    variable Tooltip
    global two_ttInfo

    if {[info exists Tooltip($w,$idx)]} {
	set two_ttInfo $Tooltip($w,$idx)
    } elseif {$two_ttInfo != ""} {
	set two_ttInfo ""
    }
    return
}

# ::two::tooltip::Enter --
#
#	Mouse has entered the hot region - show info and start
#	waiting for shwoing the tip.
#
# Arguments:
#	w		Widget path
#	tag		tag within this widget

proc ::two::tooltip::Enter {w {tag ""}} {
    variable Tooltip
    global two_ttInfo

    if {[string length $tag]} {
        if {[string length $Tooltip($w,info,$tag)]} {
            set two_ttInfo $Tooltip($w,info,$tag)
        }
        if {[string length $Tooltip($w,tip,$tag)]} {
            set Tooltip(a_delay) \
                [::after $Tooltip(delay) [namespace code [list Show $w $tag]]]
        }
    } else {
        if {[string length $Tooltip($w,info)]} {
            set two_ttInfo $Tooltip($w,info)
        }
        if {[string length $Tooltip($w,tip)]} {
            set Tooltip(a_delay) \
                [::after $Tooltip(delay) [namespace code [list Show $w]]]
        }
    }
}

# ::two::tooltip::Show --
#
#	show tooltip for widget $w
#
# Arguments:
#	w		Widget path
#	tag		tag within this widget

proc ::two::tooltip::Show {w {tag ""}} {
    variable Tooltip

    if {!$Tooltip(enabled)} {
        return
    }

    # Mouse still above widget?
    if {![string match $w [winfo containing [winfo pointerx .] [winfo pointery .]]]} {
	return
    }
    if {[string length $tag]} {
        ._two_tooltip.l configure -text $Tooltip($w,tip,$tag)
    } else {
        ._two_tooltip.l configure -text $Tooltip($w,tip)
    }
    # Calculate position
    if {[string length $tag]} {
        set x [expr {5 + [winfo pointerx $w]}]
        set y [expr {5 + [winfo pointery $w] + $Tooltip(offset)}]
    } else {
        set x [winfo pointerx $w]
        set y [expr {[winfo rooty $w] + [winfo height $w] + $Tooltip(offset)}]
        if {$y < 0} {set y 0}
    }
    set w [winfo reqwidth  ._two_tooltip.l]
    set h [winfo reqheight ._two_tooltip.l]

    # Make shure that the whole tooltip is visible
    if {$y + $h > $Tooltip(scr_h)} {
	set y [expr {$Tooltip(scr_h) - $h}]
	incr x $Tooltip(offset)
    }
    if {$x + $w > $Tooltip(scr_w)} {
	set x [expr {$Tooltip(scr_w) - $w}]
    }

    # Remove tooltip after duration:
    set Tooltip(a_duration) \
    	[::after $Tooltip(duration) [namespace code Hide]]

    # really show tooltip
    # The update command is needed on windows.
    # BUT: maybe another binding (e.g. on cursor leave) has fired in
    # the meantime, and thus the Hide command is called. This results
    # in a stale tooltip... $Tooltip(visible) is used to circumvent this.
    set Tooltip(visible) 1
    wm geometry  ._two_tooltip +$x+$y
    update
    if {$Tooltip(visible)} {
	wm deiconify ._two_tooltip
	raise ._two_tooltip
    }
    return
}

# ::two::tooltip::Hide --
#
#	remove toplevel Tooltip widget and cancel "after"
#	if info is 1, then also remove tooltip info

proc ::two::tooltip::Hide { {info 0} } {
    variable Tooltip
    global two_ttInfo

    set Tooltip(visible) 0
    if {[winfo exists ._two_tooltip]} {
	wm withdraw ._two_tooltip
    }
    catch {::after cancel $Tooltip(a_delay)}
    catch {::after cancel $Tooltip(a_duration)}
    if {$info && [string length $two_ttInfo]} {
        set two_ttInfo ""
    }
    return
}

namespace import ::two::tooltip::*
