# libw.tcl --
#
#	Contains all sub-packages of the two library. Might be a convenient
#	starting point.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoLib 1.0

# Sources all other components. Propably a good starting point
# for using LibW...

# util.tcl --
#
#	This file contains several utility procedures used by
#	the two library. Only part of these are imported
#	into the global namespace.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoUtil 1.0

# Workaround: Tk needs this procedure for tk_messageBox
# but won't load it form inside a namespace (bug in tcl8.0 up to p1)
catch {auto_load tkMessageBox}

namespace eval ::two::util {
    namespace export two_*
    variable Util
    
    # variable Util
    #
    # clipboard -	Contents of replacement clipboard for two_noImportExport
    # PWD -		Current working directory for two_kshPwd
    # OLDPWD -	Previous working directory dor two_kshPwd
}

# ::two::util::two_parseArgs --
#
#	This procedures extracts an argument "arg" and a value
#	from "arg_list", removing both argument and value. The value
#	is stored in "val".
#
# Arguments:
#
#	argsListName	Call by Reference; name of tcl list with arguments
#	valName		Call by Reference; name of variable; takes the value
#	arg		The Argument to be extracted
#
# Results:
#	1 if "arg" was found in "arg_list", otherwise 0.

proc ::two::util::two_parseArgs {argsListName valName arg} {
    upvar $argsListName argsList
    upvar $valName val

    set pos [lsearch -exact $argsList $arg]
    set end [lsearch -exact $argsList "--"]
    if {$pos >= 0 && ($end < 0 || $pos < $end)} {
        set valpos [expr {$pos +1}]
        set val [lindex $argsList $valpos]
        set argsList [lreplace $argsList $pos $valpos]
	return 1
    }
    return 0
}

# ::two::util::two_parseMultiArgs --
#
#	This procedures examines a list of arguments (option-value pairs)
#	which were used when calling a procedure. All arguments whose
#	names are contained as keys in the specsArray will be removed. The
#	value of the specsArray is the name of a variable which should take
#	the value which is associated with the option.
#
# Arguments:
#
#	specsArrayName	Name of Array which should contain the specs for the
#			options
#	argsListName	Name of List of Arguments which should be processed
#
# Results:
#	List of all options which have been eaten by this proc.
#
# Side effects:
#	Modifies the contents of $argsListName.

proc ::two::util::two_parseMultiArgs {specsArrayName argsListName} {
    upvar $specsArrayName specsArray
    upvar $argsListName   argsList

    # will contain all options which have not been eaten by this proc
    set remaining ""
    # will contain the names all options that matched the description
    set matched ""

    # loop through all options, but break at "--"
    set length [llength $argsList]
    for {set idx 0} {$idx < $length} {incr idx} {
        set arg [lindex $argsList $idx]
	if {[string match "--" $arg]} {
	    eval lappend remaining [lrange $argsList $idx end]
	    break
	}
        if {[info exists specsArray($arg)]} {
	    lappend matched $arg
	    set next [lindex $argsList [incr idx]]
            uplevel [list set $specsArray($arg) $next]
	} else {
	    lappend remaining $arg
	}
    }
    set argsList $remaining
    return $matched
}

# ::two::util::two_getOpt --
#
#	Searches the database much the same way the "option get"
#	does. The return value is set to $default if the option
#	get command returns an empty string.
#
# Arguments:
#	w		Widget path
#	name		Name of option to te queried
#	class		Option class
#	default		Default value (if "option get" returns empty value)
#
# Results:
#	The option or the default value.

proc ::two::util::two_getOpt {w name class default} {
    set retval [option get $w $name $class]
    if {![string length $retval]} {
        set retval $default
    }
    return $retval
}

# ::two::util::two_noExportImport --
#
#	Disables all copying to and from the application. Disabling the
#	export function is easy through "exportSelection none", but
#	importing the seletion is a bit trickier. It is achieved here
#	through replacing the "selection" and "clipboard" commands.
#	Caveat: You'll have to call this function before instantiating
#	any widget in order to make the "option add" work.
#
# Side Effects:
#	Replaces the tcl clipboard and selection procedures with
#	fake ones. Those replacements intentinally do only work
#	within the application.

proc ::two::util::two_noExportImport {} {
    variable Util
   
    set Util(clipboard) ""

    # No exportSelection
    option add *exportSelection  false  100

    # Install our replacement commands
    proc ::selection {opt args} "
	return \[uplevel [namespace code FakeSelection] \$opt \$args \]
    "
    proc ::clipboard {opt args} "
	uplevel [namespace code FakeClipboard] \$opt \$args
    "
    return
}

# ::two::util::FakeSelection --
#
#	This is a replacement for the standard tk selection command.
#	See two_noExportImport

proc ::two::util::FakeSelection {opt args} {
    variable Util

    switch $opt {
        clear {
	    set Util(clipboard) ""
	}
        get {
	    return $Util(clipboard)
	}
    }
    return
}

# ::two::util::FakeClipboard --
#
#	This is a replacement for the standard tk clipboard command.
#	See two_noExportImport

proc ::two::util::FakeClipboard {opt args} {
    variable Util

    switch $opt {
        clear {
	    set Util(clipboard) ""
	}
        append {
	    while {[llength $args]} {
	        switch -glob -- [lindex $args 0] {
	            -- {
			append Util(clipboard) [join [lrange $args 1 end] ""]
			set args ""
		    }
		    -* {
			set args [lrange $args 2 end]
		    }
		    default {
			append Util(clipboard) [join $args ""]
			set args ""
		    }
		}
	    }
	}
	default {
	    return -code error "wrong # args: should be \"clipboard option ?arg arg ...?\""
	}
    }
    return
}

# ::two::util::two_kshPwd --
#
#	Workaround for pwd: don't show /tmp_mnt/....
#	This happens on systems using the automounter, which mounts file-
#	systems in /tmp_mnt/ and puts symlinks in place. The problem here
#	is that accessing a path through /tmp_mnt/ will not re-mount the
#	filesystem, and the tcl cd/pwd combination will always show this
#	/tmp_mnt/ path (e.g. in tk_getOpenFile).
#	This implementation behaves much like the Korn shell: whenever you
#	cd into a directory through symlinks it will remember the route
#	you took. Using "cd .." will go up one level in your route, i.e.
#	remove the last part of the pathname, and not go up to the "real"
#	parent directory. Please note that different UNIX shells behave
#	differently in this respect.
#
# Side effects:
#	The tcl cd and pwd commands are renamed into the current namespace
#	and wrapped with a replacement function.

proc ::two::util::two_kshPwd {} {
    variable Util
    global tcl_platform

    if {[info exists Util(PWD)]} {
        return -code error "two_kshPwd has already been installed!"
    }

    if {[string match "unix" $tcl_platform(platform)]} {
	set Util(PWD)    [exec /bin/ksh -c "pwd"]
	set Util(OLDPWD) $Util(PWD)
	rename ::pwd _origPwd
	rename ::cd  _origCd
	proc ::pwd {} [namespace code {
	    return $Util(PWD)
	}]
	proc ::cd {args} "
	    return \[uplevel [namespace code KshCd] \$args\]
	"
    }
    return
}

# ::two::util::KshCd --
#
#	This is a replacement for the standard tk cd command.
#	See two_kshPwd

proc ::two::util::KshCd {{dir ""}} {
    variable Util
    global errorCode

    # special handling for "cd" and "cd -"
    if {![string length $dir]} {
        set dir ~
    } elseif {[string match "-" $dir]} {
        set dir $Util(OLDPWD)
    }

    # tilde substitution
    if {[string match "absolute" [file pathtype $dir]]} {
        set dir [file nativename $dir]
    }

    # calculate name of target directory
    set lpwd [file split $Util(PWD)]
    foreach part [file split $dir] {
        switch -- $part {
            . {
            }
            .. {
                set lpwd [lreplace $lpwd end end]
            }
            default {
                lappend lpwd $part
            }
        }
    }
    if {[llength $lpwd]} {
        set dir [eval file join $lpwd]
    } else {
        set dir [file nativename /]
    }

    # try to change current directory to the target
    if {[set _origCdRetCode [catch {_origCd $dir} _origCdRetVal]]} {
        return -code $_origCdRetCode -errorcode $errorCode $_origCdRetVal
    }

    # cd succeeded; update state
    set Util(OLDPWD) $Util(PWD) 
    set Util(PWD)    $dir
    return $_origCdRetVal
}

# ::two::util::two_invoke --
#
#	Visually invoke button or set focus to next widget if this function
#	is used on a label.
#
# Arguments:
#	w		Widget path to button
#
# Side effects:
#	The button is flashed and the invoked, or the focus is set to the
#	next widget.

proc ::two::util::two_invoke {w} {
    switch -exact -- [winfo class $w] {
	Button {
	    if {![string match "disabled" [$w cget -state]]} {
    		$w configure -state active -relief sunken
    		update idletasks
    		after 100
    		$w configure -state normal -relief raised
		$w invoke
	    }
	}
	Label {
	    set next [tk_focusNext $w]
	    if {[string match Entry [winfo class $next]]} {
		$next selection range 0 end
		$next icursor end
	    }
	    focus $next
	}
	default {
            $w invoke
	}
    }
    return
}

# ::two::util::two_push --
#
#	Push value on stack. Stack is a tcl list. Returns value.
#
# Arguments:
#	stackName	Name of TCL list variable
#	value		Value to be pushed on the stack
#
# Results:
#	The element which is pushed on the stack is returned as well.


proc ::two::util::two_push {stackName value} {
    upvar 1 $stackName stack

    lappend stack $value
    return $value
}

# ::two::util::two_pop --
#
#	Take topmost value from on stack. Stack is a tcl list. Returns value.
#
# Arguments:
#	stackName	Name of TCL list variable
#
# Results:
#	The topmost element is returned.

proc ::two::util::two_pop {stackName} {
    upvar 1 $stackName stack
    
    if {![info exists stack]} {
        return {}
    }
    set value [lindex $stack end]
    set stack [lreplace $stack end end]
    return $value
}

# ::two::util::two_now --
#
#	returns current time as formatted string
#
# Results:
#	the current time...

proc ::two::util::two_now {} {
    return [clock format [clock seconds]]
}

# ::two::util::two_ct --
#
#	returns the current toplevel, i.e. the toplevel which
#	currently contains the focus. If the application doesn't
#	have the focus, "." is returned
#
# Results:
#	A widget path

proc ::two::util::two_ct {} {
    set f [focus]
    if {[string length $f]} {
        return [winfo toplevel $f]
    } else {
        return "."
    }
}

# ::two::util::two_mb --
#
# thin wrapper around tk_messageBox which sets the parent widget to [two_ct]
#
# Arguments:
#	See man page

proc ::two::util::two_mb {args} {
    if {![two_parseArgs args p "-parent"] && ![two_parseArgs args p "-p"]} {
        set p [two_ct]
    }
    return [uplevel tk_messageBox -parent $p $args]
}

# ::two::util::two_center --
#
#	center widget w on top of parent p

proc ::two::util::two_center {w {p ""}} {
    wm withdraw  $w
    if {[string length $p]} {
        set p [winfo toplevel $p]
        wm deiconify $p
    } else {
        set p .
    }
    # center on screen
    update idletasks
    set rw [winfo reqwidth  $w]
    if {[winfo width $w] > $rw} {
        set rw [winfo width $w]
    }
    set rh [winfo reqheight $w]
    if {[winfo height $w] > $rh} {
        set rh [winfo height $w]
    }
    set x [expr {[winfo screenwidth  $p]/2 - $rw/2 - [winfo vrootx $p]}]
    set y [expr {[winfo screenheight $p]/2 - $rh/2 - [winfo vrooty $p]}]
    wm geom      $w ${rw}x${rh}+$x+$y
    wm deiconify $w
    return
}

# ::two::util::two_show --
#
#	show window w (i.e. raise/deiconify/show centered)

proc ::two::util::two_showCenter {w {p ""}} {

    if {[winfo ismapped $w]} {
        raise $w
    } elseif {[string match "iconic" [wm state $w]]} {
        wm deiconify $w
    } else {
        two_center $w $p
    }
    return
}

# ::two::util::two_busy
#
#	block interaction with a toplevel and show a "watch" cursor.
#	This function uses a transparent frame which is placed on
#	top of the toplevel widget.
#
# Know Bug:
#	Transparent frames don't propagate redraw events, so the
#	underlying widgets don't repaint at all. Looks like a
#	blocked application. Won't make any difference if your
#	app really blocks, but if it doesn't... it's ugly.
#
# Arguments:
#	busy		True to make the toplevels busy, false to
#			remove the busy cursor.
#	args		List of Paths to widgets which should be
#			disabled (preferrably toplevels)

proc ::two::util::two_busy {busy args} {
    foreach w $args {
        if {[string compare "." $w] == 0} {
            set f .twoBusyCursorFrame
        } else {
            set f $w.twoBusyCursorFrame
        }
        if {$busy} {
            if {[winfo exists $f]} {
                raise $f
            } else {
                frame $f -background {} -borderwidth 0 -cursor watch
                place $f -x 0 -y 0 -relwidth 1 -relheight 1
            }
        } else {
            if {[winfo exists $f]} {
                destroy $f
            }
        }
    }
    update idletasks
}


# ::two::util::two_readFile --
#
#	Read a given a file and return it's contents. This proc
#	combines the open, read and close calls into one short
#	procedure; errors are "forwarded" (use catch...). One
#	other thing that is handled within this proc is the
#	possibility of slow automounters etc. making an open
#	command fail. In any case, the open is tried twice
#	withing 200 ms.
#
# Arguments:
#	args		?-nonewline? ?-binary? fname
#
# Returns:
#	The contents of this file

proc ::two::util::two_readFile {args} {
    global errorCode

    # parse options
    set fname [lindex $args end]
    set nonewline 0
    set bin 0
    set opts [lreplace $args end end]
    foreach opt $opts {
        switch -exact -- $opt {
            -nonewline {
                set nonewline 1
            }
            -binary {
                set bin 1
            }
            default {
                return -code error "unknown option \"$opt\": must be -nonewline or -binary"
            }
        }
    }

    # Open file: try twice
    set err [catch {set fid [open $fname]}]
    if {$err} {
        # didn't work... try again after 200 ms to work around slow automounter
	after 200
        set err [catch {set fid [open $fname]} errorMsg]
        if {$err} {
            # second try failed as well
            return -code $err -errorcode $errorCode $errorMsg
        }
    }

    # read the file
    # Improve performance: TCL's default buffer size is 4096 and this is used
    # to read the file in "slices" and append them to the txt variable. There
    # are two methods to optimize this: set buffersize to something larger
    # (but it needs to be smaller than 1,000,000), e.g. 2^18, or use the real
    # size of the file as parameter to the open command (only available if
    # -nonewline ain't specified). Timing on a HP C160 (HP-UX 10.20) with two
    # large files (/stand/vmunix: ~6M, /opt/aCC/lib/lib++.a ~1M) reveal:
    #                   6M       1M
    # plain:      64058498  1744060
    # fconfigure:  2661062   151780
    # file size:    794698   115280
    if {$bin} {
        fconfigure $fid -translation binary
    }
    if {$nonewline} {
        fconfigure $fid -buffersize 262144
        set err [catch {set txt [read -nonewline $fid]} errorMsg]
    } else {
        set err [catch {set fs [file size $fname]} errorMsg]
        if {! $err} {
            set err [catch {set txt [read $fid $fs]} errorMsg]
        }
    }
    
    # closing should always work
    close $fid
    # handle error from above (reading file)
    if {$err} {
        return -code $err -errorcode $errorCode $errorMsg
    }
    
    # usually you would insert "return $txt" at this palce, but the following
    # should be faster (said to be due to less error checking?)
    set txt
}

namespace import ::two::util::two_noExportImport
namespace import ::two::util::two_kshPwd
namespace import ::two::util::two_invoke
namespace import ::two::util::two_ct
namespace import ::two::util::two_mb
namespace import ::two::util::two_center
namespace import ::two::util::two_showCenter
namespace import ::two::util::two_busy
namespace import ::two::util::two_readFile

# framework.tcl --
#
#	This file contains the basic procedure needed by the
#	TWO compound widgets framework.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoFramework 1.0

namespace eval ::two::framework {
    namespace export two_*
    variable OptSpecs
    variable Aliases
    variable TemplateCreate
    variable TemplateManipulate

    # variable OptSpecs
    #	Array of Option descriptions for a widget. The key is "name,-opt"
    #	with name being the name of the widget (not the pathname!).
    #	The value is a list of 2 or 3 elements:
    #	{database} {class}
    #	{database} {class} {setProc}
    #	{database} {class} {setProc} {changeProc}
    
    array set OptSpecs ""
    
    # variable Aliases
    #	Array of Option aliases for a widget. The key is "name,-opt"
    #	with name being the name of the widget (not the pathname!).
    #	The value is the canonical name of the option.
    #	Additinal Entries are only "name" to list all aliases within
    #	the named widget
    
    array set Aliases ""
}

# variable ::two::framework::TemplateCreate --
#
#	This is the template for the procedure which will be created
#	by two_widget. It holds the setup code for creating new widgets.
#
# The following substitutions take place:
#	%state% -	Name of variable used to store widget config
#	%w% -		Name of widget that will be created (not the pathname!)
#	%base% -	Name of base widget
#	%init% -	Initinalize state array with default values
#	%create% -	Additional startup code
#	%options% -	List of all additinal options
#	%spec% -	Array of option specs (with fully qualified namespace)
#	%alias% -	Array of aliases (with fully qualified namespace)
#	%unknown% -	Handler proc for unknown options
#
# Pre-defined Variables:
#	Variables pre-defined for use in Template are $w (widget path) and $args.
#	There might be more from widget's option spec (fixed args).

set ::two::framework::TemplateCreate {
    variable %state%
    
    # Initialize the default values for widget config
    %init%

    # will contain all remaining options
    set remaining ""

    # keys will be the names all options that matched
    array set matched ""

    # loop through all options, but break at "--"
    set delimit [lsearch -exact $args "--"]
    if {$delimit > 0} {
        incr delimit -1
        set argList [lrange $args 0 $delimit]
    } else {
        set argList $args
    }
    foreach {arg value} $argList {
        if {[info exists %alias%(%w%,$arg)]} {
	    # just an alias to another option
	    set arg $%alias%(%w%,$arg)
	}
        if {[info exists %spec%(%w%,$arg)]} {
	    # real option: remember match
            set matched($arg) $value
	} else {
	    # left over for base widget
	    lappend remaining $arg $value
	}
    }
    # the rest....
    if {$delimit > 0} {
        lappend remaining [lrange $args $delimit end]
    }

    # build base widget
    %base% $w
    
    # set values in widget's state array
    # and query option database
    foreach arg %options% {
        set spec $%spec%(%w%,$arg)
        if {[info exists matched($arg)]} {
	    # this option was explicitly set
	    set %state%($w,$arg) $matched($arg)
	    if {[llength $spec] > 2} {
	        eval [lindex $spec 2]
	    }
	} else {
	    # query the option database
	    set opt [option get $w [lindex $spec 0] [lindex $spec 1]]
	    if {[string length $opt]} {
		set %state%($w,$arg) $opt
		if {[llength $spec] > 2} {
	            eval [lindex $spec 2]
		}
	    }
	}
    }

    # Rename the created base widget...
    rename $w _two$w

    # Configure the remaining options
    if [string length $remaining] {
	%unknown%
    }

    # ...and install our replacement command
    namespace eval :: "
        proc [list $w] {opt args} {
            uplevel [list %w%Manipulate] $w \$opt \$args
        }
    "

    # catch widget destruction in order to clean up Text
    bind $w <Destroy> [list [list %w%Clear] $w]
    
    # do additional startup processing
    %create%

    # Return the widget path
    return $w
}

# variable ::two::framework::TemplateManipulate --
#
#	This is the template for the procedure which will be created
#	used to manipulate the new widget created by two_text.
#
# The following substitutions take place:
#	%state% -	Name of variable used to store widget config
#	%w% -		Name by which the widget was created (not the pathname!)
#	%manip% -	switch body portion to match widget's commands
#	%options% -	List of all additinal options
#	%aliases% -	List of all additinal aliases
#	%spec% -	Array of option specs (with fully qualified namespace)
#	%alias% -	Array of aliases (with fully qualified namespace)
#	%aliases% -	List of all aliases for in the current widget's api
#	%default% -	Array of option's default values
#	%unknown% -	Handler proc for unknown options
#	%wns% -		Name of namespace in which the widget will be created
#
# Pre-defined Variables:
#	Variables pre-defined for use in Template are $w (widget path),
#	$cmd (widget command) and $args (additional arguments).

set ::two::framework::TemplateManipulate {
    variable %state%
    variable %default%

    # See if this option is for us...
    switch -exact -- $cmd {
        cget {
	    if {[llength $args] != 1} {
	        return -code error "Wrong # of args in \"$w cget $args\""
	    }
	    set arg [lindex $args 0]
            if {[info exists %alias%($w,$arg)]} {
		# just an alias to another option
		set arg $%alias%($w,$arg)
	    }
            if {[info exists %state%($w,$arg)]} {
		# known option
		return $%state%($w,$arg)
	    } else {
		# left over for base widget
		return [uplevel [list %wns%::_two$w] cget $arg]
	    }
	}
        conf -
	confi -
	config -
	configu -
	configur -
	configure {
	    if {![llength $args]} {
	        # return current configuration
	        set retval [_two$w configure]
		foreach opt %options% {
		    set spec $%spec%(%w%,$opt)
		    lappend retval [list \
		        $opt             \
			[lindex $spec 0] \
			[lindex $spec 1] \
			$%default%($opt)  \
			$%state%($w,$opt) \
		    ]
		}
		foreach opt $%alias%(%w%) {
		    set spec $%spec%(%w%,$%alias%(%w%,$opt))
		    lappend retval [list \
		        $opt             \
			[lindex $spec 0]
		    ]
		}
		return $retval
	    }
	    set remaining ""
	    foreach {arg value} $args {
        	if {[info exists %alias%(%w%,$arg)]} {
		    # just an alias to another option
		    set arg $%alias%(%w%,$arg)
		}
        	if {[info exists %spec%(%w%,$arg)]} {
		    set spec $%spec%(%w%,$arg)
		    # real option: react on it
		    set %state%($w,$arg) $value
		    if {[llength $spec] > 2} {
	        	eval [lindex $spec 2]
			if {[llength $spec] > 3} {
	        	    eval [lindex $spec 3]
			}
		    }
		} else {
		    # left over for base widget
		    lappend remaining $arg $value
		}
	    }
	    if {[string length $remaining]} {
	        return  [%unknown%]
	    }
	}
        %manip%
	default {
            return [uplevel [list %wns%::_two$w] $cmd $args]
	}
    }
    return
}

# variable ::two::framework::TemplateClear --
#
#	This is the template for the procedure which will
#	clean up state values on widget destruction.
#
# The following substitutions take place:
#	%state% -	Name of variable used to store widget config
#	%clear% -	Addional cleanup code
#
# Pre-defined Variables:
#	Variables pre-defined for use in Template is $w (widget path).

set ::two::framework::TemplateClear {
    variable %state%

    %clear%
    
    # clear state
    foreach key [array names %state% "$w,*"] {
        unset %state%($key)
    }
    # remove the original widget command
    rename $w ""
    return
}

# ::two::framework::two_widget --
#
#	Utility procedure for the compound widget framework. Creates a
#	new procedure just like proc does, but with additinal option
#	parsing abilities (including Tk's Option Database), creating
#	and renaming of the "back-end" widget and setting preparing
#	for cleanup on widget destruction.
#	The optList is as list of option descriptions, each being a
#	a list themselves; following is a list of examples
#	{ {arg} }
#	{ {unknown} {proc} }
#	{ {-option} {cname} }
#	{ {-option} {database} {class} {default} }
#	{ {-option} {database} {class} {default} {setProc} }
#	{ {-option} {database} {class} {default} {setProc} {changeProc} }
#	The enties are:
#	arg -		A single fixed argument, like in normal proc def.
#	unknown -	Use this *keyword* to name the proc used
#			when changeing the configuration (.w configure)
#			and the option is not contained in this spec
#			This proc should be defined as "proc x {w args} {}"
#	-option -	The name of a option (including -)
#	cname -		Link to the canonical name of an option (NOT the class!).
#	database -	The name of the option in the option database
#	class -		The class of the option in the option database
#	default -	The default value
#	setProc -	Proc eval'ed whenever the value is set (or changed)
#	changeProc -	Proc eval'ed whenever the value is changed (configure)
#
# Arguments:
#	wname -		name of the compound widget procedure
#	stateVarName -	name of namespace variable carrying widget config
#	base -		name of base widget; additional options are appended!
#	optList -	List of lists describing the options (see manpage)
#	create -	code doing additional initialization
#	manipulate -	code used for the widget command
#	clear -		code used for additional clean-up
#
# Side Effects:
#	Creates a new procedure in caller's environment, the name
#	being the one given by "wname", and the argument list
#	"w args".
#
# Results:
#	Empty string.

proc ::two::framework::two_widget {
    wname stateVarName base optList
    {create     ""}
    {manipulate ""}
    {clear      ""}
} {
    variable OptSpecs
    variable Aliases
    variable DefaultsArray
    variable TemplateCreate
    variable TemplateManipulate
    variable TemplateClear
    
    set unknownOptHandler   {_two$w configure $remaining}
    array set DefaultsArray ""
    set Aliases($wname) ""
    set defaultsInit    ""
    set options         ""
    set fixedArgs       w
    
    # the namespace of the framework
    set fns [namespace current]
    # the namespace from which this proc has been called
    set cns [uplevel {namespace current}]
    # figure out in which namespace the proc should be created
    set wns [namespace qualifiers $wname]
    if {![string length $wns]} {
        if {![string match "::" $cns]} {
            set wns $cns
        }
    } elseif {![string match "::*" $wns]} {
        if {![string match "::" $cns]} {
            set wns $cns::$wns
        } else {
            set wns ::$wns
        }
    }

    # some sanity checks
    foreach {var content} [list wname $wname stateVarName $stateVarName base $base] {
	if {![string length $content]} {
            return -code error "value for \"$var\" missing"
	}
    }

    # transform the optList specification into a set of
    # lists and arrays in order to speed up later processing
    foreach spec $optList {
        switch [llength $spec] {
	    1 {
	      	# additional arg for widget command
		lappend fixedArgs [lindex $spec 0]
	    }
	    2 {
	    	# alias to another option or default proc for unknown option changes
	        if {[string match "unknown" [lindex $spec 0]]} {
		    set unknownOptHandler "[list eval [list [lindex $spec 1]]] \[list \$w\] \$remaining"
		} else {
		    if {![string match "-*" [lindex $spec 0]] } {
			return -code error "First element of two_widget optSpec \"$spec\" should start with \"-\""
		    }
		    if {![string match "-*" [lindex $spec 1]] } {
			return -code error "Alias (2. element) in two_widget optSpec \"$spec\" should start with \"-\""
		    }

		    lappend Aliases($wname) [lindex $spec 0]
		    set Aliases($wname,[lindex $spec 0]) [lindex $spec 1]
		}
	    }
	    4 -
	    5 -
	    6 {
	    	# option
	    	set opt [lindex $spec 0]

		if {![string match "-" [string index $opt 0]]} {
		    return -code error "First element of two_widget optSpec \"$spec\" should start with \"-\""
		}
		
    		# build array of option specs for later use
		# when parsing the actual options in the widget proc
		# The name is mangled into the key, and the default value removed
        	set OptSpecs($wname,$opt) \
		    [concat [lrange $spec 1 2] [lrange $spec 4 5]]

		# build list of options
		lappend options $opt
		
		# build list of default configs for array initialization
		append defaultsInit "\$w,$opt [list [lindex $spec 3]] \\\n        "
		set    DefaultsArray($opt)     [lindex $spec 3]
	    }
	    default {
		return -code error "Wrong # of args: should in two_widget optSpec \"$spec\""
	    }
	}
    }
    
    if {[string length $defaultsInit]} {
        set init "array set [list $stateVarName] \[list $defaultsInit\]"
    } else {
        set init "# none"
    }
    if {![string length $create]} {
        set create "# none"
    }

    # complete the creation procedure wproc
    set wproc ""
    foreach {code token} [split $TemplateCreate %] {
        append wproc $code
	switch -exact -- $token {
	    state   { append wproc $stateVarName      }
	    w       { append wproc $wname             }
	    base    { append wproc $base              }
	    init    { append wproc $init              }
	    options { append wproc [list $options]    }
	    spec    { append wproc ${fns}::OptSpecs   }
	    alias   { append wproc ${fns}::Aliases    }
	    unknown { append wproc $unknownOptHandler }
	    create  { append wproc $create            }
	}
    }

    # complete the manipulation procedure mproc
    set mproc ""
    foreach {code token} [split $TemplateManipulate %] {
        append mproc $code
	switch -exact -- $token {
            state   { append mproc $stateVarName         }
            w       { append mproc $wname                }
            options { append mproc [list $options]       }
            spec    { append mproc ${fns}::OptSpecs      }
            alias   { append mproc ${fns}::Aliases       }
            default { append mproc ${fns}::DefaultsArray }
            unknown { append mproc $unknownOptHandler    }
            manip   { append mproc $manipulate           }
	    wns     { append mproc $wns                  }
	}
    }

    # complete the cleanup procedure cproc
    set cproc ""
    foreach {code token} [split $TemplateClear %] {
        append cproc $code
	switch -exact -- $token {
	    state   { append cproc $stateVarName }
	    clear   { append cproc $clear        }
	}
    }

    # ...and finally create the proc
    # Note: $wname includes an absoulte namespace qualifier
    lappend fixedArgs args
    proc $wname             $fixedArgs   $wproc
    proc ${wname}Manipulate {w cmd args} $mproc
    proc ${wname}Clear      w            $cproc
}

namespace import ::two::framework::*

#source pre.tcl
# opt.tcl --
#
#	Set several options in the option database to make tk apps
#	more beatiful, especially on systems using CDE.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoOpt 1.0
package require twoUtil

# catch as font might already exist
catch {font create twoTxt -family courier -size 10}
if {[string match "unix" $tcl_platform(platform)]} {
    # Fonts
    catch {font create twoGui -family helvetica -size 10 -weight bold}
    catch {font create twoLbl -family helvetica -size 10}

    # General settings: colors like M$,
    # borderWidth like CDE, etc.
    option add *foreground			black
    option add *background			grey77
    option add *disabledForeground 		grey51
    option add *font				twoGui
    option add *borderWidth			1
    option add *highlightThickness		0
    option add *highlightBackground		grey77
    option add *highlightColor			black
    option add *selectColor			yellow
    option add *insertWidth			3
    option add *insertBorderWidth		1
    option add *insertBackground		red
    option add *insertOffTime			400
    option add *insertOnTime			800

    option add *Frame.relief			flat
    option add *Frame.borderWidth		0

    option add *Button.relief			raised
    option add *Button.highlightThickness	2
    option add *Button.highlightBackground	grey77
    option add *Button.anchor			center

    option add *Radiobutton.relief		flat
    option add *RadioButton.highlightThickness	2
    option add *Radiobutton.anchor		w
    option add *Radiobutton.selectColor		yellow

    option add *Checkbutton.relief		flat
    option add *Checkbutton.highlightThickness	2
    option add *Checkbutton.anchor		w
    option add *Checkbutton.selectColor		yellow

    option add *Text.background			white
    option add *Text.foreground			black
    option add *Text.font			twoTxt
    option add *Text.hardTab			1
    
    option add *Listbox.background		white
    option add *Listbox.foreground		black
    option add *Listbox.font			twoTxt
    option add *Listbox.selectBackground	blue4
    option add *Listbox.selectForeground	white
#    option add *Listbox.borderWidth		0

    option add *Entry.relief                 	sunken
    option add *Entry.background		white
    option add *Entry.foreground		black
    option add *Entry.font			twoTxt

    option add *Label.font			twoLbl

    option add *Canvas.background		white

    option add *Scrollbar.takeFocus		0
    option add *Scrollbar.width			11

    option add *TwoToolBar.relief		raised
    option add *TwoToolBar.borderWidth		1

    # tk_getOpenFile and tk_getSaveFile use button
    # as label (in order to have "-state disabled")
    option add *TkFDialog.f3.lab.font		twoLbl

    # Workaround: bg color of toplevel can't be set by option at runtime
    . configure -bg grey77

    # Workaround: tk_messageBox doesn't bind Escape key
    bind Dialog <Map> {+
	bind %W <Escape> {
            if {[winfo exists %W.cancel]} {
        	two_invoke %W.cancel
	    } elseif {[winfo exists %W.abort]} {
		two_invoke %W.abort
	    } elseif {[winfo exists %W.no]} {
		two_invoke %W.no
	    } elseif {[winfo exists %W.ok]} {
		two_invoke %W.ok
	    }
	}
    }
} else {
    option add *Text.font	twoTxt
    option add *Listbox.font	twoTxt
    option add *Entry.font	twoTxt
}

# icons.tcl --
#
#	Several icons which may be useful for toolbars.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoIcons 1.0

image create bitmap twoNew -data {
\#define new_width 20
\#define new_height 20
static unsigned char new_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xe0, 0x07, 0x00,
  0x20, 0x0c, 0x00, 0x20, 0x14, 0x00, 0x20, 0x24, 0x00, 0x20, 0x7c, 0x00,
  0x20, 0x40, 0x00, 0x20, 0x40, 0x00, 0x20, 0x40, 0x00, 0x20, 0x40, 0x00,
  0x20, 0x40, 0x00, 0x20, 0x40, 0x00, 0x20, 0x40, 0x00, 0xe0, 0x7f, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoOpen -data {
\#define open_width 20
\#define open_height 20
static unsigned char open_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x38, 0x00,
  0x00, 0x44, 0x01, 0x00, 0x80, 0x01, 0x38, 0xc0, 0x01, 0xc4, 0x1f, 0x00,
  0x04, 0x10, 0x00, 0x04, 0x10, 0x00, 0x84, 0xff, 0x03, 0x44, 0x55, 0x01,
  0xa4, 0xaa, 0x00, 0xd4, 0x57, 0x00, 0xac, 0x2a, 0x00, 0xfc, 0x1f, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoSave -data {
\#define save_width 20
\#define save_height 20
static unsigned char save_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xf8, 0xff, 0x01,
  0x28, 0x40, 0x01, 0x28, 0xc0, 0x01, 0x28, 0x40, 0x01, 0x28, 0x40, 0x01,
  0x28, 0x40, 0x01, 0x28, 0x40, 0x01, 0xc8, 0x3f, 0x01, 0x08, 0x00, 0x01,
  0xc8, 0x7f, 0x01, 0xc8, 0x4f, 0x01, 0xc8, 0x4f, 0x01, 0xc8, 0x4f, 0x01,
  0xf0, 0xff, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoSaveAll -data {
\#define saveall_width 20
\#define saveall_height 20
static unsigned char saveall_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xc0, 0xff, 0x03, 0x40, 0x81, 0x02,
  0xf0, 0xff, 0x03, 0x50, 0xa0, 0x02, 0xfc, 0xff, 0x02, 0x14, 0xa8, 0x02,
  0x14, 0xb8, 0x02, 0x14, 0xa8, 0x02, 0x14, 0xa8, 0x02, 0x14, 0xa8, 0x02,
  0xe4, 0xa7, 0x02, 0x04, 0xa0, 0x03, 0xe4, 0xaf, 0x00, 0xe4, 0xe9, 0x00,
  0xe4, 0x29, 0x00, 0xf8, 0x3f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoPrint -data {
\#define print_width 20
\#define print_height 20
static unsigned char print_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0xff, 0x00,
  0x40, 0x80, 0x00, 0x40, 0x5f, 0x00, 0x20, 0x40, 0x00, 0xa0, 0xef, 0x01,
  0x10, 0xa0, 0x02, 0xf8, 0x5f, 0x03, 0x04, 0xa0, 0x02, 0xfc, 0x7f, 0x02,
  0x04, 0x40, 0x01, 0x04, 0xc0, 0x01, 0xfc, 0x7f, 0x01, 0x08, 0xa0, 0x00,
  0xf0, 0x7f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoExit -data {
\#define exit_width 20
\#define exit_height 20
static unsigned char exit_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xe0, 0x7f, 0x00,
  0xe0, 0x40, 0x00, 0x20, 0x47, 0x00, 0x20, 0x4c, 0x00, 0x20, 0x48, 0x00,
  0x20, 0x48, 0x00, 0x20, 0x48, 0x00, 0x20, 0x4b, 0x00, 0x20, 0x4a, 0x00,
  0x20, 0x48, 0x00, 0x20, 0x48, 0x00, 0x20, 0x48, 0x00, 0xe0, 0x78, 0x00,
  0x80, 0x0b, 0x00, 0x00, 0x0c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoUndo -data {
\#define undo_width 20
\#define undo_height 20
static unsigned char undo_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x1f, 0x00, 0xe4, 0x7f, 0x00,
  0xfc, 0xf0, 0x00, 0x3c, 0xc0, 0x01, 0x3c, 0x80, 0x01, 0x7c, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x80, 0x1f, 0x00, 0x80, 0x1f, 0x00, 0x00, 0x00, 0x00,
  0x00, 0xe0, 0x03, 0x18, 0xc0, 0x03, 0x38, 0xc0, 0x03, 0xf0, 0xf0, 0x03,
  0xe0, 0x7f, 0x02, 0x80, 0x1f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoRedo -data {
\#define redo_width 20
\#define redo_height 20
static unsigned char redo_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x1f, 0x00, 0xe0, 0x7f, 0x02,
  0xf0, 0xf0, 0x03, 0x38, 0xc0, 0x03, 0x18, 0xc0, 0x03, 0x00, 0xe6, 0x03,
  0x00, 0x06, 0x00, 0x80, 0x1f, 0x00, 0x80, 0x1f, 0x00, 0x00, 0x06, 0x00,
  0x7c, 0x06, 0x00, 0x3c, 0x80, 0x01, 0x3c, 0xc0, 0x01, 0xfc, 0xf0, 0x00,
  0xe4, 0x7f, 0x00, 0x80, 0x1f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoCut -data {
\#define cut_width 20
\#define cut_height 20
static unsigned char cut_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x08, 0x00,
  0x80, 0x08, 0x00, 0x80, 0x08, 0x00, 0x80, 0x0d, 0x00, 0x00, 0x05, 0x00,
  0x00, 0x07, 0x00, 0x00, 0x02, 0x00, 0x00, 0x07, 0x00, 0x00, 0x1d, 0x00,
  0xc0, 0x25, 0x00, 0x20, 0x25, 0x00, 0x20, 0x25, 0x00, 0x20, 0x19, 0x00,
  0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoCopy -data {
\#define copy_width 20
\#define copy_height 20
static unsigned char copy_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x00,
  0x84, 0x01, 0x00, 0x84, 0x02, 0x00, 0xb4, 0x3f, 0x00, 0x04, 0x61, 0x00,
  0xf4, 0xa1, 0x00, 0x04, 0xed, 0x01, 0xf4, 0x01, 0x01, 0x04, 0x7d, 0x01,
  0xfc, 0x01, 0x01, 0x00, 0x7d, 0x01, 0x00, 0x01, 0x01, 0x00, 0xff, 0x01,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoPaste -data {
\#define paste_width 20
\#define paste_height 20
static unsigned char paste_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x07, 0x00,
  0xf8, 0x7c, 0x00, 0x7c, 0xdb, 0x00, 0x2c, 0xb0, 0x00, 0xf4, 0xdf, 0x00,
  0xac, 0xaa, 0x00, 0x54, 0xff, 0x00, 0xac, 0xc1, 0x00, 0x54, 0x41, 0x01,
  0xac, 0xdd, 0x01, 0x54, 0x01, 0x01, 0xac, 0x7d, 0x01, 0xf8, 0x01, 0x01,
  0x00, 0xff, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoMark -data {
\#define mark_width 20
\#define mark_height 20
static unsigned char mark_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 0x1f, 0x00, 0x56, 0x15, 0x00,
  0xaa, 0x1a, 0x00, 0x56, 0x15, 0x00, 0xaa, 0x1a, 0x00, 0x56, 0x15, 0x00,
  0xaa, 0x1a, 0x00, 0xfe, 0x1f, 0x00, 0x00, 0xf0, 0x03, 0x00, 0x50, 0x03,
  0x00, 0xb0, 0x02, 0x00, 0x50, 0x03, 0x00, 0xb0, 0x02, 0x00, 0x50, 0x03,
  0x00, 0xb0, 0x02, 0x00, 0xf0, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoFind -data {
\#define find_width 20
\#define find_height 20
static unsigned char find_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x28, 0x00, 0x00,
  0x4c, 0x00, 0x00, 0x94, 0x00, 0x00, 0x28, 0x01, 0x00, 0x50, 0x02, 0x00,
  0xa0, 0x7c, 0x00, 0x40, 0xf1, 0x00, 0x80, 0x38, 0x00, 0x00, 0x1d, 0x00,
  0x00, 0x8f, 0x00, 0x00, 0x47, 0x01, 0x00, 0x43, 0x01, 0x00, 0xe2, 0x03,
  0x00, 0x20, 0x02, 0x00, 0x20, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoRepl -data {
\#define replace_width 20
\#define replace_height 20
static unsigned char replace_bits[] = {
  0x00, 0x00, 0x00, 0x08, 0x00, 0x00, 0x14, 0x00, 0x00, 0x26, 0x00, 0x00,
  0x4a, 0x00, 0x00, 0x94, 0x00, 0x00, 0x28, 0x01, 0x00, 0x50, 0x3e, 0x00,
  0xa0, 0x78, 0x00, 0x40, 0x1c, 0x00, 0x80, 0x8e, 0x03, 0x80, 0x87, 0x04,
  0x80, 0x83, 0x03, 0x80, 0x91, 0x04, 0x00, 0xa9, 0x04, 0x00, 0xa8, 0x03,
  0x00, 0x7c, 0x00, 0x00, 0x44, 0x00, 0x00, 0x44, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoComp -data {
\#define cc_width 20
\#define cc_height 20
static unsigned char cc_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3c, 0x00, 0x00,
  0x64, 0xfc, 0x00, 0xa4, 0x80, 0x00, 0x24, 0xe1, 0x03, 0xe4, 0xc3, 0x01,
  0x04, 0x82, 0x00, 0x04, 0x02, 0x00, 0x04, 0x02, 0x00, 0x04, 0xe2, 0x01,
  0x04, 0x32, 0x03, 0x04, 0x12, 0x02, 0x04, 0x12, 0x02, 0x04, 0x32, 0x03,
  0xfc, 0xe3, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoLink -data {
\#define ld_width 20
\#define ld_height 20
static unsigned char ld_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x78, 0x00, 0x00, 0xcc, 0x00, 0x00,
  0x84, 0xff, 0x00, 0x84, 0x80, 0x00, 0xcc, 0x80, 0x00, 0x78, 0xe0, 0x01,
  0x00, 0xb0, 0x03, 0x00, 0x50, 0x03, 0x00, 0xb0, 0x02, 0x00, 0x70, 0x03,
  0x78, 0xe0, 0x01, 0xcc, 0x80, 0x00, 0x84, 0x80, 0x00, 0x84, 0xff, 0x00,
  0xcc, 0x00, 0x00, 0x78, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoExec -data {
\#define exec_width 20
\#define exec_height 20
static unsigned char exec_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0xff, 0x00, 0x04, 0x80, 0x01,
  0x84, 0x87, 0x02, 0xc4, 0x8d, 0x02, 0xc4, 0x8a, 0x02, 0x44, 0x8d, 0x02,
  0xc4, 0x8e, 0x02, 0x84, 0x87, 0x02, 0x04, 0x80, 0x01, 0xfc, 0xff, 0x00,
  0x40, 0x08, 0x00, 0xf0, 0xff, 0x00, 0x08, 0x80, 0x01, 0xfc, 0x7f, 0x01,
  0x04, 0xc0, 0x00, 0xfc, 0x7f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoStop -data {
\#define stop_width 20
\#define stop_height 20
static unsigned char stop_bits[] = {
  0x00, 0x00, 0x00, 0xe0, 0x7f, 0x00, 0x20, 0x40, 0x00, 0x20, 0x46, 0x00,
  0x20, 0x4f, 0x00, 0x20, 0x4f, 0x00, 0x20, 0x46, 0x00, 0x20, 0x40, 0x00,
  0x20, 0x46, 0x00, 0x20, 0x49, 0x00, 0x20, 0x49, 0x00, 0x20, 0x46, 0x00,
  0x20, 0x40, 0x00, 0x20, 0x46, 0x00, 0x20, 0x49, 0x00, 0x20, 0x49, 0x00,
  0x20, 0x46, 0x00, 0x20, 0x40, 0x00, 0xe0, 0x7f, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoHelp -data {
\#define help_width 20
\#define help_height 20
static unsigned char help_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x1f, 0x00,
  0x40, 0x30, 0x00, 0x20, 0x27, 0x00, 0xa0, 0x29, 0x00, 0xa0, 0x24, 0x00,
  0x40, 0x12, 0x00, 0x00, 0x09, 0x00, 0x00, 0x05, 0x00, 0x00, 0x05, 0x00,
  0x00, 0x05, 0x00, 0x00, 0x06, 0x00, 0x00, 0x09, 0x00, 0x00, 0x09, 0x00,
  0x00, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoHome -data {
\#define home_width 20
\#define home_height 20
static unsigned char home_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x28, 0x07, 0x00, 0xb8, 0x0d, 0x00,
  0xf8, 0x1a, 0x00, 0x78, 0x35, 0x00, 0xb8, 0x6a, 0x00, 0x58, 0xd5, 0x00,
  0xac, 0xaa, 0x01, 0xfe, 0xff, 0x03, 0x04, 0x00, 0x01, 0xf4, 0x00, 0x01,
  0x94, 0x3c, 0x01, 0x94, 0x3c, 0x01, 0xf4, 0x34, 0x01, 0x04, 0x3c, 0x01,
  0x04, 0x3c, 0x01, 0x04, 0x3c, 0x01, 0xfc, 0xff, 0x01, 0x00, 0x00, 0x00,
  };
}
image create bitmap twoBack -data {
\#define back_width 20
\#define back_height 20
static unsigned char back_bits[] = {
  0x00, 0x18, 0x00, 0x00, 0x1c, 0x00, 0x00, 0x16, 0x00, 0x00, 0x1b, 0x00,
  0x80, 0x15, 0x00, 0xc0, 0x1a, 0x00, 0x60, 0xf5, 0x03, 0xb0, 0xaa, 0x02,
  0x58, 0x55, 0x03, 0xac, 0xaa, 0x02, 0xfc, 0xff, 0x03, 0xf8, 0xff, 0x03,
  0xf0, 0xff, 0x03, 0xe0, 0xff, 0x03, 0xc0, 0xff, 0x03, 0x80, 0x1f, 0x00,
  0x00, 0x1f, 0x00, 0x00, 0x1e, 0x00, 0x00, 0x1c, 0x00, 0x00, 0x18, 0x00,
  };
}
image create bitmap twoNext -data {
\#define next_width 20
\#define next_height 20
static unsigned char next_bits[] = {
  0x80, 0x01, 0x00, 0x80, 0x03, 0x00, 0x80, 0x06, 0x00, 0x80, 0x0d, 0x00,
  0x80, 0x1a, 0x00, 0x80, 0x35, 0x00, 0xfc, 0x6a, 0x00, 0x54, 0xd5, 0x00,
  0xac, 0xaa, 0x01, 0x54, 0x55, 0x03, 0xfc, 0xff, 0x03, 0xfc, 0xff, 0x01,
  0xfc, 0xff, 0x00, 0xfc, 0x7f, 0x00, 0xfc, 0x3f, 0x00, 0x80, 0x1f, 0x00,
  0x80, 0x0f, 0x00, 0x80, 0x07, 0x00, 0x80, 0x03, 0x00, 0x80, 0x01, 0x00,
  };
}

# array.tcl --
#
#	Extended array handling code. The two_array command knows the
#	options "load" and "save" and uses a more human-readable
#	formatting (key:value pairs one on a row, with several characters
#	being masked)
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoArray 1.0
package require twoUtil

namespace eval ::two::array {
    namespace export two_array
    namespace import ::two::util::two_now
}

# two_array --
#
#	wrapper around tcl array, which is exteded with a
#	load and save command.
#
# Arguments:
#	see manpage

proc ::two::array::two_array {opt args} {
    switch -exact -- $opt {
        load {
	    return [eval LoadArray $args]
	}
	save {
	    return [eval SaveArray $args]
	}
	default {
	    return [uplevel array $opt $args]
	}
    }
    return
}


# ::two::array::LoadArray --
#
#	Read and parse simple array file: set elements of array "arrayName"
#	with "key:value" pairs from array file "fileName".
#	If force is set and no array file is found then there will be an error
#	message on stderr, an abort message on stdout and the program will exit.
#
# Arguments:
#	fileName	Name of array file
#	arrayName	name of array which will contain the configuration
#	force		1 if config file is required to be already there
#
# Results:
#	1 if successfull, error otherwise

proc ::two::array::LoadArray {fileName arrayName {force 0}} {
    upvar 2 $arrayName a

    if [ catch {
        set arrayFile [two_readFile -nonewline $fileName]
    } errorMsg ] {
        if {$force} {
            # We were told that the array file must exist
            puts stderr "Failed to read `$fileName' at [two_now]\n  -> `$errorMsg'"
            puts "aborted: [two_now]"
            exit
        } else {
            return 0
        }
    } else {
        # Read config file
	foreach line [split $arrayFile \n] {
	    foreach {key val} [split $line ":"] {break}
	    # undo coding
	    regsub -all "%n" $key "\n" key
	    regsub -all "%n" $val "\n" val
	    regsub -all "%d" $key ":"  key
	    regsub -all "%d" $val ":"  val
	    regsub -all "%p" $key "%"  key
	    regsub -all "%p" $val "%"  val
	    # set array
            set a($key) $val
	}
    }

    return 1
}

# ::two::array::SaveArray --
#
#	Save array to file: the elements of array "arrayName"
#	will be written as "key:value" pairs to array file "fileName".
#	Will put an error message on stderr if writing failes.
#
# Arguments:
#	fileName	Name of array file
#	configName	name of array which will contain the configuration
#
# Results:
#	1 if successfull, error otherwise

proc ::two::array::SaveArray {fileName arrayName} {
    upvar 2 $arrayName a
    set ok 1

    if [ catch {
        set arrayFile [open $fileName w]
        foreach key [lsort [array names a]] {
	    set val $a($key)
	    # coding: "%" => "%p" and ":" => "%d", "\n" => "%n"
	    # so only one ":" per line
	    regsub -all "%" $key  "%p" key
	    regsub -all "%" $val  "%p" val
	    regsub -all ":" $key  "%d" key
	    regsub -all ":" $val  "%d" val
	    regsub -all "\n" $key "%n" key
	    regsub -all "\n" $val "%n" val
	    # write to file
            puts $arrayFile "$key:$val"
        }
        close $arrayFile
    } errorMsg ] {
        # error message if write failed
        set ok 0
    	puts stderr "Failed to write `$fileName' at [two_now]\n  -> `$errorMsg'"
    }
    return $ok
}

namespace import ::two::array::*


# autosb.tcl --
#
#	Automatic scrollbar. Behaves like a regular scrollbar, but
#	disappears if not needed.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoAutosb 1.0
package require twoUtil

option add *TwoAutoSb.borderWidth 0 widgetDefault
option add *TwoAutoSb.takeFocus   0 widgetDefault

namespace eval ::two::autosb {
    namespace export two_autosb
    variable Autosb
    
    # variable Autosb
    #
    # $w,shown -	true if scrollbar is on
    # $w,blockS -	true if scrollbar has recently been added and should
    #		stay the same for some time to avoid flicker
    # $w,blockH -	dito for recently hidden scrollbar
}

# ::two::autosb::two_autosb --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.sb"); the original widget procedure
#	is renamed to "_two$w" (like "_cb.tl.sb") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::autosb::two_autosb {w args} {
    variable Autosb

    # Initialise the globals
    array set Autosb [list \
        $w,shown  0 \
        $w,blockS 0 \
        $w,blockH 0 \
    ]

    # Container frame
    frame $w -class TwoAutoSb

    # the original scrollbar
    uplevel [list scrollbar $w.sb] $args
    
    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Autosb
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # Return the correct result...
    return $w
}

# ::two::autosb::Widget --
#
#	replacement for widget command.
#	One tricky part here is that we block changes in the state
#	of the autosb (i.e. show/hide) for 200 ms after each change.
#	This is needed e.g. for horizintal scrollbars in text widgets:
#	Take a horizontal scrollbar whose heigth is at least that of
#	a text line, so showing the scrollbar will cover the bottom
#	visible line. Make this bottom line the only one which is
#	long enough to require a horizontal scrollbar. So, showing
#	the scrollbar removes the need for a scrollbar, and making it
#	disappear again creates the need for a scrollbar. Bummer -
#	infinit flicker. The 200 ms blocking time is just a wild
#	guess, on a very slow machine (or slow X connection, maybe
#	with -sync) this might not work. But in those cases the whole
#	application would be ratther "jumpy".
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::autosb::Widget {w opt args} {
    variable Autosb

    # See if this option is for us...
    if {[string match "set" $opt] && [llength $args] == 2} {
	set off [expr {[lindex $args 0] == 0 && [lindex $args 1] == 1}]
	if {$off && $Autosb($w,shown) && !$Autosb($w,blockS)} {
	    pack forget $w.sb
	    _two$w configure -width 1 -height 1
	    set Autosb($w,shown)  0
	    set Autosb($w,blockS) 1
	    after 200 [list set [namespace current]::Autosb($w,blockS) 0]
	} elseif {!$off && !$Autosb($w,shown) && !$Autosb($w,blockH)} {
	    pack $w.sb -fill both -expand 1
	    set Autosb($w,shown)  1
	    set Autosb($w,blockH) 1
	    after 200 [list set [namespace current]::Autosb($w,blockH) 0]
	}
    }
    return [uplevel [list $w.sb $opt] $args]
}

# ::two::autosb::_Destroy --
#
#	Call this procedure when the autosb widget is destroyed.
#	Will destroy the original widget command.
#
# Arguments:
#	w		Path to Widget

proc ::two::autosb::_Destroy {w} {
    variable Autosb

    # clear Autosb
    foreach key [array names Autosb "$w,*"] {
        unset Autosb($key)
    }
    # remove the original widget command
    rename $w ""
    return
}

namespace import ::two::autosb::*

# button.tcl --
#
#	Wrappers around several tk buttons and the tk label. Adds
#	some aliases for widget options and parses the label setting
#	for "&" to generate key bindings.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoButton 1.0
package require twoUtil

namespace eval ::two::button {
    namespace export two_*
    variable Button
    
    namespace import ::two::util::two_parseArgs
}

# ::two::button::two_button --
#
#	extends button; see AnyButton

proc ::two::button::two_button {args} {
    return [eval AnyButton button $args]
}

# ::two::button::two_label --
#
#	extends label; see AnyButton

proc ::two::button::two_label {args} {
    return [eval AnyButton label $args]
}

# ::two::button::two_checkbutton --
#
#	extends checkbutton; see AnyButton

proc ::two::button::two_checkbutton {args} {
    return [eval AnyButton checkbutton $args]
}

# ::two::button::AnyButton --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".b"); the original widget procedure
#	is renamed to "_two$w" (like "_two.b") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#
# Arguments:
#	cmd		What command should be used? (button, checkbutton, label)
#	w		Widget path
#	args		Widget arguments and options

proc ::two::button::AnyButton {cmd w args} {
    # several ways to specify button text

    set idx  -1
    set text ""
    two_parseArgs args text "-text"
    two_parseArgs args text "-label"
    two_parseArgs args text "-l"
    two_parseArgs args text "-t"
    if {[string length $text]} {
	set idx [string first "&" $text]
	if {$idx >= 0} {
	    set underline $idx
	    # ignore "&&"
	    while {[string match "&" [string index $text [expr {$idx + 1}]]]} {
		set base [expr {$idx + 2}]
		set idx  [string first "&" [string range $text $base end]]
		if {$idx < 0} {
		    break
		} else {
		    set underline [expr {$underline + $idx + 1}]
		    incr idx $base
		}
	    }
	}
	if {$idx >= 0} {
	    regsub -all -- {&([^&])} $text {\1} text
	    lappend args -text $text -underline $idx
	} else {
	    lappend args -text $text
	}
    }

    # Create button
    uplevel [list $cmd $w] $args

    # Create binding
    if {$idx >= 0} {
        set key "[string index $text $idx]"
	set tl [winfo toplevel $w]
	bind $tl <Alt-Key-[string tolower $key]> [list two_invoke $w]
	bind $tl <Alt-Key-[string toupper $key]> [list two_invoke $w]
    }

    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "

    # catch widget destruction in order to clean up Combobox
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # Return the correct result...
    return $w
}

# ::two::button::Widget --
#	replacement for widget command
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::button::Widget {w opt args} {
    variable Button

    # See if this option is for us...
    switch -exact -- $opt {
	cget {
	    switch -exact -- [lindex $args 0] {
	        -label -
		-l       { return [eval [list _two$w $opt] [lreplace $args 0 0 "-text"]] }
		default  { return [eval [list _two$w $opt] $args] }
	    }
	}
	configure {
	    set text ""
	    two_parseArgs args text "-text"
	    two_parseArgs args text "-label"
	    two_parseArgs args text "-l"
	    if {[string length $text]} {
        	set idx [string first "&" $text]
		if {$idx >= 0} {
		    # remove old binding
    		    if {[info exists Button($w,bind)]} {
		        bind [winfo toplevel $w] <$Button($w,bind)> {#nothing}
		    }
		    regsub "&" $text "" text
		    lappend args -text $text -underline $idx
		    # Create binding
		    set key "Alt-Key-[string tolower [string index $text $idx]]"
                    bind [winfo toplevel $w] <$key> [list two_invoke $w]
		    set Button($w,bind) $key
		} else {
		    lappend args -text $text
		}
	    }
	    return [eval [list _two$w $opt] $args]
	}
	default {
            return [eval [list _two$w $opt] $args]
	}
    }
    return
}

# ::two::button::_Destroy --
#
#	Call this procedure when the two_menu widget is destroyed.
#	Will destroy the original widget command.
#
# Arguments:
#	w		Path to Widget

proc ::two::button::_Destroy {w} {
    # remove the original widget command
    rename $w ""
    return
}

namespace import ::two::button::*

# bugmail.tcl --
#
#	bgerror replacement which enables the user to mail the stack
#	trace (and some more information) to the author of the
#	application (or whoever has been configured).
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoBugmail 1.0
package require twoUtil

option add *TwoBugMail.message.txt.wrapLength   4i      widgetDefault
option add *TwoBugMail.message.bmp.bitmap       error   widgetDefault
option add *TwoBugMail.details.txt.relief       sunken  widgetDefault
option add *TwoBugMail.details.txt.borderWidth  1       widgetDefault
option add *TwoBugMail.details.txt.width        60      widgetDefault
option add *TwoBugMail.details.txt.height       20      widgetDefault
option add *TwoBugMail.txt.relief               sunken  widgetDefault
option add *TwoBugMail.txt.borderWidth          1       widgetDefault
option add *TwoBugMail.txt.width                40      widgetDefault
option add *TwoBugMail.txt.height               10      widgetDefault

namespace eval ::two::bugmail {
    namespace export two_bugmail
    variable Bugmail
    
    namespace import ::two::util::two_parseArgs
    
    # variable Bugmail
    #
    # from -	mail from
    # to -		mail to
    # cc -		mail cc
    # bcc -		mail bcc
    # subject -	mail subject
    # detail -	enable/disble viewing of error details
    # version -	additinal infos about the version of the software used
    # info -	more info about current version
    # installed -	true if bugmail is installed
    # button -	name of button pressed in dialog
    
    array set Bugmail {
        from		""
        to		""
        cc		""
        bcc		""
        subject		"[Error] %M"
        detail		1
        version		"unknown"
        installed	0
        button		""
        info		""
    }
    
    # variable BugmailMsg
    #
    
    array set BugmailMsg {
        errorTitle	"Internal error"
        errorIntro	"Internal error: "
        bugTitle	"Bug report"
        bugInfo	"You can mail a bug report to the author of this Software."
        mail	"Mail bug report"
        send	"Send mail"
        description	"Please enter a short description for the error. When and where did it happen?"
        details	"Details"
        ok		"Ok"
        cancel	"Cancel"
        ignore	"Ignore"
    }
}

# ::two::bugmail::two_bugmail --
#
#	create widget ._bugmail if it doesn't exist and parse argumets.
#
# Arguments:
#	opt		Method
#	args		Options

proc ::two::bugmail::two_bugmail {opt args} {
    variable Bugmail
    variable BugmailMsg

    switch -exact -- $opt {
        install {
	    two_parseArgs args Bugmail(from)    "-from"
	    two_parseArgs args Bugmail(to)      "-to"
	    two_parseArgs args Bugmail(cc)      "-cc"
	    two_parseArgs args Bugmail(bcc)     "-bcc"
	    two_parseArgs args Bugmail(subject) "-subject"
	    two_parseArgs args Bugmail(detail)  "-detail"
	    two_parseArgs args Bugmail(version) "-version"
	    two_parseArgs args Bugmail(info)    "-info"
	    if {[two_parseArgs args msg "-messages"]} {
	        array set BugmailMsg $msg
	    }
	    if [llength $args] {
	        error "Unknown arguments \"$args\""
	    }
	    if {!$Bugmail(installed)} {
		namespace eval :: "
        	    proc bgerror {error} {
			[namespace code ErrorHandler] \$error
		    }
		"
		set Bugmail(installed) 1
	    }
	}
	configure {
	    two_parseArgs args Bugmail(from)    "-from"
	    two_parseArgs args Bugmail(to)      "-to"
	    two_parseArgs args Bugmail(cc)      "-cc"
	    two_parseArgs args Bugmail(bcc)     "-bcc"
	    two_parseArgs args Bugmail(subject) "-subject"
	    two_parseArgs args Bugmail(detail)  "-detail"
	    two_parseArgs args Bugmail(version) "-version"
	    two_parseArgs args Bugmail(info)    "-info"
	    if {[two_parseArgs args msg "-messages"]} {
	        array set BugmailMsg $msg
	    }
	    if {[llength $args]} {
	        error "Unknown arguments \"$args\""
	    }
	}
	cget {
	    switch -exact -- $args {
	        -from     { return $Bugmail(from)    }
	        -to       { return $Bugmail(to)      }
	        -cc       { return $Bugmail(cc)      }
	        -bcc      { return $Bugmail(bcc)     }
	        -subject  { return $Bugmail(subject) }
	        -detail   { return $Bugmail(detail)  }
	    	-version  { return $Bugmail(version) }
	    	-info     { return $Bugmail(info)    }
	    	-messages { return [array get BugmailMsg]   }
		default   { error "Unkown option \"$args\"" }
	    }
	}
	remove {
	    if {$Bugmail(installed)} {
	        rename bgerror ""
		set Bugmail(installed) 0
	    }
	}
	comment {
	    ErrorHandler
	}
	default {
	    error "Unknown method \"$opt\""
	}
    }
    return
}

# ::two::bugmail::ErrorHandler --
#
#	takes error and shows description. O.k., it's a lengthy procedure:-)
#	Depending on configuration, this handler will allow the user to
#	send a bug report via e-mail. Currently, only UNIX using sendmail
#	is supported; directly implementing SMTP would need to configure
#	the smtp host - having to configure something in order to be able to
#	send a bug report isn't the best solution.
#	If no error is given, the user can still send a bug report.
#	See comments in code for single steps...
#
# Arguments:
#	err		Short error message

proc ::two::bugmail::ErrorHandler { {err ""} } {
    variable Bugmail
    variable BugmailMsg
    global errorInfo argv0 argv tcl_patchLevel tk_patchLevel

    # ------------------
    # Show error message
    # ------------------

    # bug text
    if {[string length $err]} {
	set title   $BugmailMsg(errorTitle)
	set text    $BugmailMsg(errorIntro)$err
	set details $errorInfo
	set errmsg  $err
	set ok      $BugmailMsg(ok)
    } else {
        set title   $BugmailMsg(bugTitle)
	set text    $BugmailMsg(bugInfo)
	set details "none"
	set errmsg  "Bug report"
	set ok      $BugmailMsg(cancel)
    }

    set w .twoBugmailWindow
    # remove old instances
    catch {destroy $w}
    # build widget
    toplevel     $w -class TwoBugMail
    wm minsize   $w 430 150
    wm title     $w $title
    wm iconname  $w $title
    wm transient $w .

    # 3 parts, details originally not visible    
    pack [frame $w.message -relief raised -bd 1] -side top -expand 1 -fill both
    frame $w.details -relief raised -bd 1
    pack [frame $w.buttons -relief raised -bd 1] -side bottom -fill both
    set msg $w.message
    set but $w.buttons
    set det $w.details
    
    # upper part: error message
    pack [label $msg.bmp] -side left -padx 3m -pady 3m
    pack [label $msg.txt -justify left -text $text] \
        -side right -expand 1 -fill both -padx 3m -pady 3m

    # middle part: details
    grid [text $det.txt -xs "$det.sx set" -ys "$det.sy set" -wrap none -takefocus 0] \
         -row 0 -column 0 -sticky nsew
    grid [scrollbar $det.sy -command "$det.txt yview"] \
         -row 0 -column 1 -sticky ns
    grid [scrollbar $det.sx -command "$det.txt xview" -orient horizontal] \
         -row 1 -column 0 -sticky ew
    grid rowconfigure    $det 0 -weight 1
    grid columnconfigure $det 0 -weight 1
    bindtags $det.txt [list $det.txt $w all]

    # fill in bug info (stack trace)
    $det.txt insert 1.0 $details
    $det.txt mark set insert 1.0

    # lower part: buttons
    button $but.ok -text $ok -underline 0 -default active -command \
        "[namespace code set] Bugmail(button) ok"
    button $but.ignore -text $BugmailMsg(ignore) -underline 0 -command \
        "[namespace code set] Bugmail(button) ignore"
    button $but.details -text $BugmailMsg(details) -underline 0 -command "
        pack $msg -fill x -expand 0
	pack forget $but.details
	pack $det -side bottom -expand 1 -fill both
    "
    button $but.mail -text $BugmailMsg(mail) -underline 0 -command \
        "[namespace code set] Bugmail(button) mail"
    if {$Bugmail(detail) && [string length $err]} {
	pack $but.ok $but.details $but.mail \
            -side left -padx 5 -fill x -expand 1
    } else {
	pack $but.ok $but.mail \
            -side left -padx 5 -fill x -expand 1
    }
    # ignore not packages: "return -code break" doesn't seem to work:-(

    # bindings for buttons
    foreach b {ok ignore details mail} {
        set key [string tolower [string index [$but.$b cget -text] 0]]
        bind $w <Alt-$key> "two_invoke $but.$b"
    }
    bind $w <Return>  "two_invoke $but.ok"
    bind $w <Escape>  "two_invoke $but.ignore"
    bind $w <Destroy> "[namespace code set] Bugmail(button) ignore"

    # Center the window on the screen.
    wm withdraw $w
    update idletasks
    set x [expr {[winfo screenwidth  $w]/2 - [winfo reqwidth  $w]/2 - [winfo vrootx $w]}]
    set y [expr {[winfo screenheight $w]/2 - [winfo reqheight $w]/2 - [winfo vrooty $w]}]
    wm geom $w +$x+$y
    wm deiconify $w

    # Grab and claim focus
    grab  $w
    focus $w
    
    # Wait for response
    tkwait variable [namespace current]::Bugmail(button)
    
    # User didn't want to send mail:
    if {[string match "ok" $Bugmail(button)]} {
	# Clean up
	catch {
            bind $w <Destroy> {}
            destroy $w
	}
        return
    } elseif {[string match "ignore" $Bugmail(button)]} {
	catch {
            bind $w <Destroy> {}
            destroy $w
	}
        return
    }

    # ---------------
    # Send bug report
    # ---------------

    # adopt window
    $msg.txt configure -text $BugmailMsg(description)
    pack $msg -fill x -expand 0
    $det.txt delete 1.0 end
    $det.txt configure -takefocus 1
    bindtags $det.txt [list $det.txt Text $w all]

    $but.ok     configure -text $BugmailMsg(send)
    $but.ignore configure -text $BugmailMsg(cancel)

    # clear former bindings
    foreach binding [bind $w] {
        bind $w $binding {}
    }

    # new bindings
    foreach b {ok ignore} {
        set key [string tolower [string index [$but.$b cget -text] 0]]
        bind $w <Alt-$key> "two_invoke $but.$b"
    }
    bind $det.txt <Escape> "two_invoke $but.ignore"
    focus $det.txt
    pack $det -side bottom -expand 1 -fill both
    pack forget $but.ok $but.ignore $but.details $but.mail
    pack $but.ok $but.ignore \
        -side left -padx 5 -fill x -expand 1

    # Wait for response
    tkwait variable [namespace current]::Bugmail(button)

    # get Users's description
    set description [$det.txt get 1.0 "end - 1c"]

    # Clean up
    catch {
        bind $w <Destroy> {}
        destroy $w
    }

    # User didn't want to send mail:
    if {[string match "ignore" $Bugmail(button)]} {
        return
    }

    # Send mail
    set fd [open "| /usr/lib/sendmail -oi -t" "w"]
    if {[string length $Bugmail(to)]} {
        puts $fd  "To: $Bugmail(to)"
    }
    if {[string length  $Bugmail(from)]} {
        puts $fd "From: $Bugmail(from)"
    }
    if {[string length $Bugmail(cc)]} {
        puts $fd  "Cc: $Bugmail(cc)"
    }
    if {[string length $Bugmail(bcc)]} {
        puts $fd  "Cc: $Bugmail(bcc)"
    }
    set subject "Error: $errmsg"
    if {[string length   $Bugmail(subject)]} {
	regsub -all {%M} $Bugmail(subject) $errmsg subject
    }
    puts $fd \
"Subject: $subject
Date: [clock format [clock seconds]]

The following error appeared:
-----------------------------------------------------------------
$errmsg
-----------------------------------------------------------------

Where and when:
-----------------------------------------------------------------
Prog: $argv0 $argv
Vers: $Bugmail(version)
Time: [clock format [clock seconds]]
Host: [info host]
PWD:  [pwd]
Tcl:  $tcl_patchLevel
Tk:   $tk_patchLevel
-----------------------------------------------------------------

Author-supplied info:
-----------------------------------------------------------------
$Bugmail(info)
-----------------------------------------------------------------

User-supplied comment:
-----------------------------------------------------------------
$description
-----------------------------------------------------------------

Stack trace:
-----------------------------------------------------------------
$details
-----------------------------------------------------------------
"

    close $fd

    return
}

namespace import ::two::bugmail::*

# combobox.tcl --
#
#	Combobox widget. A combobox is like an editable option menu.
#	Consists of an entry, a button and a dropdown list.
#
# Copyright (c) 1998, 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoCombobox 1.0
package require twoUtil
package require twoAutosb

option add *TwoComboBox.borderWidth               2       widgetDefault
option add *TwoComboBox.drop.borderWidth          0
option add *TwoComboBox.drop.highlightThickness   1
option add *TwoComboBox.drop.highlightColor       black
option add *TwoComboBox.drop.relief               flat
option add *TwoComboBox.relief                    sunken  widgetDefault
option add *TwoComboBox.Button.highlightThickness 1
option add *TwoComboBox.Entry.borderWidth         0

# Workaround: Tk needs these procedures for focus traversion
# but won't load it form inside the namespace (bug in tcl8.0 up to p1)
catch {auto_load tk_focusNext}
catch {auto_load tk_focusPrev}

namespace eval ::two::combobox {
    namespace export two_combobox
    variable Combobox
    
    namespace import ::two::util::two_parseArgs
    
    # variable Combobox
    #
    # current -	path to currently opened combobox
    # buttonPress -	remember if buttonPress-1 has happend after opening drop
    # $w,command -	command that will be executed when opening the combobox
    # $w,state -	current state {normal|fixed|disabled}
    # $w,takefocus - should the combobox take part in keyboard focus traversal?
    # $w,style -	look of the combobox {windows|unix}
    
    array set Combobox {
        Drag	0
        buttonPress 0
        current	""
    }
}

image create bitmap comboBoxArrowWin -data {
\#define cbwin_width 9
\#define cbwin_height 10
static unsigned char cbwin_bits[] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x7c, 0x00, 0x38, 0x00,
  0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
}

image create bitmap comboBoxArrowX -data {
\#define cbarrow_width 9
\#define cbarrow_height 12
static unsigned char cbarrow_bits[] = {
  0x38, 0x00, 0x38, 0x00, 0x38, 0x00, 0x38, 0x00, 0xff, 0x01, 0xfe, 0x00,
  0x7c, 0x00, 0x38, 0x00, 0x10, 0x00, 0x00, 0x00, 0xff, 0x01, 0xff, 0x01,
  };
}

# ::two::combobox::two_combobox --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.cb"); the original widget procedure
#	is renamed to "_two$w" (like "_two.tl.cb") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#	combobox creates a toplevel widget "$w.drop" for the drop box.
#	Arguments are passed to entry widget.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::combobox::two_combobox {w args} {
    variable Combobox
    global tcl_platform

    # Initialise the globals
    array set Combobox "
	$w,command  {}
	$w,state    normal
	$w,style    $tcl_platform(platform)
	$w,takefocus 1
    "
    
    # This is a multi-style widget
    set ::two::SetStyleCmds($w) [list $w configure -style]

    # See if this option is for us...
    two_parseArgs args Combobox($w,command)   "-command"
    two_parseArgs args Combobox($w,state)     "-state"
    two_parseArgs args Combobox($w,style)     "-style"
    two_parseArgs args Combobox($w,takefocus) "-takefocus"

    # Container frame
    frame $w -class TwoComboBox
    set tl [winfo toplevel $w]
    uplevel [list entry $w.entry] $args
    if {![string match "unix" $Combobox($w,style)]} {
        button $w.button -image comboBoxArrowWin
    } else {
        button $w.button -image comboBoxArrowX
    }
    switch -exact -- $Combobox($w,state) {
	fixed {
            $w.entry configure  -state "disabled"
	    $w.button configure -takefocus $Combobox($w,takefocus)
	}
	disabled {
            $w.entry  configure -state "disabled"
            $w.button configure -state "disabled"
	}
	normal -
	default {
            $w.entry  configure -takefocus $Combobox($w,takefocus)
	    $w.button configure -takefocus 0
	}
    }
    pack $w.button -fill both -side right
    pack $w.entry -expand true -fill both -side left
    # Use our own bindings
    bindtags $w.entry  [list $w.entry  TwoComboBoxEntry  Entry $tl all]
    bindtags $w.button [list $w.button TwoComboBoxButton       $tl all]

    # Accompaining dropdown (toplevel widget)
    set d $w.drop
    toplevel $d -class TwoComboBox
    wm withdraw $d
    wm overrideredirect $d 1
    two_autosb $d.sx -command [list $d.lb xview] -orient horizontal
    two_autosb $d.sy -command [list $d.lb yview]
    listbox $d.lb -exportselection 0 \
        -xscrollcommand [list $d.sx set] \
        -yscrollcommand [list $d.sy set]
    frame $d.handle -height 4 -cursor bottom_side -bd 0

    grid $d.lb $d.sy -sticky nsew
    grid $d.sx       -sticky ew
    grid $d.handle   -sticky ew -columnspan 2
    grid rowconfigure    $d 0 -weight 1
    grid columnconfigure $d 0 -weight 1

    # Use our own bindings
    # Caveat: combobox bindings have to appear after normal widget bindings
    bindtags $d.lb [list $d.lb Listbox TwoComboBoxList TwoComboBox $tl all]
    # Handle for drop down list uses only its own bindings
    bindtags $d.handle [list $d.handle TwoComboBoxHandle $tl all]

    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Combobox
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # Return the correct result...
    return $w
}

# ::two::combobox::Widget --
#
#	replacement for widget command
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::combobox::Widget {w opt args} {
    variable Combobox

    # See if this option is for us...
    switch -exact -- $opt {
	configure {
	    two_parseArgs args Combobox($w,command)   "-command"
	    two_parseArgs args Combobox($w,state)     "-state"
	    two_parseArgs args Combobox($w,takefocus) "-takefocus"
	    switch -exact -- $Combobox($w,state) {
		fixed {
        	    $w.entry  configure -state "disabled" -takefocus 0
        	    $w.button configure -state "normal"   -takefocus $Combobox($w,takefocus)
		}
		disabled {
        	    $w.entry  configure -state "disabled" -takefocus 0
        	    $w.button configure -state "disabled" -takefocus 0
		}
		normal -
	        default {
        	    $w.entry  configure -state "normal" -takefocus $Combobox($w,takefocus)
        	    $w.button configure -state "normal" -takefocus 0
		}
	    }
	    if {[two_parseArgs args Combobox($w,style) "-style"]} {
		if {![string match "unix" $Combobox($w,style)]} {
		    $w.button configure -image comboBoxArrowWin
		} else {
		    $w.button configure -image comboBoxArrowX
		}
	    }

	    if {[llength $args]} {
		return [eval [list _two$w $opt] $args]
	    }
	}
	cget {
	    switch -exact -- $args {
	        -command { return $Combobox($w,command) }
	        -state   { return $Combobox($w,state)   }
	        -style   { return $Combobox($w,style)   }
		default  { return [eval [list _two$w $opt] $args] }
	    }
	}
	setFocus {
	    switch -exact -- $Combobox($w,state) {
		fixed {
		    focus $w.button
		}
		disabled {
		    focus $w
		}
		normal -
	        default {
        	    focus $w.entry
		}
	    }
	}
	get {
	    return [$w.entry get]
	}
	set {
    	    set state [$w.entry cget -state]
	    $w.entry configure -state normal
	    $w.entry delete 0 end
	    $w.entry insert 0 [lindex $args 0]
	    $w.entry configure -state $state
	}
	setlist {
	    $w.drop.lb delete 0 end
	    eval [list $w.drop.lb insert end] $args
	}
	getlist {
	    return [$w.drop.lb get 0 end]
	}
	add {
	    eval [list $w.drop.lb insert end] $args
	}
	open {
	    Open $w
	}
	entry {
	    return $w.entry
	}
	listbox {
	    return $w.drop.lb
	}
	default {
            return [eval [list _two$w $opt] $args]
	}
    }
    return
}

# ::two::combobox::_Destroy --
#
#	Call this procedure when the combobox widget is destroyed.
#	Will clean widget-specific Elements of Combobox.
#
# Arguments:
#	w		Path to Widget

proc ::two::combobox::_Destroy {w} {
    variable Combobox

    unset ::two::SetStyleCmds($w)

    # clear Combobox
    foreach key [array names Combobox "$w,*"] {
        unset Combobox($key)
    }
    # remove the original widget command
    rename $w ""
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoComboBox-class bindings
#-------------------------------------------------------------------------

# Entry widget in combobox

bind TwoComboBoxEntry <1> {
    if {[string match "fixed" $::two::combobox::Combobox([winfo parent %W],state)]} {
        ::two::combobox::Open [winfo parent %W]
    }
}
bind TwoComboBoxEntry <Next> {
    ::two::combobox::Open [winfo parent %W]
}

# drop down button

bind TwoComboBoxButton <1> {
    if {![string match "disabled" [%W cget -state]]} {
        %W configure -relief sunken
        ::two::combobox::Open [winfo parent %W]
        set ::two::combobox::Combobox(buttonPress) 0
    }
}
bind TwoComboBoxButton <space> {
    %W configure -relief sunken
    ::two::combobox::Open [winfo parent %W]
    set ::two::combobox::Combobox(buttonPress) 0
}
bind TwoComboBoxButton <Next> {
    %W configure -relief sunken
    ::two::combobox::Open [winfo parent %W]
    set ::two::combobox::Combobox(buttonPress) 0
}
bind TwoComboBoxButton <Enter> {
    if {![string match "disabled" [%W cget -state]]} {
        %W configure -state active
    }
}
bind TwoComboBoxButton <Leave> {
    if {![string match "disabled" [%W cget -state]]} {
        %W configure -state normal
    }
}

# Listbox widget in dropped down combobox

bind TwoComboBoxList <Return> {
    ::two::combobox::Close $::two::combobox::Combobox(current)
    break
}
bind TwoComboBoxList <Escape> {
    %W selection clear 0 end
    ::two::combobox::Close $::two::combobox::Combobox(current)
    break
}

# Handle for resizing drop-down list

bind TwoComboBoxHandle <Enter> {
    ::two::combobox::HandleMotion %W %x
}
bind TwoComboBoxHandle <Motion> {
    ::two::combobox::HandleMotion %W %x
}
bind TwoComboBoxHandle <1> {
    ::two::combobox::HandleB1 %W %x %y
}
bind TwoComboBoxHandle <ButtonRelease-1> {
    set ::two::combobox::Combobox(Drag) 0
}

bind TwoComboBoxHandle <B1-Motion> {
    ::two::combobox::HandleB1Motion %W %x %X %y %Y
}

# Bindings to the frames only trigger when the listbox is
# opened (dropped down)

bind TwoComboBox <ButtonPress> {
    set w [winfo containing [winfo pointerx .] [winfo pointery .]]
    if {[string match $w $::two::combobox::Combobox(current).button]} {
        set ::two::combobox::Combobox(buttonPress) 1
    }
}
bind TwoComboBox <ButtonRelease-1> {
    set w [winfo containing [winfo pointerx .] [winfo pointery .]]
    if {$::two::combobox::Combobox(buttonPress) ||
        ![string match $w $::two::combobox::Combobox(current).button]} {
        if {[winfo exists $::two::combobox::Combobox(current)]} {
            ::two::combobox::Close $::two::combobox::Combobox(current)
        }
    }
}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::combobox::HandleMotion --
#
#	handle mouse cursor motion events within handle frame. Set
#	the mouse cursor shape according to the relative position.
#
# Arguments:
#	h		Path to handle
#	x		relative X Position of mouse cursor within handle (%x)

proc ::two::combobox::HandleMotion {h x} {
    variable Combobox

    if {!$Combobox(Drag)} {
        if {$x > [winfo width $h] - 10} {
            $h configure -cursor bottom_right_corner
        } elseif {$x < 10} {
            $h configure -cursor bottom_left_corner
        } else {
            $h configure -cursor bottom_side
        }
    }
    return
}

# ::two::combobox::HandleB1 --
#
#	handle mosue button press within within handle frame.
#
# Arguments:
#	h		Path to handle
#	x		relative X Position of mouse cursor within handle (%x)
#	y		relative Y Position (%y)

proc ::two::combobox::HandleB1 {h x y} {
    variable Combobox

    set p [winfo toplevel $h]
    set ht [$p cget -highlightthickness]
    set Combobox(Drag) 1
    set Combobox(Cursor) [$h cget -cursor]
    set Combobox(dxl) [expr {$x + $ht}]
    set Combobox(dxr) [expr {[winfo width  $h] - $x + $ht}]
    set Combobox(x) [expr {[winfo rootx $p] + [winfo width  $h] + $x + 3 * $ht}]
    set Combobox(y) [expr {[winfo rooty $p] - [winfo height $h] + $y - $ht}]

    return
}

# ::two::combobox::HandleB1Motion --
#
#	handle interactive resizing of dropdown list.
#	This relies on the mouse cursor's shape set in the
#	HandleMotion procedure!
#
# Arguments:
#	h		Path to handle
#	relX		relative X Position of mouse cursor within border (%x)
#	absX		absolute X Positino (%X)
#	relY		relative Y Position (%y)
#	absY		absolute Y Position (%Y)

proc ::two::combobox::HandleB1Motion {h relX absX relY absY} {
    variable Combobox

    if {$Combobox(Drag)} {
        set p [winfo parent $h]

	if {[string match "*right*" $Combobox(Cursor)]} {
	    set width [expr {$relX + $Combobox(dxr)}]
	} else {
	    set width [winfo width $p]
	}
	if {[string match "*left*" $Combobox(Cursor)]} {
	    set x [expr {$absX - $Combobox(dxl)}]
	    set width [expr {$Combobox(x) - $absX}]
	} else {
	    set x [winfo rootx $p]
	}
        set y [winfo rooty $p]
        set height [expr {$absY - $Combobox(y)}]
        if {$width < 50} {
            set width 50
        }
        if {$height < 50} {
            set height 50
        }
        wm geometry $p ${width}x${height}+${x}+${y}
    }
    return
}

# ::two::combobox::Open --
#
#	Open the combobox, i.e. show the dropdown listbox.
#
# Arguments:
#	c		Pathname of combobox widget

proc ::two::combobox::Open {c} {
    variable Combobox
    
    if {[string match "disabled" $Combobox($c,state)]} {
        return
    }

    set Combobox(current) $c
    if {[string length $Combobox($c,command)]} {
        namespace eval :: $Combobox($c,command)
    }

    set x [winfo rootx $c]
    set y [expr {[winfo rooty $c] + [winfo height $c]}]
    set w [winfo width $c]
    set h 200

    wm geometry $c.drop "${w}x${h}+${x}+${y}"
    update idletasks
    wm deiconify $c.drop
    raise $c.drop
    focus $c.drop.lb
    grab set -global $c.drop
    return
}

# ::two::combobox::Close --
#
#	Close the combobox, i.e. withdraw the dropdown listbox and
#	insert the current selection into the entry field.
#
# Arguments:
#	c		Pathname of combobox widget

proc ::two::combobox::Close {c} {
    variable Combobox
    
    if {[string length [$c.drop.lb curselection]]} {
    	set state [$c.entry cget -state]
	$c.entry configure -state normal
        $c.entry delete 0 end
        $c.entry insert end [$c.drop.lb get active]
	$c.entry configure -state $state
    }
    grab release $c.drop
    wm withdraw  $c.drop
    focus [tk_focusNext $c]
    $c.button configure -relief raised
    return
}

namespace import ::two::combobox::*

# historybox.tcl --
#
#	Wrapper around the combobox widget to provide history
#	functionality: all commands entered will be remembered
#	and added to the drop down list. You can even navigate
#	the history with the cursor keys.
#
# Copyright (c) 1998, 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoHistorybox 1.0
package require twoUtil
package require twoCombobox

namespace eval ::two::historybox {
    namespace export two_historybox
    variable Historybox
    
    namespace import ::two::util::two_parseArgs
    
    # variable Historybox
    #
    # $w,current -	index of currently selected element from history
    # $w,max -	maximum number of entries in history
    # $w,hist -	list of previously entered commands
    # $w,command -	command to invoke when user presses "return"
    
    array set Historybox {
    }
}

# ::two::historybox::two_historybox --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.cb"); the original widget procedure
#	is renamed to "_two$w" (like "_two.tl.cb") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#	combobox creates a toplevel widget "$w.drop" for the drop box.
#	Arguments are passed to entry widget.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::historybox::two_historybox {w args} {
    variable Historybox

    # Initialise the globals
    array set Historybox "
	$w,command {}
	$w,max     32
	$w,hist    {}
	$w,current -1
    "
    
    # See if this option is for us...
    two_parseArgs args Historybox($w,command) "-command"
    two_parseArgs args Historybox($w,max)     "-max"

    # the hard work is done by the combobox widget
    uplevel [list two_combobox $w] $args [list -command [namespace code [list SetList $w]]]

    # ... with the help of some bindings
    set e [$w entry]
    bind $e <Return> [list ::two::historybox::Invoke $w]
    bind $e <Up>     [list ::two::historybox::Prev   $w]
    bind $e <Down>   [list ::two::historybox::Next   $w]

    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Historybox
    # BEWARE: Here we need to ADD our binding as the underlying
    # combobox has already added it's own destruction handler.
    bind $w <Destroy> +[namespace code [list _Destroy $w]]

    # Return the correct result...
    return $w
}

# ::two::historybox::Widget --
#
#	replacement for widget command
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::historybox::Widget {w opt args} {
    variable Historybox

    # See if this option is for us...
    switch -exact -- $opt {
        add2hist {
            Add2hist $w
        }
        reset {
            set Historybox($w,current) -1
        }
	configure {
	    two_parseArgs args Historybox($w,command)  "-command"
	    if {[two_parseArgs args Historybox($w,max) "-max"]} {
	        Truncate $w
	    }

	    if {[llength $args]} {
		return [eval [list _two$w $opt] $args]
	    }
	}
	cget {
	    switch -exact -- $args {
	        -command { return $Historybox($w,command) }
	        -max     { return $Historybox($w,max)     }
		default  { return [eval [list _two$w $opt] $args] }
	    }
	}
	default {
            return [eval [list _two$w $opt] $args]
	}
    }
    return
}

# ::two::historybox::_Destroy --
#
#	Call this procedure when the Historybox widget is destroyed.
#	Will clean widget-specific Elements of Historybox.
#
# Arguments:
#	w		Path to Widget

proc ::two::historybox::_Destroy {w} {
    variable Historybox

    # clear Historybox
    foreach key [array names Historybox "$w,*"] {
        unset Historybox($key)
    }
    # remove the original widget command
    rename $w ""
    return
}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::historybox::Add2hist
#
#	The current contents gets appended to the history (if different from
#	the most recent entry)
#
# Arguments:
#	w		Path to widget

proc ::two::historybox::Add2hist {w} {
    variable Historybox

    set txt [_two$w get]

    # different from previous contents?
    if {[string compare $txt [lindex $Historybox($w,hist) 0]] != 0} {
        # prepend to history
        set Historybox($w,hist) [linsert $Historybox($w,hist) 0 $txt]
        Truncate $w
    }
    set History($w,current) -1
    return
}

# ::two::historybox::Invoke --
#
#	Evaluate the command ($w configure -command <cmd>). The current
#	contents also gets appended to the history (if different from
#	the most recent entry)
#
# Arguments:
#	w		Path to widget

proc ::two::historybox::Invoke {w} {
    variable Historybox

    Add2hist $w

    # any command to call?
    if {[string length $Historybox($w,command)]} {
        namespace eval :: $Historybox($w,command)
    }
    return
}

# ::two::historybox::Truncate --
#
#	Truncate the histrory to a maximum of ($w configure -max <max>)
#	elements.
#
# Arguments:
#	$w		Path to widget

proc ::two::historybox::Truncate {w} {
    variable Historybox
    
    set max $Historybox($w,max)
    incr max -1
    set Historybox($w,hist) [lrange $Historybox($w,hist) 0 $max]
    return
}

# ::two::historybox::SetList --
#
#	Set list for dropdown listbox from history
#
# Arguments:
#	w		Path to widget

proc ::two::historybox::SetList {w} {
    variable Historybox

    eval [list $w setlist] $Historybox($w,hist)
}

# ::two::historybox::Next --
#
#	go to next element of the history
#
# Arguments:
#	$w		Path to widget

# go to next element of the target history
proc ::two::historybox::Next {w} {
    variable Historybox

    if {$Historybox($w,current) > 0} {
	incr Historybox($w,current) -1
	_two$w set [lindex $Historybox($w,hist) $Historybox($w,current)]
    } else {
	set Historybox($w,current) -1
        _two$w set ""
    }
}

# ::two::historybox::Prev --
#
#	go to previous element of the history
#
# Arguments:
#	$w		Path to widget

proc ::two::historybox::Prev {w} {
    variable Historybox

    if {$Historybox($w,current) < [llength $Historybox($w,hist)] -1} {
	incr Historybox($w,current)
	_two$w set [lindex $Historybox($w,hist) $Historybox($w,current)]
    }
}

namespace import ::two::historybox::*

# html.tcl --
#
#	HTML Widget. This does not create a toplevel, it just provides
#	the hypertext widget to be used e.g. within a help system.
#	This widget uses a subset of HTML 2.0 :-)
#
# Copyright (c) 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoHtml 1.0
package require twoUtil
package require twoTooltip
package require http

# This might interfere with other parts trying to configure the http
# useragent - but it is a shortcoming of the http library itself to
# make configurations interpreter-wide:-(
::http::config -useragent "TWO simple html widget 1.0"

# Create fonts (catch as they might already exist)
catch {font create twoHtmlH1   -family times     -size 18 -weight bold}
catch {font create twoHtmlH2   -family times     -size 14 -weight bold}
catch {font create twoHtmlH3   -family times     -size 12 -weight bold -slant italic}
catch {font create twoHtmlH4   -family times     -size 12 -weight bold}
catch {font create twoHtmlH5   -family times     -size 12 -slant italic}
catch {font create twoHtmlH6   -family times     -size 12}
catch {font create twoHtmlP    -family helvetica -size 12}
catch {font create twoHtmlPB   -family helvetica -size 12 -weight bold}
catch {font create twoHtmlPI   -family helvetica -size 12              -slant italic}
catch {font create twoHtmlPBI  -family helvetica -size 12 -weight bold -slant italic}
catch {font create twoHtmlPC   -family courier   -size 12}
catch {font create twoHtmlPBC  -family courier   -size 12 -weight bold}
catch {font create twoHtmlPIC  -family courier   -size 12              -slant italic}
catch {font create twoHtmlPBIC -family courier   -size 12 -weight bold -slant italic}
catch {font create twoHtmlS    -family times     -size 1}

namespace eval ::two::html {
    namespace export two_html
    variable Html
    variable Visited
    variable Preload
    
    namespace import ::two::util::two_parseArgs
    
    # variable Html
    #
    # $w,hist -	history as tcl list (elements are a list of URL and Title)
    # $w,hcurrent -	index of current page in the history
    # $w,cbase -	current base url for relative urls
    # $w,home -	home page
    # $w,nonce -    nonce to give each link tag a new name
    # $w,active -   active tag (needed to properly handle button press on hyperlinks)
    # $w,activefg - normal foreground color of active tag
    # $w,h,<p> -	html text for htmlPath <p> (pre-set)
    # $w,images -	names of all temprary images which are included in the widget
    # $w,ltags -	names of all temprary tags used for links
    # $w,cursor -	Original cursor while dragging with control-B2
    # mouseMoved -	Non-zero means the mouse has moved a significant
    #		amount since the button went down (so, for example,
    #		start dragging out a selection).
    # x, y -	Last known mouse coordinates for scanning and auto-scanning.
    
    # variable Visited
    #
    # hashtable that holds with all keys which to already visited html pages.
    
    # variable Preload
    #
    # hashtable that holds with all pages which are pre-set (not to be
    # taken from URL).
}

image create bitmap TwoHtmlBullet -data {
\#define bullet_width 10
\#define bullet_height 10
static char bullet_bits[] = {
   0x78, 0x00, 0xfc, 0x00, 0xe6, 0x01, 0xc3, 0x03, 0xc3, 0x03, 0xe7, 0x03,
   0xff, 0x03, 0xfe, 0x01, 0xfc, 0x00, 0x78, 0x00};
}

image create bitmap TwoHtmlBrokenImage -data {
\#define broken_width 26
\#define broken_height 32
static unsigned char broken_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0xfe, 0xff, 0x01, 0x00, 0xfe, 0xff, 0x03, 0x00,
   0x06, 0x80, 0x07, 0x00, 0x06, 0x80, 0x0f, 0x00, 0x06, 0x80, 0x1d, 0x00,
   0x86, 0x87, 0x3b, 0x00, 0xc6, 0x8c, 0x75, 0x00, 0xc6, 0x8c, 0xeb, 0x00,
   0x06, 0x86, 0xd5, 0x01, 0x06, 0x83, 0xab, 0x01, 0x06, 0x83, 0xff, 0x01,
   0x06, 0x80, 0xff, 0x01, 0x06, 0x03, 0x80, 0x01, 0x06, 0x03, 0x80, 0x01,
   0x06, 0x00, 0x80, 0x01, 0x06, 0x00, 0x80, 0x01, 0x86, 0x87, 0x87, 0x01,
   0xc6, 0xcc, 0x8c, 0x01, 0xc6, 0xcc, 0x8c, 0x01, 0x06, 0x06, 0x86, 0x01,
   0x06, 0x03, 0x83, 0x01, 0x06, 0x03, 0x83, 0x01, 0x06, 0x00, 0x80, 0x01,
   0x06, 0x03, 0x83, 0x01, 0x06, 0x03, 0x83, 0x01, 0x06, 0x00, 0x80, 0x01,
   0x06, 0x00, 0x80, 0x01, 0x06, 0x00, 0x80, 0x01, 0xfe, 0xff, 0xff, 0x01,
   0xfe, 0xff, 0xff, 0x01, 0x00, 0x00, 0x00, 0x00};
}

# ::two::html::two_html --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".html"); the original widget procedure
#	is renamed to "_two$w" (like "_two.html") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::html::two_html {w args} {
    variable Html

    # Initialise the globals
    array set Html "
        $w,home		index.html
        $w,linkcolor	blue2
        $w,alinkcolor	red
        $w,vlinkcolor	purple4
        $w,hist		{}
        $w,cbase	{}
        $w,hcurrent	0
        $w,nonce	0
        $w,active	{}
        $w,images	{}
        $w,ltags	{}
    "
    # See if this option is for us...
    two_parseArgs args Html($w,home)       "-home"
    two_parseArgs args Html($w,linkcolor)  "-linkcolor"
    two_parseArgs args Html($w,alinkcolor) "-alinkcolor"
    two_parseArgs args Html($w,vlinkcolor) "-vlinkcolor"
    set Html($w,cbase) $Html($w,home)

    # Create text widget
    uplevel [list text $w -wrap word -insertwidth 0] $args
#    uplevel text $w -wrap word $args
    
    # only needed for grabs to really go nowhere:-)
    frame $w.grab

    # Only use our own bindings
    bindtags $w [list $w TwoHtml [winfo toplevel $w] all]
#    bindtags $w [list $w Text [winfo toplevel $w] all]
    
    # Configure tags
    # SP: Spacing between paragraphs
    # S: Small font
    # HR: Horizontal rule
    # H1, H2, H3: Heading (Level 1-6)
    # LI1, LI2: List elements (Level 1-6)
    # DT1, DT2: Definition Term (Level 1-6)
    # DD1, DD2: Definition Description (Level 1-6)
    # DC: DT and DC in one line for compact definition lists
    # CEN: Center
    # REF: Hyper Reference (only the base tag, more a created on demand)
    # CUR: Navigaton cursor frame (for tab navigation)
    # P, PB, PI, PBI, PC, PBC, PIC, PBIC: normal Paragraphs (P) with
    #     Bold (B), Italics (I) and/or Courier (C) fonts
    $w tag configure SP  -spacing1 10
    $w tag configure HR  -font twoHtmlS  -background black -bgstipple gray25

    $w tag configure H1  -font twoHtmlH1 -justify center -spacing1 30 -spacing3 30
    $w tag configure H2  -font twoHtmlH2 -spacing1 20 -spacing3 20
    $w tag configure H3  -font twoHtmlH3 -spacing1 20 -spacing3 20
    $w tag configure H4  -font twoHtmlH4 -spacing1 20
    $w tag configure H5  -font twoHtmlH5 -spacing1 20
    $w tag configure H6  -font twoHtmlH6 -spacing1 20

#    $w tag configure UL
#    $w tag configure OL
#    $w tag configure DL

    $w tag configure LI1 -tabs {0.2i center 0.4i}                -lmargin2 0.4i
    $w tag configure LI2 -tabs {0.6i center 0.8i} -lmargin1 0.4i -lmargin2 0.8i
    $w tag configure LI3 -tabs {1.0i center 1.2i} -lmargin1 0.8i -lmargin2 1.2i
    $w tag configure LI4 -tabs {1.4i center 1.6i} -lmargin1 1.2i -lmargin2 1.6i
    $w tag configure LI5 -tabs {1.8i center 2.0i} -lmargin1 1.6i -lmargin2 2.0i
    $w tag configure LI6 -tabs {2.2i center 2.4i} -lmargin1 2.0i -lmargin2 2.4i

    $w tag configure DC1 -tabs {0.4i}                -lmargin2 0.4i
    $w tag configure DC2 -tabs {0.8i} -lmargin1 0.4i -lmargin2 0.8i
    $w tag configure DC3 -tabs {1.2i} -lmargin1 0.8i -lmargin2 1.2i
    $w tag configure DC4 -tabs {1.6i} -lmargin1 1.2i -lmargin2 1.6i
    $w tag configure DC5 -tabs {2.0i} -lmargin1 1.6i -lmargin2 2.0i
    $w tag configure DC6 -tabs {2.4i} -lmargin1 2.0i -lmargin2 2.4i

#    $w tag configure DT1
    $w tag configure DT2 -lmargin1 0.4i -lmargin2 0.4i
    $w tag configure DT3 -lmargin1 0.8i -lmargin2 0.8i
    $w tag configure DT4 -lmargin1 1.2i -lmargin2 1.2i
    $w tag configure DT5 -lmargin1 1.6i -lmargin2 1.6i
    $w tag configure DT6 -lmargin1 2.0i -lmargin2 2.0i

    $w tag configure DD1 -lmargin1 0.4i -lmargin2 0.4i
    $w tag configure DD2 -lmargin1 0.8i -lmargin2 0.8i
    $w tag configure DD3 -lmargin1 1.2i -lmargin2 1.2i
    $w tag configure DD4 -lmargin1 1.6i -lmargin2 1.6i
    $w tag configure DD5 -lmargin1 2.0i -lmargin2 2.0i
    $w tag configure DD6 -lmargin1 2.4i -lmargin2 2.4i
    
    $w tag configure PRE -wrap none
    $w tag configure CEN -justify center
    $w tag configure REF -underline true
    $w tag configure CUR -borderwidth 2 -relief solid

    foreach fnt {S P PB PI PBI PC PBC PIC PBIC} {
        $w tag configure $fnt -font twoHtml$fnt
    }
    
    # common bindings for all REF tags
    $w tag bind REF <Enter> [namespace code {_two%W configure -cursor hand2}]
    $w tag bind REF <Leave> [namespace code {_two%W configure -cursor left_ptr}]

    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "

    # catch widget destruction in order to clean up Html
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # Return the correct result...
    return $w
}

# ::two::html::Widget --
#
#	replacement for widget command
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::html::Widget {w opt args} {
    variable Html
    variable Preload

    # See if this option is for us...
    switch -exact -- $opt {
	configure {
	    set retval ""
	    two_parseArgs args Html($w,home)       "-home"
	    two_parseArgs args Html($w,linkcolor)  "-linkcolor"
	    two_parseArgs args Html($w,alinkcolor) "-alinkcolor"
	    two_parseArgs args Html($w,vlinkcolor) "-vlinkcolor"
	    if {![string length $Html($w,cbase)]} {
	        set Html($w,cbase) $Html($w,home)
            }

            if {[llength $args]} {
		set retval [eval [list _two$w $opt] $args]
	    }
	    if {[llength $Html($w,hist)]} {
                ShowPage $w [lindex [lindex $Html($w,hist) $Html($w,hcurrent)] 0] 1
	    }
            return $retval
	}
	cget {
	    switch -exact -- $args {
	        -home       { return $Html($w,home)       }
	        -linkcolor  { return $Html($w,linkcolor)  }
	        -alinkcolor { return $Html($w,alinkcolor) }
	        -vlinkcolor { return $Html($w,vlinkcolor) }
		default     { return [eval [list _two$w $opt] $args] }
	    }
	}
	home {
	    return [ShowPage $w $Html($w,home) 0 1]
	}
	show {
	    if {[llength $args] != 1} {
	    	return -code error "wrong # args: should be \"$w show page\""
	    }
            return [ShowPage $w [lindex $args 0]]
	}
	history -
	hist {
	    set hist ""
	    foreach link $Html($w,hist) {
	        set hist [linsert $hist 0 $link]
	    }
	    return $hist
	}
	current {
	    return [expr {[llength $Html($w,hist)] - $Html($w,hcurrent)}]
	}
	goto {
	    set retval ""
	    if {[llength $args] != 1} {
	    	return -code error "wrong # args: should be \"$w goto index\""
	    }
	    set pos [lindex $args 0]
	    if {$pos <= 0} {
	        return -code error "invalid index \"$pos\""
	    }
	    if {$pos > [llength $Html($w,hist)]} {
	        return -code error "invalid index \"$pos\""
	    }
	    set pos [expr {[llength $Html($w,hist)] - $pos}]
	    
	    if {$Html($w,hcurrent) != $pos} {
                if {$Html($w,hcurrent) == [llength $Html($w,hist)] - 1} {
                    event generate $w <<NextOn>>
                }
                if {$Html($w,hcurrent) == 0} {
                    event generate $w <<PrevOn>>
                }

                set Html($w,hcurrent) $pos
                set retval [ShowPage $w \
                    [lindex [lindex $Html($w,hist) $Html($w,hcurrent)] 0] 1
                ]

                if {$Html($w,hcurrent) == [llength $Html($w,hist)] - 1} {
                    event generate $w <<NextOff>>
                }
                if {$Html($w,hcurrent) == 0} {
                    event generate $w <<PrevOff>>
                }
	    }
	    return $retval
	}
	nextState {
	    if {$Html($w,hcurrent) < [llength $Html($w,hist)] - 1} {
	        return 1
	    } else {
		return 0
	    }
	}
	next {
	    set retval ""
	    if {[llength $args] > 0} {
	    	return -code error "wrong # args: should be \"$w next\""
	    }
	    if {$Html($w,hcurrent) == 0} {
	        event generate $w <<PrevOn>>
	    }
	    if {$Html($w,hcurrent) < [llength $Html($w,hist)] - 1} {
	        incr Html($w,hcurrent)
	        set retval [ShowPage $w \
	            [lindex [lindex $Html($w,hist) $Html($w,hcurrent)] 0] 1
	        ]
	    }
	    if {$Html($w,hcurrent) == [llength $Html($w,hist)] - 1} {
	        event generate $w <<NextOff>>
	    }
	    return $retval
	}
	prevState {
	    if {$Html($w,hcurrent) > 0 } {
	        return 1
	    } else {
		return 0
	    }
	}
	prev {
	    set retval ""
	    if {[llength $args] > 1} {
	    	return -code error "wrong # args: should be \"$w prev ?steps?\""
	    }
	    if {$Html($w,hcurrent) == [llength $Html($w,hist)] - 1} {
	        event generate $w <<NextOn>>
	    }
	    if {![string length $args]} {
	        set step 1
	    } else {
	        set step [lindex $args 0]
	    }
	    if {$Html($w,hcurrent) > $step - 1} {
	        incr Html($w,hcurrent) "-$step"
	        set retval [ShowPage $w \
                    [lindex [lindex $Html($w,hist) $Html($w,hcurrent)] 0] 1
	        ]
	    }
	    if {$Html($w,hcurrent) == 0} {
	        event generate $w <<PrevOff>>
	    }
	    return $retval
	}
	refresh {
	    return [ShowPage $w \
                [lindex [lindex $Html($w,hist) $Html($w,hcurrent)] 0] 1
	    ]
	}
	add {
	    set url [lindex $args 0]
	    set Preload($url) [lindex $args 1]
	}
	append {
	    append Preload([lindex $args 0]) [lindex $args 1]
	}
	default {
            return [eval [list _two$w $opt] $args]
	}
    }
    return
}

# ::two::html::_Destroy --
#
#	Call this procedure when the html widget is destroyed.
#	Will clean widget-specific Elements of Html and destroy
#	the original widget command.
#
# Arguments:
#	w		Path to Widget

proc ::two::html::_Destroy {w} {
    variable Html

    # clear Html
    foreach key [array names Html "$w,*"] {
        unset Html($key)
    }
    # remove the original widget command
    rename $w ""
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoHtml-class bindings
#-------------------------------------------------------------------------

bind TwoHtml <1>     {focus %W}

bind TwoHtml <Control-2> {
    if {!$tk_strictMotif} {
        set ::two::html::Html(%W,cursor) [::two::html::_two%W cget -cursor]
        ::two::html::_two%W configure -cursor fleur
        ::two::html::_two%W scan mark %x %y
        set ::two::html::Html(x) %x
        set ::two::html::Html(y) %y
        set ::two::html::Html(mouseMoved) 0
    }
}
bind TwoHtml <Control-B2-Motion> {
    if {!$tk_strictMotif} {
        if {(%x != $::two::html::Html(x)) || (%y != $::two::html::Html(y))} {
            set ::two::html::Html(mouseMoved) 1
        }
        if {$::two::html::Html(mouseMoved)} {
            ::two::html::_two%W scan dragto %x %y
        }
    }
}
bind TwoHtml <Control-ButtonRelease-2> {
    if {!$tk_strictMotif} {
        ::two::html::_two%W configure -cursor $::two::html::Html(%W,cursor)
    }
}

bind TwoHtml <Up>    {%W yview scroll -1 unit}
bind TwoHtml <Down>  {%W yview scroll  1 unit}
bind TwoHtml <Prior> {%W yview scroll -1 page}
bind TwoHtml <Next>  {%W yview scroll  1 page}

bind TwoHtml <Return> {
    ::two::html::SelectShow %W
}
bind TwoHtml <space> {
    ::two::html::SelectShow %W
}
bind TwoHtml <Left> {
    ::two::html::SelectPrev %W
}
bind TwoHtml <Right> {
    ::two::html::SelectNext %W
}
bind TwoHtml <Tab> {
    ::two::html::SelectNext %W
    # don't change focus
    break
}
bind TwoHtml <Shift-Tab> {
    ::two::html::SelectPrev %W
    # don't change focus
    break
}
bind TwoHtml <Control-Tab> {
    # Needed only to keep <Tab> binding from triggering;  doesn't
    # have to actually do anything.
}
bind TwoHtml <Shift-Control-Tab> {
    # Needed only to keep <Tab> binding from triggering;  doesn't
    # have to actually do anything.
}
bind TwoHtml <FocusIn> {
    ::two::html::_two%W tag configure CUR -relief solid
}
bind TwoHtml <FocusOut> {
    ::two::html::_two%W tag configure CUR -relief flat
}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::html::MakeAbsUrl --
#
#	Takes an url and tries to make an absoulte url from it. If it is
#	relative, then it will be taken relative to the current base url.
#
# Arguments:
#	w		Path to html widget
#	varName		Name of variable containing the url
#
# Sideeffects:
#	The Variable $varame is possibly modified to an absolute url

proc ::two::html::MakeAbsUrl {w varname} {
    upvar $varname url
    variable Html

    # Only file: and http: allowed, and both are 5 chars:-)
    set urlProto [string tolower [string range $url 0 4]]
    if {[string match "http:" $urlProto]} {
        # nothing to do
    } elseif {[string match "file:" $urlProto]} {
        # using a protocol specifier should guarantee absoute filenames.
        # No further processing needed, just remove the "file:" part
        set url [string range $url 5 end]
    } else {
        # o.k., url could be relative. Now, what is the reference?
        regexp {^[^\#]*} $Html($w,cbase) ref
        # Only if the url has any chars should it be relative, else
        # take the ref (the previous location) directly
        if {[string length $url]} {
            # get the protocol descriptor of the reference file
            set refProto [string tolower [string range $ref 0 4]]
            # is the reference an html file?
            if {[string match "http:" $refProto]} {
                # extract base from url
                regexp -nocase {^(http://[^/]+)(/.*)?$} $ref trash base add
                if {[string match "/*" $url]} {
                    set url "$base$url"
                } elseif {[string match "#*" $url]} {
                    set url "$ref$url"
                } else {
                    if {![string match "*/" $add]} {
                        regsub {/+[^/]+$} $add {} add
                    }
                    append base $add
                    set url "$base/$url"
                }
            } else {
                # no html, so assume file. Remove the "file:" Protocol
                # selector (only if present)
                if {[string match "file:" $refProto]} {
                    set ref [string range $ref 5 end]
                }
                regexp {^[^\#]*} $ref ref

                if {[string match "#*" $url]} {
                    set url "$ref$url"
                } else {
                    # the file command handles absoulte vs. relative file names
                    # properly...
                    set url [file join [file dirname $ref] $url]
                }

                # ... but leaves the ".." in the filename. This would make the
                # filename grow forever and make matching filenames against
                # the built-in list of files impossilbel.
                set fname ""
                if {[string match "absolute" [file pathtype $ref]]} {
                    set top /
                } else {
                    set top [file split [pwd]]
                }

                foreach part [file split $url] {
                    switch -- $part {
                        . {
                        }
                        .. {
                            if {![llength $fname]} {
                                set fname $top
                                set top /
                            }
                            set fname [lreplace $fname end end]
                        }
                        default {
                            lappend fname $part
                        }
                    }
                }
                if {[llength $fname]} {
                    set url [eval file join $fname]
                } else {
                    set url [file nativename /]
                }
            }
        } else {
            set url $ref
        }
    }
}

# ::two::html::ShowPage --
#
#	show page accessed by "url" in the html widget "w". Also handles
#	the history: pages are appended to the history (this is skipped
#	if the fromhist parameter is set).
#
# Arguments:
#	w		Path to html widget
#	url		Name of html page which should be shown
#	fromhistory	true if page is accessed through history
#	isabsolute	true if this page is meant as absoulte url (home...)
#
# Returns:
#	Empty string if successfull, or the error message

proc ::two::html::ShowPage {w url {fromhistory 0} {isabsolute 0}} {
    variable Html
    variable History
    variable Visited
    global TwoHtmlTitle
    global two_ttInfo

    #
    # Step 1: turn relative url to absolute url
    #

    # extract anchor part
    regexp "(\[^\#\]*)\#?(.*)" $url trash url anchor

    # only navigate within current page?
    if {![string length $url] && [string length $anchor]} {
        if {[catch {_two$w yview "A_$anchor"} err]} {
            set two_ttInfo $err
            return $err
        } else {
            # current base url...
            regexp {^[^\#]*} $Html($w,cbase) url
            append url \# $anchor
            set Visited($w,$url) ""
            return ""
        }
    }

    if {!$isabsolute && !$fromhistory} {
        MakeAbsUrl $w url
    }

    #
    # Step 2: load and display html code
    #

    grab $w.grab
    after idle [namespace code [list _two$w configure -cursor watch]]
    
    set html ""

    # is this in fact just an image?
    switch -glob -- $url {
        *.gif -
        *.jpg -
        *.jpeg -
        *.xbm -
        *.xpm -
        *.png {
            set html "<HTML><HEAD>
              <TITLE>$url</TITLE>
              </HEAD><BODY>
              <IMG SRC=\"$url\">
              </BODY></HTML>"
        }
        default {
            if {[catch {set html [GetUrl $w url]} err]} {
                set two_ttInfo $err
                _two$w configure -cursor left_ptr
                grab release $w.grab
                two_mb -type ok -default ok -icon error -title "HTML Widget" \
                    -message $err -parent $w
                return $err
            }
        }
    }
    set Html($w,cbase) $url
    set title [RenderHtml $w $html]

    _two$w configure -cursor left_ptr
    grab release $w.grab

    # show requested anchor
    if {[string length $anchor]} {
        catch {_two$w yview "A_$anchor"}
    }
    set TwoHtmlTitle $title
    set Visited($w,$url) ""
    
    #
    # Step 3: add to history
    #

    if {!$fromhistory} {
        if {![llength $Html($w,hist)]} {
            set Html($w,hist) [list [list $url $title]]
        } elseif {![string match $url \
            [lindex [lindex $Html($w,hist) $Html($w,hcurrent)] 0]
        ]} {
            if {$Html($w,hcurrent) == 0} {
                event generate $w <<PrevOn>>
            }
            if {$Html($w,hcurrent) <= [llength $Html($w,hist)]} {
                set Html($w,hist) [lrange $Html($w,hist) 0 $Html($w,hcurrent)]
                incr Html($w,hcurrent)
                event generate $w <<NextOff>>
            }
            lappend Html($w,hist) [list $url $title]
        }
    }

    set two_ttInfo ""
    return ""
}

# ::two::html::GetUrl --
#
#	Retrieve a file through the url. The url must be handled over
#	by reference and not by value, as a redirect might well
#	change it.
#
# Arguments:
#	w		Path to html widget
#	urlvar		Name variable with Url to retrieve
#
# Returns:
#	Contents of that url

proc ::two::html::GetUrl {w urlvar} {
    variable Html
    variable Preload
    global two_ttInfo
    upvar $urlvar url

    set content ""

    if {[info exists Preload($url)]} {
        # Fine, this page is already pre-loaded...
        set content $Preload($url)
    } else {
        # try to retrieve the page and render it
        # or return with an error message
        if {[string match "http:" [string tolower [string range $url 0 4]]]} {
            while {1} {
                # http... set timeout to 8 sec.
                # might throw an error... fine
                set token [::http::geturl $url -timeout 8000]
                # redirect?
                array set meta [set ${token}(meta)]
                if {[info exists meta(Location)]} {
                    # string trim due to bug in http package
                    set url [string trimleft $meta(Location)]
                    unset $token
                    unset meta
                } else {
                    break
                }
            }
            set content [::http::data $token]
            unset $token
        } else {
            # might throw an error... fine, this will be reported
            # to the calling function which should handle it.
            set content [two_readFile -binary $url]
        }
    }
    set two_ttInfo $url
    update idletasks

    # this should be "return $content", but this is faster!?
    set content
}

# ::two::html::RenderHtml --
#
#	Render and display an html page (and return the title).
#
# Arguments:
#	w		Path to html widget
#	html		Name of html page which should be shown
#
# Returns:
#	The Title of the HTML file (if it is set), else an empty
#	string

proc ::two::html::RenderHtml {w html} {
    variable Html
    variable Visited

    # clean up
    _two$w delete 1.0 end
    set Html($w,nonce) 0
    if {[llength $Html($w,images)]} {
        eval image delete $Html($w,images)
        set Html($w,images) {}
    }
    if {[llength $Html($w,ltags)]} {
        two_tooltip delete $w
        eval [list _two$w tag delete] $Html($w,ltags)
        set Html($w,ltags) {}
    }

    # removing comments would be easy as of tcl8.1, due to the
    # non greedy .*?, but trying to be compatible to tcl8.0 makes
    # this unavailable: <!--.*?-->
    set rest $html
    set html ""
    while {[set idx [string first "<!--" $rest]] >= 0} {
        append html [string range $rest 0 [expr "$idx -1"]]
        set rest [string range $rest [expr "$idx +4"] end]
        if {[set idx [string first "-->" $rest]] >= 0} {
            set rest [string range $rest [expr "$idx +3"] end]
        } else {
            append html $rest
            set rest ""
        }
    }
    append html $rest

    # TODO: embedded TCL <SCRIPT NAME="TwoTcl">code</SCRIPT>
    # would be quite nice:-)

    # Split into parts: the first one is pure text, all subsequent parts are
    # a tag and the text up to the next tag or to the end of the file
    set parts [split $html "<"]

    # the first part is NOT a tag; insert it directly if it ain't whitespace
    set part [string trim [lindex $parts 0]]
    if {[string length $part]} {
        _two$w insert end $part P
    }

    # state info:
    set Para	""	;# current paragraph format (list, filled as stack!)
    set Char	""	;# character format (B=Bold, I=Italics, C=Courier)
    set Link	""	;# currently within an anchor? (contains the tag to add)
    set Name	""	;# Name of anchor which should be added
    set Indent	""	;# additinal paragraph format for indendation
    set Level	0	;# indendation level
    set Show	1	;# 1 if text is to be shown, else 0
    set Center1	0	;# true if next paragraph is top be centred
    set Center2	0	;# true if centering untill next /CENTER
    # OL($Level) used for current position within ordered lists

    # only used once at start of next paragraph:
    set Break	0	;# Linebreak: 0 = none, 1 = if needed, 2 = forced
    set Space	0	;# if true, then the next paragraph need to have spacing
    set Tags	"P"	;# tags from previous line
    
    # The <TITLE> of the HTML file
    set Title ""

    # Work through the parts
    foreach part [lrange $parts 1 end] {
    	# Split that part into tag and remaining txt
    	set tmp [split $part ">"]
    	set tag [lindex $tmp 0]
    	set txt [join [lrange $tmp 1 end] ">"]
    	set img ""

    	# normalize the tag and split it into tag and options
    	set tmp [split [string trim $tag] " "]
    	set tag [string toupper [lindex $tmp 0]]
    	set opt [join [lrange $tmp 1 end] " "]
    	
    	# make list of option-value-pairs from opt
    	set opt [SplitParam $opt]

        # current column is:
        set Col [lindex [split [_two$w index "end-1c"] "."] 1]
        
        #
        # Step 1: analyse tags
        #
        
        switch -exact -- $tag {
            HEAD {
                set Show 0
            }
            /HEAD {
                set Show 1
            }
            TITLE {
                set Title [string trim $txt]
            }
            BASE {
                set base [HasOpt $opt HREF]
                # The MakeBasUrl proc expects the base url to be a real URL.
                # So turn the base into a real url - with the proper base:-)
                append base "/dummy.html"
                if {[string length $base]} {
                    set Html($w,cbase) $base
                }
            }
            BR {
                incr Break 2
            }
            P {
                incr Break 2
                set Space 1
                set Center1 [HasOpt $opt ALIGN CENTER]
            }
            /P {
                incr Break
                set Space 1
                set Center1 0
            }
            H1 -
            H2 -
            H3 -
            H4 -
            H5 -
            H6 {
                incr Break
                lappend Para $tag
                set Center [HasOpt $opt ALIGN CENTER]
            }
            /H1 -
            /H2 -
            /H3 -
            /H4 -
            /H5 -
            /H6 {
                incr Break
                set Para [lreplace $Para end end]
            }
            HR {
    	        if {$Col} {
                    if {[string match " " [_two$w get "insert -1c" insert]]} {
                        _two$w delete "insert -1c" insert
                    }
    	            _two$w insert end "\n" $Tags
    	        }
                _two$w insert end "\n" S
                if {$Space} {
                    _two$w tag add "SP" "insert linestart"
                    set Space 0
                }
                set Tags {S HR}
                _two$w insert end " " $Tags
                set Col 1
                incr Break 2
                set Space 1
            }
            PRE {
                incr Break
                set Space 1
                lappend Para $tag
                append Char "C"
            }
            /PRE {
                incr Break
                set Space 1
                set Para [lreplace $Para end end]
                regsub "C" $Char "" Char
                if {[string match "\n" [_two$w get "insert -1c" insert]] &&
                    [string match PRE [lindex [_two$w tag names "insert -1c"] end]]
                } {
                    _two$w delete "insert -1c" insert
                }
            }
            CENTER {
                incr Break
                incr Center2
            }
            /CENTER {
                incr Break
                incr Center2 -1
                if {$Center2 < 0} {set Center2 0}
            }
            UL -
            DIR -
            MENU -
            BLOCKQUOTE {
                incr Break
                set Space 1
                incr Level
                if {$Level > 6} {
                    set Indent DD6
                } else {
                    set Indent DD$Level
                }
                lappend Para UL
            }
            OL {
                incr Break
                set Space 1
                incr Level
                set OL($Level) [HasOpt $opt START]
                if {![string length $OL($Level)] || [catch {expr $OL($Level)}]} {
                    set OL($Level) 1
                }
                lappend Para $tag
                if {$Level > 6} {
                    set Indent DD6
                } else {
                    set Indent DD$Level
                }
            }
            DL {
                incr Break
                set Space 1
                incr Level
                if {[string match "COMPACT" [string toupper [lindex $opt 0]]]} {
                    lappend Para DLC
                    if {$Level > 6} {
                        set Indent LI6
                    } else {
                        set Indent LI$Level
                    }
                } else {
                    lappend Para $tag
                    if {$Level > 6} {
                        set Indent DD6
                    } else {
                        set Indent DD$Level
                    }
                }
            }
            /UL -
            /DIR -
            /MENU -
            /OL -
            /DL -
            /BLOCKQUOTE {
                incr Break
                set Space 1
                incr Level -1
                if {$Level < 0} {set Level 0}
                set Para [lreplace $Para end end]
                if {$Level > 6} {
                    set Indent DD6
                } elseif {$Indent > 0} {
                    set Indent DD$Level
                }
            }
            LI {
                set Indent LI$Level
                # Can't use the "incr Break" as we insert additional text
                # right here, but the Break would occur afterwards...
                set Break 0
    	        if {$Col && [string match " " [_two$w get "insert -1c" insert]]} {
                    _two$w delete "insert -1c" insert
                    incr Col -1
                }
                if {$Col} {
                    _two$w insert end "\n" $Tags
    	        }
                set txt [string trimleft $txt]
                if {$Level > 6} {
                    set li LI6
                } else {
                    set li LI$Level
                }
                if {[string match OL [lindex $Para end]]} {
                    _two$w insert end "\t$OL($Level).\t" "$li P"
                    incr OL($Level)
                } else {
                    _two$w insert end "\t" "$li P"
                    _two$w image create end -image TwoHtmlBullet
                    _two$w tag add "$li P" "insert -1c" insert
                    _two$w insert end "\t" "$li P"
                }
                set Col 3
                if {$Space} {
                    _two$w tag add "SP" "insert linestart"
                    set Space 0
                }
                if {$Level > 6} {
                    set Indent DD6
                } else {
                    set Indent DD$Level
                }
            }
            DT {
                if {[string match "DLC" [lindex $Para end]]} {
                    set Break 0
                    if {$Col && [string match " " [_two$w get "insert -1c" insert]]} {
                        _two$w delete "insert -1c" insert
                        incr Col -1
                    }
                    if {$Col} {
                        _two$w insert end "\n" $Tags
                    }
                    if {$Level > 6} {
                        set Indent LI6
                    } else {
                        set Indent LI$Level
                    }
                    _two$w insert end "\t" "$Indent P"
                    if {$Space} {
                        _two$w tag add "SP" "insert linestart"
                        set Space 0
                    }
                } else {
                    incr Break
                    if {$Level > 6} {
                        set Indent DT6
                    } else {
                        set Indent DT$Level
                    }
                }
            }
            DD {
                if {[string match "DLC" [lindex $Para end]]} {
                    if {$Level > 6} {
                        set Indent LI6
                    } else {
                        set Indent LI$Level
                    }
                    if {[string match " " [_two$w get "insert -1c" insert]]} {
                        _two$w delete "insert -1c" insert
                    }
                    _two$w insert end "\t" "$Indent P"
                    if {$Space} {
                        _two$w tag add "SP" "insert linestart"
                        set Space 0
                    }
                } else {
                    incr Break
                    if {$Level > 6} {
                        set Indent DD6
                    } else {
                        set Indent DD$Level
                    }
                }
            }
            IMG {
                set img [HasOpt $opt SRC]
                set alt [HasOpt $opt ALT]
                if {[string first "&" $alt] >= 0} {
                    set alt [DecodeCharEntities $alt]
                }
            }
            A {
                set Name [HasOpt $opt NAME]
                set href [HasOpt $opt HREF]
                set abs $href
                MakeAbsUrl $w abs
                if {[string length $href]} {
                    set Link link$Html($w,nonce)
                    incr Html($w,nonce)
                    lappend Html($w,ltags) $Link
    
                    if {![info exists Visited($w,$abs)]} {
                        _two$w tag configure $Link -foreground $Html($w,linkcolor)
                    } else {
                        _two$w tag configure $Link -foreground $Html($w,vlinkcolor)
                    }
                    _two$w tag bind $Link <1> [namespace code "
                        set Html($w,active) $Link
                        set Html($w,activefg) \[_two$w tag cget $Link -foreground\]
                        _two$w tag configure $Link -foreground $Html($w,alinkcolor)
                    "]
                    _two$w tag bind $Link <ButtonRelease-1> [namespace code "
                        _two$w tag configure $Link -foreground \$Html($w,activefg)
                        if {\[lsearch -exact \[_two$w tag names @%x,%y\] \$Html($w,active)\] != -1} {
                            $w show $href
                        }
                    "]
                    two_tooltip $w -tag $Link -info $href
                }
            }
            /A {
                set Link ""
            }
            B -
            STRONG {
                append Char "B"
            }
            /B -
            /STRONG {
                regsub "B" $Char "" Char
            }
            I -
            EM -
            CITE -
            VAR {
                append Char "I"
            }
            /I -
            /EM -
            /CITE -
            /VAR {
                regsub "I" $Char "" Char
            }
            ADDRESS {
                incr Break
                append Char "I"
            }
            /ADDRESS {
                incr Break
                regsub "I" $Char "" Char
            }
            TT -
            CODE -
            KBD -
            SAMP {
                append Char "C"
            }
            /TT -
            /CODE -
            /KBD -
            /SAMP {
                regsub "C" $Char "" Char
            }
    	}
    	
    	#
    	# Step 2: add the text snippets to the widget
    	#

    	if {$Show} {
    	    # current paragraph format
            set p [lindex $Para end]

            # remove newlines and collapse multiple spaces
            # if this text is not marked as preformatted
            if {![string match "PRE" $p]} {
                regsub -all "\[\n\t \]+" $txt { } txt
            }

            # cut off any leading whitespace at beginning of a new line
            # (in preforatted text only remove leading newlines)
            if {$Break} {
                # remove trailing space before newline in order to make centered
                # text really centered...
                if {$Col && [string match " " [_two$w get "insert -1c" insert]]} {
                    _two$w delete "insert -1c" insert
                }
                if {[string match "PRE" $p]} {
                    set txt [string trimleft $txt "\n"]
                } else {
                    set txt [string trimleft $txt]
                }
            }

            # Forced line break: add newline in any case
    	    if {$Break >= 2} {
                if {$Col} {
                    _two$w insert end "\n" $Tags
                } else {
                    _two$w insert end "\n" "P"
                }
                set Break 0
            }
            
            # Format and insert the current piece of text
            if {[string length $txt] || [string length $img]} {
                # text should start at a new line, but cursor currently
                # is in another column
                if {$Break && $Col} {
                    _two$w insert end "\n" $Tags
                }
                set Break 0
                
                # paragraph format
                if {[string length $p]} {
                    set Tags [list $p]
                } else {
                    set Tags ""
                }
                if {$Center1 || $Center2} {
                    lappend Tags CEN
                }
                if {[string length $Indent]} {
                    lappend Tags $Indent
                }
                
                # character format
                if {![string match "H?" $p]} {
                    set c "P"
                    if {[string match "*B*" $Char]} {append c "B"}
                    if {[string match "*I*" $Char]} {append c "I"}
                    if {[string match "*C*" $Char]} {append c "C"}
                    lappend Tags $c
                }
                if {[string length $Link]} {
                    lappend Tags REF $Link
                }
                if {[string length $Name]} {
                    _two$w mark set "A_$Name" insert
                    _two$w mark gravity "A_$Name" left
                    set Name ""
                }

                # insert an image (before text as image is no container tag
                # and thus might well be followed by some text)
                if {[string length $img]} {
                    AddImage $w $img
                    foreach t $Tags {
                        _two$w tag add $t "insert -1c" insert
                    }
                    if {$Space} {
                        set Space 0
                        _two$w tag add "SP" "insert -1c" insert
                    }
                    set imgtag img$Html($w,nonce)
                    incr Html($w,nonce)
                    lappend Html($w,ltags) $imgtag
                    _two$w tag add $imgtag "insert -1c" insert
                    if {[string length $Link]} {
                        two_tooltip $w -tag $imgtag -tip $alt -info $href
                    } else {
                        two_tooltip $w -tag $imgtag -tip $alt -info $img
                    }
                }

                # really insert the text
                if {[string length $txt]} {
                    # transform some character classes; only if there are any:-)
                    if {[string first "&" $txt] >= 0} {
                        set txt [DecodeCharEntities $txt]
                    }
                    if {$Space} {
                        set Space 0
                        set pos [_two$w index insert]
                        _two$w insert end $txt $Tags
                        _two$w tag add "SP" $pos "$pos + 1c"
                    } else {
                        _two$w insert end $txt $Tags
                    }
                }
            }
        }
    }
    return $Title
}

# ::two::html::AddImage --
#
#	Add an image to the Html widget
#
# Arguments:
#	w		Path to html widget
#	img		Path to image file

proc ::two::html::AddImage {w img} {
    variable Html

    # special "protocol" for already existing images
    if {[string match "icon:*" [string tolower $img]]} {
        if {[catch {
            _two$w image create end -image [string range $img 5 end]
        }]} {
            _two$w image create end -image TwoHtmlBrokenImage
        }
    } else {
        MakeAbsUrl $w img
        set imgData ""
        catch {set imgData [GetUrl $w img]}

        # try to build the image
        if {[string length $imgData] &&
            (
            ![catch {set name [image create bitmap -data $imgData]}] ||
            ![catch {set name [image create photo  -data $imgData]}] ||
            ![catch {set name [image create photo  -data [base64_encode $imgData]]}]
            )
        } {
            _two$w image create end -image $name
            lappend Html($w,images) $name
        } else {
            _two$w image create end -image TwoHtmlBrokenImage -align center
        }
    }
}

# ::two::html::SelectNext --
#
#	Set html page internal Focus to next link in the page.
#
# Arguments:
#	w		Path to html widget

proc ::two::html::SelectNext {w} {
    variable Html

    # any links in here?
    if {[llength [_two$w tag ranges REF]]} {
        # Where are we now?
        if {[llength [_two$w tag ranges CUR]]} {
            set curBeg [_two$w index CUR.first]
            set curEnd [_two$w index CUR.last]
        } else {
            set curBeg 1.0
            set curEnd 1.0
        }
        # look for next position
        set next [_two$w tag nextrange REF $curEnd end]
        if {![llength $next]} {
            set next [_two$w tag nextrange REF 1.0 $curBeg]
        }
        # found any? Select this link
        if {[llength $next]} {
            _two$w tag remove CUR 1.0 end
            _two$w tag add CUR [lindex $next 0] [lindex $next 1]
            _two$w see [lindex $next 0]
        }
    }
}

# ::two::html::SelectPrev --
#
#	Set html page internal Focus to previous link in the page.
#
# Arguments:
#	w		Path to html widget

proc ::two::html::SelectPrev {w} {
    variable Html

    # any links in here?
    if {[llength [_two$w tag ranges REF]]} {
        # Where are we now?
        if {[llength [_two$w tag ranges CUR]]} {
            set curBeg [_two$w index CUR.first]
            set curEnd [_two$w index CUR.last]
        } else {
            set curBeg 1.0
            set curEnd 1.0
        }
        # look for previous position
        set prev [_two$w tag prevrange REF $curBeg 1.0]
        if {![llength $prev]} {
            set prev [_two$w tag prevrange REF end $curEnd]
        }
        # found any? Select this link
        if {[llength $prev]} {
            _two$w tag remove CUR 1.0 end
            _two$w tag add CUR [lindex $prev 0] [lindex $prev 1]
            _two$w see [lindex $prev 0]
        }
    }
}

# ::two::html::SelectShow --
#
#	Show page whose link has been selected.
#
# Arguments:
#	w		Path to html widget

proc ::two::html::SelectShow {w} {
    variable Html

    # anything selected?
    if {[llength [_two$w tag ranges CUR]]} {
        # get position
        set pos [_two$w bbox CUR.first]
        foreach {xmin ymin width height} $pos {break}
        set x [expr {$xmin + $width  / 2}]
        set y [expr {$ymin + $height / 2}]
        # create the same events as if the mouse pointer were used
        event generate $w <Enter>           -x $x -y $y
        event generate $w <ButtonPress-1>   -x $x -y $y
        event generate $w <ButtonRelease-1> -x $x -y $y
    }
}

# ::two::html::SplitParam --
#
#	Split the parameters of a tag into a paired list
#	(option, value). Options without values are handled
#	as if the value were empty.
#
# Arguments:
#	param		Parameter string from tag
#
# Returns:
#	A single list with an even number of elements. Elements are
#	pairs of option and value: {opt1} {val1} {opt2} {val2}...

proc ::two::html::SplitParam {param} {
    set opt ""
    set param [string trim $param]

    while {[string length $param]} {
        # get the next key
        regexp {^([^= ]*)( *=? *)(.*)$} $param trash key sep val
        set key [string trim $key]
        if {[string length $key]} {
            set key [string toupper $key]
            if {[string match "*=*" $sep]} {
                # separator given, so we have to extract the value
                set val [string trim $val]
                if {[string match "\"*" $val]} {
                    # quoted... so look for closing quote
                    # TODO: right now, masked quotes (backslash!) are not
                    # handled properly: they act as normal end-quotes
                    regexp {^\"([^\"]*)\"?(.*)$} $val trash val param
                } else {
                    # no quotes, so just take next word
                    regexp {^([^ ]*) ?(.*)$} $val trash val param
                }
                lappend opt $key $val
            } else {
                # no separator, so just take this key and continue
                set param $val
                lappend opt $key ""
            }
            set param [string trim $param]
        } else {
            # no key, so we're finished
            set param ""
        }
    }
    return $opt
}

# ::two::html::HasOpt --
#
#	Query a list of option-value-pairs if a specified option is
#	set and return that value. Or, if an additional value is
#	given to the command, reuturn true if this option is set
#	and false otherwise.
#
# Arguments:
#	opt		key value pairs like those from SplitParam
#	searchKey	the key to search for
#	args		optional; the value which we look for
#
# Returns:
#	Two parameters given: the value to this key or, if no such
#	key is available, an empty string.
#	Three parameters given: True if key-value pair is found,
#	else False.

proc ::two::html::HasOpt {opt searchKey args} {
    if {[llength $args]} {
        set searchVal [string toupper [join $args]]
        set matchVal 1
        set retval 0
    } else {
        set matchVal 0
        set retval ""
    }
    set searchKey [string toupper $searchKey]

    foreach {key val} $opt {
        if {[string compare [string toupper $key] $searchKey]==0} {
            if {$matchVal} {
                if {[string compare [string toupper $val] $searchVal]==0} {
                    set retval 1
                } else {
                    set retval 0
                }
            } else {
                set retval $val
            }
        }
    }
    return $retval
}

# ::two::html::DecodeCharEntities --
#
#	Decode character entities to ISO latin-1 (e.g. &auml; to )
#
# Arguments:
#
#	encoded -		the encoded text
#
# Retruns:
#	Decoded text

proc ::two::html::DecodeCharEntities {encoded} {
    variable CharEntities

    set txt ""
    foreach part [split $encoded "&"] {
        set l [string first ";" $part]
        set char [string range $part 0 [expr {$l - 1}]]
        if {[string match {\#[0-9]*} $char]} {
            append txt [format %c [string range $char 1 end]]
        } elseif {[info exists CharEntities($char)]} {
            append txt $CharEntities($char)
        } elseif {[string length $char]} {
            append txt "&" $char ";"
        }
        append txt [string range $part [incr l] end]
    }
    # could be "return $txt" as well:-)
    set txt
}

variable ::two::html::CharEntities
array set ::two::html::CharEntities {
    lt     "<"
    gt     ">"
    amp    "&"
    quot   \"
    nbsp   " "

    pound  ""
    yen    ""
    brvbar ""
    sect   ""
    copy   ""
    laquo  ""
    not    ""
    shy    ""
    reg    ""
    deg    ""
    plusmn ""
    sup1   ""
    sup2   ""
    sup3   ""
    acute  ""
    micro  ""
    para   ""
    middot ""
    raquo  ""
    frac14 ""
    frac12 ""
    frac34 ""
    iquest ""

    auml   ""
    ouml   ""
    uuml   ""
    Auml   ""
    Ouml   ""
    Uuml   ""
    szlig  ""

    aelig  ""
    AElig  ""

    aacute ""
    acirc  ""
    agrave ""
    eacute ""
    ecirc  ""
    egrave ""
    oslash ""
    oacute ""
    ocirc  ""
    ograve ""
    uacute ""
    ucirc  ""
    ugrave ""
}


# ::two::html::base64_encode --
#
# 	do base-64 encoding (see RFC 2045 for details). This version
#	works on strings; it transform the input string given as
#	argument into the base64-encoded version which is returned.
#	CAVEAT EMPTOR: When reading the input data from whatever
#	source (like a file), make sure that you don't use any kind
#	of EOL translation => fconfigure $fh -translation binary
#
# Arguments:
# 	orig -		binary data which is to be encoded
#
# Results:
#	base64-encoded string

proc ::two::html::base64_encode {orig} {

    set base64en {
        A B C D E F G H I J K L M N O P Q R S T U V W X Y Z
	a b c d e f g h i j k l m n o p q r s t u v w x y z
	0 1 2 3 4 5 6 7 8 9 + /
    }
    set pos 0
    set out ""

    foreach {i0 i1 i2} [split $orig {}] {
        set in "$i0$i1$i2"
        set num [string length $in]
	binary scan \000$in\000\000 I b32
	for {set idx 18} {$idx >= (3-$num)*6} {incr idx -6} {
	    append out [lindex $base64en [expr {($b32>>$idx) & 0x3F}]]
	}
	if {$num < 3} {
	    for {set idx 0} {$idx < 3-$num} {incr idx} {
	        append out "="
	    }
	    append out "\n"
	    set pos 0
	} elseif {$pos == 17} {
	    append out "\n"
	    set pos 0
	} else {
	    incr pos
	}
    }
    return $out
}

namespace import ::two::html::*

# splitter.tcl --
#
#	Provides an "interactive geometry manager". Slave widgets are
#	arranged in rows or columns, with the user being able to modify
#	the size of each such pane.
#
# Copyright (c) 1998, 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoSplitter 1.0
package require twoUtil

#option add *Splitter.background                grey51
#option add *Splitter.background                black
option add *TwoSplitterHandle.borderWidth         0
option add *TwoSplitterHandle.relief              raised
option add *TwoSplitterHandle.borderWidth         1

namespace eval ::two::splitter {
    namespace export two_splitter
    variable Splitter
    
    namespace import ::two::util::two_parseArgs
    
    # variable Splitter
    #
    # $w,s,style -	style of slave (unix|windows)
    # $w,s,dyn -	dynamic resizing on/off
    # $w,s,h -	path to handle for this slave
    # $w,s,fract -	current fraction
    # $m,m,orient -	orientation (horizontal or vertical)
    # $m,m,slaves -	list of slaves
    # $m,m,len -	number of slaves
    # $h -		slave associated with this handle
    # style -	default style
    # width -	default width
    # orient -	default orientation
    # cursor -	cursor for non-dragable area on unix-style handle
    # drag -	currently dragging any splitter
    # mod -		the currently dragged splitter has been modified
    
    array set Splitter "
        style	$tcl_platform(platform)
        dynamic	1
        orient	vertical
        cursor	left_ptr
        drag	0
        mod		0
    "
}

# ::two::splitter::two_splitter --
#
#	The splitter is implemented as a layout manager; it doesn't subclass
#	any existing widget.
#
# Arguments:
#	opt		Widget command...
#	args		optional args

proc ::two::splitter::two_splitter {opt args} {
    variable Splitter

    switch -glob -- $opt {
	.* {
	    set orient  ""
	    set before  ""
	    set after   ""
	    set style   $Splitter(style)
	    set dynamic $Splitter(dynamic)
	    two_parseArgs args orient  "-orient"
	    two_parseArgs args style   "-style"
	    two_parseArgs args dynamic "-dynamic"
	    two_parseArgs args after   "-after"
	    two_parseArgs args before  "-before"
	    set m [winfo parent $opt]
	    set new [linsert $args 0 $opt]

	    # settings for master
	    if {[string length $orient]} {
	        set Splitter($m,m,orient) $orient
	    } elseif {![info exists Splitter($m,m,orient)]} {
	        set Splitter($m,m,orient) $Splitter(orient)
	    }

	    # New splitter?
	    if {![info exists Splitter($m,m,slaves)]} {
	        set Splitter($m,m,slaves) ""
	    } else {
	        # any of those widgets already managed by splitter?
		foreach w $new {
		    if {[info exists Splitter($w,s,fract)]} {
		        two_splitter forget $w
		    }
		}
	    }

	    # look for position where to insert new widgets
	    set idx [llength $Splitter($m,m,slaves)]
	    if {[string length $before]} {
	        if {![winfo exists $before]} {
		    return -code error "bad window path name \"$before\""
		}
	        set idx [lsearch -exact $Splitter($m,m,slaves) $before]
		if {$idx < 0} {
		    return -code error "window \"$before\" isn't managed by splitter"
		}
	    }
	    if {[string length $after]} {
	        if {![winfo exists $after]} {
		    return -code error "bad window path name \"$after\""
		}
	        set idx [lsearch -exact $Splitter($m,m,slaves) $after]
		if {$idx >= 0} {
		    incr idx
		} else {
		    return -code error "window \"$after\" isn't managed by splitter"
		}
	    }

            # are there already some slaves?
	    if {[llength $Splitter($m,m,slaves)]} {
	        set len [llength [linsert $args 0 $opt]]
	        set mul [expr {(double($Splitter($m,m,len)) / double($Splitter($m,m,len) + $len))}]
	        foreach w [lrange $Splitter($m,m,slaves) 0 [expr {$idx - 1}]] {
		    # reduce fraction
		    set Splitter($w,s,fract) [expr {$Splitter($w,s,fract) * $mul}]
		}
	        foreach w [lrange $Splitter($m,m,slaves) $idx [expr {$Splitter($m,m,len) - 2}]] {
		    # increase fraction
		    set Splitter($w,s,fract) [expr {1 - (1-$Splitter($w,s,fract))*$mul}]
		}
	    }
	    
            # process all widgets on command line
	    foreach w $new {
	        # set widget config
		set Splitter($w,s,style)  $style
		set Splitter($w,s,dyn)    $dynamic
		set Splitter($m,m,slaves) [linsert $Splitter($m,m,slaves) $idx $w]
		incr idx
		bind $w <Destroy> "+two_splitter forget $w"
		
                # This is a multi-style widget
                set ::two::SetStyleCmds($w) [list two_splitter configure $w -style]
	    }
	    # re-set master
	    Config $m
	}
	configure {
	    if {[llength $args] < 1} {
	        return -code error "wrong # args: should be \"two_splitter configure path ?options?\""
	    }
	    set w    [lindex $args 0]
	    set m    [winfo parent $w]
	    if {[llength $args] == 1} {
	        return [list \
		    [list -orient  $Splitter($m,m,orient)]  \
		    [list -style   $Splitter($w,s,style)]   \
		    [list -dynamic $Splitter($w,s,dyn)] \
		    [list -fract   $Splitter($w,s,fract)]   \
		]
	    }
	    set args [lrange $args 1 end]
	    set or   [two_parseArgs args Splitter($m,m,orient) "-orient"]
	    set st   [two_parseArgs args Splitter($w,s,style)  "-style"]
	    two_parseArgs args Splitter($w,s,dyn) "-dynamic"
	    two_parseArgs args Splitter($w,s,fract)   "-fract"
	    if {[llength $args]} {
	        return -code error "unknown option \"$args\""
	    }
	    if {$or} {
	        foreach slave [lrange $Splitter($m,m,slaves) 0 [expr $Splitter($m,m,len) - 2]] {
	            if {[info exists Splitter($slave,s,h)]} {
		        unset Splitter([lindex $Splitter($slave,s,h) 0])
	        	foreach h $Splitter($slave,s,h) {
			    destroy $h
			}
		    }
		}
	    } elseif {$st} {
	        if {[info exists Splitter($w,s,h)]} {
	            unset Splitter([lindex $Splitter($w,s,h) 0])
	            foreach h $Splitter($w,s,h) {
			destroy $h
		    }
		}
	    }
	    Config $m
	}
	cget {
	    if {[llength $args] != 2} {
	        return -code error "wrong # args: should be \"two_splitter cget option\""
	    }
	    set w   [lindex $args 0]
	    set m   [winfo  parent $w]
	    set opt [lindex $args 1]
	    switch -exact -- $opt {
	        -orient  { return $Splitter($m,m,orient)  }
	        -style   { return $Splitter($w,s,style)   }
	        -dynamic { return $Splitter($w,s,dyn) }
		-fract   { return $Splitter($w,s,fract)   }
		default  { return -code error "unknown option \"$opt\"" }
	    }
	}
	slaves {
	    set m    [lindex $args 0]
	    set args [lrange $args 1 end]
	    return $Splitter($m,m,slaves)
	}
	forget {
	    foreach w $args {
                unset ::two::SetStyleCmds($w)
	        place forget $w
	
		if {[regsub "two_splitter forget $w" [bind $w <Destroy>] "" newbinding]} {
		    bind $w <Destroy> $newbinding
		}
	        set m [winfo parent $w]
		# remove all handles (if any remain)
	        foreach slave [lrange $Splitter($m,m,slaves) 0 [expr $Splitter($m,m,len) - 2]] {
	            if {[info exists Splitter($slave,s,h)]} {
		        unset Splitter([lindex $Splitter($slave,s,h) 0])
			foreach h $Splitter($slave,s,h) {
		            destroy $h
			}
			unset Splitter($slave,s,h)
		    }
		}
		# remove slave from master's list of slaves
		set pos [lsearch -exact $Splitter($m,m,slaves) $w]
                set Splitter($m,m,slaves) [lreplace $Splitter($m,m,slaves) $pos $pos]
		# correct fractions of remaining slaves
	        set len [llength $Splitter($m,m,slaves)]
		if {$len == 1} {
		    set slave [lindex $Splitter($m,m,slaves) 0]
		    set Splitter($slave,s,fract) 1.0
		} elseif {$len>0 && $pos >= 0} {
		    set prev [lindex $Splitter($m,m,slaves) [expr $pos - 1]]
		    set mul [expr {(double($len) + 1) / double($len) }]
		    foreach slave [lrange $Splitter($m,m,slaves) 0 [expr {$pos - 1}]] {
			set Splitter($slave,s,fract) [expr {$Splitter($slave,s,fract) * $mul}]
		    }
		    foreach slave [lrange $Splitter($m,m,slaves) $pos [expr {$Splitter($m,m,len) - 1}]] {
			set Splitter($slave,s,fract) [expr {1 - ((1- $Splitter($slave,s,fract)) * $mul)}]
		    }
		}
		# remove state of slave
		foreach key [array names Splitter "$w,s,*"] {
        	    unset Splitter($key)
		}
		# last slave?
		if {$len == 0} {
		    foreach key [array names Splitter "$m,m,*"] {
        		unset Splitter($key)
		    }
		}
		# re-set master
		if {[info exists Splitter($m,m,slaves)]} {
		    Config $m
		}
	    }
	}
	default {
            return -code error "bad option \"$opt\": must be cget, configure, forget, slaves or a widget path"
	}
    }
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoSplitter-class bindings
#-------------------------------------------------------------------------

# Button 1: record state
bind TwoSplitter <1> {
    ::two::splitter::SplitterB1 %W %x %y
}

# B1-Release: drag finished
bind TwoSplitter <B1-ButtonRelease> {
    ::two::splitter::SplitterB1Release %W
}

# B1-Motion: resize slaves and/or drag handle
bind TwoSplitter <B1-Motion> {
    ::two::splitter::SplitterB1Motion %W %X %Y
}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::splitter::SplitterB1 --
#
#	handle button press within splitter frame.
#
# Arguments:
#	h		Path splitter
#	x		relative X Position of mouse cursor within splitter (%x)
#	y		relative Y Position (%y)

proc ::two::splitter::SplitterB1 {h x y} {
    variable Splitter

    set w $Splitter($h)
    set m [winfo parent $w]
    set Splitter(drag) 1
    set Splitter(mod)  0
    set Splitter(w)    $w
    set Splitter(m)    $m
    set Splitter(hor)  [string match "$Splitter($m,m,orient)*" "horizontal"]
    set Splitter(size) [expr {$Splitter(hor) ? [winfo width $m] : [winfo height $m]}]
    set Splitter(min)  [expr {int($Splitter($w,s,min) * $Splitter(size) + 10)}]
    set Splitter(max)  [expr {int($Splitter($w,s,max) * $Splitter(size) - 10)}]
    set Splitter(off)  [expr {$Splitter(hor) ? $x - 2 : $y - 2}]
    set Splitter(pos)  [expr {$Splitter(hor) ? [winfo rootx $m] : [winfo rooty $m]}]
    if {![string match "unix" $Splitter($w,s,style)]} {
	set Splitter(bg) [$h cget -bg]
	$h configure -bg black
    }
    return
}

# ::two::splitter::SplitterB1Motion --
#
#	handle mouse motion while button pressed within splitter frame.
#
# Arguments:
#	h		Path splitter
#	absX		relative X Position of mouse cursor within splitter (%x)
#	absY		relative Y Position (%y)

proc ::two::splitter::SplitterB1Motion {h absX absY} {
    variable Splitter

    if {!$Splitter(drag)} {
        return
    }

    set Splitter(mod) 1
    set w $Splitter(w)
    set m $Splitter(m)
    if {$Splitter(hor)} {
        set x [expr {$absX - $Splitter(pos) - $Splitter(off)}]
	if {$x < $Splitter(min)} {
	    set Splitter(x) $Splitter(min)
	} elseif {$x > $Splitter(max)} {
	    set Splitter(x) $Splitter(max)
	} else {
	    set Splitter(x) $x
	}
        if {!$Splitter($w,s,dyn)} {
	    if {[string match "unix" $Splitter($w,s,style)]} {
		place $h -x [expr {$Splitter(x) - 2}] -relx 0.0
        	place ${h}_sep -x [expr {$Splitter(x) + 1}] -relx 0.0
	    } else {
		place $h -x $Splitter(x) -relx 0.0
	    }
	}
    } else {
        set y [expr {$absY - $Splitter(pos) - $Splitter(off)}]
	if {$y < $Splitter(min)} {
	    set Splitter(y) $Splitter(min)
	} elseif {$y > $Splitter(max)} {
	    set Splitter(y) $Splitter(max)
	} else {
	    set Splitter(y) $y
	}
        if {!$Splitter($w,s,dyn)} {
	    if {[string match "unix" $Splitter($w,s,style)]} {
		place %W -y [expr {$Splitter(y) - 2}] -rely 0.0
        	place %W_sep -y [expr {$Splitter(y) + 1}] -rely 0.0
	    } else {
		place %W -y $Splitter(y) -rely 0.0
	    }
	}
    }
    if {$Splitter($w,s,dyn)} {
        set w $Splitter(w)
	if {$Splitter(hor)} {
	    set Splitter($w,s,fract) [expr {double($Splitter(x)) / double([winfo width $m])}]
	} else {
	    set Splitter($w,s,fract) [expr {double($Splitter(y)) / double([winfo height $m])}]
	}
        # TODO: only place the 2 neighbour slaves
	Config $m
    }
    return
}

# ::two::splitter::SplitterB1Release --
#
#	handle button release within splitter frame.
#
# Arguments:
#	h		Path splitter

proc ::two::splitter::SplitterB1Release {h} {
    variable Splitter

    set w $Splitter(w)
    set m $Splitter(m)
    set Splitter(drag) 0
    if {![string match "unix" $Splitter($w,s,style)]} {
	$h configure -bg $Splitter(bg)
    }
    if {!$Splitter(mod)} {
        return
    }
    if {$Splitter(hor)} {
        if {[string match "unix" $Splitter($w,s,style)] && !$Splitter($w,s,dyn)} {
	    incr Splitter(x) 2
	}
        set Splitter($w,s,fract) [expr {double($Splitter(x)) / double([winfo width $m])}]
    } else {
        if {[string match "unix" $Splitter($w,s,style)] && !$Splitter($w,s,dyn)} {
	    incr Splitter(y) 2
	}
        set Splitter($w,s,fract) [expr {double($Splitter(y)) / double([winfo height $m])}]
    }
    # TODO: only place the 2 neighbour slaves
    Config $m
    return
}

# ::two::splitter::Config --
#
#	recalculate settings of panned frame
#
# Arguments:
#	m		Path to master widget

proc ::two::splitter::Config {m} {
    variable Splitter

    set len [llength $Splitter($m,m,slaves)]
    set hor [string match "$Splitter($m,m,orient)*" "horizontal"]
    set idx 0
    set fract [expr 1.0 / $len]
    set Splitter($m,m,len) $len
    set oldfract 0.0
    foreach w [lrange $Splitter($m,m,slaves) 0 [expr $len - 1]] {
        # calculate required fraction...
        if {![info exists Splitter($w,s,fract)]} {
	    set Splitter($w,s,fract) $fract
	} else {
	    set fract $Splitter($w,s,fract)
	}
	if {$idx == $len - 1} {
	    set fract 1.0
	}
	# remember index
	set Splitter($w,s,idx) $idx
	# build/configure handle
	if {$idx != 0} {
	    if {[string match "." $m]} {
	        set h ".__h$idx"
	    } else {
	        set h "$m.__h$idx"
	    }
	    # splitter has wrong type?
	    if {[string match "unix" $Splitter($oldw,s,style)]} {
	        if {![winfo exists ${h}_sep] && [winfo exists $h]} {
	            destroy $h
		}
	    } else {
	        if {[winfo exists ${h}_sep]} {
	            destroy $h
		}
	    } 
	    if {![winfo exists $h]} {
		set cursor [expr {$hor ? "sb_h_double_arrow" : "sb_v_double_arrow"}]

                if {[string match "unix" $Splitter($oldw,s,style)]} {
	            frame $h -class TwoSplitterHandle -cursor $cursor
		    frame ${h}_sep -class TwoSplitterSep -bd 1 -relief sunken
		    raise $h
		    bindtags $h [list $h TwoSplitter [winfo toplevel $m] all]
		    set Splitter($oldw,s,h) [list $h ${h}_sep]
		} else {
	            frame $h -class TwoSplitter -cursor $cursor
		    set Splitter($oldw,s,h) [list $h]
		}
	    }
	    set Splitter($h) $oldw
	}
        # offset for pane sizes depending on style and position
	if {$idx == $len - 1} {
            # last pane
	    set off1 0
	    set off2 0
	    set off3 -1
	    set off4 -2
	} else {
	    set off1 -1 ;# first pane in unix style
	    set off2 -2 ;# first pane in windows style
	    set off3 -2 ;# subsequent pane in unix style
	    set off4 -4 ;# subsequent pane in windows style
	}
	# place children and splitter handles
	# o.k., this is lengthy code with a lot of cut-and-paste:-(
	if {$hor} {
	    if {$idx == 0} {
	        # the first pane
        	if {[string match "unix" $Splitter($w,s,style)]} {
	            place $w -rely 0.0 -relheight 1.0 -relx $oldfract -relwidth [expr {$fract - $oldfract}] -x 0 -width $off1
		} else {
	            place $w -rely 0.0 -relheight 1.0 -relx $oldfract -relwidth [expr {$fract - $oldfract}] -x 0 -width $off2
		}
	    } else {
	        # subsequent panes
        	if {[string match "unix" $Splitter($oldw,s,style)]} {
	            place $w -rely 0.0 -relheight 1.0 -relx $oldfract -relwidth [expr {$fract - $oldfract}] -x 1 -width $off3
		    place $h -y -12 -rely 1.0 -height 8 -relx $oldfract -x -4 -width 8
		    place ${h}_sep -relheight 1.0 -relx $oldfract -x -1 -width 2 -height 4
		} else {
	            place $w -rely 0.0 -relheight 1.0 -relx $oldfract -relwidth [expr {$fract - $oldfract}] -x 2 -width $off4
		    place $h -rely 0.0 -relheight 1.0 -relx $oldfract -x -2 -width 4
		}
	    }
	} else {
	    if {$idx == 0} {
	        # the first pane
        	if {[string match "unix" $Splitter($w,s,style)]} {
	            place $w -relx 0.0 -relwidth 1.0 -rely $oldfract -relheight [expr {$fract - $oldfract}] -y 0 -height $off1
		} else {
	            place $w -relx 0.0 -relwidth 1.0 -rely $oldfract -relheight [expr {$fract - $oldfract}] -y 0 -height $off2
		}
	    } else {
	        # subsequent panes
        	if {[string match "unix" $Splitter($oldw,s,style)]} {
	            place $w -relx 0.0 -relwidth 1.0 -rely $oldfract -relheight [expr {$fract - $oldfract}] -y 1 -height $off3
		    place $h -x -12 -relx 1.0 -width 8 -rely $oldfract -y -4 -height 8
		    place ${h}_sep -relwidth 1.0 -rely $oldfract -y -1 -height 2 -width 4
		} else {
	            place $w -relx 0.0 -relwidth 1.0 -rely $oldfract -relheight [expr {$fract - $oldfract}] -y 2 -height $off4
		    place $h -relx 0.0 -relwidth 1.0 -rely $oldfract -y -2 -height 4
		}
	    }
	}
	# set limits
	if {$len == 1} {
	    set Splitter($w,s,min) 0.0
	    set Splitter($w,s,max) 1.0
	} elseif {$idx == 0} {
            set Splitter($w,s,min) 0.0
	} elseif {$idx == $Splitter($m,m,len) - 1} {
	    set Splitter($w,s,min) $oldfract
            set Splitter($oldw,s,max) 1.0
	} else {
	    set Splitter($w,s,min) $oldfract
            set Splitter($oldw,s,max) $fract
	}
	# remember for next loop
	set oldw     $w
	set oldfract $fract
	
	# calculate fract for next pane (all subsequent panes are equally sized by default)
	if {$idx == $len - 1} {
	    set fract 1.0
	} else {
	    set fract [expr {$fract + (1 - $fract) / ($len - $idx - 1)}]
	}

	incr idx
    }
    return
}

namespace import ::two::splitter::*

# tooltip.tcl --
#
#	Showing tooltips... Tooltips can be applied on buttons and
#	Menu entries. They can provide a short tip message wich appears
#	after the mouse resting for some time over the button (not
#	available on menu entries!). Additionally, the global variable
#	two_ttInfo is used to hold a more detailed description which
#	could be used in the status line of an application.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoTooltip 1.0
package require twoUtil

option add *TwoToolTip.borderWidth       1
option add *TwoToolTip.background        black
option add *TwoToolTip.Label.foreground  black
option add *TwoToolTip.Label.background  lightYellow
option add *TwoToolTip.Label.wrapLength  250
option add *TwoToolTip.Label.padX        3
option add *TwoToolTip.Label.padY        3
option add *TwoToolTip.Label.justify     center

set two_ttInfo ""

namespace eval ::two::tooltip {
    namespace export two_tooltip
    variable Tooltip
    global two_ttInfo
    
    namespace import ::two::util::two_parseArgs
    
    # variable Tooltip
    #
    # enabled -	true if tooltips are enabled (only affects toplevel)
    # visible -	true if a tooltip is currently visible
    # offset -	y-distance from widget
    # delay -	delay between mouse enter event and showing of tooltip
    # duration -    how long is the tooltip visible
    # a_delay -	id returned by "after $Tooltip(delay)"
    # a_duration -	id returned by "after $Tooltip(duration)"
    # scr_w -	screen width (ensure that whole tooltip is on screen)
    # scr_h -	screen height (ensure that whole tooltip is on screen)
    # $w,tip -	tooltip for widget $w
    # $w,info -	info text for widget $w
    # $w,$id -	info text for index $id if $w is a menu
    
    array set Tooltip "
        enabled	1
        visible	0
        offset	4
        delay	500
        duration	5000
        scr_w	[winfo screenwidth .]
        scr_h	[winfo screenheight .]
        visible     0
        a_delay     0
        a_duration  0
    "
}

# ::two::tooltip::two_tooltip --
#
#	Add tooltip to widget w. Create toplevel "._two_tooltip" if it doesn't exist.
#	Enable/disable tooltip. Configure tooltips.
#
# Arguments:
#	w		Widget path
#	args		Tooltip arguments and options (see manpage)

proc ::two::tooltip::two_tooltip {w args} {
    variable Tooltip

    switch -exact -- $w {
	delay {
	    if {[llength $args] > 1} {
	    	return -code error "wrong # args: should be \"two_tooltip delay ?ms?\""
	    } elseif {[llength $args] == 1} {
		return [set Tooltip(delay) [lindex $args 0]]
	    } else {
		return $Tooltip(delay)
	    }
	}
	duration {
	    if {[llength $args] > 1} {
	    	return -code error "wrong # args: should be \"two_tooltip duration ?ms?\""
	    } elseif {[llength $args] == 1} {
		return [set Tooltip(duration) [lindex $args 0]]
	    } else {
		return $Tooltip(duration)
	    }
	}
	1 -
	on -
	enable {
	    set Tooltip(enabled) 1
	}
	0 -
	off -
	disable {
	    set Tooltip(enabled) 0
	    Hide 1
	}
	delete {
	    Hide
            foreach key [array names Tooltip "$w,*"] {
                unset Tooltip($key)
            }
	}
	add {
	    Create
	    eval Add $w [lrange $args 1 end]
	}
	default {
	    Create
	    eval Add $w $args
	}
    }
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoToolTipButton-class bindings
#-------------------------------------------------------------------------

bind TwoToolTipButton <Enter>      {::two::tooltip::Enter %W}
bind TwoToolTipButton <Leave>      {::two::tooltip::Hide 1}
bind TwoToolTipButton <Any-Key>    {::two::tooltip::Hide}
bind TwoToolTipButton <Any-Button> {::two::tooltip::Hide}

# And a binding to the tooltip itself:
bind TwoToolTip <Enter> {::two::tooltip::Hide 1}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::tooltip::Create --
#
#	Create the toplevel if it doesn't yet exist
#

proc ::two::tooltip::Create {} {
    if {![winfo exists ._two_tooltip]} {
        toplevel            ._two_tooltip -class TwoToolTip
        wm overrideredirect ._two_tooltip 1
        wm positionfrom     ._two_tooltip program
        wm withdraw         ._two_tooltip
        pack [label         ._two_tooltip.l -highlightthickness 0 -bd 0]
    }
}

# ::two::tooltip::Add --
#
#	Add a tooltip to a widget or menu. Uses bindtags TwoToolTipButton for
#	buttons, but needs to bind directly to menus.
#
# Arguments:
#	w		Widget path
#	args		?-index index? ?-tip m1? ?-info m2?

proc ::two::tooltip::Add {w args} {
    variable Tooltip

    set id  ""
    set tag ""
    set m1  ""
    set m2  ""
    set istext 0
    two_parseArgs args id  "-index"
    two_parseArgs args tag "-tag"
    two_parseArgs args m1  "-tip"
    two_parseArgs args m2  "-info"
    if {[llength $args]} {
	return -code error "bad argument \"$args\" for \"two_tooltip\": must be -index index, -tip text or -info text"
    }
    if {![string length $m2]} {
        set m2 $m1
    }
    if {[string length $id]} {
        # try to find out if this really is a menu
	if {[catch {$w cget -tearoff}]} {
	    return -code error "widget \"$w\" is no menu!"
	}
	# get numercial index
	set id [$w index $id]
	# set tip
	set Tooltip($w,$id)  $m2
    } elseif {[string length $tag]} {
        # try to find out if this really is a text or a canvas widget
	if {![catch {$w tag bind dummy}]} {
	    # Appears to be a text widget
	    set istext 1
	} elseif {[catch {$w bind dummy}]} {
	    return -code error "widget \"$w\" is no text or canvas widget!"
	}
	# set tip
	set Tooltip($w,tip,$tag)  $m1
	set Tooltip($w,info,$tag) $m2
    } else {
	set Tooltip($w,tip)  $m1
	set Tooltip($w,info) $m2
    }
    
    set Tooltip(visible) 0

    if {[string length $id]} {
        # Menu
	# We have to bind directly to the menu and use $w instead of %W:
	# The tk8.0 menu system uses clones with varying widget names!
	# Overwrite binding (don't use {+ code}) to avoid multiple binding
	bind $w <Motion> [namespace code "MenuSelect $w \[%W index @%x,%y\]"]
    } elseif {[string length $tag]} {
        if {$istext} {
            $w tag bind $tag <Enter>      [namespace code [list Enter %W $tag]]
            $w tag bind $tag <Leave>      [namespace code {Hide 1}]
            $w tag bind $tag <Any-Key>    [namespace code Hide]
            $w tag bind $tag <Any-Button> [namespace code Hide]
        } else {
            $w bind $tag <Enter>      [namespace code [list Enter %W $tag]]
            $w bind $tag <Leave>      [namespace code {Hide 1}]
            $w bind $tag <Any-Key>    [namespace code Hide]
            $w bind $tag <Any-Button> [namespace code Hide]
        }
    } else {
        # Normal button
	bindtags $w [concat TwoToolTipButton [bindtags $w]]
    }
    return
}

# ::two::tooltip::MenuSelect --
#
#	Binding for changing the selection within a menu
#
# Arguments:
#	w1		Widget path for original widget
#	idx		active index

proc ::two::tooltip::MenuSelect {w idx} {
    variable Tooltip
    global two_ttInfo

    if {[info exists Tooltip($w,$idx)]} {
	set two_ttInfo $Tooltip($w,$idx)
    } elseif {$two_ttInfo != ""} {
	set two_ttInfo ""
    }
    return
}

# ::two::tooltip::Enter --
#
#	Mouse has entered the hot region - show info and start
#	waiting for shwoing the tip.
#
# Arguments:
#	w		Widget path
#	tag		tag within this widget

proc ::two::tooltip::Enter {w {tag ""}} {
    variable Tooltip
    global two_ttInfo

    if {[string length $tag]} {
        if {[string length $Tooltip($w,info,$tag)]} {
            set two_ttInfo $Tooltip($w,info,$tag)
        }
        if {[string length $Tooltip($w,tip,$tag)]} {
            set Tooltip(a_delay) \
                [::after $Tooltip(delay) [namespace code [list Show $w $tag]]]
        }
    } else {
        if {[string length $Tooltip($w,info)]} {
            set two_ttInfo $Tooltip($w,info)
        }
        if {[string length $Tooltip($w,tip)]} {
            set Tooltip(a_delay) \
                [::after $Tooltip(delay) [namespace code [list Show $w]]]
        }
    }
}

# ::two::tooltip::Show --
#
#	show tooltip for widget $w
#
# Arguments:
#	w		Widget path
#	tag		tag within this widget

proc ::two::tooltip::Show {w {tag ""}} {
    variable Tooltip

    if {!$Tooltip(enabled)} {
        return
    }

    # Mouse still above widget?
    if {![string match $w [winfo containing [winfo pointerx .] [winfo pointery .]]]} {
	return
    }
    if {[string length $tag]} {
        ._two_tooltip.l configure -text $Tooltip($w,tip,$tag)
    } else {
        ._two_tooltip.l configure -text $Tooltip($w,tip)
    }
    # Calculate position
    if {[string length $tag]} {
        set x [expr {5 + [winfo pointerx $w]}]
        set y [expr {5 + [winfo pointery $w] + $Tooltip(offset)}]
    } else {
        set x [winfo pointerx $w]
        set y [expr {[winfo rooty $w] + [winfo height $w] + $Tooltip(offset)}]
        if {$y < 0} {set y 0}
    }
    set w [winfo reqwidth  ._two_tooltip.l]
    set h [winfo reqheight ._two_tooltip.l]

    # Make shure that the whole tooltip is visible
    if {$y + $h > $Tooltip(scr_h)} {
	set y [expr {$Tooltip(scr_h) - $h}]
	incr x $Tooltip(offset)
    }
    if {$x + $w > $Tooltip(scr_w)} {
	set x [expr {$Tooltip(scr_w) - $w}]
    }

    # Remove tooltip after duration:
    set Tooltip(a_duration) \
    	[::after $Tooltip(duration) [namespace code Hide]]

    # really show tooltip
    # The update command is needed on windows.
    # BUT: maybe another binding (e.g. on cursor leave) has fired in
    # the meantime, and thus the Hide command is called. This results
    # in a stale tooltip... $Tooltip(visible) is used to circumvent this.
    set Tooltip(visible) 1
    wm geometry  ._two_tooltip +$x+$y
    update
    if {$Tooltip(visible)} {
	wm deiconify ._two_tooltip
	raise ._two_tooltip
    }
    return
}

# ::two::tooltip::Hide --
#
#	remove toplevel Tooltip widget and cancel "after"
#	if info is 1, then also remove tooltip info

proc ::two::tooltip::Hide { {info 0} } {
    variable Tooltip
    global two_ttInfo

    set Tooltip(visible) 0
    if {[winfo exists ._two_tooltip]} {
	wm withdraw ._two_tooltip
    }
    catch {::after cancel $Tooltip(a_delay)}
    catch {::after cancel $Tooltip(a_duration)}
    if {$info && [string length $two_ttInfo]} {
        set two_ttInfo ""
    }
    return
}

namespace import ::two::tooltip::*

# menu.tcl --
#
#	Enhanced menu command: alias names for menu entries, handling
#	of "&" to set "underline", tooltips.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoMenu 1.0
package require twoUtil
package require twoTooltip

namespace eval ::two::menu {
    namespace export two_menu
    variable Menu
    
    namespace import ::two::util::two_parseMultiArgs
    
    # variable Menu
    #
    # $w,alias -	index position of alias "alias"
}

# ::two::menu::two_menu --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.xm"); the original widget procedure
#	is renamed to "_two$w" (like "_two.tl.xm") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::menu::two_menu {w args} {
    # Create menu
    uplevel [list menu $w] $args

    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Combobox
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # Return the correct result...
    return $w
}

# ::two::menu::Widget --
#
#	replacement for widget command
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

variable ::two::menu::MenuSpec
array set ::two::menu::MenuSpec {
    -alias	alias
    -tip	tip
    -info	info
    -x		xaccel
    -xaccel	xaccel
    -text	label
    -l		label
    -label	label
}

proc ::two::menu::Widget {w opt args} {
    variable Menu
    # TCL Bug? Using variable would break upvar in called proc:-(
    upvar MenuSpec MenuSpec

    set alias  ""
    set tip    ""
    set info   ""
    set xaccel ""
    set label ""
    # See if this option is for us...
    switch -exact -- $opt {
	add -
	insert -
	entryconfigure {
	    two_parseMultiArgs MenuSpec args
	    # label was specified
	    if {[string length $label]} {
	        set idx [string first "&" $label]
		if {$idx >= 0} {
		    set underline $idx
		    # ignore "&&"
		    while {[string match "&" [string index $label [expr {$idx + 1}]]]} {
			set base [expr {$idx + 2}]
		        set idx  [string first "&" [string range $label $base end]]
			if {$idx < 0} {
			    break
			} else {
			    set underline [expr {$underline + $idx + 1}]
			    incr idx $base
			}
		    }
		}
		if {$idx >= 0} {
		    regsub -all -- {&([^&])} $label {\1} label
		    lappend args -label $label -underline $underline
		} else {
		    lappend args -label $label
		}
	    }
	    # calculate index for insert/entryconfigure
	    if {![string match "add" $opt]} {
	        set index [lindex $args 0]
		if {[string length $alias]} {
	            # new alias
	            set Menu($w,$alias) $index
		} else {
	            if [info exists Menu($w,$index)] {
                        # there already is an alias
			set index $Menu($w,$index)
			set args [concat $index [lrange $args 1 end]]
		    }
		}
	    }
	    # create entry
	    if {[llength $args]} {
		set retval [eval [list _two$w $opt] $args]
	    }
            
	    # calculate index for "add"
	    if {[string match "add" $opt]} {
	        set index [_two$w index end]
		if {[string length $alias]} {
	            set Menu($w,$alias) $index
		}
            }

	    if {[string length $xaccel] == 1} {
	        _two$w entryconfigure $index -accelerator "Ctrl+$xaccel"
                # eval postcommand to set proper state of entries
		# then invoke entry (invoke respects state!)
		# Can't use [winfo toplevel $w] as this would return
		# a false toplevel:-(
		bind . <Control-Key-$xaccel> "
		    eval \[[list $w] cget -postcommand\]
		    [list $w invoke $index]
		    break
		"
	    } elseif {[string match "F*" $xaccel]} {
	        _two$w entryconfigure $index -accelerator "$xaccel"
		bind . <$xaccel> "
		    eval \[[list $w] cget -postcommand\]
		    [list $w invoke $index]
		    break
		"
	    }
	    if {[string length $tip] || [string length $info]} {
	        two_tooltip $w -index $index -tip $tip -info $info
	    }
	    return $retval
	}
	entrycget {
	    switch -exact -- [lindex $args 1] {
	        -tip    { return -code error "Sorry, not yet implemented" }
	        -info   { return -code error "Sorry, not yet implemented" }
		-alias  { return -code error "Sorry, not yet implemented" }
	        -text   { return [eval [list _two$w $opt] [lreplace $args 1 1 "-label"]] }
		default { return [eval [list _two$w $opt] $args] }
	    }
	}
	activate -
	delete -
	index -
	invoke -
	postcascade -
	type -
	yposition {
	    set arglist ""
	    foreach a $args {
                # is this an alias?
		if {[info exists Menu($w,$a)]} {
		    lappend arglist $Menu($w,$a)
		} else {
		    lappend arglist $a
		}
	    }
	    return [uplevel [list _two$w $opt] $arglist]
	}
	default {
            return [uplevel [list _two$w $opt] $args]
	}
    }
    return
}

# ::two::menu::_Destroy --
#
#	Call this procedure when the two_menu widget is destroyed.
#	Will destroy the original widget command.
#
# Arguments:
#	w		Path to Widget

proc ::two::menu::_Destroy {w} {
    # remove the original widget command
    rename $w ""
    return
}

namespace import ::two::menu::*

# mdi.tcl --
#
#	MDI Manager widget. This widget can hold an arbitrary number
#	of child widgets which look like KDE or Win95. The user can
#	manipulate them in the usual way.
#
# Copyright (c) 1998, 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoMdi 1.0
package require twoUtil
package require twoMenu

# Workaround: Tk needs these procedures for focus traversion
# but won't load it form inside the namespace (bug in tcl8.0 up to p1)
catch {auto_load tk_focusNext}
catch {auto_load tk_focusPrev}

#-------------------------------------------------------------------------
# naming convention:
# $m, master -	Main MDI widget, contains slaves.
# $b, border -	Direct child of master, for internal use only. Provides
#		border for slave widget. Caveat: [winfo slaves $m] will
#		return the borders (and popup menus), not the slaves.
# $s, slave -	The tcl/tk frame widget which is returned to the MDI
#		user. More details: see doc for AddSlave procedure
#-------------------------------------------------------------------------

option add *TwoMdiMaster.background                grey51
option add *TwoMdi.relief                          raised
option add *TwoMdi.cursor                          left_ptr
option add *TwoMdiTitle.Button.highlightThickness  0
option add *TwoMdiTitle.Button.background          grey77
#option add *MdiTitle.Button.activeBackground       grey77

if {[string match "unix" $tcl_platform(platform)]} {
    option add *MdiMaster.borderWidth		0
} else {
    option add *MdiMaster.relief		sunken
    option add *MdiMaster.borderWidth		2
}

namespace eval ::two::mdi {
    namespace export two_mdi
    variable Mdi
    
    namespace import ::two::util::two_parseArgs
    
    # variable Mdi
    #
    # $w,slavebg -	background color of slave window
    # $w,slavebd -	border width of slave window
    # $w,font -	font of title of slave window
    # $w,titlebg -	background color of inactive title
    # $w,titlefg -	foreground color of inactive title
    # $w,activebg -	background color of active title
    # $w,activefg -	foreground color of active title
    # nonce -	nonce to give any toplevel a new number
    # doubleclick -	time (in ms) for double clicks (used in slave's system menu)
    # offset -	current offset for new slaves
    # offset0 -	start of offset
    # step -	relative offset for next slave
    # slaveW -	minimum initial width of new slaves
    # slaveH -	minimum initial height of new slaves
    # borderW -	width of slave window's border
    # Drag -	1 if drag is going on
    # Cursor -	name of current cursor; used to determine which edge is drag'ed
    # Master -	current master
    # Master,w -	width of current master
    # Master,h -	height of current master
    # Master,x1 -	left x of current master
    # Master,y1 -	left y of current master
    # Master,x2 -	right x of current master
    # Master,y2 -	right y of current master
    # Border -	path to border of current slave window
    # x -		distance from right border at start of drag
    # y -		distance from bottom border at start of drag
    # xx -		helper value for x dragging of window
    # yy -		helper value for y dragging of window
    # $m,style -	style of mdi (windows or unix)
    # $m,icon -	prefix of icons according to current style
    # $m,arrangeicons -	if 1, then icons will be moved on resize
    # $m,xscrollcommand -	command to use on x scroll
    # $m,yscrollcommand -	command to use on y scroll
    # $m,slaves -	list of all slaves (in arbitrary order)
    # $m,xOffset -	current x offset of slave's bounding box to the left of the master
    # $m,yOffset -	current y offset of slave's bounding box to the top of the master
    # $m,maxWidth -	width of bounding box of all slaves and the master himself
    # $m,maxHeight -	height of bounding box of all slaves and the master himself
    # $m,icons -	list of positions of all minimized slaves of $m
    # $m,max -	1 if all slaves of master $m are maximized
    # $m,active -	path name of currently active slave of master $m
    # $m,resizeQ -	Id of an after idle script for resizing (proc QueueResize)
    # $b,min -	1 if slave windget is minimized (
    # $b,place -	records "place info" for normal slave
    # $b,icon -	position of icon (index, not coordinates)
    # $b,title -	string to display as title
    # $b,x -	x position of slave's border
    # $b,y -	y position of slave's border
    # $b,lastmenu -	time of last invocation of slave menu (to detect double-B1)
    
    array set Mdi {
        nonce	0
        doubleclick	500
        offset	10
        offset0	10
        step        20
        slaveW      300
        slaveH      200
        borderW	3
        Drag	0
        x		0
        y		0
    }
}

#-------------------------------------------------------------------------
# Bitmaps
#-------------------------------------------------------------------------

# window handles, Windows Look
image create bitmap mdiWinMin -foreground black -data {
    \#define winmin_width 13
    \#define winmin_height 11
    static unsigned char winmin_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0xf8, 0x01, 0xf8, 0x01, 0x00, 0x00, };
}
image create bitmap mdiWinMax -foreground black -data {
    \#define winmax_width 13
    \#define winmax_height 11
    static unsigned char winmax_bits[] = {
      0x00, 0x00, 0xfc, 0x07, 0xfc, 0x07, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04,
      0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0xfc, 0x07, 0x00, 0x00, };
}
image create bitmap mdiWinRest -foreground black -data {
    \#define winrest_width 13
    \#define winrest_height 11
    static unsigned char winrest_bits[] = {
      0x00, 0x00, 0xf0, 0x03, 0xf0, 0x03, 0x10, 0x02, 0xfc, 0x02, 0xfc, 0x02,
      0x84, 0x03, 0x84, 0x00, 0x84, 0x00, 0xfc, 0x00, 0x00, 0x00, };
}
image create bitmap mdiWinClose -foreground black -data {
    \#define winclose_width 13
    \#define winclose_height 11
    static unsigned char winclose_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0x18, 0x06, 0x30, 0x03, 0xe0, 0x01, 0xc0, 0x00,
      0xe0, 0x01, 0x30, 0x03, 0x18, 0x06, 0x00, 0x00, 0x00, 0x00, };
}

# window handles, KDE Look
image create bitmap mdiKdeMin -foreground white -background grey44 -data {
    \#define kdemin_main_width 16
    \#define kdemin_main_height 16
    static unsigned char kdemin_main_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x80, 0x03, 0x80, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
} -maskdata {
    \#define kdemin_mask_width 16
    \#define kdemin_mask_height 16
    static unsigned char kdemin_mask_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x80, 0x03, 0x80, 0x02, 0x80, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
}
image create bitmap mdiKdeMax -foreground white -background grey44 -data {
    \#define kdemax_main_width 16
    \#define kdemax_main_height 16
    static unsigned char kdemax_main_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0xfc, 0x1f, 0x04, 0x00, 0x04, 0x08, 0x04, 0x08,
      0x04, 0x08, 0x04, 0x08, 0x04, 0x08, 0x04, 0x08, 0x04, 0x08, 0xf4, 0x0f,
      0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
} -maskdata {
    \#define kdemax_mask_width 16
    \#define kdemax_mask_height 16
    static unsigned char kdemax_mask_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0xfc, 0x1f, 0xfc, 0x1f, 0x0c, 0x18, 0x0c, 0x18,
      0x0c, 0x18, 0x0c, 0x18, 0x0c, 0x18, 0x0c, 0x18, 0x0c, 0x18, 0xfc, 0x1f,
      0xfc, 0x1f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
}
image create bitmap mdiKdeRest -foreground white -background grey44 -data {
    \#define kderest_main_width 16
    \#define kderest_main_height 16
    static unsigned char kderest_main_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xf8, 0x1f, 0x08, 0x10, 0x08, 0x10,
      0x08, 0x10, 0x08, 0x10, 0x08, 0x10, 0x08, 0x10, 0x08, 0x10, 0x08, 0x10,
      0xf8, 0x1f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
} -maskdata {
    \#define kderest_mask_width 16
    \#define kderest_mask_height 16
    static unsigned char kderest_mask_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0xfc, 0x1f, 0xfc, 0x1f, 0x0c, 0x18, 0x0c, 0x18,
      0x0c, 0x18, 0x0c, 0x18, 0x0c, 0x18, 0x0c, 0x18, 0x0c, 0x18, 0xfc, 0x1f,
      0xfc, 0x1f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
}
image create bitmap mdiKdeClose -foreground white -background grey44 -data {
    \#define kdeclose_main_width 16
    \#define kdeclose_main_height 16
    static unsigned char kdeclose_main_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0x04, 0x10, 0x04, 0x08, 0x08, 0x04, 0x10, 0x02,
      0x20, 0x01, 0x40, 0x00, 0x40, 0x00, 0x20, 0x01, 0x10, 0x02, 0x08, 0x04,
      0x04, 0x08, 0x04, 0x10, 0x00, 0x00, 0x00, 0x00, };
} -maskdata {
    \#define kdeclose_mask_width 16
    \#define kdeclose_mask_height 16
    static unsigned char kdeclose_mask_bits[] = {
      0x00, 0x00, 0x00, 0x00, 0x0c, 0x30, 0x1c, 0x38, 0x38, 0x1c, 0x70, 0x0e,
      0xe0, 0x07, 0xc0, 0x03, 0xc0, 0x03, 0xe0, 0x07, 0x70, 0x0e, 0x38, 0x1c,
      0x1c, 0x38, 0x0c, 0x30, 0x00, 0x00, 0x00, 0x00, };
}

# text icon (shown as upper left handle in window titlebar)
image create bitmap mdiText -foreground white -data {
\#define text_width 11
\#define text_height 13
static unsigned char text_bits[] = {
  0x7f, 0x00, 0xc1, 0x00, 0x41, 0x01, 0x59, 0x02, 0xc1, 0x07, 0x01, 0x04,
  0xf9, 0x04, 0x01, 0x04, 0x01, 0x04, 0xf9, 0x04, 0x01, 0x04, 0x01, 0x04,
  0xff, 0x07, };
}

#-------------------------------------------------------------------------
# procedures
#-------------------------------------------------------------------------

# ::two::mdi::two_mdi --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.win"); the original widget procedure
#	is renamed to "_two$w" (like "_two.tl.win") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#
# Arguments:
#	m		Path to master widget
#	args		Widget arguments and options

proc ::two::mdi::two_mdi {m args} {
    variable Mdi
    global tcl_platform

    incr Mdi(nonce)

    # This is a multi-style widget
    set ::two::SetStyleCmds($m) [list $m configure -style]

    # Initialise the globals
    array set Mdi "
	$m,slavebg   grey77
	$m,slavebd   2
	$m,font      {}
	$m,titlebg   grey51
	$m,titlefg   grey77
	$m,activebg  blue4
	$m,activefg  white
	$m,max       0
	$m,icons     {}
	$m,active    {}
        $m,style     $tcl_platform(platform)
	$m,arrangeicons   1
	$m,xscrollcommand {}
	$m,yscrollcommand {}
	$m,slaves    {}
    "

    # See if this option is for us...
    two_parseArgs args Mdi($m,slavebg)   "-slavebg"
    two_parseArgs args Mdi($m,slavebd)   "-slavebd"
    two_parseArgs args Mdi($m,font)      "-font"
    two_parseArgs args Mdi($m,titlebg)   "-titlebg"
    two_parseArgs args Mdi($m,titlefg)   "-titlefg"
    two_parseArgs args Mdi($m,activebg)  "-activebg"
    two_parseArgs args Mdi($m,activefg)  "-activefg"
    two_parseArgs args Mdi($m,style)     "-style"
    two_parseArgs args Mdi($m,arrangeicons)   "-arrangeicons"
    two_parseArgs args Mdi($m,xscrollcommand) "-xscrollcommand"
    two_parseArgs args Mdi($m,yscrollcommand) "-yscrollcommand"
    
    if {[string match "unix" $Mdi($m,style)]} {
        set Mdi($m,icon) "mdiKde"
    } else {
        set Mdi($m,icon) "mdiWin"
    }

    # Create master widget
    uplevel [list frame $m] $args -class TwoMdiMaster -takefocus 0
    
    # Create popup menu for window icon. All slaves windows use this menu,
    # there is no need for separate "system" menus for each slave.
    two_menu $m.menu -tearoff 0
    $m.menu add command -label "Restore" -alias rest -state disabled \
        -command [namespace code {Restore $Mdi(Border)}]
    $m.menu add command -label "Minimize" -alias min \
        -command [namespace code {Minimize $Mdi(Border)}]
    $m.menu add command -label "Maximize" -alias max \
        -command [namespace code {Maximize $Mdi(Border)}]
    $m.menu add command -label "Lower" -alias lower \
        -command [namespace code {LowerSlave $Mdi(Border)}]
    $m.menu add separator
    $m.menu add command -label "Close" -alias close \
        -command [namespace code {DeleteSlave $Mdi(Border)}]

    # Create root popup menu
    two_menu $m.root -tearoff 0
    $m.root add command -label "Shuffle Up" -alias up -state disabled \
        -command [namespace code "Shuffle up $m"]
    $m.root add command -label "Shuffle Down" -alias down -state disabled \
        -command [namespace code "Shuffle down $m"]

    # Use our own bindings
    bindtags $m [list $m TwoMdi TwoMdiMaster [winfo toplevel $m] all]

    # Move widget command
    rename $m _two$m

    # Install our replacement command
    namespace eval :: "
        proc $m {opt args} {
	    uplevel [namespace code Widget] $m \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Mdi
    bind $m <Destroy> [namespace code "_Destroy $m"]

    # Return the correct result...
    return $m
}

# ::two::mdi::Widget --
#
#	replacement for widget command
#
# Arguments:
#	m		Path to real master widget
#	opt		Widget command...
#	arg		optional args

proc ::two::mdi::Widget {m opt args} {
    variable Mdi

    # See if this option is for us...
    switch -exact -- $opt {
	configure {
	    set retval ""
	    two_parseArgs args Mdi($m,slavebg)   "-slavebg"
	    two_parseArgs args Mdi($m,slavebd)   "-slavebd"
	    two_parseArgs args Mdi($m,font)      "-font"
	    two_parseArgs args Mdi($m,titlebg)   "-titlebg"
	    two_parseArgs args Mdi($m,titlefg)   "-titlefg"
	    two_parseArgs args Mdi($m,activebg)  "-activebg"
	    two_parseArgs args Mdi($m,activefg)  "-activefg"
	    set st [two_parseArgs args Mdi($m,style)          "-style"]
	    set ai [two_parseArgs args Mdi($m,arrangeicons)   "-arrangeicons"]
	    set rx [two_parseArgs args Mdi($m,xscrollcommand) "-xscrollcommand"]
	    set ry [two_parseArgs args Mdi($m,yscrollcommand) "-yscrollcommand"]
	    if {$ai || $rx || $ry} {
	        Resize $m
	    }
	    if {[llength $args]} {
		set retval [eval [list _two$m $opt] $args]
	    }
	    if {[string match "unix" $Mdi($m,style)]} {
		set Mdi($m,icon) "mdiKde"
	    } else {
		set Mdi($m,icon) "mdiWin"
	    }
	    foreach s $Mdi($m,slaves) {
		ConfigureSlave $s
		if {$st} {
		    SetSlaveStyle $s
		}
	    }
	    return $retval
	}
	cget {
	    switch -exact -- [lindex $args 0] {
	        -slavebg   { return $Mdi($m,slavebg)   }
	        -slavebd   { return $Mdi($m,slavebd)   }
	        -font      { return $Mdi($m,font)      }
	        -titlebg   { return $Mdi($m,titlebg)   }
	        -titlefg   { return $Mdi($m,titlefg)   }
	        -activebg  { return $Mdi($m,activebg)  }
	        -activefg  { return $Mdi($m,activefg)  }
	        -style     { return $Mdi($m,style)     }
	        -arrangeicons   { return $Mdi($m,arrangeicons)   }
		-xscrollcommand { return $Mdi($m,xscrollcommand) }
		-yscrollcommand { return $Mdi($m,yscrollcommand) }
		default   { return [eval [list _two$m $opt] $args] }
	    }
	}
	new -
	addSlave {
	    return [eval [list AddSlave $m] $args]
	}
	slaves {
	    set slaves ""
	    foreach slave $Mdi($m,slaves) {
	        lappend slaves $slave.w.s
	    }
	    return $slaves
	}
	entryconfigure {
	    set slave [lindex $args 0]
	    set b     [winfo parent [winfo parent $slave]]
	    set args  [lrange $args 1 end]

	    if {[two_parseArgs args Mdi($b,title) "-title"]} {
	        $b.w.t.label configure -text $Mdi($b,title)
	    }
	    two_parseArgs args Mdi($b,closecmd) "-closecmd"
	    if {[two_parseArgs args spec "-pos"]} {
	        SetSlavePos $m $b $spec
	    }
	    if {[llength $args]} {
	        eval [list $b.w.s configure] $args
	    }
	}
	entrycget {
	    if {[llength $args] != 2} {
	        return -code error "wrong # args: should be \"two_mdi entrycget slave option\""
	    }
	    set slave [lindex $args 0]
	    set b     [winfo parent [winfo parent $slave]]
	    set arg   [lindex $args 1]
	    switch -exact -- $arg {
	        -title     { return $Mdi($b,title)    }
	        -closecmd  { return $Mdi($b,closecmd) }
	        -pos       {
	            if $Mdi($b,min) {
	                set state min
	                set pos $Mdi($b,place)
	            } elseif $Mdi($m,max) {
	                set state max
	                set pos $Mdi($b,place)
	            } else {
                        set state normal
                        set pos [place info $b]
	            }
	            return [list $state $pos]
	        }
		default    { return [uplevel [list $slave cget] $arg] }
	    }
	}
	tile {
	    Tile $m
	}
	cascade {
	    Cascade $m
	}
	active {
	    # might not exist if deleted and no new focus set
	    if {[winfo exists $Mdi($m,active)]} {
		return $Mdi($m,active).w.s
	    } else {
	        return ""
	    }
	}
	delete {
	    if {[llength $args] > 1} {
	    	return -code error "wrong # args: should be \"two_mdi delete slave\""
	    }
	    # delete slave window
	    return [DeleteSlave [winfo parent [winfo parent [lindex $args 0]]]]
	}
	show {
	    if {[llength $args] > 1} {
	    	return -code error "wrong # args: should be \"two_mdi show slave\""
	    }
	    set b [winfo parent [winfo parent [lindex $args 0]]]
	    set m [winfo parent $b]
	    if {$Mdi($b,min)} {
		if {$Mdi($m,max)} {
	            Maximize $b
		} else {
	            RestoreMin $b
		}
	    } else {
		raise $b
                SetFocus $b.w.s
	    }
	}
	maximize {
	    # get active child
	    set slave [$m active]
	    if {[string length $slave]} {
	        Maximize [winfo parent [winfo parent $slave]]
	    }
	}
	minimize {
	    if {![llength $args]} {
                foreach b [AllSlaveBorders $m] {
		    Minimize $b
		}
	    } else {
		foreach slave $args {
	            Minimize [winfo parent [winfo parent $slave]]
		}
	    }
	}
	restore {
	    if {![llength $args]} {
                foreach b [AllSlaveBorders $m] {
		    Restore $b
		}
	    } else {
		foreach slave $args {
	            Restore [winfo parent [winfo parent $slave]]
		}
	    }
	}
	xview {
	    eval [list Xview $m] $args
	}
	yview {
	    eval [list Yview $m] $args
	}
	default {
            return [eval [list _two$m $opt] $args]
	}
    }
    return
}

# ::two::mdi::_Destroy --
#
#	Call this procedure when the Mdi widget is destroyed.
#	Will clean widget-specific Elements of Mdi and destroy
#	the original widget command.
#
# Arguments:
#	m		Path to master widget

proc ::two::mdi::_Destroy {m} {
    variable Mdi

    unset ::two::SetStyleCmds($m)

    # clear Mdi
    foreach key [array names Mdi "$m,*"] {
        unset Mdi($key)
    }
    # remove the original widget command
    rename $m ""
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoMdi-class bindings
#-------------------------------------------------------------------------

# ---
# Bindings for master widget
# ---

# any mouse button gives root popup
# use some displacement from mouse coordinates to avoid problems
# with grab when clicking in fast sequence
bind TwoMdi <ButtonPress> {
    tk_popup %W.root %X [expr {%Y + 1}]
}

# resize master will lead to recalculation of scrollbars
bind TwoMdi <Configure> {
    ::two::mdi::Resize %W
}

# ---
# Bindings for button in unix-style slave widget
# ---

bind TwoMdiButton <Enter> {
    %W configure -relief raised
}

bind TwoMdiButton <Leave> {
    %W configure -relief flat
}

# ---
# Bindings for frame of slave widget
# ---

# Focus in: make active window
bind TwoMdiBorder <FocusIn> {
    ::two::mdi::FocusIn %W
}

# Motion: set shape of mouse cursor
bind TwoMdiBorder <Motion> {
    ::two::mdi::BorderMotion %W %x %y
}
# Enter: set shape of mouse cursor
bind TwoMdiBorder <Enter> {
    ::two::mdi::BorderMotion %W %x %y
}
# Enter from inside
bind TwoMdiSubFrame <Leave> {
    set b [winfo parent %W] 
    ::two::mdi::BorderMotion $b \
        [expr {%X - [winfo rootx $b]}] \
	[expr {%Y - [winfo rooty $b]}]
}

# Button 1: record cursor and mdi master widget
bind TwoMdiBorder <1> {
    ::two::mdi::BorderB1 %W %x %y
}

# B1-Release: drag finished
bind TwoMdiBorder <B1-ButtonRelease> {
    set ::two::mdi::Mdi(Drag) 0
}

# B1-Motion: resize slave
bind TwoMdiBorder <B1-Motion> {
    ::two::mdi::BorderB1Motion %W %x %X %y %Y
}

# Alternative Impl. to reduce number of redraws
# (Idea taken from a news posting in comp.lang.tcl by Frdric BONNET)
#variable b1motion {}
#bind TwoMdiBorder <B1-Motion> [namespace code {
#    after cancel $b1motion
#    set b1motion [after 50 [namespace code {BorderB1Motion %W %x %X %y %Y}]]
#}]

# right mouse button gives icon menu
bind TwoMdiBorder <3> {
    ::two::mdi::SlavePopup %W %X %Y
}

# ---
# Bindings for title of slave widget
# ---

# Double-One: restore icons and maximize normal slaves
bind TwoMdiTitle <Double-1> {
    ::two::mdi::TitleDouble1 %W
}

# Button 1: retrieve paths of elements of this slave
# and record start of move and mdi master widget
bind TwoMdiTitle <1> {
    ::two::mdi::TitleB1 %W %x %y
}

# B1-Motion: move slave
bind TwoMdiTitle <B1-Motion> {
    ::two::mdi::TitleB1Motion %W %X %Y
}

# B1-Release: move finished; call resize to set scrollbars
# (min and max values may have changed through a widget reaching
# beyond the border)
bind TwoMdiTitle <B1-ButtonRelease> {
    %W configure -cursor left_ptr
    set ::two::mdi::Mdi(Drag) 0
    ::two::mdi::Resize [winfo parent [winfo parent [winfo parent [winfo parent %W]]]]
}

# right mouse button gives icon menu
bind TwoMdiTitle <3> {
    ::two::mdi::SlavePopup [winfo parent [winfo parent [winfo parent %W]]] %X %Y
}

# ---
# Bindings for icon in title of slave widget
# ---

bind TwoMdiIcon <1> {
    ::two::mdi::TitleIcon1 %W %t
}

# ---
# Bindings for container for users's widget
# ---

# if user directly destroys his slave widget:
# clean up things nevertheless
bind TwoMdiSlave <Destroy> {
    ::two::mdi::DeleteSlave [winfo parent [winfo parent %W]] 1
}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::mdi::FocusIn --
#	react on focus entering the slave
#
# Arguments:
#	b		Path to slave widget's border

proc ::two::mdi::FocusIn {b} {
    variable Mdi
    # don't do anything if this widget is minimized
    if {!$Mdi($b,min)} {
	set m [winfo parent $b]
	# did we already have an active slave?
	set active $Mdi($m,active)
	# make sure this widget isn't already the active one
	if {![string match $b $active]} {
	    set Mdi($m,active) $b
	    if {[winfo exists $active]} {
        	InactiveSlave $active $m
	    }
	    event generate $m <<ActiveOn>>
	    ActiveSlave $b $m
	    raise $b
	    # make sure that the focus is set to a child
	    # which wants it!
	    SetFocus $b.w.s
	}
    }
}

# ::two::mdi::BorderMotion --
#
#	handle mouse cursor motion events within border frame. Set
#	the mouse cursor shape according to the relative position.
#
# Arguments:
#	b		Path to slave widget's border
#	x		relative X Position of mouse cursor within border (%x)
#	y		relative Y Position (%y)

proc ::two::mdi::BorderMotion {b x y} {
    variable Mdi

    set m [winfo parent $b]
    if {!$Mdi(Drag)} {
        if {$Mdi($m,max) || $Mdi($b,min)} {
       	    $b configure -cursor ""
	} elseif {$x < 10} {
            if {$y < 10} {
        	$b configure -cursor top_left_corner
            } elseif {$y > [winfo height $b] - 10} {
        	$b configure -cursor bottom_left_corner
	    } else {
        	$b configure -cursor left_side
	    }
	} elseif {$x > [winfo width $b] - 10} {
            if {$y < 10} {
        	$b configure -cursor top_right_corner
            } elseif {$y > [winfo height $b] - 10} {
        	$b configure -cursor bottom_right_corner
	    } else {
        	$b configure -cursor right_side
	    }
	} elseif {$y < 10} {
            $b configure -cursor top_side
	} else {
            $b configure -cursor bottom_side
	}
    }
    return
}

# ::two::mdi::BorderB1 --
#
#	handle button press within within border frame.
#
# Arguments:
#	b		Path to slave widget's border
#	x		relative X Position of mouse cursor within border (%x)
#	y		relative Y Position (%y)

proc ::two::mdi::BorderB1 {b x y} {
    variable Mdi

    set m [winfo parent $b]
    if {! ($Mdi($m,max) || $Mdi($b,min))} {
	set Mdi(Drag)     1
	set Mdi(Cursor)   [$b cget -cursor]
	set Mdi(Master,w) [winfo width  $m]
	set Mdi(Master,h) [winfo height $m]
	set Mdi(x)        [expr {[winfo width  $b] - $x}]
	set Mdi(y)        [expr {[winfo height $b] - $y}]
	set Mdi(xx)       [expr {$x + [winfo rootx $m] + [$m cget -bd]}]
	set Mdi(yy)       [expr {$y + [winfo rooty $m] + [$m cget -bd]}]
	set Mdi(Master)   "_two$m"

	SetFocus $b.w.s
    }
    return
}

# ::two::mdi::BorderB1Motion --
#
#	handle interactive resizing of slave. This relies on the mouse
#	cursor's shape set in motion procedure!
#
# Arguments:
#	b		Path to slave widget's border
#	relX		relative X Position of mouse cursor within border (%x)
#	absX		absolute X Positino (%X)
#	relY		relative Y Position (%y)
#	absY		absolute Y Position (%Y)

proc ::two::mdi::BorderB1Motion {b relX absX relY absY} {
    variable Mdi

    set m [winfo parent $b]
    if {$Mdi(Drag) && (! ($Mdi($m,max) || $Mdi($b,min)))} {
	if {[string match "*right*" $Mdi(Cursor)]} {
	    set x [expr {$relX + $Mdi(x)}]
	    if {$x > 20} {
        	set x [expr {($x + [winfo x $b]) < $Mdi(Master,w) ? $x : $Mdi(Master,w) - [winfo x $b]}]
		place $b -width $x
	    }
	} elseif {[string match "*left*" $Mdi(Cursor)]} {
            set x [expr {($absX - $Mdi(xx)) < 0 ? 0 : ($absX - $Mdi(xx))}]
            set w [expr {[winfo width $b] + [winfo x $b] - [$m cget -bd] - $x}]
	    if {$w > 20} {
		place $b -x $x -width $w
		set Mdi($b,x) $x
	    }
	}
	# no elseif: top-right etc.
	if {[string match "*bottom*" $Mdi(Cursor)]} {
	    set y [expr {$relY + $Mdi(y)}]
	    if {$y > 20} {
        	set y [expr {($y + [winfo y $b]) < $Mdi(Master,h) ? $y : $Mdi(Master,h) - [winfo y $b]}]
		place $b -height $y
	    }
	} elseif {[string match "*top*" $Mdi(Cursor)]} {
            set y [expr {($absY - $Mdi(yy)) < 0 ? 0 : ($absY - $Mdi(yy))}]
            set h [expr {[winfo height $b] + [winfo y $b] - [$m cget -bd] - $y}]
	    if {$h > 20} {
		place $b -y $y -height $h
		set Mdi($b,y) $y
	    }
	}
    }
    return
}

# ::two::mdi::TitleB1 --
#
#	handle button press within within title frame.
#
# Arguments:
#	t		Path to slave widget's border
#	x		relative X Position of mouse cursor within border (%x)
#	y		relative Y Position (%y)

proc ::two::mdi::TitleB1 {t x y} {
    variable Mdi

    set b [winfo parent [winfo parent [winfo parent $t]]]
    set m [winfo parent $b]
    if {!$Mdi($m,max) && !$Mdi($b,min)} {
        set Mdi(Drag)      1
	set Mdi(Master)    "_two$m"
	set Mdi(Border)    $b
	set Mdi(Master,x1) [expr {[winfo rootx $m] + [$b cget -bd] - [$m cget -bd]}]
	set Mdi(Master,y1) [expr {[winfo rooty $m] + [$b cget -bd] - [$m cget -bd]}]
	set Mdi(Master,x2) [expr {$Mdi(Master,x1) + [winfo width  $m]}]
	set Mdi(Master,y2) [expr {$Mdi(Master,y1) + [winfo height $m]}]
	set Mdi(xx)        [expr {$x + 2 + [winfo x $t] + $Mdi(borderW) + $Mdi(Master,x1)}]
	set Mdi(yy)        [expr {$y + 2 + [winfo y $t] + $Mdi(borderW) + $Mdi(Master,y1)}]
	$t configure -cursor fleur
    }

    SetFocus $b.w.s
    return
}

# ::two::mdi::TitleB1Motion --
#
#	handle interactive resizing of slave. This relies on the mouse
#	cursor's shape set in motion procedure!
#
# Arguments:
#	t		Path to slave widget's border
#	absX		absolute X Positino (%X)
#	absY		absolute Y Position (%Y)

proc ::two::mdi::TitleB1Motion {t absX absY} {
    variable Mdi

    if {$Mdi(Drag)} {
	# check bounds: cursor within master
	if {$absX < $Mdi(Master,x1)} {
	    set x $Mdi(Master,x1)
	} elseif {$absX > $Mdi(Master,x2)} {
	    set x $Mdi(Master,x2)
	} else {
	    set x $absX
	}

	if {$absY < $Mdi(Master,y1)} {
	    set y $Mdi(Master,y1)
	} elseif {$absY > $Mdi(Master,y2)} {
	    set y $Mdi(Master,y2)
	} else {
	    set y $absY
	}

	place $Mdi(Border) -x [expr {$x - $Mdi(xx)}] -y [expr {$y - $Mdi(yy)}]
    }
    return
}

# ::two::mdi::TitleDouble1 --
#
#	handle doubleclick in title bar of slave.
#
# Arguments:
#	w		Path to Mdi child's title bar

proc ::two::mdi::TitleDouble1 {w} {
    variable Mdi

    set f [winfo parent [winfo parent [winfo parent $w]]]
    set m [winfo parent $f]
    if {$Mdi($f,min)} {
        RestoreMin $f
    } elseif {$Mdi($m,max)} {
        RestoreMax $f
    } else {
        Maximize $f
    }
}

# ::two::mdi::TitleIcon1 --
#
#	handle click on icon in title bar (doubleclick as well)
#
# Arguments:
#	w		Path to icon in Mdi child's title bar
#	t		Time field of click event

proc ::two::mdi::TitleIcon1 {w t} {
    variable Mdi
    set l [winfo parent $w]
    set b [winfo parent [winfo parent $l]]

    if {($Mdi($b,lastmenu) != 0) && ($t - $Mdi($b,lastmenu) < $Mdi(doubleclick))} {
        set Mdi($b,lastmenu) $t
        DeleteSlave $b
    } else {
        set Mdi($b,lastmenu) $t
        SlavePopup $b [winfo rootx $l] [expr {[winfo rooty $l] + [winfo height $l]}]
    }
}

# ::two::mdi::Resize --
#
#	update scrollbars on resize of master
#
# Arguments:
#	m		Path to Mdi master widget

proc ::two::mdi::Resize {m} {
    variable Mdi

    # return if no scrolling is needed
    if {$Mdi($m,max)} {
        return
    }

    # move icons
    if {$Mdi($m,arrangeicons) && [llength $Mdi($m,slaves)]} {
	set Mdi($m,icons) ""
	set iconX 0
	set iconY 0
	set iconH [expr {[winfo reqheight [lindex $Mdi($m,slaves) 0].w.t] \
	                + 2 * ($Mdi(borderW) + $Mdi($m,slavebd))}]
    }

    # claculate bounding box of all slaves and master
    # and position of icons
    set minX 0
    set minY 0
    set maxX [winfo width  $m]
    set maxY [winfo height $m]
    foreach b [AllSlaveBorders $m] {
        if {$Mdi($b,min)} {
	    if {$Mdi($m,arrangeicons)} {
		place $b -x [expr {150 * $iconX}] -y [expr { - $iconH * $iconY}] \
	    	    -rely 1 -width 150 -height $iconH -relwidth 0 -relheight 0 -anchor sw
		incr iconX
		if [expr {150 * ($iconX + 1)  > $maxX}] {
		    # position out of window
		    set  iconX 0
		    incr iconY
		}
	    }
	} elseif {[string length $Mdi($m,xscrollcommand)] || [string length $Mdi($m,yscrollcommand)]} {
	    set x [winfo x $b]
	    set y [winfo y $b]
	    set w [winfo width  $b]
	    set h [winfo height $b]
            if {$x < $minX} {set minX $x}
            if {$y < $minY} {set minY $y}
            if {$x + $w > $maxX} {set maxX [expr {$x + $w}]}
            if {$y + $h > $maxY} {set maxY [expr {$y + $h}]}
	    set Mdi($b,x) $x
	    set Mdi($b,y) $y
	}
    }
    # set value of scrollbars
    set w [winfo width  $m]
    set h [winfo height $m]
    # configure x scroll
    if {[string length $Mdi($m,xscrollcommand)]} {
	set Mdi($m,xOffset)  [expr { $minX < 0  ? 0 - $minX : 0}]
        set Mdi($m,maxWidth) [expr {($maxX > $w ? $maxX     : $w) + $Mdi($m,xOffset)}]

	set minFract [expr { $Mdi($m,xOffset)       / double($Mdi($m,maxWidth))}]
	set maxFract [expr {($Mdi($m,xOffset) + $w) / double($Mdi($m,maxWidth))}]

        eval $Mdi($m,xscrollcommand) [list $minFract $maxFract]
    }
    # configure y scroll
    if {[string length $Mdi($m,yscrollcommand)]} {
	set Mdi($m,yOffset)   [expr { $minY < 0  ? 0 - $minY : 0}]
        set Mdi($m,maxHeight) [expr {($maxY > $h ? $maxY     : $h) + $Mdi($m,yOffset)}]

	set minFract [expr { $Mdi($m,yOffset)       / double($Mdi($m,maxHeight))}]
	set maxFract [expr {($Mdi($m,yOffset) + $h) / double($Mdi($m,maxHeight))}]

        eval $Mdi($m,yscrollcommand) [list $minFract $maxFract]
    }
    return
}

# ::two::mdi::QueueResize --
#
#	Insert a resize into the event loop. If there is already one,
#	it will be removed.
#
# Arguments:
#	m		Path to Mdi master widget

proc ::two::mdi::QueueResize {m} {
    variable Mdi
    
    if {[info exists Mdi($m,resizeQ)]} {
        after cancel $Mdi($m,resizeQ)
    }
    set Mdi($m,resizeQ) [after idle [namespace code "
        [list Resize $m]
        [list unset Mdi($m,resizeQ)]
    "]]
}

# ::two::mdi::Xview --
#
#	set x view on scroll
#
# Arguments:
#	m		Path to Mdi master widget
#	args		Where to scroll to

proc ::two::mdi::Xview {m args} {
    variable Mdi
    # return if no scrolling needed
    if {$Mdi($m,max)} {
        return
    }
    set w [winfo width $m]
    switch -exact -- [lindex $args 0] {
	moveto {
	    set offset [expr {[lindex $args 1] * $Mdi($m,maxWidth)}]
	}
	scroll {
	    switch -exact -- [lindex $args 2] {
	        units {
		    set offset [expr {$Mdi($m,xOffset) + 10 * [lindex $args 1]}]
		}
		pages {
		    set offset [expr {$Mdi($m,xOffset) + [lindex $args 1] * $w / 2}]
		}
		default {
		    return -code error "wrong # args: should be \"$m xview scroll number units|pages\""
		}
	    }
        }
	default {
	    return -code error "unknown option \"[lindex $args 0]\": must be moveto or scroll"
	}
    }
    # suitable range for offset
    if {$offset < 0} {
	set offset 0
    }
    set maxOffset [expr {$Mdi($m,maxWidth) - $w}]
    if {$offset > $maxOffset} {
        set offset $maxOffset
    }
    # move widgets
    foreach b $Mdi($m,slaves) {
	if {!$Mdi($b,min)} {
	    set x [expr {$Mdi($b,x) - $offset + $Mdi($m,xOffset)}]
	    place $b -x $x
	    set Mdi($b,x) $x
	}
    }

    # reset scrollbar position
    set minFract [expr { $offset       / double($Mdi($m,maxWidth))}]
    set maxFract [expr {($offset + $w) / double($Mdi($m,maxWidth))}]
    eval $Mdi($m,xscrollcommand) [list $minFract $maxFract]
    set Mdi($m,xOffset) $offset
    return
}

# ::two::mdi::Yview --
#
#	set y view on scroll
#
# Arguments:
#	m		Path to Mdi master widget
#	args		Where to scroll to

proc ::two::mdi::Yview {m args} {
    variable Mdi
    # return if no scrolling needed
    if {$Mdi($m,max)} {
        return
    }
    set h [winfo height $m]
    switch -exact -- [lindex $args 0] {
	moveto {
	    set offset [expr {[lindex $args 1] * $Mdi($m,maxHeight)}]
	}
	scroll {
	    switch -exact -- [lindex $args 2] {
	        units {
		    set offset [expr {$Mdi($m,yOffset) + 10 * [lindex $args 1]}]
		}
		pages {
		    set offset [expr {$Mdi($m,yOffset) + [lindex $args 1] * $h / 2}]
		}
		default {
		    return -code error "wrong # args: should be \"$m xview scroll number units|pages\""
		}
	    }
        }
	default {
	    return -code error "unknown option \"[lindex $args 0]\": must be moveto or scroll"
	}
    }
    # suitable range for offset
    if {$offset < 0} {
	set offset 0
    }
    set maxOffset [expr {$Mdi($m,maxHeight) - $h}]
    if {$offset > $maxOffset} {
        set offset $maxOffset
    }
    # move widgets
    foreach b $Mdi($m,slaves) {
	if {!$Mdi($b,min)} {
	    set y [expr {$Mdi($b,y) - $offset + $Mdi($m,yOffset)}]
	    place $b -y $y
	    set Mdi($b,y) $y
	}
    }

    # reset scrollbar position
    set minFract [expr { $offset       / double($Mdi($m,maxHeight))}]
    set maxFract [expr {($offset + $h) / double($Mdi($m,maxHeight))}]
    eval $Mdi($m,yscrollcommand) [list $minFract $maxFract]
    set Mdi($m,yOffset) $offset
    return
}

# ::two::mdi::SlavePopup --
#
#	show popup menu for slave window, also used as icon menu.
#	Make sure there is some some displacement from mouse coordinates to
#	avoid problems with grab when clicking in fast sequence
#
# Arguments:
#	b		Path to slave widget's border
#	x, y		Position of popup

proc ::two::mdi::SlavePopup {b x y} {
    variable Mdi
    
    if {$y <= [winfo pointery $b]} {
        set y [expr {[winfo pointery $b] + 1}]
    }

    set m [winfo parent $b]
    set Mdi(Border) $b
    set Mdi(Master) $m

    $m.menu entryconfigure 0 -state \
        [expr {$Mdi($b,min)||$Mdi($m,max) ? "normal" : "disabled"}]
    $m.menu entryconfigure 1 -state \
        [expr {$Mdi($b,min) ? "disabled" : "normal"}]
    $m.menu entryconfigure 2 -state \
        [expr {$Mdi($m,max) ? "disabled" : "normal"}]
    if { [string match $Mdi($m,active) $b] && ([llength [AllSlaveBorders $m 1]] > 1) } {
        $m.menu entryconfigure 3 -state "normal"
    } else {
        $m.menu entryconfigure 3 -state "disabled"
    }
    tk_popup $m.menu $x $y
    return
}

# ::two::mdi::Minimize --
#
#	minimize slave window
#
# Arguments:
#	b		Path to Slave windows's border

proc ::two::mdi::Minimize {b} {
    variable Mdi

    set m [winfo parent $b]

    if {!$Mdi($m,max)} {
        set Mdi($b,place) [place info $b]
    }

    if {$Mdi($m,arrangeicons)} {
	# get stacking order right
	set topIcon ""
	foreach s [AllSlaveBorders $m] {
	    if {$Mdi($s,min)} {
		set topIcon $s
	    }
	}
	if {[string length $topIcon]} {
	    raise $b $topIcon
	} else {
	    lower $b
	}
    } else {
        # search for first free position
	set x   0
	set y   0
	set max [winfo width $m]
	set h   [expr {[winfo height $b.w.t] + 2 * ($Mdi(borderW) + $Mdi($m,slavebd))}]
	while 1 {
	    if {[lsearch -exact $Mdi($m,icons) "$x,$y"] >= 0} {
        	# position already taken
		incr x
	    } elseif {150 * ($x + 1) > $max} {
		# position out of window
		set  x 0
		incr y
	    } else {
		# found pos
		lappend Mdi($m,icons) "$x,$y"
		set     Mdi($b,icon)  "$x,$y"
		# set x and y to widget coordinates
		set x [expr {150 * $x}]
		set y [expr {- $h * $y}]
		break
	    }
	}
 	place $b -x $x -y $y -rely 1 -width 150 -height $h -relwidth 0 -relheight 0 -anchor sw
 	lower $b
    }

    # flag current widget as minimized
    set Mdi($b,min) 1

    # take away focus
    if {[string match $b $Mdi($m,active)]} {
        # $b is already marked as minimized and won't show up in the following list
        set next [lindex [AllSlaveBorders $m 1] end]
	if {[string length $next]} {
            SetFocus $next.w.s
	    # nothing more to do - the <FocusIn> binding will care about state
	} else {
	    # there ain't no remaining slave
	    # and we don't want a <FocusOut> binding, so handle this case here...
            # Find next widget which is willing to take the focus
	    set next [tk_focusNext $m]
	    # Is this, for any reason, within the mdi? Shouldn't be, as there
	    # is really no slave remaining which should take the focus
	    if {[string match "$m*" $next]} {
	        focus [winfo toplevel $m]
	    } else {
	        focus $next
	    }
	    InactiveSlave $b $m
	    set Mdi($m,active) ""
	    event generate $m <<ActiveOff>>
	}
    }

    $b.w.t.min configure -image "$Mdi($m,icon)Rest" \
	-command [namespace code "RestoreMin $b"]
    $b.w.t.max configure -image "$Mdi($m,icon)Max" \
	-command [namespace code "Maximize $b"]

    if {$Mdi($m,max)} {
        RestoreMax $b
    }
    # recalculate scrollbar settings
    # (in arrangeicons mode also icon positons)
    QueueResize $m
    return
}

# ::two::mdi::Maximize --
#
#	maximize slave window (and all others)
#
# Arguments:
#	b		Path to slave windows's border

proc ::two::mdi::Maximize {b} {
    variable Mdi

    set m [winfo parent $b]

    if {$Mdi($m,max)} {
        # maximize when already maximized can happen through
	# "master show slave"
	# in this case we only have to work on this slave
	set candidates $b
    } else {
        # maximize from normal state:  work on all normal
	# (i.e. non-minimized) slaves
        set candidates $Mdi($m,slaves)
	set Mdi($m,max) 1
    }

    foreach s $candidates {
	if {!$Mdi($s,min)} {
	    set Mdi($s,place) [place info $s]
	}
	if {$Mdi($s,min) && $s == $b} {
	    set Mdi($s,min) 0
	}
	if {!$Mdi($s,min)} {
	    place $s -x 0 -y 0 -rely 0 -width 0 -height 0 -relwidth 1 -relheight 1 -anchor nw
	    $s.w.t.min configure -image "$Mdi($m,icon)Min" \
		-command [namespace code "Minimize $s"]
	    $s.w.t.max configure -image "$Mdi($m,icon)Rest" \
		-command [namespace code "RestoreMax $s"]
        }
    }
    
    # give focus
    SetFocus $b.w.s

    # recalculate scrollbar settings
    if {[string length $Mdi($m,yscrollcommand)]} {
        eval $Mdi($m,yscrollcommand) 0 1
    }
    if {[string length $Mdi($m,xscrollcommand)]} {
        eval $Mdi($m,xscrollcommand) 0 1
    }
    return
}

# ::two::mdi::Restore --
#
#	restore slave window to former size. Calls RestoreMin
#	or RestoreMax, depending on current state.
#
# Arguments:
#	b		Path to slave windows's border

proc ::two::mdi::Restore {b} {
    variable Mdi
    
    if {$Mdi($b,min)} {
	RestoreMin $b
    } elseif {$Mdi([winfo parent $b],max)} {
	RestoreMax $b
    }
    return
}

# ::two::mdi::RestoreMin --
#
#	restore size and position of slave window.
#	Call this when restoring from minimized state.
#
# Arguments:
#	b		Path to slave windows's border

proc ::two::mdi::RestoreMin {b} {
    variable Mdi

    set m [winfo parent $b]

    # remove from list of icons
    if {!$Mdi($m,arrangeicons)} {
        set icon [lsearch -exact $Mdi($m,icons) $Mdi($b,icon)]
        set Mdi($m,icons) [lreplace $Mdi($m,icons) $icon $icon]
    }

    eval [list place configure $b] $Mdi($b,place)
    $b.w.t.min configure -image "$Mdi($m,icon)Min" \
	-command [namespace code "Minimize $b"]
    $b.w.t.max configure -image "$Mdi($m,icon)Max" \
	-command [namespace code "Maximize $b"]
    set Mdi($b,min) 0

    # give focus; first, set focus to toplevel in order to
    # really trigger FocusIn binding. If focus were already,
    # for whatever reason, inside the minimized slave, things
    # would go boom...
    focus [winfo toplevel $m]
    SetFocus $b.w.s
    
    # Make sure that the mdi master doesn't think we are in maximized
    # mode: this could happen if we had indeed some maximized slaves
    # and the were all closed. Opening new windows should put them in
    # maximized state, so this is o.k., but if we want to restore a
    # minimized window, then forget about maximization.
    set Mdi($m,max) 0
    # reset scrollbar values
    QueueResize $m
    return
}

# ::two::mdi::RestoreMax --
#
#	restore size and place of slave window and maybe all others.
#	Call this when restoring from maximized state.
#
# Arguments:
#	b		Path to slave windows's border

proc ::two::mdi::RestoreMax {b} {
    variable Mdi

    set m [winfo parent $b]
    set Mdi($m,max) 0

    foreach b $Mdi($m,slaves) {
	if {!$Mdi($b,min)} {
            eval [list place configure $b] $Mdi($b,place)
	    $b.w.t.min configure -image "$Mdi($m,icon)Min" \
	        -command [namespace code "Minimize $b"]
	    $b.w.t.max configure -image "$Mdi($m,icon)Max" \
	        -command [namespace code "Maximize $b"]
	}
    }
    # reset scrollbar values
    QueueResize $m
    return
}

# ::two::mdi::LowerSlave --
#
#	move slave to bottom of stacking order.
#	Deals with focus: FocusIn binding may trigger after popupmenu
#	and raise the window.
#
# Arguments:
#	b		Path to slave window's border

proc ::two::mdi::LowerSlave {b} {
    set m  [winfo parent $b]
    set bb [AllSlaveBorders $m 1]
    set l  [llength $bb]

    if {$l > 1} {
	lower $b
	SetFocus [lindex $bb [expr {$l - 2}]].w.s
    }
    return
}

# ::two::mdi::Shuffle --
#
#	modify stacking order. Shuffle up will turn lowest slave to top,
#	while Shuffle down will lower the toplevel
#
# Arguments:
#	dir		direction: "up" or "down"
#	m		Path to master widget

proc ::two::mdi::Shuffle {dir m} {
    variable Mdi

    set bb [AllSlaveBorders $m 1]
    set l  [llength $bb]
    if {$l > 1} {
	if {[string match "up" $dir]} {
            # don't use $Mdi($m,slaves): here we take advantage of
	    # of the stacking order of the slaves!
	    SetFocus "[lindex $bb 0].w.s"
	} elseif {[string match "down" $dir]} {
	    LowerSlave [lindex $bb [expr {$l - 1}]]
	} else {
            return -code error "unknown option \"$dir\": must be up or down"
	}
    }
    return
}

# ::two::mdi::NoScrolling --
#
#	Resize all slaves and update geometry in order to disable
#	Scrolling in Mdi master. This is needed for commands that
#	know that they stack all slaves in a way that will remove
#	the need for scrolling and needs to know the size of the
#	Master *after* that change. Scrollbars might be automatic,
#	so we really have to make the change and see how large the
#	Mdi is afterwards.
#
# Arguments:
#	m		master widget
#	bb		List of all affected slaves

proc ::two::mdi::NoScrolling {m bb} {
    variable Mdi

    # Any scrollbars around? And if so, is there a changes for them
    # to really scroll the Mdi?
    if {([string length $Mdi($m,xscrollcommand)] &&
	   ($Mdi($m,yOffset) != 0 ||
	    $Mdi($m,maxWidth) > [winfo width $m])
    )|| ([string length $Mdi($m,yscrollcommand)] &&
	   ($Mdi($m,xOffset) != 0 ||
	    $Mdi($m,maxHeight) > [winfo height $m])
    )} {
	# O.k., we'll have to take make sure that all geometry is updated
	# as if the slaves were smaller than the Master. This is needed
	# in case we have auto scrollbars which would vanish *after*
	# calculating the size the slave and thus leave a gap behind.

        # Settings for all slaves:
    	foreach b $bb {
	    set Mdi($b,x) 0
	    set Mdi($b,y) 0
	    place $b -x 0 -y 0 -width 1 -height 1
	}
	# recalculate scrollbar settings
	if {[string length $Mdi($m,yscrollcommand)]} {
	    set Mdi($m,yOffset)   0
            set Mdi($m,maxHeight) [winfo height $m]
            eval $Mdi($m,yscrollcommand) 0 1
	}
	if {[string length $Mdi($m,xscrollcommand)]} {
	    set Mdi($m,xOffset)  0
            set Mdi($m,maxWidth) [winfo width $m]
            eval $Mdi($m,xscrollcommand) 0 1
	}
	update idletasks
    }
    return
}

# ::two::mdi::Tile --
#
#	Tile all slave windows in master.
#	1-3 slaves: place one above the other
#	n*3 slaves: place in n columns
#	other: fill last column starting from the bottom
#
# Arguments:
#	m		master widget

proc ::two::mdi::Tile {m} {
    variable Mdi

    # get list of all slaves; don't use $Mdi($m,slaves)
    # because we use the stacking order of the slaves!
    set bb     [AllSlaveBorders $m 1]
    set count  [llength $bb]
    if {$count > 0} {
	if {$Mdi($m,max)} {
	    RestoreMax [lindex $bb 0]
	}

	NoScrolling $m $bb

	set width  [expr {[winfo width  $m] - 1}]
	set height [expr {[winfo height $m] - 1}]
	set rows   [expr {$count >= 3 ? 3 : $count}]
	set cols   [expr {int( ($count + 2) / 3 ) }]
	set add    [expr {$rows * $cols - $count}]
	set h      [expr {$height / $rows}]
        set w      [expr {$width / ($add > 0 ? $cols - 1 : $cols)}]

	set x 0
	set y 0
	for {set i [expr {$count - 1}]} {$i >= 0} {incr i -1} {
	    set b [lindex $bb $i]
	    if {$x == 0 && $y == $add} {
		set w [expr {$width / $cols} ]
		incr cols
	    }
	    set Mdi($b,x) [expr {$x * $w}]
	    set Mdi($b,y) [expr {$y * $h}]
	    place $b -x $Mdi($b,x) -y $Mdi($b,y) -width $w -height $h
	    incr x
	    if {$x >= $cols - 1} {
		set x 0
		incr y
	    }
	}
    }
    return
}

# ::two::mdi::Cascade --
#
#	Cascade the slave widgets (stack them with some displacement).
#
# Arguments:
#	m		master widget

proc ::two::mdi::Cascade {m} {
    variable Mdi

    # get list of all slaves; don't use $Mdi($m,slaves):
    # we need the stacking order of the slaves
    set bb     [AllSlaveBorders $m 1]
    set count  [llength $bb]

    if {$count > 0} {
	if {$Mdi($m,max)} {
	    RestoreMax [lindex $bb 0]
	}

	NoScrolling $m $bb

	# calculate width and height of slave
	# each should be 75% of master, but at least have a minimum size
	set width  [winfo width  $m]
	set height [winfo height $m]
	set slaveW [expr {0.75 * $width}]
	set slaveH [expr {0.75 * $height}]
	if {$slaveW < $Mdi(slaveW)} {
            set slaveW $Mdi(slaveW)
	}
	if {$slaveH < $Mdi(slaveH)} {
            set slaveH $Mdi(slaveH)
	}

        set offset $Mdi(offset0)
	foreach b $bb {
            place $b -x $offset -y $offset -width $slaveW -height $slaveH
	    incr offset $Mdi(step)
            if {$offset > $width - $slaveW || $offset > $height - $slaveH } {
        	set offset $Mdi(offset0)
	    }
	}
    }
    QueueResize $m
    return
}

# ::two::mdi::AllSlaveBorders --
#
#	Returns list of all slaves of master widget. If $noMin is set to true,
#	then only normal slaves (no matter if maximized or not) are included.
#
# Arguments:
#	m		master
#	noMin		don't include minimized slaves if set to true

proc ::two::mdi::AllSlaveBorders {m {noMin 0}} {
    variable Mdi

    set bb ""
    foreach b [winfo children $m] {
        if {![string match $b "$m.menu"] &&
	    ![string match $b "$m.root"] &&
	    !($noMin && $Mdi($b,min))
	} {
	    lappend bb $b
	}
    }
    return $bb
}


# ::two::mdi::ConfigureSlave --
#
#	Configure settings according to Mdi($m,...) for slave widget
#
# Arguments:
#	b		Path to Slave window's border

proc ::two::mdi::ConfigureSlave {b} {
    variable Mdi

    # Master
    set m [winfo parent $b]

    # Frame
    $b configure -bg $Mdi($m,slavebg) -bd $Mdi($m,slavebd)

    # Title: active slave?
    if {[string match $b $Mdi($m,active)]} {
        ActiveSlave $b $m
    } else {
        InactiveSlave $b $m
    }
    # Title: font and text
    if {[string length $Mdi($m,font)]} {
	$b.w.t.label configure -text $Mdi($b,title) -font $Mdi($m,font)
    } else {
	$b.w.t.label configure -text $Mdi($b,title)
    }

    # Spaceing
    $b.w.d configure -bg $Mdi($m,slavebg)
    return
}

# ::two::mdi::SetSlavePos --
#
#	Set the position of a slave widget according to the
#	spec as generated by entrycget -pos. This spec is a
#	list: first element is the state {min|max|normal}, the
#	second is the output of [place info] for the border
#	frame (for the normal state).
#
# Arguments:
#	m		Path to mdi master widget
#	b		Path to Slave window's border
#	spec		Spec for state and position

proc ::two::mdi::SetSlavePos {m b spec} {
    variable Mdi

    set state [lindex $spec 0]
    set pos   [lindex $spec 1]
    if {[string match "min" $state] && !$Mdi($b,min)} {
        Minimize $b
    } elseif {[string match "max" $state] && !$Mdi($m,max)} {
        Maximize $b
    }
    if {[string match "min" $state] || [string match "max" $state] } {
        set Mdi($b,place) $pos
    } else {
        eval [list place configure $b] $pos
    }
    array set conf $pos
    set Mdi($b,x) $conf(-x)
    set Mdi($b,y) $conf(-y)
    QueueResize $m
}


# ::two::mdi::SetSlaveStyle --
#
#	sets slave's style according to the style (windows/unix)
#
# Arguments:
#	b		Path to Slave window's border

proc ::two::mdi::SetSlaveStyle {b} {
    variable Mdi

    set m  [winfo parent   $b]
    set tl [winfo toplevel $m]
    
    if {[string match "unix" $Mdi($m,style)]} {
        bindtags $b.w.t.close [list $b.w.t.close TwoMdiButton Button $tl all]
        bindtags $b.w.t.max   [list $b.w.t.max   TwoMdiButton Button $tl all]
        bindtags $b.w.t.min   [list $b.w.t.min   TwoMdiButton Button $tl all]
	$b.w.t.close configure -relief flat -image mdiKdeClose
	$b.w.t.max   configure -relief flat
	$b.w.t.min   configure -relief flat
        pack $b.w.t.close -side right -padx 0 -pady 0 -fill y
        pack $b.w.t.max   -side right         -pady 0 -fill y
        pack $b.w.t.min   -side right         -pady 0 -fill y
        pack $b.w.t.icon  -side left  -padx 2 -pady 0 -fill y
        pack $b.w.t.label -side left -expand 1              -fill both
    } else {
        bindtags $b.w.t.close [list $b.w.t.close Button $tl all]
        bindtags $b.w.t.max   [list $b.w.t.max   Button $tl all]
        bindtags $b.w.t.min   [list $b.w.t.min   Button $tl all]
	$b.w.t.close configure -relief raised -image mdiWinClose
	$b.w.t.max   configure -relief raised
	$b.w.t.min   configure -relief raised
        pack $b.w.t.close -side right -padx 2 -pady 2 -fill none
        pack $b.w.t.max   -side right         -pady 2 -fill none
        pack $b.w.t.min   -side right         -pady 2 -fill none
        pack $b.w.t.icon  -side left  -padx 2 -pady 2 -fill y
        pack $b.w.t.label -side left -expand 1              -fill both
    }

    if {$Mdi($m,max)} {
	$b.w.t.min configure -image "$Mdi($m,icon)Min"
	$b.w.t.max configure -image "$Mdi($m,icon)Rest"
    } elseif {$Mdi($b,min)} {
	$b.w.t.min configure -image "$Mdi($m,icon)Rest"
	$b.w.t.max configure -image "$Mdi($m,icon)Max"
    } else {
	$b.w.t.max configure -image "$Mdi($m,icon)Max"
	$b.w.t.min configure -image "$Mdi($m,icon)Min"
    }
    return
}

# ::two::mdi::ActiveSlave --
#
#	Configure color settings to show active state of slave
#
# Arguments:
#	b		Path to slave window's border
#	m		Path to master

proc ::two::mdi::ActiveSlave {b m} {
    variable Mdi
    $b.w.t       configure -bg $Mdi($m,activebg)
    $b.w.t.icon  configure -bg $Mdi($m,activebg)
    $b.w.t.label configure -bg $Mdi($m,activebg) -fg $Mdi($m,activefg)
    return
}

# ::two::mdi::InactiveSlave --
#
#	Configure color settings to show inactive state of slave
#
# Arguments:
#	b		Path to slave window's border
#	m		Path to master

proc ::two::mdi::InactiveSlave {b m} {
    variable Mdi
    $b.w.t       configure -bg $Mdi($m,titlebg)
    $b.w.t.icon  configure -bg $Mdi($m,titlebg)
    $b.w.t.label configure -bg $Mdi($m,titlebg) -fg $Mdi($m,titlefg)
    return
}

# ::two::mdi::SetFocus --
#
#	Set focus to the first child of $w which takes focus
#	or, if there is none, to $w. If any child of $w already
#	has the focus: don't mess with it.
#
# Arguments:
#	w		Path to widget

proc ::two::mdi::SetFocus {w} {
    if {![string match "$w.*" [focus]]} {
	set child [FocusNextChild $w]
	if {[string match "$w*" $child]} {
            focus $child
	} else {
	    focus $w
	}
    }
    return
}

# ::two::mdi::FocusNextChild --
#	Look for the first child of the widget "w" which is willing
#	to take the focus (like in keyboard traversal). Skip
#	toplevels!
#	(modified version of tk_focusNext from tk 8.0.3)
#
# Arguments:
#	w		Path to widget

proc ::two::mdi::FocusNextChild {w} {
    set cur $w
    while {1} {

	# Descend to just before the first child of the current widget.

	set parent $cur
	set children [winfo children $cur]
	set i -1

	# Look for the next sibling that isn't a top-level.

	while {1} {
	    incr i
	    if {$i < [llength $children]} {
		set cur [lindex $children $i]
		if {[string compare [winfo toplevel $cur] $cur] == 0} {
		    continue
		} else {
		    break
		}
	    }

	    # No more siblings, so go to the current widget's parent.
	    # If it's the current widget, break out of the loop, otherwise
	    # look for its next sibling.

	    set cur $parent
	    if {[string compare $w $cur] == 0} {
		break
	    }
	    set parent [winfo parent $parent]
	    set children [winfo children $parent]
	    set i [lsearch -exact $children $cur]
	}
	if {[string compare $cur $w] == 0} {
	    return $cur
	}
	set errCode [catch {$cur cget -takefocus} focusOk]
	if {![string length $focusOk]} {
	    set focusOk 1
	}
	# This is a bit of a HACK: a TwoEditor widget might be
	# split, and then one of the clones is active, i.e. had the
	# focus before the whole editor lost the focus.
	if {[string match TwoEditor [$cur cget -class]]} {
	    return $::two::editor::Editor($cur,active)
	}
	if {!$errCode && $focusOk} {
	    return $cur
	}
    }
}

# ::two::mdi::Animate --
#
#	show anomation for window size changes (minimize/maximize/restore)
#	BETTER NOT USE IT - it's too slow :-(
#
# Arguments:
#	m		Master widget
#	x1		start x position
#	y1		start y position
#	w1		start width
#	h1		start height
#	x2,y2,w2,h2	dito for end

proc ::two::mdi::Animate {m x1 y1 w1 h1 x2 y2 w2 h2} {
    variable Mdi
    set a $m._anim
    if {[winfo exists $a]} {
        return
    }
    frame $a -bg $Mdi($m,activebg)
    set dx [expr {($x2 - $x1) / 5}]
    set dy [expr {($y2 - $y1) / 5}]
    set dw [expr {($w2 - $w1) / 5}]
    set dh [expr {($h2 - $h1) / 5}]

    after idle [namespace code _Animate] $a $x1 $y1 $dx $dy $w1 $h1 $dw $dh 4
    return
}

# ::two::mdi::_Animate --
#
#	helper function for aninate.

proc ::two::mdi::_Animate {a x1 y1 dx dy w1 h1 dw dh count} {
    if {$count == 0} {
        destroy $a
    } else {
	incr x1 $dx
	incr y1 $dy
	incr w1 $dw
	incr h1 $dh
	incr count -1
	place $a -x $x1 -y $y1 -width $w1 -height $h1
	update
	after idle [namespace code _Animate] $a $x1 $y1 $dx $dy $w1 $h1 $dw $dh $count
    }
    return
}

# ::two::mdi::DeleteSlave --
#
#	Delete slave widget from MDI master. Clean up $Mdi.
#	Maybe ask closecmd beforehand.
#	Returns 1 if really deleted
#
# Arguments:
#	b		Path to slave window's border
#	force		If set to 1, no closecmd will be called

proc ::two::mdi::DeleteSlave {b {force 0}} {
    variable Mdi

    set m [winfo parent $b]

    # run closecmd if there is one
    if {!$force && [string length $Mdi($b,closecmd)]} {
        set retval [namespace eval :: $Mdi($b,closecmd) [list $b.w.s]]
        if {![regexp -nocase {^(1|true|yes|ok)$} $retval]} {
	    return 0
	}
    }

    # remove from list of slaves
    set pos [lsearch -exact $Mdi($m,slaves) $b]
    set Mdi($m,slaves) [lreplace $Mdi($m,slaves) $pos $pos]

    # Disable the first two root menu entries, i.e. the shuffle operations
    # stop making sense?
    if {[llength $Mdi($m,slaves)] == 1} {
        $m.root entryconfigure 0 -state disabled
        $m.root entryconfigure 1 -state disabled
    }

    if {!$Mdi($m,arrangeicons) && $Mdi($b,min)} {
	# remove from list of icons
	set icon [lsearch -exact $Mdi($m,icons) $Mdi($b,icon)]
	set Mdi($m,icons) [lreplace $Mdi($m,icons) $icon $icon]
    }

    # send an event
    event generate $b.w.s <<Closeing>>

    # remove binding from inner container widget to avoid infinite loop
    # (would call DeleteSlave:-)
    bindtags $b.w.s nothing

    # finally destroy widget
    place forget $b
    destroy $b
    
    # set focus if needed
    if {[string match "." [focus]]} {
        set next [lindex [AllSlaveBorders $m 1] end]
	if {[string length $next]} {
	    # yes, there is another slave to recieve the focus
            SetFocus $next.w.s
	} else {
	    # sorry, this was the last slave which is willing to take focus
	    set Mdi($m,active) ""
            event generate $m <<ActiveOff>>
	}
    } else {
        # No need to reset focus, but this might as well
        # have been the last slave willing to take the focus:-)
        if {[llength [AllSlaveBorders $m 1]] < 1} {
            event generate $m <<ActiveOff>>
        }
    }
    # maybe reorder icons
    if {$Mdi($m,arrangeicons) && $Mdi($b,min)} {
        Resize $m
    }
    # clear Mdi
    foreach key [array names Mdi "$b,*"] {
        unset Mdi($key)
    }

    # resize: now we might have to set the scrollbars different
    Resize $m

    return 1
}

# ::two::mdi::AddSlave --
#
#	Add slave widget to MDI master
#	Widget Hierarchy:
#	$m                          Frame:  master widget
#	+- $m.#                     Frame:  border widget (for internal use); # = nonce = $b
#	|  +- $m.#.w                Frame:  provide space (padding!) for border
#	|  |  +- $m.#.w.t           Frame:  title bar
#	|  |  |  +- $m.#.w.t.icon   Label:  icon in title bar
#	|  |  |  +- $m.#.w.t.label  Label:  text in title bar
#	|  |  |  +- $m.#.w.t.min    Button: minimize or restore from min
#	|  |  |  +- $m.#.w.t.max    Button: maximize or restore from max
#	|  |  |  +- $m.#.w.t.close  Button: close slave
#	|  |  +- $m.#.w.d           Frame:  spaceing (distance) between title and content
#	|  |  +- $m.#.w.s           Frame:  container for real slave = $s (returned to user)
#	+- $m.menu                  Menu:   Slave window menu
#	+- $m.root                  Menu:   Master root menu
#
# Arguments:
#	m		Path to master
#	args		Additional args

proc ::two::mdi::AddSlave {m args} {
    variable Mdi

    # call resize to reset scrollbar values
    # Otherwise we would have to take the current displacement into account
    Resize $m

    # Name of new border
    set b $m.$Mdi(nonce)
    lappend Mdi($m,slaves) $b
    
    # Enable the first two root menu entries, i.e. the shuffle operations
    # start making sense?
    if {[llength $Mdi($m,slaves)] == 2} {
        $m.root entryconfigure 0 -state normal
        $m.root entryconfigure 1 -state normal
    }

    # Set defaults
    set Mdi($b,title)    "New $Mdi(nonce)"
    set Mdi($b,closecmd) ""
    set Mdi($b,min)      0
    set Mdi($b,lastmenu) 0

    # See if there are some args for us
    two_parseArgs args Mdi($b,title)    "-title"
    two_parseArgs args Mdi($b,closecmd) "-closecmd"
    two_parseArgs args posSpec		"-pos"

    # window border is built by a simple frame
    frame $b -class TwoMdi
    # with our additinal bindings
    set tl [winfo toplevel $m]
    bindtags $b [list $b TwoMdiBorder TwoMdiFrame $tl all]

    # another frame within window border to glue components
    # and provide padding space (to show the border)
    pack [frame $b.w -bd 0 -class TwoMdi] \
        -side bottom -expand true -fill both \
        -padx $Mdi(borderW) -pady $Mdi(borderW)
    bindtags $b.w [list $b.w TwoMdiSubFrame $tl all]

    # title
    pack [frame  $b.w.t -bd 0 -class TwoMdiTitle] -side top -fill x
    button $b.w.t.close -takefocus 0 -command [namespace code "DeleteSlave $b"]
    button $b.w.t.max   -takefocus 0 -command [namespace code "Maximize $b"]
    button $b.w.t.min   -takefocus 0 -command [namespace code "Minimize $b"]
    label  $b.w.t.icon  -image mdiText
    label  $b.w.t.label -justify left -anchor w -bd 0
    # frame gets no bindings (Problem: class name equals binding)
    bindtags $b.w.t nothing
    # label and icon get our own bindings
    bindtags $b.w.t.label [list $b.w.t.label TwoMdiTitle $tl all]
    bindtags $b.w.t.icon  [list $b.w.t.icon  TwoMdiIcon  $tl all]
    # configure and pack the elements according to the current style
    SetSlaveStyle $b
    
    # spaceing
    pack [frame $b.w.d -height $Mdi(borderW) -bd 0 -class TwoMdi] -side top -fill x
    bindtags $b.w.d nothing
    
    # slave: container for user's widget
    set s $b.w.s
    pack [uplevel 2 [list frame $s -bd 0 -class TwoMdiSlave] $args] \
        -side bottom -expand true -fill both
    bindtags $s [list $s TwoMdiSlave $tl all]
    
    # configure slave widget
    ConfigureSlave $b

    # do we have a position?
    if {[info exists posSpec]} {
        SetSlavePos $m $b $posSpec
    } else {
        # calculate width and height of slave
        # each should be 75% of master, but at least have a minimum size
        set width  [winfo width  $m]
        set height [winfo height $m]
        set slaveW [expr {0.75 * $width}]
        set slaveH [expr {0.75 * $height}]
        if {$slaveW < $Mdi(slaveW)} {
            set slaveW $Mdi(slaveW)
        }
        if {$slaveH < $Mdi(slaveH)} {
            set slaveH $Mdi(slaveH)
        }
    
        # calculate new offset
        if {$Mdi(offset) > $width - $slaveW || $Mdi(offset) > $height - $slaveH } {
            set Mdi(offset) $Mdi(offset0)
        }
    
        # set new offset as widget position for scrollbar calculation
        set Mdi($b,x) $Mdi(offset)
        set Mdi($b,y) $Mdi(offset)
        
        # place window within master
        if {$Mdi($m,max)} {
            set Mdi($b,place) "-x $Mdi(offset) -y $Mdi(offset) -rely 0 \
                -width $slaveW -height $slaveH -relwidth 0 -relheight 0 -anchor nw"
            place $b -x 0 -y 0 -rely 0 -width 0 -height 0 -relwidth 1 -relheight 1 -anchor nw
            $b.w.t.min configure -image "$Mdi($m,icon)Min" \
                -command [namespace code "Minimize $b"]
            $b.w.t.max configure -image "$Mdi($m,icon)Rest" \
                -command [namespace code "RestoreMax $b"]
        } else {
            place $b -x $Mdi(offset) -y $Mdi(offset) -width $slaveW -height $slaveH
        }
    }
    # resize again: now we might have to set the scrollbars differently
    QueueResize $m

    incr Mdi(offset) $Mdi(step)
    incr Mdi(nonce)
    return $s
}

namespace import ::two::mdi::*

# text.tcl --
#
#	Modified text widget bindings. This widget behaves more like
#	windows text widgets. No emacs bindings are used to be able
#	to use those keys for the application's shortcuts. Has unlimited
#	undo and redo capability.
#
# Copyright (c) 1998, 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# parts of this code are based on the tk4.2 library file "text.tcl"
# SCCS: @(#) text.tcl 1.46 96/08/23 14:07:32
# Copyright (c) 1992-1994 The Regents of the University of California.
# Copyright (c) 1994-1996 Sun Microsystems, Inc.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoText 1.0
package require twoUtil
package require twoFramework

namespace eval ::two::text {
    namespace export two_text
    variable Tooltip
    
    # variable Text
    #
    # $w,cursor -	Original cursor while dragging with control-B2
    # $w,undoSp -	Pointer to the undo stack
    # $w,undo0Sp -	Pointer to the stack frame where the document is
    #			considered unmodified
    # $w,u_a<n> -	Undo command (d = delete, i = insert, r = replace)
    # $w,u_b<n> -	First undo argument (start)
    # $w,u_c<n> -	Second undo argument (end / text)
    # $w,r_?<n> -	Dito, but redo command; ? = a,b,c
    # $w,p_? -   	   Dito, but previous command and *no* count (no stack here);
    #			? = a,b,c
    # $w,redoSp -	Pointer to the redo stack
    # $w,mod -		True if editor has been modified
    # $w,nappend -	Don't append next key to undo stack (needed after save)
    # $w,ai -		If 1 then any Return will indent the new line just like
    #			the previous one
    # $w,istep -	Number of chars for indenting
    # $w,istring -	String consisting of $w,istep number of spaces
    # $w,xsel -		Export selection setting; we don't use tk's exportselection directly as
    #			we might have the same selection in several cloned texts!
    # $w,clones -	List of Paths to all cloned widgets, including the current widget
    # $w,prevPos -	Used when moving up or down lines via the keyboard.
    #			Keeps track of the previous insert position, so
    #			we can distinguish a series of ups and downs, all
    #			in a row, from a new up or down.
    # $w,prevCol -	Same as above, but only the column
    # word -		Characters allowed in a word (used by control-left/right)
    # afterId -		If non-null, it means that auto-scanning is underway
    #			and it gives the "after" id for the next auto-scan
    #			command to be executed.
    # char -		Character position on the line;  kept in order
    #			to allow moving up or down past short lines while
    #			still remembering the desired position.
    # mouseMoved -	Non-zero means the mouse has moved a significant
    #			amount since the button went down (so, for example,
    #			start dragging out a selection).
    # pressX - 		Horizontal position of last ButtonPress-1
    # selectMode -	The style of selection currently underway:
    #			char, word, or line.
    # x, y -		Last known mouse coordinates for scanning and auto-scanning.
    
    array set Text {
        word		{[_a-zA-Z0-9]}
        nonword		{[^_a-zA-Z0-9]}
        afterId		{}
        pressX		0
        char		0
        mouseMoved	0
        selectMode	char
    }
}

# Notice:
# -------
# In many places, the original widget proc "::two::text::_two$w" is
# used instead of the replacement "$w". Often, this is just for performance
# reasons (no need to go through another level of command parsing).
# In the two_text widget, there is another reason for doing this: the
# widget may have clones, and using the $w interface will usually change
# the settings in all of the clones. This is undesiable in many places
# in internal widget code and thus is worked around through directly
# acessing the original widget.


# ::two::text::two_text --
#
#	Create widget, parse arguments and install our own widget procedure.
#	To increase speed, "_two$w" is called directly (instead of "$w") whenever
#	it is known that only native options are used.
#
# Arguments:
#	w -		Widget path
#	args -		Widget arguments and options

two_widget ::two::text::two_text Text text {
    {-ai         -autoindent}
    {-istep      -indentstep}
    {-exportsel  -exportselection}
    {-xsel       -exportselection}
    {-autoindent
        autoIndent
	AutoIndent
	0
	{}
	{UpdateClones $w -autoindent}
    }
    {-overwrite
        overWrite
	OverWrite
	0
	{SetOverwrite $w; ShowOverwrite $w}
	{UpdateClones $w -overwrite}
    }
    {-indentstep
        indentStep
	IndentStep
	4
	{SetIndent $w}
	{UpdateClones $w -indentstep istring}
    }
    {-exportselection
        exportSelection
	ExportSelection
	1
	{}
	{UpdateClones $w -exportselection}
    }
    {-hardtab
        hardTab
	HardTab
	0
	{}
	{UpdateClones $w -hardtab}
    }
    {unknown Configure}
} {
    # indent step defaults to 4
    if {![info exists Text($w,istring)]} {
        set Text($w,istring) "    "
    }

    # Initialise the "globals"
    array set Text "
        $w,undoSp	0
	$w,undo0Sp	0
	$w,redoSp	0
	$w,p_a		{}
	$w,p_b		{}
	$w,p_c		{}
	$w,mod		0
	$w,nappend	0
	$w,clones	$w
        $w,prevPos	1.0
        $w,prevCol	0
    "

    # use our own bindings
    bindtags $w [list $w TwoText [winfo toplevel $w] all]
    
    # set the anchor mark
    $w mark set anchor insert

    # introduce the insert tag (distinct from the insert mark!)
    # used to highlight the character which is being overwritten
    # in overwrite mode
    $w tag configure insert -relief solid -borderwidth 1

    # Install our own selection handle
    proc _two$w,handle {offset length} "
        return \[[namespace code Selection] $w \$offset \$length\]
    "
    selection handle $w [namespace code _two$w,handle]
} {
    afterLoad {
	foreach clone $Text($w,clones) {
	    UndoReset $clone
	    RedoReset $clone
            set Text($clone,mod) 0
	    event generate $clone <<ModOff>>
	    event generate $clone <<SelOff>>
            event generate $clone <<UndoOff>>
	}
    }
    afterSave {
	foreach clone $Text($w,clones) {
	    set Text($clone,undo0Sp) $Text($clone,undoSp)
	    set Text($clone,nappend) 1
            set Text($clone,mod) 0
	    event generate $clone <<ModOff>>
	    event generate $clone <<SelOff>>
	}
    }
    clone {
	if {[llength $args] != 1} {
	    return -code error "wrong # args: should be \"$w clone newTextName\""
	}
	set clone [lindex $args 0]
	lappend Text($w,clones) $clone

	# create clone widget
	uplevel [list two_text $clone]

	# copy config
	foreach conf [_two$w configure] {
	    if {[llength $conf] == 5} {
		if {[string compare [lindex $conf 3] [lindex $conf 4]] != 0} {
		    if {![string match *scrollcommand [lindex $conf 0]]} {
		        _two$clone configure [lindex $conf 0] [lindex $conf 4]
		    }
		}
	    }
	}

	# copy state
	foreach key [array names Text "$w,*"] {
	    regsub "^$w,(.*)\$" $key "$clone,\\1" cloneKey
            set Text($cloneKey) $Text($key)
	}

	# copy contents
	_two$clone delete 1.0 end
	_two$clone insert 1.0 [_two$w get 1.0 end-1c]

	# copy marks
	eval [list _two$clone mark unset] [_two$clone mark names]
	foreach mark [_two$w mark names] {
	    _two$clone mark set     $mark [_two$w index        $mark]
	    _two$clone mark gravity $mark [_two$w mark gravity $mark]
	}

	# copy tags
	eval [list _two$clone tag delete] [_two$clone tag names]
	foreach tag [_two$w tag names] {
	    foreach cfg [_two$w tag configure $tag] {
	        # only copy the default value:
	        # $cfg consists of 2 or 5 elements
	        # 2 -> a synonym: argvName, dbName(synonym name)
	        # 5 -> argvName, dbName, dbClass, defValue, currentValue
	        if {[string length [lindex $cfg 4]]} {
		    _two$clone tag configure $tag [lindex $cfg 0] [lindex $cfg 4]
		}
	    }
	    if {[llength [_two$w tag ranges $tag]]} {
		eval [list _two$clone tag add $tag] [_two$w tag ranges $tag]
	    }
	}
	return $clone
    }
    delete {
	set oldPos [_two$w index insert]
	set oldSel [llength [_two$w tag ranges sel]]
	set oldSp  $Text($w,undoSp)
	set oldSp0 $Text($w,undo0Sp)
	set i1 [_two$w index [lindex $args 0]]
	if {[llength $args] < 2} {
	    set i2 $i1
	} else {
	    set i2 [_two$w index [lindex $args 1]]
	}
	foreach clone $Text($w,clones) {
	    set retval [Delete $clone $i1 $i2]
	    if {$oldSp == $oldSp0 && $Text($clone,undoSp) != $oldSp} {
        	set Text($clone,mod) 1
        	event generate $clone <<ModOn>>
	    }
	    if {!$oldSp && $Text($clone,undoSp)} {
        	event generate $clone <<UndoOn>>
	    }
	    if {$oldSel && ![llength [_two$clone tag ranges sel]]} {
                event generate $clone <<SelOff>>
	    }
	}
	if {![string match $oldPos [_two$w index insert]]} {
	    SendPos $w
	} elseif {$Text($w,-overwrite)} {
            ShowOverwrite $w
        }

	return $retval
    }
    goto {
	set oldPos [_two$w index insert]
	set retval [$w mark set insert [lindex $args 0]]
        if {[llength [_two$w tag ranges sel]]} {
            $w tag remove sel 1.0 end
	}
	$w see insert
        if {![string match $oldPos [_two$w index insert]]} {
            SendPos $w
        } elseif {$Text($w,-overwrite)} {
            ShowOverwrite $w
        }
	return $retval
    }
    indent {
	# get numerical representation of insert
	set pos [_two$w index insert]
	set oldPos $pos
	set oldSel [llength [_two$w tag ranges sel]]
	set oldSp  $Text($w,undoSp)
	set oldSp0 $Text($w,undo0Sp)
	foreach clone $Text($w,clones) {
	    Indent $clone $pos [lindex $args 0]
	    if {$oldSp == $oldSp0 && $Text($clone,undoSp) != $oldSp} {
        	set Text($clone,mod) 1
        	event generate $clone <<ModOn>>
	    }
	    if {!$oldSp && $Text($clone,undoSp)} {
        	event generate $clone <<UndoOn>>
	    }
	    if {$oldSel && ![llength [_two$clone tag ranges sel]]} {
                event generate $clone <<SelOff>>
	    }
	}
	if {![string match $oldPos [_two$w index insert]]} {
	    SendPos $w
	} elseif {$Text($w,-overwrite)} {
            ShowOverwrite $w
        }
    }
    insert {
	set oldPos [_two$w index insert]
	set oldSp  $Text($w,undoSp)
	set oldSp0 $Text($w,undo0Sp)
	# get numerical representation of position
	set i [_two$w index [lindex $args 0]]
	set s [lrange $args 1 end]
	if {![string length [lindex $s 0]] && [llength $s] == 1} {
            return
	}
	set i [_two$w index $i]
	foreach clone $Text($w,clones) {
	    set retval [eval [list Insert $clone $i] $s]
	    if {$oldSp == $oldSp0 && $Text($clone,undoSp) != $oldSp} {
        	set Text($clone,mod) 1
        	event generate $clone <<ModOn>>
	    }
	    if {!$oldSp && $Text($clone,undoSp)} {
        	event generate $clone <<UndoOn>>
	    }
	}
	if {![string match $oldPos [_two$w index insert]]} {
	    SendPos $w
	} elseif {$Text($w,-overwrite)} {
            ShowOverwrite $w
        }
	return $retval
    }
    insertReplace {
        # anything selected? if yes: trying to insert text *within* selection
	# will replace the whole selection
        if {[llength [_two$w tag ranges sel]] &&
	    [string length [lindex $args 1]] &&
    	    [_two$w compare [lindex $args 0] >= sel.first] &&
	    [_two$w compare [lindex $args 0] <= sel.last]
	} {
	    return [eval [list $w replace sel.first sel.last] [lrange $args 1 end]]
	}
	# nothing selected...
	if {$Text($w,-overwrite) && [$w compare insert != "insert lineend"]} {
	    # we are overwriting and NOT at the end of a line
	    # get number of characters that are to be inserted
	    set s ""
	    foreach {txt tag} [lrange $args 1 end] {
	        append s $txt
	    }
	    set len [string length [string trimright $s "\n"]]
	    return [eval [list $w replace insert "insert + $len c"] [lrange $args 1 end]]
	}
	# nothing special: just insert the text
	return [eval [list $w insert] $args]
    }
    mark {
	set oldPos [_two$w index insert]
	# only some of the mark subcommands apply to the other clones
	foreach clone $Text($w,clones) {
            if {[string compare $clone $w] != 0} {
		switch -exact -- [lindex $args 0] {
		    set {
		        if {![string match insert [lindex $args 1]]} {
                            _two$clone mark set [lindex $args 1] \
                                [_two$w index [lindex $args 2]]
			}
		    }
		    gravity -
		    unset {
		        eval [list _two$clone mark] $args
		    }
		}
	    }
	}
	set retval [eval [list _two$w mark] $args]
	if {[string match "set"    [lindex $args 0]] &&
	    [string match "insert" [lindex $args 1]]
	} {
	    if {![string match $oldPos [_two$w index insert]]} {
		SendPos $w
	    } elseif {$Text($w,-overwrite)} {
                ShowOverwrite $w
            }
	}
	return $retval
    }
    modState {
        return $Text($w,mod)
    }
    paste {
        # what text to paste?
        if {[llength $args] > 1} {
            # given on command line
            set txt [join [lrange $args 1 end] ""]
        } else {
            # none given, so take primary selection
            set txt [selection get -displayof $w]
        }
        # anything selected? if yes: trying to paste text *within* the selection
	# will replace the whole selection
        if {[llength [_two$w tag ranges sel]] &&
	    [string length [lindex $args 1]] &&
    	    [_two$w compare [lindex $args 0] >= sel.first] &&
	    [_two$w compare [lindex $args 0] <= sel.last]
	} {
	    return [$w replace sel.first sel.last $txt]
	}
	# nothing selected...
	return [$w insert [lindex $args 0] $txt]
    }
    redo {
        if {!$Text($w,redoSp)} {
            return 0
        }
	set oldPos [_two$w index insert]
	set oldSp  $Text($w,undoSp)
	set oldSel [llength [_two$w tag ranges sel]]
	foreach clone $Text($w,clones) {
	    Redo $clone
	    if {$Text($clone,undoSp) == $Text($clone,undo0Sp)} {
        	set Text($clone,mod) 0
		event generate $clone <<ModOff>>
	    } elseif {$Text($clone,undoSp) == $Text($clone,undo0Sp) + 1} {
        	set Text($clone,mod) 1
		event generate $clone <<ModOn>>
	    }
	    if {!$Text($clone,redoSp)} {
		event generate $clone <<RedoOff>>
	    }
	    if {!$oldSp} {
		event generate $clone <<UndoOn>>
	    }
	    if {$oldSel && ![llength [_two$clone tag ranges sel]]} {
                event generate $clone <<SelOff>>
	    }
	}
        # show cursor
        _two$w see insert

	if {![string match $oldPos [_two$w index insert]]} {
	    SendPos $w
	} elseif {$Text($w,-overwrite)} {
            ShowOverwrite $w
        }
    }
    redoState {
        if {$Text($w,redoSp) > 0} {
	    return 1
	} else {
	    return 0
	}
    }
    replace {
	set oldPos [_two$w index insert]
	set oldSel [llength [_two$w tag ranges sel]]
	set oldSp  $Text($w,undoSp)
	set oldSp0 $Text($w,undo0Sp)
	set i1 [_two$w index [lindex $args 0]]
	if {[llength $args] < 2} {
	    set i2 $i1
	} else {
	    set i2 [_two$w index [lindex $args 1]]
	}
	set s [lrange $args 2 end]
	foreach clone $Text($w,clones) {
	    set retval [eval [list Replace $clone $i1 $i2] $s]
	    if {$oldSp == $oldSp0 && $Text($clone,undoSp) != $oldSp} {
        	set Text($clone,mod) 1
        	event generate $clone <<ModOn>>
	    }
	    if {!$oldSp && $Text($clone,undoSp)} {
        	event generate $clone <<UndoOn>>
	    }
	    if {$oldSel && ![llength [_two$clone tag ranges sel]]} {
                event generate $clone <<SelOff>>
	    }
	}
	if {![string match $oldPos [_two$w index insert]]} {
	    SendPos $w
	} elseif {$Text($w,-overwrite)} {
            ShowOverwrite $w
        }
	return $retval
    }
    see {
        set pos [lindex $args 0]
        # make sure the line isn't the bottom line...
        _two$w see "$pos + 1 line"
        _two$w see $pos
    }
    selState {
	if {[llength [_two$w tag ranges sel]]} {
	    return 1
	} else {
	    return 0
	}
    }
    tag {
        set sel1 [llength [_two$w tag ranges sel]]
	foreach clone $Text($w,clones) {
            # only some of the tag subcommands apply to the other clones
            if {[string compare $clone $w] != 0} {
		switch -exact -- [lindex $args 0] {
		    add {
		        set newargs [lrange $args 0 1]
		        foreach idx [lrange $args 2 end] {
		            lappend newargs [_two$w index $idx]
		        }
                        eval [list _two$clone tag] $newargs
		    }
		    remove -
		    bind -
		    configure -
		    lower -
		    raise -
		    delete {
			eval [list _two$clone tag] $args
		    }
		}
	    }
	}
	# but all apply to the primary widget
	set retval [eval [list _two$w tag] $args]
        set sel2 [llength [_two$w tag ranges sel]]
	foreach clone $Text($w,clones) {
	    if {$sel1 && !$sel2} {
                event generate $clone <<SelOff>>
	    } elseif {!$sel1 && $sel2} {
                event generate $clone <<SelOn>>
	    }
	}
	if {$sel2 && $Text($w,-exportselection)} {
	    selection own $w
	}
	return $retval
    }
    undo {
	if {!$Text($w,undoSp)} {
	    return 0
	}
	set oldPos [_two$w index insert]
	set oldSel [llength [_two$w tag ranges sel]]
	set oldSp  $Text($w,redoSp)
	foreach clone $Text($w,clones) {
	    Undo $clone
	    if {$Text($clone,undoSp) == $Text($clone,undo0Sp)} {
        	set Text($clone,mod) 0
		event generate $clone <<ModOff>>
	    } elseif {$Text($clone,undoSp) == $Text($clone,undo0Sp) - 1} {
        	set Text($clone,mod) 1
		event generate $clone <<ModOn>>
	    }
	    if {!$Text($clone,undoSp)} {
		event generate $clone <<UndoOff>>
	    }
	    if {!$oldSp} {
		event generate $clone <<RedoOn>>
	    }
	    if {$oldSel && ![llength [_two$clone tag ranges sel]]} {
                event generate $clone <<SelOff>>
	    }
	}
        # show cursor
        _two$w see insert

	if {![string match $oldPos [_two$w index insert]]} {
	    SendPos $w
	} elseif {$Text($w,-overwrite)} {
            ShowOverwrite $w
        }
    }
    undoState {
	if {$Text($w,undoSp) > 0} {
	    return 1
	} else {
	    return 0
	}
    }
} {
    # remove reference from cloned widgets
    set pos [lsearch -exact $Text($w,clones) $w]
    set others [lreplace $Text($w,clones) $pos $pos]
    foreach other $others {
        set Text($other,clones) $others
    }
}

#-------------------------------------------------------------------------
# The code below creates the TwoText-class bindings
#-------------------------------------------------------------------------

bind TwoText <FocusIn> {
    ::two::text::SendPos %W
    ::two::text::SetOverwrite %W
}
bind TwoText <1> {
    set ::two::text::Text(selectMode)  char
    set ::two::text::Text(mouseMoved)  0
    set ::two::text::Text(pressX)      %x

    ::two::text::SetCursor %W [::two::text::ClosestGap %W %x %y]
    %W mark set anchor [::two::text::_two%W index insert]

    if {[string match "normal" [::two::text::_two%W cget -state]]} {
        focus %W
    }
    if {[llength [::two::text::_two%W tag ranges sel]]} {
        %W tag remove sel 1.0 end
    }
}
bind TwoText <B1-Motion> {
    set ::two::text::Text(x) %x
    set ::two::text::Text(y) %y
    ::two::text::SelectTo %W %x %y
}
bind TwoText <Double-1> {
    set ::two::text::Text(selectMode) word
    ::two::text::SelectTo %W %x %y
}
bind TwoText <Triple-1> {
    set ::two::text::Text(selectMode) line
    ::two::text::SelectTo %W %x %y
}
bind TwoText <Shift-1> {
    ::two::text::SelectTo %W %x %y
}
bind TwoText <Double-Shift-1> {
    ::two::text::SelectTo %W %x %y
}
bind TwoText <Triple-Shift-1> {
    ::two::text::SelectTo %W %x %y
}
bind TwoText <B1-Leave> {
    set ::two::text::Text(x) %x
    set ::two::text::Text(y) %y
    ::two::text::AutoScan %W
}
bind TwoText <B1-Enter> {
    ::two::text::CancelRepeat
}
bind TwoText <ButtonRelease-1> {
    ::two::text::CancelRepeat
}
bind TwoText <2> {
    if {!$tk_strictMotif} {
        ::two::text::_two%W scan mark %x %y
        set ::two::text::Text(x) %x
        set ::two::text::Text(y) %y
        set ::two::text::Text(mouseMoved) 0
    }
}
bind TwoText <B2-Motion> {
    if {!$tk_strictMotif} {
        if {(%x != $::two::text::Text(x)) || (%y != $::two::text::Text(y))} {
            set ::two::text::Text(mouseMoved) 1
        }
    }
}
bind TwoText <ButtonRelease-2> {
    if {[string match "normal" [::two::text::_two%W cget -state]]} {
        focus %W
    }
    if {!$::two::text::Text(mouseMoved) || $tk_strictMotif} {
        set ::two::text::Text(%W,nappend) 1
        ::two::text::_two%W mark set insert [::two::text::ClosestGap %W %x %y]
        %W mark set anchor [::two::text::_two%W index insert]
        ::two::text::SendPos %W
	if {[llength [::two::text::_two%W tag ranges sel]]} {
	    %W insert insert \
	        [::two::text::_two%W get sel.first sel.last]
	} else {
	    catch {%W insert insert \
	        [selection get -displayof %W]
            }
	}
    }
}
bind TwoText <Control-2> {
    if {!$tk_strictMotif} {
        set ::two::text::Text(%W,cursor) [%W cget -cursor]
        %W configure -cursor fleur
        ::two::text::_two%W scan mark %x %y
        set ::two::text::Text(x) %x
        set ::two::text::Text(y) %y
        set ::two::text::Text(mouseMoved) 0
    }
}
bind TwoText <Control-B2-Motion> {
    if {!$tk_strictMotif} {
        if {(%x != $::two::text::Text(x)) || (%y != $::two::text::Text(y))} {
            set ::two::text::Text(mouseMoved) 1
        }
        if {$::two::text::Text(mouseMoved)} {
            ::two::text::_two%W scan dragto %x %y
        }
    }
}
bind TwoText <Control-ButtonRelease-2> {
    if {!$tk_strictMotif} {
        %W configure -cursor $::two::text::Text(%W,cursor)
    }
}
bind TwoText <Left> {
    ::two::text::SetCursor %W insert-1c
}
bind TwoText <Right> {
    ::two::text::SetCursor %W insert+1c
}
bind TwoText <Up> {
    ::two::text::SetCursor %W [::two::text::UpDownLine %W -1]
}
bind TwoText <Down> {
    ::two::text::SetCursor %W [::two::text::UpDownLine %W 1]
}
bind TwoText <Shift-Left> {
    ::two::text::KeySelect %W [::two::text::_two%W index {insert - 1c}]
}
bind TwoText <Shift-Right> {
    ::two::text::KeySelect %W [::two::text::_two%W index {insert + 1c}]
}
bind TwoText <Shift-Up> {
    ::two::text::KeySelect %W [::two::text::UpDownLine %W -1]
}
bind TwoText <Shift-Down> {
    ::two::text::KeySelect %W [::two::text::UpDownLine %W 1]
}
bind TwoText <Control-Left> {
    ::two::text::SetCursor %W [::two::text::WordLeftStart %W insert]
}
bind TwoText <Control-Right> {
    ::two::text::SetCursor %W [::two::text::WordRightStart %W insert]
}
bind TwoText <Control-Up> {
    ::two::text::SetCursor %W [::two::text::PrevPara %W insert]
}
bind TwoText <Control-Down> {
    ::two::text::SetCursor %W [::two::text::NextPara %W insert]
}
bind TwoText <Shift-Control-Left> {
    if {
        ![llength [::two::text::_two%W tag ranges sel]] ||
        [::two::text::_two%W compare insert == sel.first]
    } {
        # no selection or cursor at start of selection, extending it to the left
        # -> expand sel to start of next word to the left
        ::two::text::KeySelect %W [::two::text::WordLeftStart %W insert]
    } elseif {[::two::text::_two%W compare insert == sel.last]} {
        # currently at end of selection; trying to shorten from the right
        # -> go to end of next word to the left
        set newpos [::two::text::WordLeftEnd %W insert]
        if {[::two::text::_two%W compare $newpos > sel.first]} {
            # o.k., not gone past the beginning
            ::two::text::KeySelect %W $newpos
        } else {
            # stop at start of selection
            ::two::text::KeySelect %W [::two::text::_two%W index sel.first]
        }
    } else {
        # sel present, but cursor is not at start or end of selection,
        # extending the selection to the left
        # -> go to start of current word and extend sel to start of next word to the left
        ::two::text::KeySelect %W [::two::text::WordLeftStart %W \
            [::two::text::WordStart %W insert]]
    }
}
bind TwoText <Shift-Control-Right> {
    if {
        ![llength [::two::text::_two%W tag ranges sel]] ||
        [::two::text::_two%W compare insert == sel.last]
    } {
        ::two::text::KeySelect %W [::two::text::WordRightEnd %W insert]
    } elseif {[::two::text::_two%W compare insert == sel.first]} {
        set newpos [::two::text::WordRightStart %W insert]
        if {[::two::text::_two%W compare $newpos < sel.last]} {
            ::two::text::KeySelect %W $newpos
        } else {
            ::two::text::KeySelect %W [::two::text::_two%W index sel.last]
        }
    } else {
        ::two::text::KeySelect %W [::two::text::WordRightEnd %W \
            [::two::text::WordEnd %W insert]]
    }
}
bind TwoText <Shift-Control-Up> {
    ::two::text::KeySelect %W [::two::text::PrevPara %W insert]
}
bind TwoText <Shift-Control-Down> {
    ::two::text::KeySelect %W [::two::text::NextPara %W insert]
}
bind TwoText <Prior> {
    ::two::text::SetCursor %W [::two::text::ScrollPages %W -1]
}
bind TwoText <Shift-Prior> {
    ::two::text::KeySelect %W [::two::text::ScrollPages %W -1]
}
bind TwoText <Next> {
    ::two::text::SetCursor %W [::two::text::ScrollPages %W 1]
}
bind TwoText <Shift-Next> {
    ::two::text::KeySelect %W [::two::text::ScrollPages %W 1]
}
bind TwoText <Control-Prior> {
    ::two::text::_two%W xview scroll -1 page
}
bind TwoText <Control-Next> {
    ::two::text::_two%W xview scroll 1 page
}

bind TwoText <Home> {
    ::two::text::SetCursor %W {insert linestart}
}
bind TwoText <Shift-Home> {
    ::two::text::KeySelect %W {insert linestart}
}
bind TwoText <End> {
    ::two::text::SetCursor %W {insert lineend}
}
bind TwoText <Shift-End> {
    ::two::text::KeySelect %W {insert lineend}
}
bind TwoText <Control-Home> {
    ::two::text::SetCursor %W 1.0
}
bind TwoText <Control-Shift-Home> {
    ::two::text::KeySelect %W 1.0
}
bind TwoText <Control-End> {
    ::two::text::SetCursor %W {end - 1 char}
}
bind TwoText <Control-Shift-End> {
    ::two::text::KeySelect %W {end - 1 char}
}
bind TwoText <Control-KeyPress-plus> {
    %W indent 1
}
bind TwoText <Control-KeyPress-greater> {
    %W indent 1
}
bind TwoText <Control-KeyPress-minus> {
    %W indent -1
}
bind TwoText <Control-KeyPress-less> {
    %W indent -1
}
bind TwoText <Tab> {
    if {[llength [::two::text::_two%W tag ranges sel]]} {
        %W indent 1
    } else {
	%W insertReplace insert \t
    }
    # don't change focus
    break
}
bind TwoText <Shift-Tab> {
    if {[llength [::two::text::_two%W tag ranges sel]]} {
        %W indent -1
    } else {
	%W insertReplace insert \t
    }
    # don't change focus
    break
}
bind TwoText <Control-Tab> {
    # Needed only to keep <Tab> binding from triggering;  doesn't
    # have to actually do anything.
}
bind TwoText <Shift-Control-Tab> {
    # Needed only to keep <Tab> binding from triggering;  doesn't
    # have to actually do anything.
}
bind TwoText <Return> {
    ::two::text::Newline %W
}
bind TwoText <Insert> {
    ::two::text::ToggleOverwrite %W
}
bind TwoText <Delete> {
    if {[llength [::two::text::_two%W tag ranges sel]] &&
        [::two::text::_two%W compare insert >= sel.first] &&
        [::two::text::_two%W compare insert <= sel.last]
    } {
	%W delete sel.first sel.last
    } elseif {[::two::text::_two%W compare insert != end-1c]} {
	%W delete insert insert+1c
    }
    ::two::text::_two%W see insert
}
bind TwoText <BackSpace> {
    if {[llength [::two::text::_two%W tag ranges sel]] &&
        [::two::text::_two%W compare insert >= sel.first] &&
        [::two::text::_two%W compare insert <= sel.last]
    } {
	%W delete sel.first sel.last
    } elseif {[::two::text::_two%W compare insert != 1.0]} {
	%W delete insert-1c insert
    }
    ::two::text::_two%W see insert
}
bind TwoText <Control-Delete> {
    if {[llength [::two::text::_two%W tag ranges sel]] &&
        [::two::text::_two%W compare insert >= sel.first] &&
        [::two::text::_two%W compare insert < sel.last]
    } {
	%W delete insert sel.last
    } elseif {[::two::text::_two%W compare insert != end-1c]} {
	%W delete insert "insert lineend"
    }
    ::two::text::_two%W see insert
}
bind TwoText <Control-BackSpace> {
    if {[llength [::two::text::_two%W tag ranges sel]] &&
        [::two::text::_two%W compare insert > sel.first] &&
        [::two::text::_two%W compare insert <= sel.last]
    } {
	%W delete sel.first insert
    } elseif {[::two::text::_two%W compare insert != 1.0]} {
	%W delete "insert linestart" insert
    }
    ::two::text::_two%W see insert
}
bind TwoText <KeyPress> {
    if {[string length %A]} {
	%W insertReplace insert %A
        if {[llength [::two::text::_two%W tag ranges sel]]} {
            %W tag remove sel 1.0 end
	}
        ::two::text::_two%W see insert
    }
}

# Ignore all Alt, Meta, and Control keypresses unless explicitly bound.
# Otherwise, if a widget binding for one of these is defined, the
# <KeyPress> class binding will also fire and insert the character,
# which is wrong.  Ditto for <Escape>.

bind TwoText <Alt-KeyPress>     { #nothing }
bind TwoText <Meta-KeyPress>    { #nothing }
bind TwoText <Control-KeyPress> { #nothing }
bind TwoText <Escape>           { #nothing }
bind TwoText <KP_Enter>         { #nothing }

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------
# Please note that calling the widget function (as returned by %W)
# will modify all cones as well and properly set undo/redo etc.,
# while using _two%W applies directly to the widget itself.
# This is useful for "mark set insert" only!

# ::two::text::SetIndent --
#
#	Set the string representation of the indentation step.
#	This is used when indenting a portion of text; it
#	consists of n whitespaces, where n ist the indent step.
#
# Arguments:
#	w		Path to widget

proc ::two::text::SetIndent {w} {
    variable Text

    set Text($w,istring) ""
    for {set x 0} {$x < $Text($w,-indentstep)} {incr x} {
	append Text($w,istring) " "
    }
}

# ::two::text::Newline --
#
#	Insert a newline character.
#
# Arguments:
#	w		Path to widget

proc ::two::text::Newline {w} {
    variable Text

    $w insertReplace insert \n

    if {$Text($w,-autoindent)} {
        set prevLine [_two$w index {insert - 1 lines linestart} ]
	set nrSpace 0
	_two$w search -count nrSpace -regexp "^( |\t)*" $prevLine insert
    	if {$nrSpace > 0} {
    	    $w insert insert [_two$w get $prevLine "$prevLine + $nrSpace chars"]
    	}
    }
    _two$w see insert
}


# ::two::text::UpdateClones --
#
#	Copy a specific setting to all clones
#
# Arguments:
#	w -		Path to widget
#	args -		Name of all options which should be updated all clones

proc ::two::text::UpdateClones {w args} {
    variable Text

    foreach opt $args {
	foreach clone $Text($w,clones) {
            set Text($clone,$opt) $Text($w,$opt)
	}
    }
}

# ::two::text::Configure --
#
#	Set configuration options on all clones. Only used for
#	default text-widget-Options. This is a bit hairy, as
#	the scrollbar settings should not be identical on
#	all clones!
#
# Arguments:
#	w -		Path to widget
#	args -		remaining options string....

proc ::two::text::Configure {w args} {
    variable Text

    # loop through all options, but break at "--"
    set delimit [lsearch -exact $args "--"]
    if {$delimit > 0} {
        incr delimit -1
        set argList [lrange $args 0 $delimit]
    } else {
        set argList $args
    }

    # remove scrollbar settings....
    set remaining ""
    foreach {arg value} $argList {
        if {[string match -xs* $arg] || [string match -ys* $arg]} {
	    set retval [_two$w configure $arg $value]
	} else {
	    lappend remaining $arg $value
	}
    }
    # the rest....
    if {$delimit > 0} {
        lappend remaining [lrange $args $delimit end]
    }

    if {[llength $remaining]} {
	if {[info exists Text($w,clones)]} {
            foreach clone $Text($w,clones) {
	        set retval [eval [list _two$clone configure] $remaining -exportselection 0]
	    }
	} else {
            set retval [eval [list _two$w configure] $remaining -exportselection 0]
	}
    }
    return $retval
}

# ::two::text::ClosestGap --
#
#	Given x and y coordinates, this procedure finds the closest boundary
#	between characters to the given coordinates and returns the index
#	of the character just after the boundary.
#
# Arguments:
#	w		The text window.
#	x		X-coordinate within the window.
#	y		Y-coordinate within the window.

proc ::two::text::ClosestGap {w x y} {
    set pos [_two$w index @$x,$y]
    set bbox [_two$w bbox $pos]
    if {![string compare $bbox ""]} {
	return $pos
    }
    if {($x - [lindex $bbox 0]) < ([lindex $bbox 2]/2)} {
	return $pos
    }
    return [_two$w index "$pos + 1 char"]
}

# ::two::text::SelectTo --
#
#	This procedure is invoked to extend the selection, typically when
#	dragging it with the mouse.  Depending on the selection mode (character,
#	word, line) it selects in different-sized units.  This procedure
#	ignores mouse motions initially until the mouse has moved from
#	one character to another or until there have been multiple clicks.
#
# Arguments:
#	w		The text window in which the button was pressed.
#	x		Mouse x position.
#	y 		Mouse y position.

proc ::two::text::SelectTo {w x y} {
    variable Text

    set cur [ClosestGap $w $x $y]

    if {![llength [_two$w tag ranges sel]]} {
        set anchor [_two$w index insert]
        $w mark set anchor $anchor
	if {[_two$w compare $cur < insert]} {
	    $w tag add sel $cur [_two$w index insert]
	} elseif {[_two$w compare $cur > insert]} {
	    $w tag add sel [_two$w index insert] $cur
	}
    } else {
        set anchor [_two$w index anchor]
    }

    if {[_two$w compare $cur != $anchor] || (abs($Text(pressX) - $x) >= 3)} {
	set Text(mouseMoved) 1
    }

    switch -exact -- $Text(selectMode) {
	char {
	    if {[_two$w compare $cur < anchor]} {
		set first $cur
		set last $anchor
	    } else {
		set first $anchor
		set last $cur
	    }
	}
	word {
	    if {[_two$w compare $cur < $anchor]} {
		set first [WordStart $w $cur]
		set last  [WordEnd   $w $anchor]
	    } else {
		set first [WordStart $w $anchor]
		set last  [WordEnd   $w $cur]
	    }
	}
	line {
	    if {[_two$w compare $cur < $anchor]} {
		set first [_two$w index "$cur linestart"]
		set last  [_two$w index "$anchor -1c lineend +1c"]
	    } else {
		set first [_two$w index "$anchor linestart"]
		set last  [_two$w index "$cur lineend +1c"]
	    }
	}
    }
    if {$Text(mouseMoved) || ![string match "char" $Text(selectMode)]} {
        if {[llength [_two$w tag ranges sel]]} {
            if {[_two$w compare $cur > $anchor]} {
                _two$w mark set insert $last
            } else {
                _two$w mark set insert $first
            }
            SendPos $w
        }
        if {[_two$w compare $first != $last]} {
            $w tag remove sel 1.0 $first
            $w tag add    sel $first $last
            $w tag remove sel $last end
	} else {
            $w tag remove sel 1.0 end
	}
    }
    return
}

# ::two::text::AutoScan --
#
#	This procedure is invoked when the mouse leaves a text window
#	with button 1 down.  It scrolls the window up, down, left, or right,
#	depending on where the mouse is (this information was saved in
#	Text(x) and Text(y)), and reschedules itself as an "after"
#	command so that the window continues to scroll until the mouse
#	moves back into the window or the mouse button is released.
#
# Arguments:
#	w		The text window.

proc ::two::text::AutoScan {w} {
    variable Text
    if {![winfo exists $w]} return
    if {$Text(y) >= [winfo height $w]} {
	_two$w yview scroll 2 units
    } elseif {$Text(y) < 0} {
	_two$w yview scroll -2 units
    } elseif {$Text(x) >= [winfo width $w]} {
	_two$w xview scroll 2 units
    } elseif {$Text(x) < 0} {
	_two$w xview scroll -2 units
    } else {
	return
    }
    SelectTo $w $Text(x) $Text(y)
    set Text(afterId) [after 50 [namespace code AutoScan] $w]
    return
}

# ::two::text::SetCursor --
#
#	Move the insertion cursor to a given position in a text.  Also
#	clears the selection, if there is one in the text, and makes sure
#	that the insertion cursor is visible.  Also, don't let the insertion
#	cursor appear on the dummy last line of the text.
#
# Arguments:
#	w		The text window.
#	pos		The desired new position for the cursor in the window.

proc ::two::text::SetCursor {w pos} {
    variable Text

    set Text(selectMode) char

    if {[_two$w compare $pos == end]} {
	set pos "end -1c"
    }
    _two$w mark set insert $pos
    _two$w see insert
    SendPos $w

    if {[llength [_two$w tag ranges sel]]} {
	$w tag remove sel 1.0 end
    }
    return
}

# ::two::text::KeySelect --
#
#	This procedure is invoked when stroking out selections using the
#	keyboard.  It moves the cursor to a new position, then extends
#	the selection to that position.
#
# Arguments:
#	w		The text window.
#	new		A new position for the insertion cursor (the cursor hasn't
#			actually been moved to this position yet).

proc ::two::text::KeySelect {w new} {
    variable Text

    if {![llength [_two$w tag ranges sel]]} {
	$w mark set anchor insert
    }
    # move anchor to same position as if selecting with the mouse
    # (those select modes only are started by double or triple clicking)
    switch -exact -- $Text(selectMode) {
        word {
            if {[_two$w compare $new < anchor]} {
                set anchor [WordEnd $w anchor]
            } else {
                set anchor [WordStart $w anchor]
            }
        }
        line {
            if {[_two$w compare $new < anchor]} {
                set anchor [_two$w index "anchor -1c lineend +1c"]
            } else {
                set anchor [_two$w index "anchor linestart"]
            }
        }
        default {
            set anchor anchor
        }
    }

    if {[_two$w compare $new < $anchor]} {
        set first $new
        set last  $anchor
    } else {
        set first $anchor
        set last  $new
    }
    $w tag remove sel 1.0    $first
    $w tag add    sel $first $last
    $w tag remove sel $last  end

    _two$w mark set insert $new
    _two$w see insert
    SendPos $w
    return
}

# ::two::text::Delete --
#
#	Delete a string from the text, ranging from start to end
#
# Arguments:
#	w		The text window in which to delete text
#	start		The first character to delete
#	end		The last character to delete

proc ::two::text::Delete {w start end} {
    variable Text

    # should we really delete the text?
    if {[string match "disabled" [_two$w cget -state]]} {
        return
    }

    # Workaround: selection in empty editor...aaarrgh :-(
    if {[string match "1.0" [_two$w index end-1c]]} {
        # delete selection
        _two$w delete 1.0 end
        return
    }

    # kill redo stack
    if {$Text($w,redoSp) != 0} {
        RedoReset $w
    }

    # Convert start and end into their numerical representation
    set start [_two$w index $start]
    if {[string match $start [_two$w index end]]} {
	set start [_two$w index "end - 1c"]
    }
    set end [_two$w index $end]
    if {[string match $end [_two$w index end]]} {
	set end [_two$w index "end - 1c"]
    }

    # if deleting single characters: only put single lines on undo stack
    if {[string match "\n" [_two$w get $start $end]]} {
        set Text($w,nappend) 1
    }

    # push reverse command on the undo stack; d=delete, i=insert
    if { [string match "i"  $Text($w,p_a)] &&
         [string match $end $Text($w,p_b)] &&
	 !$Text($w,nappend)
    } {
    	incr Text($w,undoSp) -1
	set Text($w,u_b$Text($w,undoSp)) $start
	set Text($w,u_c$Text($w,undoSp)) [_two$w get $start $end]$Text($w,u_c$Text($w,undoSp))
    	set Text($w,p_b) "$start"
    } else {
        set Text($w,nappend) 0
	set Text($w,u_a$Text($w,undoSp)) "i"
	set Text($w,u_b$Text($w,undoSp)) $start
	set Text($w,u_c$Text($w,undoSp)) [_two$w get $start $end]
    	set Text($w,p_a) "i"
    	set Text($w,p_b) $start
    }
    incr Text($w,undoSp)

    # Finally delete the text...
    _two$w delete $start $end
    return
}

# ::two::text::Insert --
#
#	Insert a string into a text at the point of the insertion cursor.
#	If there is a selection in the text, and it covers the point of the
#	insertion cursor, then delete it
#
# Arguments:
#	w		The text window in which to insert the string
#	pos		Numerical Position of where to insert the text
#	args		List of character - tag pairs. Usually just a single char

proc ::two::text::Insert {w pos args} {
    variable Text

    # should we really insert the string?
    if {[string match "disabled" [_two$w cget -state]]} {
        return
    }

    # kill redo stack
    if {$Text($w,redoSp)} {
        RedoReset $w
    }

    # Convert start into numerical representation, mark it,
    # insert the text and finally retrieve the new end.
    set start [_two$w index $pos]
    if {[string match $start [_two$w index end]]} {
	set start [_two$w index "end - 1c"]
    }
    _two$w mark set _twoPriv $start
    eval [list _two$w insert $start] $args

    set end [_two$w index _twoPriv]
    if {[string match $end [_two$w index end]]} {
	set end [_two$w index "end - 1c"]
    }
    _two$w mark unset _twoPriv

    # if inserting single characters: only put single lines on undo stack
    if {[lsearch -exact $args "\n"] > -1} {
        set Text($w,nappend) 1
    }

    # push reverse command on the undo stack; d=delete, i=insert, r=replace
    if {[string match "d"    $Text($w,p_a)] &&
        [string match $start $Text($w,p_c)] &&
	!$Text($w,nappend)
    } {
    	incr Text($w,undoSp) -1
    	set Text($w,u_b$Text($w,undoSp)) $Text($w,p_b)
    	set Text($w,u_c$Text($w,undoSp)) $end
    	set Text($w,p_c) "$end"
    } elseif {[string match "r" $Text($w,p_a)] &&
        [string match $start $Text($w,p_c)] &&
	!$Text($w,nappend)
    } {
    	incr Text($w,undoSp) -1
    	set Text($w,u_b$Text($w,undoSp)) [list $Text($w,p_b) $end]
    	set Text($w,p_c) $end
    } else {
        set Text($w,nappend) 0
	set Text($w,u_a$Text($w,undoSp)) "d"
	set Text($w,u_b$Text($w,undoSp)) $start
	set Text($w,u_c$Text($w,undoSp)) $end
    	set Text($w,p_a) "d"
    	set Text($w,p_b) $start
    	set Text($w,p_c) $end
    }
    incr Text($w,undoSp)
    return
}

# ::two::text::Replace --
#
#	Replace a portion of the text at the point of the insertion cursor.
#	If there is a selection in the text, and it covers the point of the
#	insertion cursor, then delete it
#
# Arguments:
#	w		The text window in which to insert the string
#	start		Index of first character to be replaced
#	end		Index of last character to be replaced
#	args		List of character - tag pairs. Usually just a single char

proc ::two::text::Replace {w start end args} {
    variable Text

    # should we really insert the string?
    if {[string match "disabled" [_two$w cget -state]]} {
        return
    }

    # kill redo stack
    if {$Text($w,redoSp)} {
        RedoReset $w
    }

    # Convert start into numerical representation, delete the text,
    # mark it the start position,
    # insert the text and finally retrieve the new end.
    set start [_two$w index $start]
    if {[string match $start [_two$w index end]]} {
	set start [_two$w index "end - 1c"]
    }
    set orig [_two$w get $start $end]
    _two$w delete $start $end
    _two$w mark set _twoPriv $start
    eval [list _two$w insert $start] $args
    set end [_two$w index _twoPriv]
    if {[string match $end [_two$w index end]]} {
	set end [_two$w index "end - 1c"]
    }
    _two$w mark unset _twoPriv

    # push reverse command on the undo stack; r=replace
    # Don't append anything here...
    set Text($w,nappend) 0
    set Text($w,u_a$Text($w,undoSp)) "r"
    set Text($w,u_b$Text($w,undoSp)) [list $start $end]
    set Text($w,u_c$Text($w,undoSp)) $orig
    set Text($w,p_a) "r"
    set Text($w,p_b) $start
    set Text($w,p_c) $end
    incr Text($w,undoSp)
    return
}

# ::two::text::UpDownLine --
#
#	Returns the index of the character one line above or below the
#	insertion cursor.  There are two or three tricky things here.  First,
#	we want to maintain the original column across repeated operations,
#	even though some lines that will get passed through don't have
#	enough characters to cover the original column.  Second, don't
#	try to scroll past the beginning or end of the text. And third,
#	if user selected to virtually expand hard-tabs, we have to calculate
#	the right column.
#
# Arguments:
#	w		The text window in which the cursor is to move.
#	n		The number of lines to move: -1 for up one line,
#			+1 for down one line.

proc ::two::text::UpDownLine {w n} {
    variable Text

    # Current Position
    set i [_two$w index insert]
    scan $i "%d.%d" line col

    # calculate new position
    set newLine [expr {$line + $n}]
    if {$Text($w,-hardtab)} {
        # We need to calculate the virtual position by expanding hard tabs
        # to 8 spaces (or less).
        
        # Cursor still where it was last time? -> use cached column
        if {[string match $i $Text($w,prevPos)]} {
            set col $Text($w,prevCol)
        } else {
            # get column with tabs expanded
            set col 0
            foreach char [split [_two$w get "insert linestart" insert] ""] {
                if {[string compare "\t" $char] == 0} {
                    incr col [expr {8 - ($col % 8)}]
                } else {
                    incr col
                }
            }
            # cache result
            set Text($w,prevCol) $col
        }

	# Now get real column within new line:
	set maxCol [expr {$col * 8}]
	set ncol 0 ;# new column number with tabs expanded
	set rcol 0 ;# real column number
	foreach char [split [_two$w get $newLine.0 $newLine.$maxCol] ""] {
	    if {[string compare "\t" $char] == 0} {
	        incr ncol [expr {8 - ($ncol % 8)}]
	    } else {
	        incr ncol
	    }
	    incr rcol
	    if {$ncol >= $col} break
	}
	set new [_two$w index [expr {$line + $n}].$rcol]
    } else {
        # Cursor still where it was last time? -> use cached column
        if {[string match $i $Text($w,prevPos)]} {
            set col $Text($w,prevCol)
        } else {
            set Text($w,prevCol) $col
        }
        # calculate new column
        set new [_two$w index $newLine.$col]
    }

    # did we go past the bounds? That case, we want to remember the
    # previous position (basically the row, the col ain't modified)
    if {[_two$w compare $new == end] || [_two$w compare $new == "insert linestart"]} {
	set new $i
    }
    set Text($w,prevPos) $new
    return $new
}

# ::two::text::PrevPara --
#
#	Returns the index of the beginning of the paragraph just before a given
#	position in the text (the beginning of a paragraph is the first non-blank
#	character after a blank line).
#
# Arguments:
#	w		The text window in which the cursor is to move.
#	pos		Position at which to start search.

proc ::two::text::PrevPara {w pos} {
    set pos [_two$w index "$pos linestart"]
    while {1} {
	if {([string match "\n" [_two$w get "$pos - 1 line"]] &&
	    ![string match "\n" [_two$w get $pos]]) ||
	    [string match "1.0" $pos]
        } {
	    if {[regexp -indices {^[ 	]+(.)} [_two$w get $pos "$pos lineend"] dummy index]} {
		set pos [_two$w index "$pos + [lindex $index 0] chars"]
	    }
	    if {[_two$w compare $pos != insert] || [string match "1.0" $pos]} {
		return $pos
	    }
	}
	set pos [_two$w index "$pos - 1 line"]
    }
    return
}

# ::two::text::NextPara --
#
#	Returns the index of the beginning of the paragraph just after a given
#	position in the text (the beginning of a paragraph is the first non-blank
#	character after a blank line).
#
# Arguments:
#	w		The text window in which the cursor is to move.
#	start		Position at which to start search.

proc ::two::text::NextPara {w start} {
    set pos [_two$w index "$start linestart + 1 line"]
    while {![string match "\n" [_two$w get $pos]]} {
	if {[_two$w compare $pos == end]} {
	    return [_two$w index "end - 1c"]
	}
	set pos [_two$w index "$pos + 1 line"]
    }
    while {[string match "\n" [_two$w get $pos]]} {
	set pos [_two$w index "$pos + 1 line"]
	if {[_two$w compare $pos == end]} {
	    return [_two$w index "end - 1c"]
	}
    }
    if {[regexp -indices {^[ 	]+(.)} [_two$w get $pos "$pos lineend"] dummy index]} {
	return [_two$w index "$pos + [lindex $index 0] chars"]
    }
    return $pos
}

# ::two::text::ScrollPages --
#
#	This is a utility procedure used in bindings for moving up and down
#	pages and possibly extending the selection along the way.  It scrolls
#	the view in the widget by the number of pages, and it returns the
#	index of the character that is at the same position in the new view
#	as the insertion cursor used to be in the old view.
#
# Arguments:
#	w		The text window in which the cursor is to move.
#	count		Number of pages forward to scroll;  may be negative
#			to scroll backwards.

proc ::two::text::ScrollPages {w count} {
    set bbox [_two$w bbox insert]
    set oldyview [_two$w yview]
    _two$w yview scroll $count pages
    if {![string length $bbox]} {
	return [_two$w index @[expr [winfo height $w]/2],0]
    }
    # did we really scroll?
    if {[string compare $oldyview [_two$w yview]]==0} {
        # no scrolling occured
        if {$count < 0} {
            return 1.0
        } else {
            return [_two$w index "end -1c"]
        }
    } else {
        return [_two$w index @[lindex $bbox 0],[lindex $bbox 1]]
    }
}

# ::two::text::CancelRepeat --
#
#	This procedure is invoked to cancel an auto-repeat action described
#	by Text(afterId).  It's used by several widgets to auto-scroll
#	the widget when the mouse is dragged out of the widget with a
#	button pressed.

proc ::two::text::CancelRepeat {} {
    variable Text
    after cancel $Text(afterId)
    set Text(afterId) {}
}

# ::two::text::Undo --
#
#	This procedure is used to undo the last text-modifying action
#
# Arguments:
#	w		Path to text widget

proc ::two::text::Undo {w} {
    variable Text

    # should we really undo changes?
    if {[string match "disabled" [_two$w cget -state]]} {
        return
    }

    if {!$Text($w,undoSp)} {
        return
    }

    incr Text($w,undoSp) -1
    set start $Text($w,u_b$Text($w,undoSp))

    # Do Undo and push reverse command on the redo stack
    switch $Text($w,u_a$Text($w,undoSp)) {
	i {
	    set text $Text($w,u_c$Text($w,undoSp))
	    _two$w mark set insert $start
	    _two$w insert $start $text
	    set end [_two$w index insert]

	    set Text($w,r_a$Text($w,redoSp)) "d"
	    set Text($w,r_b$Text($w,redoSp)) $start
	    set Text($w,r_c$Text($w,redoSp)) $end
	}
	d {
	    set end $Text($w,u_c$Text($w,undoSp))
	    set text [_two$w get $start $end]
	    _two$w mark set insert $start
	    _two$w delete $start $end

	    set Text($w,r_a$Text($w,redoSp)) "i"
	    set Text($w,r_b$Text($w,redoSp)) $start
	    set Text($w,r_c$Text($w,redoSp)) $text
	}
	r {
	    set start [lindex $Text($w,u_b$Text($w,undoSp)) 0]
	    set end   [lindex $Text($w,u_b$Text($w,undoSp)) 1]
	    set orig  [_two$w get $start $end]
	    _two$w delete $start $end
	    _two$w mark set insert $start
	    _two$w insert insert $Text($w,u_c$Text($w,undoSp))
	    set end [_two$w index insert]

	    set Text($w,r_a$Text($w,redoSp)) "r"
	    set Text($w,r_b$Text($w,redoSp)) [list $start $end]
	    set Text($w,r_c$Text($w,redoSp)) $orig
	}
    }

    # finish pop...
    unset Text($w,u_a$Text($w,undoSp))
    unset Text($w,u_b$Text($w,undoSp))
    unset Text($w,u_c$Text($w,undoSp))

    incr Text($w,redoSp)
    set Text($w,p_a) "u"
    return
}

# ::two::text::Redo --
#
#	This one is used to redo the last undo
#
# Arguments:
#	w		Path to text widget

proc ::two::text::Redo {w} {
    variable Text

    # should we really undo changes?
    if {[string match "disabled" [_two$w cget -state]]} {
        return
    }

    if {!$Text($w,redoSp)} {
        return
    }

    incr Text($w,redoSp) -1
    set start $Text($w,r_b$Text($w,redoSp))

    # Do Redo and push reverse command on the undo stack
    switch $Text($w,r_a$Text($w,redoSp)) {
	i {
	    set text $Text($w,r_c$Text($w,redoSp))
	    _two$w mark set insert $start
	    _two$w insert insert $text
	    set end [_two$w index insert]

	    set Text($w,u_a$Text($w,undoSp)) "d"
	    set Text($w,u_b$Text($w,undoSp)) $start
	    set Text($w,u_c$Text($w,undoSp)) $end
	}
	d {
	    set end $Text($w,r_c$Text($w,redoSp))
	    set text [_two$w get $start $end]
	    _two$w mark set insert $start
	    _two$w delete $start $end

	    set Text($w,u_a$Text($w,undoSp)) "i"
	    set Text($w,u_b$Text($w,undoSp)) $start
	    set Text($w,u_c$Text($w,undoSp)) $text
	}
	r {
	    set start [lindex $Text($w,r_b$Text($w,redoSp)) 0]
	    set end   [lindex $Text($w,r_b$Text($w,redoSp)) 1]
	    set orig  [_two$w get $start $end]
	    _two$w delete $start $end
	    _two$w mark set insert $start
	    _two$w insert insert $Text($w,r_c$Text($w,redoSp))
	    set end [_two$w index insert]

	    set Text($w,u_a$Text($w,undoSp)) "r"
	    set Text($w,u_b$Text($w,undoSp)) [list $start $end]
	    set Text($w,u_c$Text($w,undoSp)) $orig
	}
    }

    # finish pop...
    unset Text($w,r_a$Text($w,redoSp))
    unset Text($w,r_b$Text($w,redoSp))
    unset Text($w,r_c$Text($w,redoSp))

    incr Text($w,undoSp)
    set Text($w,p_a) "u"
    return
}

# ::two::text::UndoReset --
#
#	clears all the contents of the undo stack
#
# Arguments:
#	w		Path to text widget

proc ::two::text::UndoReset {w} {
    variable Text

    foreach entry [array names Text "$w,u_*"] {
        unset Text($entry)
    }
    set Text($w,undoSp) 0
    set Text($w,undo0Sp) 0
    set Text($w,nappend) 1
    return
}

# ::two::text::RedoReset --
#
#	clears all the contents of the redo stack
#
# Arguments:
#	w		Path to text widget

proc ::two::text::RedoReset {w} {
    variable Text

    foreach entry [array names Text "$w,r_*"] {
        unset Text($entry)
    }
    set Text($w,redoSp) 0
    event generate $w <<RedoOff>>
    return
}

# ::two::text::WordLeftStart --
#
#	Calculate position where to move the insertion cursor
#	if we want to go one word to the left. This is basically
#	the start of the word currently covering the insertion
#	cursor, or if there is none, the start of the next word
#	to the left.
#
# Arguments:
#	w		Path to text widget
#	idx		Current position
#
# Returns:
#	New position

proc ::two::text::WordLeftStart {w idx} {
    variable Text

    if {[_two$w compare $idx == "1.0"]} {
        return 1.0
    } elseif {
        [_two$w compare $idx == "$idx linestart"] ||
        [regexp $Text(word) [_two$w get $idx]]
    } {
        set idx "$idx -1c"
    }

    set cnt -1
    set newpos [_two$w search -count cnt -regexp -backwards -- \
        "(^|$Text(nonword))$Text(word)" $idx 1.0]

    if {$cnt == 1} {
        return $newpos
    } elseif {$cnt == 2} {
        return [_two$w index "$newpos +1c"]
    }
    return 1.0
}

# ::two::text::WordLeftEnd --
#
#	Calculate index of position immediatly after the end of the
#	next word to the left.
#
# Arguments:
#	w		Path to text widget
#	idx		Current position
#
# Returns:
#	New position

proc ::two::text::WordLeftEnd {w idx} {
    variable Text

    if {[_two$w compare $idx == "1.0"]} {
        return 1.0
    } elseif {
        [_two$w compare $idx == "$idx lineend"] ||
        [regexp $Text(word) [_two$w get "$idx -1c"]]
    } {
        set idx "$idx -1c"
    }

    set cnt -1
    set newpos [_two$w search -count cnt -regexp -backwards -- \
        "$Text(word)($Text(nonword)|\$)" $idx 1.0]

    if {$cnt > 0} {
        return [_two$w index "$newpos +1c"]
    }
    return 1.0
}

# ::two::text::WordRightStart --
#
#	Calculate position where to move the insertion cursor
#	if we want to go one word to the right. This is the start
#	of the next word to the right.
#
# Arguments:
#	w		Path to text widget
#	idx		Current position
#
# Returns:
#	New position

proc ::two::text::WordRightStart {w idx} {
    variable Text

    if {[_two$w compare $idx == "end -1c"]} {
        return [_two$w index "end -1c"]
    } elseif {[_two$w compare $idx == "$idx linestart"]} {
        set idx "$idx +1c"
    }

    set cnt -1
    set newpos [_two$w search -count cnt -regexp -- \
        "(^|$Text(nonword))$Text(word)" $idx "end -1c"]

    if {$cnt == 1} {
        return $newpos
    } elseif {$cnt == 2} {
        return [_two$w index "$newpos +1c"]
    }
    return [_two$w index "end -1c"]
}

# ::two::text::WordRightEnd --
#
#	Calculate index of position immediatly after the end of the
#	next word to the right.
#
# Arguments:
#	w		Path to text widget
#	idx		Current position
#
# Returns:
#	New position

proc ::two::text::WordRightEnd {w idx} {
    variable Text

    if {[_two$w compare $idx == "end -1c"]} {
        return [_two$w index "end -1c"]
    } elseif {[_two$w compare $idx == "$idx lineend"]} {
        set idx "$idx +1c"
    }

    set cnt -1
    set newpos [_two$w search -count cnt -regexp -- \
        "$Text(word)($Text(nonword)|\$)" $idx "end -1c"]

    if {$cnt > 0} {
        return [_two$w index "$newpos +1c"]
    }
    return [_two$w index "end -1c"]
}

# ::two::text::WordStart --
#
#	Calculate beginning of current word. This is used for selecting
#	whole words.
#
# Arguments:
#	w		Path to text widget
#	idx		Current position
#
# Returns:
#	New position

proc ::two::text::WordStart {w idx} {
    variable Text

    if {[_two$w compare $idx == 1.0]} {
        return 1.0
    } elseif {! [regexp $Text(word) [_two$w get "$idx -1c"]]} {
        return [_two$w index $idx]
    } elseif {[_two$w compare $idx == "$idx linestart"]} {
        return [_two$w index "$idx linestart"]
    }

    set cnt -1
    set newpos [_two$w search -count cnt -regexp -backwards -- \
        "(^|$Text(nonword))$Text(word)" $idx 1.0]

    if {$cnt == 1} {
        return $newpos
    } elseif {$cnt == 2} {
        return [_two$w index "$newpos +1c"]
    }
    return 1.0
}

# ::two::text::WordEnd --
#
#	Calculate end of current word. This is used for selecting
#	whole words.
#
# Arguments:
#	w		Path to text widget
#	idx		Current position
#
# Returns:
#	New position

proc ::two::text::WordEnd {w idx {next 0}} {
    variable Text

    if {[_two$w compare $idx == "end -1c"]} {
        return [_two$w index "end -1c"]
    } elseif {! [regexp $Text(word) [_two$w get $idx]]} {
        return [_two$w index $idx]
    } elseif {[_two$w compare $idx == "$idx lineend"]} {
        return [_two$w index "$idx lineend"]
    }

    set cnt -1
    set newpos [_two$w search -count cnt -regexp -- \
        "$Text(word)($Text(nonword)|\$)" "$idx -1c" "end -1c"]

    if {$cnt >= 0} {
        return [_two$w index "$newpos +1c"]
    }
    return [_two$w index "end -1c"]
}

# ::two::text::Indent --
#
#	Indent the current line (or lines if there is a multi-line selection)
#	or remove indentation, depending on $direction (value and sign).
#
# Arguments:
#	w		Path to text widget
#	pos		Index of position to indent
#	direction	Indent (>0) or remove indentation (<0)

proc ::two::text::Indent {w pos direction} {
    variable Text

    # Do we have a selection
    if {[llength [_two$w tag ranges sel]]} {
        # Store position of selection and cursor
        set first  [ _two$w index sel.first ]
        set last   [ _two$w index sel.last  ]
        set cursor [ _two$w index insert    ]
        # Retrieve line numbers
        scan [_two$w index "sel.first linestart"]-[_two$w index "sel.last-1c linestart"] "%d.0-%d.0" start end
        # Temporarily remove selection
        _two$w tag remove sel 1.0 end
        if { $direction > 0 } {
            # Indent all lines
            for {set line $start} {$line <= $end} {incr line} {
                TabToSpace $w $line.0
                Insert $w "$line.0" $Text($w,istring)
            }
        } else {
            # Remove a maximum of $count spaces from each line
            for {set line $start} {$line <= $end} {incr line} {
                TabToSpace $w $line.0
                set count 0
                _two$w search -count count -regexp \
                    "^ +" "$line.0" "$line.0 lineend"
                if {$count > $Text($w,-indentstep)} {
		    set count $Text($w,-indentstep)
		}
		if {$count > 0} {
                    Delete $w "$line.0" "$line.$count"
                }
            }
        }
	# restore selection and cursor position
	_two$w tag add sel $first $last
	_two$w mark set insert $cursor
    } else {
        TabToSpace $w $pos
        if {$direction > 0} {
            Insert $w "$pos linestart" $Text($w,istring)
        } else {
            set count 0
            _two$w search -count count -regexp \
                -- "^ +" "$pos linestart" "$pos lineend"
            if {$count > $Text($w,-indentstep)} {
	        set count $Text($w,-indentstep)
	    }
            if {$count > 0} {
                Delete $w "$pos linestart" "$pos linestart + $count chars"
            }
        }
    }
    return
}

# ::two::text::TabToSpace --
#
#	All tabs up to the first non-whitespace character are
#	replaced by spaces: used to make indentation increase/
#	decrease handle tabs in a proper way...
#
# Arguments:
#	w		Text Widget
#	pos		index if line which should be modified

proc ::two::text::TabToSpace {w pos} {
    # any tabs in here?
    set count 0
    _two$w search -count count -regexp \
        -- "^ *\t\[ \t\]*" "$pos linestart" "$pos lineend"
    if {$count > 0 } {
        # what string should we replace?
        set start [_two$w get "$pos linestart" "$pos linestart + $count c"]
        # loop through all tabs and replace them with the right
        # number of spaces, given a hardtab length of 8 chars.
        set tab 0
        while {[set first [string first \t $start]] >= 0} {
            incr tab
            set num   [expr {8 - ($first % 8)}]
            set rest  [string range $start [expr {$first + 1}] end]
            set start [string range $start 0 [expr {$first - 1}]]
            for {set idx 0} {$idx < $num} {incr idx} {
                append start " "
            }
            append start $rest
        }
        if {$tab > 0} {
            Replace $w "$pos linestart" "$pos linestart + $count c" $start
        }
    }
}


# ::two::text::ToggleOverwrite --
#
#	Toggle the overwriting state
#
# Arguments:
#	w		Widget

proc ::two::text::ToggleOverwrite {w} {
    variable Text

    if {$Text($w,-overwrite)} {
        set Text($w,-overwrite) 0
        event generate $w <<OverwriteOff>>
        _two$w tag remove insert 1.0 end
    } else {
        set Text($w,-overwrite) 1
        event generate $w <<OverwriteOn>>
        SendPos $w
    }
}

# ::two::text::SendPos --
#
#	Send a virtual event with x and y set to the current position of
#	the insertion cursor
#
# Arguments:
#	w		Widget

proc ::two::text::SendPos {w} {
    variable Text

    scan [_two$w index insert] "%d.%d" y x
    incr x
    event generate $w <<PosChanged>> -x $x -y $y
    if {$Text($w,-overwrite)} {
        ShowOverwrite $w
    }
    return
}

# ::two::text::ShowOverwrite
#
#	Show the overwrite cursor at the proper position
#
# Arguments:
#	w		Widget

proc ::two::text::ShowOverwrite {w} {
    variable Text

    _two$w tag remove insert 1.0 end

    if {$Text($w,-overwrite)} {
        # no overwrite cursor at end of line...
        if {[_two$w compare insert != "insert lineend"]} {
            # no overwrite cursor if within selection
            if {[llength [_two$w tag ranges sel]] == 0 ||
                [_two$w compare insert < sel.first] ||
                [_two$w compare insert > sel.last]
            } {
                _two$w tag add insert insert "insert +1c"
            }
        }
    }
    return
}

# ::two::text::SetOverwrite
#
#	Set the overwrite state according to current settings
#
# Arguments:
#	w		Widget

proc ::two::text::SetOverwrite {w} {
    variable Text

    if {$Text($w,-overwrite)} {
        event generate $w <<OverwriteOn>>
    } else {
        event generate $w <<OverwriteOff>>
    }
    return
}


# ::two::text::Selection --
#
#	Our own replacement selection handler. This is needed as we
#	simly set exportSelection to no in order to be able to have
#	a selection within the original widget and the clone
#	simultaneously. Without this workaround, we would only be
#	able to add the tag "sel" to one widget at a time.
#	CAVEAT: There seems to be no way to detect a "selection
#	lost" event - the text widget itself handles those events
#	with no interface at the TCL level (binding). There only
#	is the "selection own" command; thus we can only claim
#	the selection, but not remove the sel tag when we should.
#
# Arguments:
#	w		Widget

proc ::two::text::Selection {w offset length} {
    set sel ""
    foreach {start stop} [_two$w tag ranges sel] {
       append sel [_two$w get $start $stop]
    }
    return [string range $sel $offset [expr {$offset + $length}]]
}

namespace import ::two::text::*

# toolbar.tcl --
#
#	Provides a toolbar widget. Toolbars are usually placed below
#	an application's menue. They can hold several child widgets,
#	like all kinds of buttons, menus and comboboxes. They can
#	also be torn off.
#
# Copyright (c) 1998 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoToolbar 1.0
package require twoTooltip
package require twoUtil
package require twoCombobox

option add *TwoToolBar.Button.highlightThickness  0
option add *TwoToolBar.Button.takeFocus           0

# Workaround: Tk needs this procedure for cloning menus
# but won't load it form inside the namespace (bug in tcl8.0 up to p1)
catch {auto_load tkMenuDup}

namespace eval ::two::toolbar {
    namespace export two_toolbar
    variable Toolbar
    
    namespace import ::two::util::two_parseArgs
    
    # variable Toolbar
    #
    # nonce -	nonce to give any toplevel a new number
    # $w,nonce -	nonce to give any button within toolbar a new number
    # $w,tearoff -	If 1 then this toolbar can be torn off
    # $w,istorn -	1 if toolbar is currently torn of
    # $w,title -	Title string for torn-off toolbar
    # $w,#,state -	current state (on/off)
    # $w,#,value -	value for associated variable if button # is selected
    # $w,#,onvalue -	value for check'ed checkbutton
    # $w,#,offvalue -	value for uncheck'ed checkbutton
    # $w,#,variable -	variable associated with button
    # $w,#,command -	to be evalated on button press
    # $w,idx,"alias" -	index position of alias "alias"
    # $bt,state -	state of buttons
}

# Bitmap for tearoff handle
image create bitmap twoTearoff -foreground white -background black -data {
    \#define tearoff_main_width 8
    \#define tearoff_main_height 12
    static unsigned char tearoff_main_bits[] = {
      0x41, 0x08, 0x00, 0x41, 0x08, 0x00, 0x41, 0x08, 0x00, 0x41, 0x08, 0x00,
      };
} -maskdata {
    \#define tearoff_mask_width 8
    \#define tearoff_mask_height 12
    static unsigned char tearoff_mask_bits[] = {
      0x41, 0x8a, 0x10, 0x41, 0x8a, 0x10, 0x41, 0x8a, 0x10, 0x41, 0x8a, 0x10,
      };
}

# ::two::toolbar::two_toolbar --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.tb"); the original widget procedure
#	is renamed to "_two$w" (like "_two.tl.tb") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#	To handle the tearoff, toolbar creates a second widget as toplevel
#	(path: "$w.top"). All options apply to both widgets.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::toolbar::two_toolbar {w args} {
    variable Toolbar

    # Initialise the globals
    array set Toolbar "
	$w,tearoff	1
	$w,istorn	0
	$w,nonce	0
	$w,row		$w.top.row0
	$w,title	Toolbar
    "

    # See if this option is for us...
    two_parseArgs args Toolbar($w,tearoff) "-tearoff"
    two_parseArgs args Toolbar($w,title)   "-title"

    # Container frame to be able to do a simple "pack forget" on whole toolbar
    # and restore frame again at the same position
    frame $w -borderwidth 0 -highlightthickness 0
    # Toolbar frame - this one takes the arguments
    uplevel [list frame $w.tb -class TwoToolBar] $args
    pack $w.tb -fill both -expand true
    # torn-off toolbar, takes the arguments as well
    uplevel [list toplevel $w.top -class TwoToolBar] $args
    wm withdraw  $w.top
    wm title     $w.top $Toolbar($w,title)
    wm transient $w.top $w.tb
    wm resizable $w.top 0 0
    wm protocol  $w.top WM_DELETE_WINDOW [namespace code "Tearoff $w 0"]
    wm withdraw  $w.top ;# needed on windows
    # Add tearoff buttons to both widgets
    button $w.tb.0  -image twoTearoff -command [namespace code "Tearoff $w 1"]
    button $w.top.0 -image twoTearoff -command [namespace code "Tearoff $w 0"]
    if {$Toolbar($w,tearoff)} {
	pack $w.tb.0  -side left -fill y
	pack $w.top.0 -side left -fill y
    }
    pack [frame $Toolbar($w,row) -bd 0] -side top -anchor w

    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Toolbar
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # Return the correct result...
    return $w
}

# ::two::toolbar::Widget --
#
#	replacement for widget command
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::toolbar::Widget {w opt args} {
    variable Toolbar

    if {[two_parseArgs args l "-label"]} {
        lappend args -text $l
    }

    # See if this option is for us...
    switch -exact -- $opt {
	configure {
	    if {[two_parseArgs args Toolbar($w,title) "-title"]} {
	        wm title $w.top $Toolbar($w,title)
	    }
	    if {[two_parseArgs args Toolbar($w,tearoff) "-tearoff"]} {
                if {$Toolbar($w,tearoff)} {
                    pack $w.tb.0  -side left    -fill y
                    grid $w.top.0 -row 0 -col 0 -sticky ns
                } else {
                    pack forget $w.tb.0
                    grid forget $w.top.0
                }
	    }
	    if {[llength $args]} {
		return [eval [list _two$w $opt] $args]
	    }
	}
	cget {
	    switch -exact -- $args {
	        -tearoff { return $Toolbar($w,tearoff) }
	        -title   { return $Toolbar($w,title) }
		default  { return [eval [list _two$w $opt] $args] }
	    }
	}
	add {
	    if {[two_parseArgs args dummy "-off"]} {
		lappend args -state disabled
	    }
	    if {[two_parseArgs args dummy "-on"]} {
		lappend args -state normal
	    }
	    set type [lindex $args 0]
	    set args [lrange $args 1 end]
	    set tip  ""
	    set info ""
	    set doTip   [two_parseArgs args tip  "-tip"]
	    set doInfo  [two_parseArgs args info "-info"]
	    set index   [incr Toolbar($w,nonce)]
	    set buttons [list $w.tb.$index $w.top.$index]
	    if {[two_parseArgs args alias "-alias"]} {
		set Toolbar($w,idx,$alias) $index
	    }
	    switch -exact -- $type {
		separator -
		s {
	    	    foreach bt $buttons {
	            	uplevel [list frame $bt] $args -width 2 -relief sunken -bd 1
		    }
                    pack [lindex $buttons 0] -side left -fill y -padx 2 -pady 2
                    pack [lindex $buttons 1] -side left -fill y -padx 2 -pady 2 -in $Toolbar($w,row)
                }
                newline -
                n {
                    pack [uplevel [list frame [lindex $buttons 0]] $args -width 2 -relief sunken -bd 1] \
                        -side left -fill y -padx 2 -pady 2
                    pack [uplevel [list frame $w.top.sep$index] $args -height 2 -relief sunken -bd 1] \
                        -side top -fill x  -pady 2
                    set Toolbar($w,row) $w.top.row$index
                    pack [frame $Toolbar($w,row) -bd 0] -side top -anchor w
		}
		command -
		com -
		button -
		b {
	    	    foreach bt $buttons {
                        uplevel [list button $bt] $args -relief flat
                        bindtags $bt [list $bt TwoToolBarCommand Button [winfo toplevel $bt] all]
                    }
                    PackWidgets $w $buttons
		}
		combobox -
		combo {
	    	    foreach bt $buttons {
	            	uplevel [list two_combobox $bt] -takefocus 0 $args
		    }
                    PackWidgets $w $buttons
		}
		cascade -
		cas {
		    set slave ""
		    two_parseArgs args slave "-menu"
		    set state "normal"
		    if {[two_parseArgs args state "-state"]} {
			lappend args -state $state
		    }
	    	    foreach bt $buttons {
	            	uplevel [list menubutton $bt] $args [list -menu $bt.slave -relief flat]
		    	bindtags $bt [list $bt TwoToolBarMenubutton Menubutton [winfo toplevel $bt] all]
		        $slave clone $bt.slave
	        	set Toolbar($bt,state) $state
		    }
                    PackWidgets $w $buttons
		}
		checkbutton -
		check {
		    eval [list AddCheckbutton $w $index $buttons] $args
                    PackWidgets $w $buttons
		}
		radiobutton -
		radio {
		    eval [list AddRadiobutton $w $index $buttons] $args
                    PackWidgets $w $buttons
		}
		default {
		    incr Toolbar($w,nonce) -1
		    return -code error "unkown item type \"$type\""
		}
	    }
	    if {$doTip || $doInfo} {
	    	foreach bt $buttons {
	            two_tooltip $bt -tip $tip -info $info
		}
	    }
	    return $Toolbar($w,nonce)
	}
	entrybind {
	    set index    [lindex $args 0]
	    set sequence [lindex $args 1]
	    set script   [lindex $args 2]
	    if {[info exists Toolbar($w,idx,$index)]} {
	        set index $Toolbar($w,idx,$index)
	    }
	    foreach bt "$w.tb.$index $w.top.$index" {
	        bind $bt $sequence [namespace code $script]
	    }
	}
        entrycget {
	    # TODO
	    set index  [lindex $args 0]
	    set option [lrange $args 1 end]
	    if {[info exists Toolbar($w,idx,$index)]} {
	        set index $Toolbar($w,idx,$index)
	    }
	    if {$Toolbar($w,istorn)} {
		return [uplevel [list $w.top.$index cget] $option]
	    } else {
		return [uplevel [list $w.tb.$index cget] $option]
	    }
	}
	entrycommand -
	entrycmd {
	    set index  [lindex $args 0]
	    set option [lrange $args 1 end]
	    if {[info exists Toolbar($w,idx,$index)]} {
	        set index $Toolbar($w,idx,$index)
	    }
	    if $Toolbar($w,istorn) {
		return [uplevel [list $w.top.$index] $option]
	    } else {
		return [uplevel [list $w.tb.$index] $option]
	    }
	}
        entryconfigure -
	entrycfg {
	    # TODO
	    if {[two_parseArgs args dummy "-off"]} {
		lappend args -state disabled
	    }
	    if {[two_parseArgs args dummy "-on"]} {
		lappend args -state normal
	    }
	    set state ""
	    if {[two_parseArgs args state "-state"]} {
		lappend args -state $state
	    }
	    set index   [lindex $args 0]
	    set options [lrange $args 1 end]
	    set tip     ""
	    set info    ""
	    set doTip   [two_parseArgs options tip  "-tip"]
	    set doInfo  [two_parseArgs options info "-info"]
	    if {[info exists Toolbar($w,idx,$index)]} {
	        set index $Toolbar($w,idx,$index)
	    }
	    if {[two_parseArgs args alias "-alias"]} {
		set Toolbar($w,idx,$alias) $index
	    }
	    foreach bt "$w.tb.$index $w.top.$index" {
	        if {[string length $state]} {
	            set Toolbar($bt,state) $state
		}
		if {[llength $options]} {
		    uplevel [list $bt configure] $options
		}
		if {$doTip || $doInfo} {
	            two_tooltip $bt -tip $tip -info $info
		}
	    }
	}
	tearoff {
	    if {[llength $args] != 1} {
	        return -code error "wrong # args: should be \"two_toolbar tearoff boolean\""
	    }
            Tearoff $w [lindex $args 0]
	}
	default {
            return [eval [list _two$w $opt] $args]
	}
    }
    return
}

# ::two::toolbar::_Destroy --
#
#	Call this procedure when the toolbar widget is destroyed.
#	Will clean widget-specific Elements of Toolbar and destroy
#	the original widget command.
#
# Arguments:
#	w		Path to Widget

proc ::two::toolbar::_Destroy {w} {
    variable Toolbar

    # destroy accompaining toplevel
    destroy $w.top
    # clear Toolbar
    set keys ""
    lappend keys [array names Toolbar "$w,*"]
    lappend keys [array names Toolbar "$w.tb.*,*"]
    lappend keys [array names Toolbar "$w.top.*,*"]
    foreach key $keys {
        unset Toolbar($key)
    }
    # remove the original widget command
    rename $w ""
    # remove trace procedures
    foreach p [info procs traceProc,$w,*] {
        rename $p ""
    }
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoToolBar-class bindings
#-------------------------------------------------------------------------

# ---
# command button widgets
# ---
# raise button on mouse over

bind TwoToolBarCommand <Enter> {
    if {![string match "disabled" [%W cget -state]]} {
        %W configure -relief raised
    }
}

bind TwoToolBarCommand <Leave> {
    %W configure -relief flat
}

# ---
# menu button widgets
# ---
# disable menubutton if not shown to enable proper
# handling of hotkeys (and F10)

bind TwoToolBarMenubutton <Map> {
    if {![string match "disabled" $::two::toolbar::Toolbar(%W,state)]} {
	%W configure -state normal
    }
}

bind TwoToolBarMenubutton <Unmap> {
    %W configure -state disabled
}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::toolbar::Tearoff --
#
#	Tear off toolbar ("doit" == 1) or restore it ("doit" == 0)
#
# Arguments:
#	w		Path to Widget
#	doit		Tearoff (1) or restore (0)

proc ::two::toolbar::Tearoff {w doit} {
    variable Toolbar

    if {$doit} {
        set x [winfo rootx $w.tb]
        set y [winfo rooty $w.tb]
        wm geometry  $w.top +$x+$y
        update
        wm deiconify $w.top
	pack forget  $w.tb
	$w configure -height 1
	set Toolbar($w,istorn) 1
    } else {
	pack $w.tb -fill both -expand true
        wm withdraw $w.top
	set Toolbar($w,istorn) 0
    }
    # all elements are configured identically in both states
    # with combobox being the exception: it's too interactive:-)
    for {set index 0} {$index <= $Toolbar($w,nonce)} {incr index} {
        if {$Toolbar($w,istorn)} {
	    set b1 $w.tb.$index
    	    set b2 $w.top.$index
	} else {
	    set b1 $w.top.$index
    	    set b2 $w.tb.$index
	}
	set class ""
	catch {set class [$b1 cget -class]}
	if {[string match "TwoComboBox" $class]} {
	    eval [list $b2 setlist] [$b1 getlist]
	    eval [list $b2 set]     [$b1 get]
	}
    }
    return
}

# ::two::toolbar::PackWidgets --
#
#	Adds a widget to the toolbar, i.e. does the "pack" and "grid"
#	operations to add both versions of a widget to the toplevel or
#	the embedded version of the toolbar.
#
# Arguments:
#	w		Path to toolbar widget
#	buttons		widget names of the 2 widgets that should be placed
#			placed within the toolbar. First one is for the
#			embedded version, the second for the toplevel

proc ::two::toolbar::PackWidgets {w buttons} {
    variable Toolbar

    pack [lindex $buttons 0] -side left
    pack [lindex $buttons 1] -side left -in $Toolbar($w,row)
}

# ::two::toolbar::AddCheckbutton --
#
#	adds a checkbutton to the toolbar
#
# Arguments:
#	w		Path to toolbar widget
#	index		Index # of button
#	buttons		widget names of the 2 buttons that should be created
#	args		args of original widget command

proc ::two::toolbar::AddCheckbutton {w index buttons args} {
    variable Toolbar

    set Toolbar($w,$index,state)    0
    set Toolbar($w,$index,onvalue)  1
    set Toolbar($w,$index,offvalue) 0
    set Toolbar($w,$index,variable) ""
    set Toolbar($w,$index,command)  ""
    two_parseArgs args Toolbar($w,$index,onvalue)  "-onvalue"
    two_parseArgs args Toolbar($w,$index,offvalue) "-offvalue"
    two_parseArgs args Toolbar($w,$index,variable) "-variable"
    two_parseArgs args Toolbar($w,$index,command)  "-command"
    foreach bt $buttons {
	uplevel [list button $bt] $args [list -relief flat -command \
	    [list ::two::toolbar::CheckbuttonCommand $w $index]]
    }
    if {[string length $Toolbar($w,$index,variable)]} {
	upvar \#0 "$Toolbar($w,$index,variable)" var
	trace variable var w [namespace code traceProc,$w,$index]
	proc traceProc,$w,$index {name1 name2 op} "
	    variable Toolbar
            upvar \$name1 var
	    if {\[string length \$name2\]} {
		if {\[string match \$var(\$name2) \$Toolbar($w,$index,onvalue)\]} {
		    set Toolbar($w,$index,state) 1
		} elseif {\$var(\$name2) == \$Toolbar($w,$index,offvalue)} {
		    set Toolbar($w,$index,state) 0
		}
	    } else {
		if {\[string match \$var \$Toolbar($w,$index,onvalue)\]} {
		    set Toolbar($w,$index,state) 1
		} elseif {\[string match \$var \$Toolbar($w,$index,offvalue)\]} {
		    set Toolbar($w,$index,state) 0
		}
	    }
	    CheckbuttonState $w $index
	"
    }
    return
}

# ::two::toolbar::CheckbuttonCommand --
#
#	command called when checkbutton is pressed
#
# Arguments:
#	w		Path to toolbar widget
#	index		Index # of button

proc ::two::toolbar::CheckbuttonCommand {w index} {
    variable Toolbar
    set Toolbar($w,$index,state) [expr 1 - $Toolbar($w,$index,state)]
    if {[string length $Toolbar($w,$index,variable)]} {
        set var $Toolbar($w,$index,variable)
	uplevel \#0 "
	    if { $Toolbar($w,$index,state) } {
		if { ![info exists $var] || ![string match $$var $Toolbar($w,$index,onvalue)]} {
	            set $var $Toolbar($w,$index,onvalue)
		}
	    } else {
		if { ![info exists $var] || ![string match $$var $Toolbar($w,$index,offvalue)]} {
	            set $var $Toolbar($w,$index,offvalue)
		}
	    }
	"
    } else {
        # only call directly if there is no trace on var
        CheckbuttonState $w $index
    }
    if {[string length $Toolbar($w,$index,command)]} {
        uplevel \#0 $Toolbar($w,$index,command)
    }
    return
}

# ::two::toolbar::CheckbuttonState --
#
#	update state of checkbutton
#
# Arguments:
#	w		Path to toolbar widget
#	index		Index # of button

proc ::two::toolbar::CheckbuttonState {w index} {
    variable Toolbar
    foreach bt "$w.tb.$index $w.top.$index" {
	if {$Toolbar($w,$index,state)} {
            $bt configure -relief sunken
	} else {
            $bt configure -relief flat
	}
    }
    return
}

# ::two::toolbar::AddRadiobutton --
#
#	adds a radio button to the toolbar
#
# Arguments:
#	w		Path to toolbar widget
#	index		Index # of button
#	buttons		widget names of the 2 buttons that should be created
#	args		args of original widget command

proc ::two::toolbar::AddRadiobutton {w index buttons args} {
    variable Toolbar

    set Toolbar($w,$index,state)    0
    set Toolbar($w,$index,value)    $index
    set Toolbar($w,$index,variable) ""
    set Toolbar($w,$index,command)  ""
    two_parseArgs args Toolbar($w,$index,value)    "-value"
    two_parseArgs args Toolbar($w,$index,variable) "-variable"
    two_parseArgs args Toolbar($w,$index,command)  "-command"
    foreach bt $buttons {
	uplevel [list button $bt] $args [list -relief flat -command \
	    [list ::two::toolbar::RadiobuttonCommand $w $index]]
    }
    if {[string length $Toolbar($w,$index,variable)]} {
	set varname $Toolbar($w,$index,variable)
	upvar \#0 "$varname" var
	lappend Toolbar($w,radio,$varname) $index
	trace variable var w [namespace code traceProc,$w,$index]
	proc traceProc,$w,$index {name1 name2 op} "
	    variable Toolbar
            upvar \$name1 var
	    foreach i \$Toolbar($w,radio,$varname) {
		if {\[string length \$name2\]} {
		    if {\[string match \$var(\$name2) \$Toolbar($w,\$i,value)\]} {
			set Toolbar($w,\$i,state) 1
		    } else {
			set Toolbar($w,\$i,state) 0
		    }
		} else {
		    if {\[string match \$var \$Toolbar($w,\$i,value)\]} {
			set Toolbar($w,\$i,state) 1
		    } else {
			set Toolbar($w,\$i,state) 0
		    }
		}
		RadiobuttonState $w \$i
	    }
	"
    }
    return
}

# ::two::toolbar::RadiobuttonCommand --
#
#	command called when checkbutton is pressed. This will update
#	the global variable (if set) and set the state
#
# Arguments:
#	w		Path to toolbar widget
#	index		Index # of button

proc ::two::toolbar::RadiobuttonCommand {w index} {
    variable Toolbar
    set Toolbar($w,$index,state) [expr 1 - $Toolbar($w,$index,state)]
    if {[string length $Toolbar($w,$index,variable)]} {
        set var $Toolbar($w,$index,variable)
	uplevel \#0 "
	    if { ![info exists $var] || ![string match $$var $Toolbar($w,$index,value)]} {
	        set $var $Toolbar($w,$index,value)
	    }
	"
    } else {
        # only call directly if there is no trace on var
        RadiobuttonState $w $index
    }
    if {[string length $Toolbar($w,$index,command)]} {
        uplevel \#0 $Toolbar($w,$index,command)
    }
    return
}

# ::two::toolbar::RadiobuttonState --
#
#	update state of radiobutton according to contents of
#	the Toolbar variable
#
# Arguments:
#	w		Path to toolbar widget
#	index		Index # of button

proc ::two::toolbar::RadiobuttonState {w index} {
    variable Toolbar
    foreach bt "$w.tb.$index $w.top.$index" {
	if {$Toolbar($w,$index,state)} {
            $bt configure -relief sunken
	} else {
            $bt configure -relief flat
	}
    }
    return
}

namespace import ::two::toolbar::*

# editor.tcl --
#
#	Editor widget. Consists of 2 two_text widgets, managed by a
#	two_splitter to make it "splittable"
#
# Copyright (c) 1998, 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoEditor 1.0
package require twoUtil
package require twoSplitter
package require twoText
package require twoAutosb

# Create fonts (catch as they might already exist)
catch {font create twoEditorS -family courier -size 10}
catch {font create twoEditorC -family courier -size 10 -slant italic}
catch {font create twoEditorK -family courier -size 10 -weight bold}

namespace eval ::two::editor {
    namespace export two_editor
    variable Editor
    
    namespace import ::two::util::two_parseArgs
    
    # variable Editor
    #
    # nonce -		nonce to give any new file a new number
    # matchId -		id of after event handler for removing match highlighting
    # hl,active -	true if HighlightLoop is active
    # hl,widgets -	List of text widgets which currently are being highlighted
    # hl,startLen -	Length of match for highlight start delimiter
    # hl,stopLen -	Length of match for highlight stop delimiter
    # $w,file -		name of current file
    # $w,mtime -	file mtime of file on disk
    # $w,split -	true if editor is split
    # $w,new -		true if editor holds a new (i.e. unnamed) file
    # $w,children -	path names of all children of this editor (frames...)
    # $w,scrollx -	true if a horizontal scrollbar is used
    # $w,closecmd -	command that is to be called on close
    # $w,saveerrcmd -	command that is to be called if saving fails
    # $w,openerrcmd -	command that is to be called if open fails
    # $w,active -	active part (has focus or had it recently)
    # $w,style -    	look of the splitter used by the editor {windows|unix}
    # $w,match -	show matching parantheses if value is true
    # $w,high -		do keyword highlighting
    # $w,comment -	list of pairs of comment delimiters
    # $w,string -	list of string delimiters
    # $w,keywords -	list of keywords to highlight
    # $w,keywordRE -	regular expression describing "keywords"
    # $w,startRE -	Regular expression to find start of comments and strings
    # $w,stop,$s -	End delimiter of comment or string according to start $s
    # $w,tag,$s -	What tag to add according to start delimiter $s
    # $w,id -		Increased whenever highlighting is started
    # $w,isnew -	True if the widget has been created but never been mapped,
    #			so generated events would not be delivered.
    
    array set Editor "
        nonce	0
        bg		{}
        hl,active	0
        hl,widgets	{}
        style		$tcl_platform(platform)
        matchId		0
    "
}

# ::two::editor::two_editor --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.ed"); the original widget procedure
#	is renamed to "_two$w" (like "_two.tl.ed") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#	editor will create two two_text widgets and a pane.
#	Arguments are passed to both two_text widgets.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::editor::two_editor {w args} {
    variable Editor

    # Initialise the globals
    array set Editor "
        $w,file		{}
	$w,buffer	{}
	$w,mtime	0
        $w,split	0
        $w,new		1
	$w,scrollx	1
        $w,closecmd	{}
        $w,saveerrcmd	{}
        $w,openerrcmd	{}
	$w,active	$w.1.txt
	$w,children	$w.1
	$w,style        $Editor(style)
	$w,match	1
	$w,highlight	0
	$w,comment	{}
	$w,string	{}
	$w,keywords	{}
	$w,keywordRE	{}
	$w,isnew	1
    "

    # See if this option is for us...
    two_parseArgs args Editor($w,split)      "-split"
    two_parseArgs args Editor($w,scrollx)    "-scrollx"
    two_parseArgs args Editor($w,closecmd)   "-closecmd"
    two_parseArgs args Editor($w,saveerrcmd) "-saveerrcmd"
    two_parseArgs args Editor($w,openerrcmd) "-openerrcmd"
    two_parseArgs args Editor($w,file)       "-file"
    two_parseArgs args Editor($w,buffer)     "-buffer"
    two_parseArgs args Editor($w,style)      "-style"
    two_parseArgs args Editor($w,match)      "-match"
    two_parseArgs args Editor($w,highlight)  "-hightlight"
    two_parseArgs args Editor($w,comment)    "-comment"
    two_parseArgs args Editor($w,string)     "-string"
    two_parseArgs args Editor($w,keywords)   "-keywords"
    two_parseArgs args Editor($w,keywordRE)  "-keywordRE"

    # Container frame
    # Explicitly takes the focus to be able to
    # directly go to the proper child
    frame $w -class TwoEditor -bd 0 -width 500 -height 300 -takefocus 0
    
    bind $w <Map> "
        set ::two::editor::Editor([list $w],isnew) 0
        bind [list $w] <Map> {}
    "
    
    # First editor
    frame $w.1
    set e $w.1.txt
    grid [uplevel [list two_text $e -xs "$w.1.sx set" -ys "$w.1.sy set" -wrap none] $args] \
	 -row 0 -column 0 -sticky nsew
    grid [two_autosb $w.1.sy -command "$e yview"] \
	 -row 0 -column 1 -sticky ns
    grid [two_autosb $w.1.sx -command "$e xview" -orient horizontal] \
         -row 1 -column 0 -sticky ew
    grid rowconfigure    $w.1 0 -weight 1
    grid columnconfigure $w.1 0 -weight 1

    # really show scrollbar?
    if {!$Editor($w,scrollx)} {
	grid forget $w.1.sx
	$w.1.txt configure -wrap char
    }

    # bindings for matching of paranthesis
    if {$Editor($w,match)} {
	bindtags $e [concat [bindtags $e] TwoShowMatch]
    }
    $e tag configure match -background red -foreground white

    # Used for syntax highlighting
    $e tag configure string  -font twoEditorS -foreground \#00a
    $e tag configure comment -font twoEditorC -foreground \#080
    $e tag configure keyword -font twoEditorK -foreground \#800
    HighlightInit $w

    # Proxy events
    # Only do this with the first two_text widget to
    # avoid bindings being triggered multiple times
    bind $e <<ModOn>>      [list event generate $w <<ModOn>>]
    bind $e <<ModOff>>     [list event generate $w <<ModOff>>]
    bind $e <<SelOn>>      [list event generate $w <<SelOn>>]
    bind $e <<SelOff>>     [list event generate $w <<SelOff>>]
    bind $e <<UndoOn>>     [list event generate $w <<UndoOn>>]
    bind $e <<UndoOff>>    [list event generate $w <<UndoOff>>]
    bind $e <<RedoOn>>     [list event generate $w <<RedoOn>>]
    bind $e <<RedoOff>>    [list event generate $w <<RedoOff>>]
    bind $e <<PosChanged>> [list event generate $w <<PosChanged>> -x %x -y %y]

    # set active on focus in
    bind $w.1 <FocusIn> [namespace code "set Editor($w,active) $w.1.txt"]

    # pane
    two_splitter $w.1 -orient vertical -style $Editor($w,style)
    
    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Editor
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # initial state
    if {$Editor($w,split)} {
        Split $w
    }
    if {[string length $Editor($w,buffer)]} {
	New  $w $Editor($w,buffer)
    } elseif {[string length $Editor($w,file)]} {
        Open $w $Editor($w,file)
    } else {
        New $w
    }
    # Return the correct result...
    return $w
}

# ::two::editor::Widget --
#
#	replacement for widget command
#
# Arguments:
#	w	Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::editor::Widget {w opt args} {
    variable Editor
    # See if this option is for us...
    switch -- $opt {
	configure {
	    set oldsplit $Editor($w,split)
	    two_parseArgs args Editor($w,closecmd)   "-closecmd"
	    two_parseArgs args Editor($w,saveerrcmd) "-saveerrcmd"
	    two_parseArgs args Editor($w,openerrcmd) "-openerrcmd"
	    two_parseArgs args Editor($w,split)      "-split"
	    two_parseArgs args Editor($w,comment)    "-comment"
	    two_parseArgs args Editor($w,string)     "-string"
	    two_parseArgs args Editor($w,keywords)   "-keywords"
	    two_parseArgs args Editor($w,keywordRE)  "-keywordRE"
	    HighlightInit $w
	    if {[two_parseArgs args Editor($w,highlight) "-highlight"]} {
	        if {!$Editor($w,highlight)} {
		    $w.1.txt tag remove string  1.0 end
		    $w.1.txt tag remove comment 1.0 end
		    $w.1.txt tag remove keyword 1.0 end
		} else {
		    HighlightStart $w
		}
	    }
	    if {[two_parseArgs args Editor($w,match) "-match"]} {
	        foreach ed $Editor($w,children) {
		    set b [bindtags $ed.txt]
		    if {[string match TwoShowMatch [lindex $b end]]} {
		        set b [lreplace $b end end]
		    }
        	    if {$Editor($w,match)} {
			set b [concat $b TwoShowMatch]
        	    }
		    bindtags $ed.txt $b
		}
	    }
	    if {[two_parseArgs args Editor($w,scrollx) "-scrollx"]} {
	        foreach ed $Editor($w,children) {
        	    if {$Editor($w,scrollx)} {
			grid $ed.sx -row 1 -sticky nsew
			$ed.txt configure -wrap none
        	    } else {
			grid forget $ed.sx
			$ed.txt configure -wrap char
		    }
		}
	    }
	    if {[two_parseArgs args Editor($w,file) "-file"]} {
	        Open $w $Editor($w,file)
	    }
	    if {[two_parseArgs args Editor($w,buffer) "-buffer"]} {
	        New $w $Editor($w,buffer)
	    }
	    if {[two_parseArgs args Editor($w,style) "-style"]} {
	        two_splitter configure $w.1 -style $Editor($w,style)
	    }
	    if {$Editor($w,split) && !$oldsplit} {
	        Split $w
	    } elseif {$oldsplit && !$Editor($w,split)} {
	        Join $w
	    }
	    if {[llength $args]} {
		return [eval [list $w.1.txt $opt] $args]
	    }
	}
	cget {
	    switch -exact -- $args {
	        -closecmd   { return $Editor($w,closecmd)   }
	        -saveerrcmd { return $Editor($w,saveerrcmd) }
	        -openerrcmd { return $Editor($w,openerrcmd) }
	        -split      { return $Editor($w,split)      }
	        -scrollx    { return $Editor($w,scrollx)    }
	        -file       { return $Editor($w,file)       }
	        -buffer     { return $Editor($w,buffer)     }
	        -style      { return $Editor($w,style)      }
	        -match      { return $Editor($w,match)      }
	        -highlight  { return $Editor($w,highlight)  }
	        -comment    { return $Editor($w,comment)    }
	        -string     { return $Editor($w,string)     }
	        -keywords   { return $Editor($w,keywords)   }
	        -keywordRE  { return $Editor($w,keywordRE)  }
		-class      { return TwoEditor              }
		-takefocus  { return 0 }
		default     { return [uplevel $w.1.txt $opt $args] }
	    }
	}
	open {
	    Open $w [lindex $args 0]
	}
	close {
	    return [Close $w]
	}
	save {
	    Save $w [lindex $args 0]
	}
	fname -
	filename {
	    if {[string length $Editor($w,file)]} {
		return $Editor($w,file)
	    } else {
		return $Editor($w,buffer)
	    }
	}
	isNew {
            return $Editor($w,new)
	}
	isFile {
            if {[string length $Editor($w,file)]} {
                return 1
            } else {
                return 0
            }
	}
	isBuffer {
            if {[string length $Editor($w,buffer)]} {
                return 1
            } else {
                return 0
            }
	}
	setFocus {
	    if {[winfo exists $Editor($w,active)]} {
                focus $Editor($w,active)
	    } else {
                focus $w.1.txt
	    }
	}
	mtime {
	    return $Editor($w,mtime)
	}
	mtimeReset {
	    set file $Editor($w,file)
	    if {[string length $file]} {
                # We could use [file exists] and subsequently call
                # [file mtime], but this has a race condition (file could
                # have been removed in the meantime), so it's better to
                # use catch and just try to get the mtime.
		catch {set Editor($w,mtime) [file mtime $file]}
	    }
	}
	splitState {
	    return $Editor($w,split)
	}
	repaint {
	    if {$Editor($w,highlight)} {
	        HighlightStart $w
	    }
	}
	default {
	    # clones are almost equal, except for cursor
	    # position and scrolling. So we can't use
	    # $w.1.txt :-(
            return [uplevel [list $Editor($w,active) $opt] $args]
	}
    }
    return
}

# ::two::editor::_Destroy --
#
#	Call this procedure when the editor widget is destroyed.
#	Will clean widget-specific Elements of Editor and destroy
#	the original widget command.
#
# Arguments:
#	w		Path to Widget

proc ::two::editor::_Destroy {w} {
    variable Editor

    # clear Editor
    foreach key [array names Editor "$w,*"] {
        unset Editor($key)
    }
    # remove the original widget command
    rename $w ""
}

#-------------------------------------------------------------------------
# Bindings
#-------------------------------------------------------------------------

# bindings for highlighting of matching paranthesis

# removed - doesn't seem to be useful
#bind TwoShowMatch <quotedbl>     {::two::editor::MatchBackward %W \" }
#bind TwoShowMatch <grave>        {::two::editor::MatchBackward %W \` }
#bind TwoShowMatch <apostrophe>   {::two::editor::MatchBackward %W \' }

bind TwoShowMatch <parenleft>      {::two::editor::MatchNestForw %W \( \) }
bind TwoShowMatch <parenright>     {::two::editor::MatchNestBack %W \) \( }

bind TwoShowMatch <braceleft>      {::two::editor::MatchNestForw %W \{ \} }
bind TwoShowMatch <braceright>     {::two::editor::MatchNestBack %W \} \{ }

bind TwoShowMatch <bracketleft>    {::two::editor::MatchNestForw %W \[ \] }
bind TwoShowMatch <bracketright>   {::two::editor::MatchNestBack %W \] \[ }

bind TwoShowMatch <Control-period> {::two::editor::ShowMatch %W}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::editor::New --
#
#	Create a new editor with no file; we call it a "buffer".
#
# Arguments:
#	w		Path to editor widget
#	name		Name of the new buffer

proc ::two::editor::New {w {name ""}} {
    variable Editor

    if {[string length $name]} {
        set Editor($w,new) 0
    } else {
        set name "New $Editor(nonce)"
        set Editor($w,new) 1
    }
    set  Editor($w,buffer) $name
    set  Editor($w,mtime)  0
    incr Editor(nonce)

    $w.1.txt afterLoad
}

# ::two::editor::Open --
#
#	Load file into buffer. If file == "" then a new filename is created.
#
# Argumets:
#	w		Path to editor widget
#	file		Name of file to open or empty for new file

proc ::two::editor::Open {w {file ""}} {
    variable Editor

    # first close old file -> give user a chance to abort
    if {[Close $w]} {
	if {[string length $file]} {
            set Editor($w,file) $file
            set Editor($w,new)  0
	    $w.1.txt delete 1.0 end

	    # the file commands might fail due to race conditions: we should
	    # catch such errors, but only if the widget user has provided an
	    # error handler. The widget is set to the new filename nevertheless
	    # to have a predictable state.
	    # {insert end} should never fail, so we can catch the whole block:-)
            if {[string length $Editor($w,openerrcmd)]} {
	        if {[catch {
                    set Editor($w,mtime) [file mtime $file]
	            $w.1.txt insert end [two_readFile $file]
                } errmsg]} {
                    uplevel "\#0" [list $Editor($w,openerrcmd) $file $errmsg]
                }
            } else {
                set Editor($w,mtime) [file mtime $file]
                $w.1.txt insert end [two_readFile $file]
            }

	    $w.1.txt mark set insert 1.0
            $w.1.txt afterLoad
	    $w.1.txt see 1.0

	    if {$Editor($w,highlight)} {
        	HighlightStart $w
	    }
	} else {
	    New $w
	}
    }
    return
}

# ::two::editor::Save --
#
#	Save buffer into file. If file == "" then the current name is used.
#
# Argumets:
#	w		Path to editor widget
#	file		Name of file to save as or empty for current name

proc ::two::editor::Save {w {file ""}} {
    variable Editor
    global errorCode

    if {[string length $file]} {
        set Editor($w,file) $file
        set Editor($w,new)  0
    } elseif {![string length $Editor($w,file)]} {
        return -code error "No filename set!"
    } else {
        set file $Editor($w,file)
    }

    # try twice if first attempt fails - work around temporary problems
    if {[catch {set fd [open $file "w"]} errmsg]} {
        after 200
        set err [catch {set fd [open $file "w"]} errmsg]
        if {$err} {
            # second try failed as well
            if {[string length $Editor($w,saveerrcmd)]} {
                uplevel "\#0" [list $Editor($w,saveerrcmd) $file $errmsg]
                return
            } else {
                return -code $err -errorcode $errorCode $errmsg
            }
        }
    }

    puts -nonewline $fd [$w.1.txt get 1.0 "end - 1 chars"]
    close $fd

    set Editor($w,mtime) [file mtime $file]
    $w.1.txt afterSave
    
    # Keyword Highlighting
    if {$Editor($w,highlight)} {
        HighlightStart $w
    }
    return
}

# ::two::editor::Close --
#
#	Delete contents of current buffer.
#	Returns 1 if file was really closes
#
# Argumets:
#	w		Path to editor widget

proc ::two::editor::Close {w} {
    variable Editor

    # run closecmd if there is one
    if {[string length $Editor($w,closecmd)]} {
        set retval [uplevel "\#0" [list $Editor($w,closecmd) $w]]
        if {$retval} {
	    # workaround: if {!"true"} breaks, but if {"true"} works
	} else {
	    return 0
	}
    }
    set Editor($w,file) ""
    $w.1.txt delete 1.0 end
    $w.1.txt afterLoad
    return 1
}

# ::two::editor::Split --
#
#	show second two_text widget and configure pane
#
# Argumets:
#	w		Path to widget

proc ::two::editor::Split {w} {
    variable Editor

    lappend Editor($w,children) $w.2

    frame $w.2
    set e $w.2.txt
    $w.1.txt clone $e
    $e configure -xs "$w.2.sx set" -ys "$w.2.sy set" -wrap none
    grid $e -row 0 -column 0 -sticky nsew
    grid [two_autosb $w.2.sy -command "$e yview"] \
	 -row 0 -column 1 -sticky ns
    grid [two_autosb $w.2.sx -command "$e xview" -orient horizontal] \
         -row 1 -column 0 -sticky ew
    grid rowconfigure    $w.2 0 -weight 1
    grid columnconfigure $w.2 0 -weight 1

    # bindings for matching of paranthesis
    if {$Editor($w,match)} {
	bindtags $e [concat [bindtags $e] TwoShowMatch]
    }

    # really show scrollbar?
    if {!$Editor($w,scrollx)} {
	grid forget $w.2.sx
	$w.2.txt configure -wrap char
    }

    # set active on focus in
    bind $w.2 <FocusIn> [namespace code "set Editor($w,active) $w.2.txt"]

    # add to pane and configure handle
    two_splitter $w.2

    # Signal
    event generate $w <<SplitOn>>
    return
}

# ::two::editor::Join --
#
#	show second two_text widget and configure pane
#
# Argumets:
#	w		Path to widget

proc ::two::editor::Join {w} {
    variable Editor

    set Editor($w,children) $w.1

    # has the second child the focus?
    if {[string match $w.2.* [focus]]} {
        # set focus to first child
        focus $w.1.txt
    }
    # is the second child marked as having had the focus recently?
    if {[string match $w.2.* $Editor($w,active)]} {
        set Editor($w,active) $w.1.txt
    }

    # remove from pane
    two_splitter forget $w.2
    # destroy second editor
    destroy $w.2
    # Signal
    event generate $w <<SplitOff>>
    return
}

# ::two::editor::MatchBackward --
#
#	Search for previous occurence of character "char" in
#	widget "w" and highlight both the character under
#	the insertion cursor and the matchingcharacter.
#	Useful for unnested matches.
#
# Status:
#	Deprecated
# 
# Sideeffects:
#	Highlights the matching characters for one second.
#
# Arguments:
#	w -		Path to text widget
#	char -		Character which should match

proc ::two::editor::MatchBackward {w char} {
    variable Editor

    after cancel $Editor(matchId)
    $w tag remove match 1.0 end

    set found [$w search -backwards $char "insert -1c" 1.0]
    if {[string length $found]} {
	$w tag add match $found "$found +1c" "insert -1c" insert
	set Editor(matchId) [after 1000 "$w tag remove match 1.0 end"]
    }
    return
}

# ::two::editor::MatchNestForw --
#
#	Search for next matching parenthesis, bracket or
#	brace. Matching here also respects nesting of
#	opening and closing parenthesis etc. Useful for
#	binding to opening parenthesis.
#
# Sideeffects:
#	Highlights the matching characters for one second.
#
# Arguments:
#	w -		Path to text widget
#	inc -		Character which increases the level of nesting
#	dec -		Character which decreases the level of nesting

proc ::two::editor::MatchNestForw {w inc dec} {
    variable Editor

    after cancel $Editor(matchId)
    $w tag remove match 1.0 end

    set idx insert
    set end [$w index "insert + 100 lines lineend"]
    set level 1
    set nextInc [$w search $inc $idx $end]
    set nextDec [$w search $dec $idx $end]
    while {1} {
        if {![string length $nextDec]} {
	    return
        }
	if {[string length $nextInc] &&
	    [$w compare "$nextInc +1c" < "$nextDec +1c"]
	} {
	    incr level
	    set nextInc [$w search $inc "$nextInc +1c" $end]
	} else {
	    incr level -1
	    if {!$level} break
	    set nextDec [$w search $dec "$nextDec +1c" $end]
	}
    }
    $w tag add match $nextDec "$nextDec +1c" "insert -1c" insert
    set Editor(matchId) [after 1000 "$w tag remove match 1.0 end"]
}

# ::two::editor::MatchNestBack --
#
#	Search for previous matching parenthesis, bracket or
#	brace. Matching here also respects nesting of
#	opening and closing parenthesis etc. Useful for
#	binding to closing parenthesis etc.
#
# Sideeffects:
#	Highlights the matching characters for one second.
#
# Arguments:
#	w -		Path to text widget
#	inc -		Character which increases the level of nesting
#	dec -		Character which decreases the level of nesting

proc ::two::editor::MatchNestBack {w inc dec} {
    variable Editor

    after cancel $Editor(matchId)
    $w tag remove match 1.0 end

    set idx insert
    set end [$w index "insert - 100 lines linestart"]
    set level 1
    set nextInc [$w search -backwards $inc "$idx -1c" $end]
    set nextDec [$w search -backwards $dec "$idx -1c" $end]
    while {1} {
        if {![string length $nextDec]} {
	    return
        }
	if {[string length $nextInc] &&
	    [$w compare "$nextInc" > "$nextDec"]
	} {
	    incr level
	    set nextInc [$w search -backwards $inc "$nextInc" $end]
	} else {
	    incr level -1
	    if {!$level} break
	    set nextDec [$w search -backwards $dec "$nextDec" $end]
	}
    }
    $w tag add match $nextDec "$nextDec +1c" "insert -1c" insert
    set Editor(matchId) [after 1000 "$w tag remove match 1.0 end"]
}

# ::two::editor::ShowMatch --
#
#	Show matching parenthesis etc., depending character
#	at the insertion cursor (or the selected character).
#
# Sideeffects:
#	Highlights the matching characters for one second.
#
# Arguments:
#	w -		Path to text widget

proc ::two::editor::ShowMatch {w} {
    if {[llength [$w tag ranges sel]]} {
        set char [$w get sel.first]
    } else {
        set char [$w get "insert -1c"]
    }
    switch $char {
        \" -
	\` -
	\' {
	    # removed - doesn't appear to be useful:-)
	    # MatchBackward $w $char
	}
	\( {
	    MatchNestForw $w \( \)
	}
	\) {
	    MatchNestBack $w \) \(
	}
	\[ {
	    MatchNestForw $w \[ \]
	}
	\] {
	    MatchNestBack $w \] \[
	}
	\{ {
	    MatchNestForw $w \{ \}
	}
	\} {
	    MatchNestBack $w \} \{
	}
    }
}

proc ::two::editor::lremove {listName element} {
    upvar $listName l
    
    set pos [lsearch -exact $l $element]
    if {$pos > -1} {
        set l [lreplace $l $pos $pos]
    }
}

# ::two::editor::HighlightInit --
#
#	Parse highlighting spec and build lists and hashtables to
#	increase the speed of the actual highlighting.
#
# Sideeffects:
#	Sets some entries to $Editor
#
# Arguments:
#	w -		Path to text widget

proc ::two::editor::HighlightInit {w} {
    variable Editor

    if {![llength $Editor($w,string)] && ![llength $Editor($w,comment)]} {
        return
    }

    # Comment and string markers are _no_ regular expression (they
    # are used as part of an array _key_ -> no regexp possible), so
    # we have to escape special characters for use within a regexp:
    set bad {([*+|()^\\])}

    # List of markers for start and stop of string or comment
    set startMarkerList ""
    foreach startMarker $Editor($w,string) {
	set Editor($w,tag,$startMarker) string
        set Editor($w,stop,$startMarker) $startMarker
        regsub -all -- $bad $startMarker {[\1]} startMarker
        lappend startMarkerList $startMarker
    }
    foreach {commentStart commentStop} $Editor($w,comment) {
        regsub -all -- $bad $commentStop {[\1]} commentStop
	set Editor($w,tag,$commentStart) comment
        set Editor($w,stop,$commentStart) $commentStop
        regsub -all -- $bad $commentStart {[\1]} commentStart
        lappend startMarkerList $commentStart
    }
    # build the regular expression
    set Editor($w,startRE) [join $startMarkerList |]
    if {[string length $Editor($w,keywords)]} {
        set Editor($w,fieldsep) 1
	append Editor($w,startRE) {|(^|[^_a-zA-Z0-9#])(}
        append Editor($w,startRE) [join $Editor($w,keywords) |]
	append Editor($w,startRE) {)([^_a-zA-Z0-9#]|$)}
    } else {
        set Editor($w,fieldsep) 0
    }
    if {[string length $Editor($w,keywordRE)]} {
        append Editor($w,startRE) |($Editor($w,keywordRE))
    }
    return
}

# ::two::editor::HighlightStart --
#
#	Parse current content and highlight according to current
#	config (comment, string and keywords). The real work is
#	done in HighlightLoop
#
# Sideeffects:
#	Adds tags to comments, strings and keywords in current widget
#
# Arguments:
#	w -		Path to text widget

proc ::two::editor::HighlightStart {w} {
    variable Editor

#if {![info exists Editor(hl,start)]} {
#    set Editor(hl,start) [clock clicks]
#}

    # First of all, stop previous highlighting within this
    # editor (if non is on the way, nothing will happen)
    HighlightStop $w
    # List of markers for start and stop of string or comment
    if {![llength $Editor($w,string)] && ![llength $Editor($w,comment)]} {
        return
    }
    SendEvent $w <<HighlightStart>>
    $w.1.txt tag remove string  1.0 end
    $w.1.txt tag remove comment 1.0 end
    $w.1.txt tag remove keyword 1.0 end
    # set cursor
    set Editor($w,cursor) [$w.1.txt cget -cursor]
    if {[string match watch $Editor($w,cursor)]} {
        set Editor($w,cursor) xterm
    }
    $w.1.txt configure -cursor watch
    
    # Add this widget to list of widgets for highlight loop
    lappend Editor(hl,widgets) $w
    set Editor($w,highlightPos) 1.0
    # schedule highlight loop anew if there is none right now
    if {!$Editor(hl,active)} {
        set Editor(hl,active) 1
        after idle "after 1 [namespace code HighlightLoop]"
    }
    return
}

# ::two::editor::HighlightStop --
#
#	Stop any undergoing highlighting for widget w
#
# Arguments:
#	w -		Path to text widget

proc ::two::editor::HighlightStop {w} {
    variable Editor

    if {$Editor(hl,active) && [lsearch -exact $Editor(hl,widgets) $w] > -1} {
        lremove Editor(hl,widgets) $w
	if {[winfo exists $w]} {
	    $w.1.txt configure -cursor $Editor($w,cursor)
            SendEvent $w <<HighlightStop>>
	}
    }
    return
}

# ::two::editor::HighlightLoop --
#
#	This is the loop for actually perfoming the highlighting.
#	It is done as an after-idle-loop to not freeze the user
#	interface while highlighting, but each iteration does
#	5 highlightings for efficiency reasons. Highlings are
#	done in all widgets given in the Editor(hl,widgets)
#	list.
#

proc ::two::editor::HighlightLoop {} {
    variable Editor

    # List of word characters
    set chars {[_a-zA-Z0-9#]}
    # loop through all widgets which are to be highlighted
    foreach w $Editor(hl,widgets) {
        if {![winfo exists $w]} {
            # remove from list of highlighted windows
            lremove Editor(hl,widgets) $w
            continue
        }
    
        # cache RE in variable (avoid expensive Hashtable lookup)
        # and several other vars as well
        set startRE $Editor($w,startRE)
        set start $Editor($w,highlightPos)
        # We need to access the underlyign two_text widget. Usually,
        # this would be $w.txt.1 - but this goes through quite a lot
        # of code and makes the highlighting take about twice as long.
        # Non-modifying operations might well directly access the
        # TCL text widget, but modifications need to be done in all
        # clones (split editor!). Still it would be somehow better
        # to use $w.txt.1 as command, but we need the speed...
        # $txt1 -> one widget, only for non-modifying calls
        # $txt2 -> list of widgets; use whenever modifying the text
        set txt1 ::two::text::_two$w.1.txt
        if {$Editor($w,split)} {
            set txt2 [list ::two::text::_two$w.1.txt ::two::text::_two$w.2.txt]
        } else {
            set txt2 [list ::two::text::_two$w.1.txt]
        }
    
        # do 10 highlightings per widget at a time
        for {set loop 0} {$loop < 10} {incr loop} {
            # search for first candidate
            set start [$txt1 search -regexp \
                -count [namespace current]::Editor(hl,startLen) \
                -- $startRE $start end]
            if {![string length $start]} {
                # finished: there are no more candidates
                lremove Editor(hl,widgets) $w
                foreach txt $txt2 {
                    $txt configure -cursor $Editor($w,cursor)
                }
                SendEvent $w <<HighlightStop>>
                break
            }
        
            # what exactly is our candidate?
            set idx "$start + $Editor(hl,startLen) c"
            set match [$txt1 get $start $idx]
            if {[info exists Editor($w,stop,$match)]} {
                # O.k., we have a stop marker accompaining this start marker,
                # so it must be a string or a comment
                if {[IsEscaped $txt1 $start]} {
                    # escaped (backslashed) start delimiter -> ignore it:
                    # continue right after this delimiter
                    set start $idx
                } else {
                    # search for stop marker right after the start delimiter
                    while {1} {
                        # search for next candidate for stop delimiter.
                        # Need a workaround here: if stop delimiter is end-of-line,
                        # and $idx *is* at the end of the line, text search will
                        # return the end-of-line of the *next* line :-(
                        if {[string compare {$} $Editor($w,stop,$match)] == 0} {
                            set idx [$txt1 index "$idx lineend"]
                            set Editor(hl,stopLen) 0
                        } else {
                            set idx [$txt1 search -regexp \
                                -count [namespace current]::Editor(hl,stopLen) \
                                -- $Editor($w,stop,$match) $idx end]
                        }
                        if {[string length $idx]} {
                            if {$Editor(hl,stopLen)>0 && [IsEscaped $txt1 $idx]} {
                                # backslashed stop delimiter -> ignore it and
                                # continue right after the stop delimiter. We
                                # only accept this for non-zero length stop
                                # delimiters, so e.g. escaped end-of-lines are
                                # NOT covered here. There simply is no
                                # consistent to do this: different languages
                                # behave differently:-(
                                append idx " + $Editor(hl,stopLen) c"
                                continue
                            }
                            # found stop marker
                            set stop "$idx + $Editor(hl,stopLen) c"
                            foreach txt $txt2 {
                                $txt tag add $Editor($w,tag,$match) $start $stop
                            }
                            set start $stop
                            break
                        } else {
                            # there ain't no matching stop marker!?
                            # ignore this one and continue highlighting right after
                            # the start delimiter
                            append start " + $Editor(hl,startLen) c"
                            break
                        }
                    }
                }
            } else {
                # must be a keyword.
                if {$Editor($w,fieldsep)} {
                    # Keywords can be found at the beginning of the line or right after
                    # a field separator: in the second case, the field separator has
                    # been matched as well, but shouldn't get highlighted:-)
                    # Does the match start with a regular character (no field separator)?
                    # If not: set index l1 to 1
                    set l1 [expr {[string match $chars [string index $match 0]] ? 0 : 1}]
        
                    # Similar things happend at the end of a keyword: might be the lineend
                    # or a field separator.
                    # Does the match end with a regular character (no field separator)?
                    set len [string length $match]
                    incr len -1
                    set l2 [expr {[string match $chars [string index $match $len]] ? $len + 1 : $len}]
                } else {
                    # No list of keywords, but a RE given. Just ignore the fuzz
                    # about field separators: the RE should *exactly* describe
                    # the keyword (usually some kind of tag like in HTML)
                    set l1 0
                    set l2 [string length $match]
                }
                
                # add the tag and continue right after the keyword (excluding a
                # trailing field separator)
                set stop "$start + $l2 c"
                foreach txt $txt2 {
                    $txt tag add keyword "$start + $l1 c" $stop
                }
                set start $stop
            }
        }
        set Editor($w,highlightPos) $start
    }
    # give event loop a chance and then resort to the highlight loop
    if {[llength $Editor(hl,widgets)]} {
        after idle "after 1 [namespace code HighlightLoop]"
    } else {
        set Editor(hl,active) 0
#puts [expr {[clock clicks] - $Editor(hl,start)}]
#unset Editor(hl,start)
    }
    return
}

# ::two::editor::IsEscaped --
#
#	Test if the character at a given position is masked by a backslash.
#	It checks if there is an odd number of backslashes right before
#	this index (even would mean: backslash itself is escaped)
#
# Arguments:
#	txt		Path to text widget
#	idx		Position of possibly escaped character
#
# Returns:
#	1 if character is escaped, 0 otherwise.

proc ::two::editor::IsEscaped {txt idx} {
    if {[string compare \\ [$txt get "$idx -1c" $idx]]!=0} {
        return 0
    }
    set line [$txt get "$idx linestart" $idx]
    set pre  [string trimright $line \\]
    # We really should write "return [expr...]" here, but
    # this version should be faster:
    expr {([string length $line] - [string length $pre]) % 2}
}

# ::two::editor::SendEvent --
#
#	Generate an event. If the widget is new and never has been mapped
#	(e.g. eventloop not been entered since creation), events would
#	be lost. In such a case, delay the creation, otherwise create now.
#
# Arguments:
#	w -		Path to text widget
#	ev -		Name of event which should be created
#
# Side effects:
#	Adds lines to the <Map> binding

proc ::two::editor::SendEvent {w ev} {
    variable Editor

    if {$Editor($w,isnew)} {
        bind $w <Map> "+event generate [list $w $ev]"
    } else {
        event generate $w $ev
    }
}

namespace import ::two::editor::*

# shell.tcl --
#
#	A very simple terminal emulation. The "shell" prompt, commands
#	which are to be run and their in-/output use different tags.
#	The command's name can be faked.
#
# Copyright (c) 1998-99 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoShell 1.0
package require twoUtil

namespace eval ::two::shell {
    namespace export two_shell
    variable Shell
    
    namespace import ::two::util::two_parseArgs
    
    # variable Shell
    #
    # expect -		1 if expect is available, else 0
    # $w,prompt -	string to use as fake prompt
    # $w,spawnId -	used by expect
    # $w,tty -		name of terminal used by expect
    # $w,oldfocus -	remember focus when running a program
    # $w,aftercmd -	command to evaluate after completion
    # $w,destbs -   	true if destructive backspace
    # $w,batch -	List of commands which should run next
    
    set Shell(expect) [expr {![catch {package require Expect}]}]
}

# ::two::shell::two_shell --
#
#	Create widget, parse arguments and install our own widget procedure.
#	The real widget is "$w" (like ".tl.sh"); the original widget procedure
#	is renamed to "_two$w" (like "_two.tl.sh") and a call to "widget"
#	is put in place. "widget" will extract all additional options and
#	subsequently call "_two$w" with the remaining options.
#
# Arguments:
#	w		Widget path
#	args		Widget arguments and options

proc ::two::shell::two_shell {w args} {
    variable Shell

    # Initialise the globals
    array set Shell "
	$w,prompt	{shell> }
	$w,spawnId	0
	$w,tty		{}
	$w,oldfocus	{}
	$w,destbs	1
	$w,batch	{}
    "

    # See if this option is for us...
    two_parseArgs args Shell($w,prompt) "-prompt"
    two_parseArgs args Shell($w,detsbs) "-destructivebs"

    # text widget
    uplevel [list text $w] $args -takefocus 0

    # Remove all bindings
    bindtags $w [list $w TwoShellTerm [winfo toplevel $w] all]

    # Insert prompt
    $w insert end $Shell($w,prompt) prompt
    $w see end
    
    # mark beginning of command
    $w mark set command end-1c
    $w mark gravity command left

    # Move widget command
    rename $w _two$w

    # Install our replacement command
    namespace eval :: "
        proc $w {opt args} {
	    uplevel [namespace code Widget] $w \$opt \$args
	}
    "
    
    # catch widget destruction in order to clean up Shell
    bind $w <Destroy> [namespace code "_Destroy $w"]

    # Return the correct result...
    return $w
}

# ::two::shell::Widget --
#
#	replacement for widget command
#
# Arguments:
#	w		Path to real widget
#	opt		Widget command...
#	args		optional args

proc ::two::shell::Widget {w opt args} {
    variable Shell

    # See if this option is for us...
    switch -exact -- $opt {
	configure {
	    if {[two_parseArgs args Shell($w,prompt) "-prompt"]} {
	        # change all currently displayed prompts!
	        # The following works, because only one prompt per
	        # line is possible, so indexes are still valid after
	        # changing the first occurences.
	        foreach {start end} [_two$w tag ranges prompt] {
	            _two$w delete $start $end
	            _two$w insert $start $Shell($w,prompt) prompt
	        }
	    }
            two_parseArgs args Shell($w,destbs) "-destructivebs"
	    if {[llength $args]} {
		return [uplevel [list ::two::shell::_two$w $opt] $args]
	    }
	}
	cget {
	    switch -exact -- $args {
	        -prompt        { return $Shell($w,prompt) }
	        -destructivebs { return $Shell($w,destbs) }
		default        { return [uplevel [list ::two::shell::_two$w $opt] $args] }
	    }
	}
	run {
	    set Shell($w,aftercmd) ""
	    two_parseArgs args Shell($w,aftercmd) "-aftercmd"
	    # any command given?
	    if {![llength $args] || ![string length [lindex $args 0]]} {
	        return -code error "wrong # args: should be \"two_shell run command\""
	    }
	    # $fake is to be shown as command which is executed
	    set fake  [lindex $args 0]
	    set dir   [pwd]
	    set focus 1
	    two_parseArgs args fake  "-fake"
	    two_parseArgs args dir   "-dir"
	    two_parseArgs args focus "-focus"
	    if {[llength $args] > 1} {
	        return -code error "unkown option \"[join [lrange $args 1 end] { }]\""
	    }
	    Run $w [lindex $args 0] $dir $fake $focus
	}
	kill {
	    Kill $w
	}
	runState {
	    if {$Shell($w,spawnId) != 0} {
        	return 1
	    } else {
        	return 0
	    }
	}
	clear {
	    $w delete 1.0 end
	    if {$Shell($w,spawnId) == 0} {
		$w insert end $Shell($w,prompt) prompt
	    }
	}
	default {
            return [uplevel [list ::two::shell::_two$w $opt] $args]
	}
    }
    return
}

# ::two::shell::_Destroy --
#
#	Call this procedure when the shell widget is destroyed.
#	Will clean widget-specific Elements of Shell and destroy
#	the original widget command.
#
# Arguments:
#	w		Path to Widget

proc ::two::shell::_Destroy {w} {
    variable Shell

    # clear Shell
    foreach key [array names Shell "$w,*"] {
        unset Shell($key)
    }
    # remove the original widget command
    rename $w ""
    return
}

#-------------------------------------------------------------------------
# The code below creates the TwoShell-class bindings
#-------------------------------------------------------------------------

bind TwoShellActive <KeyPress> {
    ::two::shell::Input %W %A
    break
}
bind TwoShellActive <Any-Tab> {
    ::two::shell::Input %W \t
    # don't change focus
    break
}

bind TwoShellActive <Control-KeyPress> {
    if {[string match c %K]} {
        ::two::shell::Kill %W
    } else {
        ::two::shell::Input %W %A
    }
    break
}

bind TwoShellActive <Escape> {
    ::two::shell::Kill %W
    break
}

bind TwoShellActive <Return> {
    if {$::two::shell::Shell(expect)} {
        ::two::shell::Input %W \r
    } else {
        ::two::shell::Input %W \n
    }
    break
}

bind TwoShellActive <Button-1> {
    focus %W
}

# Special key sequences...
# they start with an escape: \x1b (decimal: 27)
bind TwoShellActive <Up>        {::two::shell::Input %W "\x1b\[A"}
bind TwoShellActive <Down>      {::two::shell::Input %W "\x1b\[B"}
bind TwoShellActive <Right>     {::two::shell::Input %W "\x1b\[C"}
bind TwoShellActive <Left>      {::two::shell::Input %W "\x1b\[D"}
bind TwoShellActive <End>       {::two::shell::Input %W "\x1b\[F"}
bind TwoShellActive <Home>      {::two::shell::Input %W "\x1b\[H"}
bind TwoShellActive <Insert>    {::two::shell::Input %W "\x1b\[2~"}
bind TwoShellActive <Delete>    {::two::shell::Input %W "\x1b\[3~"}
bind TwoShellActive <BackSpace> {::two::shell::Input %W "\b"}
bind TwoShellActive <Prior>     {::two::shell::Input %W "\x1b\[5~"}
bind TwoShellActive <Next>      {::two::shell::Input %W "\x1b\[6~"}

#-------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------

# ::two::shell::Input --
#
#	Insert text through currently running program
#
# Arguments:
#	w		Path to widget
#	txt		Text to send to program

proc ::two::shell::Input {w txt} {
    variable Shell

    if {[string length $txt]} {
        if {$Shell(expect)} {
            exp_send -i $Shell($w,spawnId) -- $txt
        } else {
            puts -nonewline $Shell($w,spawnId) $txt
            $w insert end $txt
            $w see end
        }
    }
}

# ::two::shell::Run --
#
#	execute a given command
#
# Argumets:
#	w		Path to widget
#	cmd		command to be executed
#	dir		working directory for command
#	fake		display fake as command
#	focus		if 1, set focus to shell window

proc ::two::shell::Run {w cmd dir fake focus} {
    variable Shell
    global env

    # already up and running?
    if {$Shell($w,spawnId) != 0} {
        # Kill $w
        # Delay for later execution
        lappend Shell($w,batch) [list $cmd $dir $fake $focus]
        return
    }

    # remember focus
    if {$focus} {
        set Shell($w,oldfocus) [focus]
        focus $w
    } else {
        set Shell($w,oldfocus) ""
    }

    # notify listener
    event generate $w <<RunningOn>>
    update idletasks

    # mark beginning of command
    $w mark set command end-1c

    # insert fake command
    $w insert end "$fake\n" command
    $w mark set insert end
    $w see insert
    
    if {$Shell(expect)} {
        # size of widget
        set f [$w cget -font]
        set rows [expr {[winfo reqheight $w] / [font metrics $f -linespace]}]
        set cols [expr {[winfo reqwidth $w]  / [font measure $f {0}]}]

	# clear env
	set envVars [list LINES COLUMNS TERM TERMCAP]
        foreach envVar $envVars {
            if {[info exists env($envVar)]} {
                set oldEnv($envVar) $env($envVar)
                unset env($envVar)
            }
        }
        set env(LINES)   $rows
        set env(COLUMNS) $cols
        # nd = Non-destructive space (cursor or carriage right)
        # cl = Clear screen and home cursor
        # ce = Clear to end of line

        # run command
        set stty_init "-tabs erase \^H columns $cols rows $rows"
        set cwd [CdPwd $w $dir]
        if {[catch [concat spawn -noecho -nottycopy $cmd]]} {
            cd $cwd
            $w insert end "Couldn't run "
            $w insert end "$fake\n" command
            Stopped $w
            return
        }
        cd $cwd
        set Shell($w,spawnId) $spawn_id
        set Shell($w,tty)     $spawn_out(slave,name)

        # don't echo anything on stdout
        log_user 0

	# re-set env to original state
        foreach envVar $envVars {
            if {[info exists oldEnv($envVar)]} {
                set env($envVar) $oldEnv($envVar)
            } elseif {[info exists env($envVar)]} {
                unset env($envVar)
            }
        }

        # read output from command
        set txt ::two::shell::_two$w
        expect_background "
            -i  $Shell($w,spawnId)
            eof {
                [namespace code Stopped] $w
            }
            {\a} {
                bell
            }
            {^\b} {
		if {$Shell($w,destbs)} {
		    $txt delete {insert - 1c} insert
		} else {
                    $txt mark set insert {insert - 1c}
		}
            }
            {^\n} {
                $txt mark set insert end
                $txt insert insert {\n}
                $txt see insert
                update idletasks
            }
            {^\r} {
                $txt mark set insert {insert linestart}
                $txt see insert
            }
            -re {^\[^\a\b\n\r\]+} {
                if {\[$txt compare insert != {end -1c}\]} {
                    $txt delete insert \"insert + \[string length \$expect_out(0,string)\] c\"
                }
                $txt insert insert \$expect_out(0,string)
                $txt see insert
            }
        "
    } else {
        # run command
        set cwd [CdPwd $w $dir]
        set Shell($w,spawnId) [open "|$cmd |& cat -u" r+]
        fconfigure $Shell($w,spawnId) -blocking 0 -buffering none
        cd $cwd

        # read output from command
        fileevent $Shell($w,spawnId) readable "
            if {\[eof $Shell($w,spawnId)\]} {
                [namespace code Kill] $w
            } else {
                $w insert end \[read $Shell($w,spawnId)\]
                $w see end
            }
        "
    }

    # Enable interaction
    bindtags $w [list $w TwoShellTerm TwoShellActive [winfo toplevel $w] all]
    _two$w configure -takefocus 1

    return
}

# ::two::shell::Kill --
#
#	stop execution of a command
#
# Argumets:
#	w		Path to widget

proc ::two::shell::Kill {w} {
    variable Shell

    if {$Shell($w,spawnId) != 0} {
        if {$Shell(expect)} {
            close -i $Shell($w,spawnId)
	} else {
            $w insert end [read $Shell($w,spawnId)]
	    catch {close $Shell($w,spawnId)}
	}
        Stopped $w
    }
    return
}

# ::two::shell::Stopped --
#
#	yes, command has been stopped. This procedure does all required
#	clean ups, state changes and evals the afterCommand
#
# Argumets:
#	w		Path to widget

proc ::two::shell::Stopped {w} {
    variable Shell
    global errorCode

    if {$Shell(expect) && $Shell($w,spawnId) != 0} {
        # Make sure the process is vanished -> avoid Zombies
        if {! [catch {set waitret [wait -i $Shell($w,spawnId)]}]} {
            # normal exiting with unusual return value?
            foreach {pid spid os ret errcl sig desc} $waitret {break}
#            if {$os == -1} {
#                $w insert end "OS error: $ret ($errorCode)\n" command
#            } elseif {$ret != 0} {
#                $w insert end "Return value: $ret\n" command
#            }
            if {[string match CHILDKILLED $errcl]} {
                $w insert end "Program killed: $desc ($sig)\n" command
            }
        }
    }
    set Shell($w,spawnId) 0

    # reset focus
    if {[string match $w [focus]] && [winfo exists $Shell($w,oldfocus)]} {
        focus $Shell($w,oldfocus)
    }

    # Make sure there is a newline
    if {![string match [$w index insert] [$w index "insert linestart"]]} {
        $w insert end "\n"
    }
    
    # Evaluate aftercmd (if set)
    if {[string length $Shell($w,aftercmd)]} {
        namespace eval :: $Shell($w,aftercmd)
    }

    # Show prompt
    $w insert end $Shell($w,prompt) prompt
    $w see end

    # Disable interaction
    bindtags $w [list $w TwoShellTerm [winfo toplevel $w] all]
    $w configure -takefocus 0
    if {[string match $w [focus]]} {
        focus [winfo parent $w]
    }

    event generate $w <<RunningOff>>
    
    # More commands in batch queue?
    if {[llength $Shell($w,batch)]} {
        set cmd [lindex $Shell($w,batch) 0]
        set Shell($w,batch) [lrange $Shell($w,batch) 1 end]
        after idle "::two::shell::Run [list $w] $cmd"
    }
    return
}

# ::two::shell::CdPwd --
#
#	Change the directory and catch errors. The return value should
#	be used to return to the current directory (using a regular cd)
#
# Argumets:
#	w		Path to widget
#	dir		Name of directory
#
# Results:
#	Returns the current dir
#
# Side effects:
#	If the cd was not successfull, call Stopped and make
#	the calling procedure return (here: return -code return)

proc ::two::shell::CdPwd {w dir} {
    set cwd [pwd]
    if {[catch {cd $dir} msg]} {
        $w insert end "$msg\n" command
        # return to start
        catch {cd $cwd}
        Stopped $w
        return -code return
    }
    return $cwd
}

namespace import ::two::shell::*

# keyrec.tcl --
#
#	Record key press events. Usefull e.g. when building new dialog
#	widgets which are in need of calls to update, but don't want
#	to loose key events.
#
# Copyright (c) 1999 Frank J. Leitner
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Header: /home/ldv/cvs/two/lib/libw.tcl,v 1.79 2000/01/13 14:51:19 fjl Exp $

package provide twoKeyrec 1.0

namespace eval ::two::keyrec {
    namespace export two_keyrec
    variable Keyrec
    
    # variable Keyrec
    #
    # w -		Path to recorder widget
    # focus -	Old focus prior to starting the recorder
    # rec -		True if currently recording
    # e -		Events recorded so far
    
    array set Keyrec {
        e	""
    }
}

# ::two::keyrec::two_keyrec --
#
#	Record events
#
# Arguments:
#	cmd		Sub-Command
#	args		arguments and options

proc ::two::keyrec::two_keyrec {cmd args} {
    variable Keyrec
    
    switch $cmd {
        init {
            if {[info exists Keyrec(w)] && [winfo exists $Keyrec(w)]} {
                destroy $Keyrec(w)
            }
            if {[llength $args]>1} {
                return -code error "Worng # arguments: should be \"two_keyrec init ?widget?\""
            } elseif  {[llength $args]==1} {
                set Keyrec(w) [lindex $args 0]
            } else {
                set Keyrec(w) .__two_keyrec
            }
            frame $Keyrec(w) -class TwoKeyrec -takefocus 0
            place $Keyrec(w) -x -10 -y -10 -width 1 -height 1
        }
        start {
            if {![info exists Keyrec(w)]} {
                two_keyrec init
            }
            set Keyrec(focus) [focus]
            set Keyrec(rec)   1
            set Keyrec(e)     ""
            focus -force $Keyrec(w)
            update idletasks
        }
        stop {
            if {![info exists Keyrec(rec)] || !$Keyrec(rec)} {
                return -code error "Must call \"two_keyrec start\" first"
            }
            update
            if {[winfo exists $Keyrec(focus)]} {
                focus -force $Keyrec(focus)
            } else {
                focus -force .
            }
            set Keyrec(rec) 0
        }
        replay {
            if {[llength $args] > 1} {
                return -code error "Wrong # arguments: should be \"two_keyrec replay ?widget?\""
            } elseif {[llength $args] == 1} {
                set w [lindex $args 0]
            } else {
                set w [list [focus]]
            }
            if {![string length $w] || ![winfo exists $w]} {
                set w .
            }
            foreach e $Keyrec(e) {
                eval [list event generate $w] $e
            }
        }
    }
    return
}

proc ::two::keyrec::addBindings {} {
    foreach sequence {
        <KeyPress> <KeyRelease>
        <Shift-KeyPress> <Shift-KeyRelease>
        <Control-KeyPress> <Control-KeyRelease>
        <Shift-Control-KeyPress> <Shift-Control-KeyRelease>
    } {
        bind TwoKeyrec $sequence "
            if \$::two::keyrec::Keyrec(rec) {
                lappend ::two::keyrec::Keyrec(e) \[list $sequence -keycode %k -keysym %K -time %t\]
            }
            break
        "
    }

    bind TwoKeyrec <FocusOut> {
        if {$::two::keyrec::Keyrec(rec) && [string length [focus]]} {
            set ::two::keyrec::Keyrec(focus) [focus]
            focus -force $::two::keyrec::Keyrec(w)
        }
    }
}
::two::keyrec::addBindings

namespace import ::two::keyrec::*

